/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/CType.java, ras, dumpviewer 1.6
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 2002
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */
/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION: C typedefinitions for access to JVM control blocks
 * ===========================================================================
 */


package com.ibm.jvm.dump;

import java.lang.*;
import java.util.*;
import java.io.*;


public class CType {


    public static final int TYPE_PRIMITIVE = 1;
    public static final int TYPE_STRUCT    = 2;
    public static final int TYPE_UNION     = 3;
    public static final int TYPE_ENUM      = 4;
    public static final int TYPE_POINTER   = 5;
    public static final int TYPE_ARRAY     = 6;
    public static final int TYPE_FUNCTION  = 7;
    public static final int TYPE_ALIAS     = 8;
    public static final int TYPE_STRUCTTAG = 9;



    int typeNum;
    int size;
    int baseType;
    String name;
    String description;
    CType aliasType;

    int numFields;
    FieldInfo fields[];

    CType derefType;
    ArrayInfo arrayInfo;

    static LinkedList declaredTypes = new LinkedList();
    static EnumValues enumValues[];

    private CType parseTypeNum(String t, int offset) {
        String tnum = t.substring(offset);
        int typeNum = Integer.parseInt(tnum);
        CType type;
        if ((type = find(typeNum)) != null)
            return type;
        else {
            type = create(typeNum);
            return type;
        }
    }

    private int parseField(String t, int offset) {
        int start = offset;
        int end = start;
        FieldInfo field = new FieldInfo();
        end = t.indexOf(':', start);
        if (end < 0)
            return t.length();
        field.name = t.substring(start, end);
        end++; start = end;
        end = t.indexOf(',', start);
        int typeNum = Integer.parseInt(t.substring(start, end));
        if ((field.type = find(typeNum)) == null)
            field.type = create(typeNum);
        end++; start = end;
        end = t.indexOf(',', start);
        field.offset = Integer.parseInt(t.substring(start, end));
        end++; start = end;
        end = t.indexOf(';', start);
        field.size = Integer.parseInt(t.substring(start, end));

        numFields++;
        FieldInfo[] oldfields = fields;
        fields = new FieldInfo[numFields];
        if (oldfields != null)
            System.arraycopy(oldfields, 0, fields, 0, numFields-1);
        fields[numFields-1] = field;
        end++; return end;
    }

    private void parseStruct(String t, int offset) {
        int start = offset+1;
        if (t.equals("s0")) {
            if (name == null)
                name = "void";
            return;
        }
        while (t.charAt(start) >= '0' && t.charAt(start) <= '9') 
            start++;
        if (size == 0)
            size = Integer.parseInt(t.substring(offset+1, start));
        numFields = 0;
        while (start < t.length()) {
            start = parseField(t, start);
        }
        
    }

    private void parseArray(String t, int offset) {
        arrayInfo = new ArrayInfo();
        int start = offset+4;  // skip ar0;
        int end;
        arrayInfo.dimensions = 1;
        arrayInfo.start = new int[1];
        arrayInfo.end = new int[1];
        end = t.indexOf(';', start);
        arrayInfo.start[0] = Integer.parseInt(t.substring(start, end));
        end++; start = end; end = t.indexOf(';', start);
        arrayInfo.end[0] = Integer.parseInt(t.substring(start, end));
        end++; start = end;
        while (end < t.length() && t.charAt(end) != ';') end++;
        arrayInfo.type = create(Integer.parseInt(t.substring(start, end)));
    }
    private void parseEnum(String t, int offset) {
        String name;
        int val;
        int numEnums = enumValues != null?enumValues.length:0;
        EnumValues newEnums[];
        int start, end;

        // skip e<typenumber>
        offset++; 
        if (t.charAt(offset) == '-') offset++;
        while (t.charAt(offset) >= '0' && t.charAt(offset) <= '9')
            offset++;
        start = offset;
        end = start;
        while (end < t.length()) {
            end = t.indexOf(':', start);
            if (end < 0 || end == start)
                break;
            name = t.substring(start, end);
            end++; start = end;
            end = t.indexOf(',', start);
            val = Integer.parseInt(t.substring(start, end));
            end++; start = end;
            numEnums++;
            newEnums =  new EnumValues[numEnums];
            if (enumValues != null)
                System.arraycopy(enumValues, 0, newEnums, 0, numEnums-1);
            enumValues = newEnums;
            enumValues[numEnums-1] = new EnumValues();
            enumValues[numEnums-1].name = name;
            enumValues[numEnums-1].value = val;
        }
    }


    private void parseStab(String description) {
        int index = 0;
        if (description == null || description.length() == 0)
            return;
        if (description.charAt(index) == 'V' || 
            description.charAt(index) == 'k')
                index++;
        switch (description.charAt(index)) {
        case '*':
            baseType = TYPE_POINTER;
            derefType = parseTypeNum(description, index+1);
            break;
        case 's':
            baseType = TYPE_STRUCT;
            parseStruct(description, index);
            break;
        case 'u':
            baseType = TYPE_UNION;
            parseStruct(description, index);
            break;
        case 'e':
            baseType = TYPE_ENUM;
            parseEnum(description, index);
            break;
        case 'a':
            baseType = TYPE_ARRAY;
            parseArray(description, index);
            break;
        case 'f':
            baseType = TYPE_FUNCTION;
            derefType = parseTypeNum(description, index+1);
            break;
        case '-':
            String tnum = description.substring(index);
            baseType = Integer.parseInt(tnum);
            break;
        case 'r':
            break;
        default:
            baseType = TYPE_ALIAS;
            aliasType = parseTypeNum(description, index);
            break;
        }
    }

    /* Create a new CType object with the specified type number to be
     * used as placeholder when a type reference is found
     */
    static CType create(int typeNum) {
        CType t;
        if ((t = find(typeNum)) == null) {
            t = new CType();
            t.typeNum = typeNum;
            declaredTypes.add(t);
        }
        return t;
    }

    /**
     * Create a new CType with the given type number, name and description
     * If a type object of this number already exists, override it.
     *
     * @param   typeNum      Unique type number
     * @param   baseType     Type number of base type
     * @param   size         Size in bytes of type
     * @param   name         Valid C type name
     * @param   description  Type description as a stab string
     * @return  New CType object
     * @exception  none
     */
    public static CType create(int typeNum, int baseType, int size, String name, String description) {
        CType t;
        if ((t = find(typeNum)) == null) {
            t = new CType();
            declaredTypes.add(t);
        }
        t.typeNum = typeNum;
        t.baseType = baseType;
 
        t.size = size;
        t.name = name;
        t.description = description;
        if (description != null && description.length() > 0) {
            t.parseStab(description);
        }
        return t;
    }


    /**
     * Find a CType by name
     *
     * @param   name         Name of type to find
     * @return  CType object of specified name, or null on failure
     * @exception  none
     */
    public static CType find(String name) {
        for (Iterator iter = declaredTypes.iterator(); iter.hasNext(); ) {
            CType t = (CType) iter.next();
            if ((t.name != null) && t.name.equals(name))
                return t;
        }
        return null;
    }

    /**
     * Find a CType by type number
     *
     * @param   typeNum      Unique type number of type to find
     * @return  CType object of specified number, or null on failure
     * @exception  none
     */
    public static CType find(int typeNum) {
        for (Iterator iter = declaredTypes.iterator(); iter.hasNext(); ) {
            CType t = (CType) iter.next();
            if (t.typeNum == typeNum)
                return t;
        }
        return null;
    }


    /* Skip alias types to find the actual type definition */
    CType getType() {
        CType t = this;
        while (t.baseType == CType.TYPE_ALIAS)
            t = t.aliasType;
        return t;
    }

    /**
     * Return type name
     *
     * @return  Name of type
     * @exception  none
     */
    public String getName() {
        return name;
    }

    /**
     * Return type size in bytes
     *
     * @return  Size of type
     * @exception  none
     */
    public int getSize() {
        if (getType() != this)
            return getType().size;
        return size;
    }

    /**
     * Return number of fields for structure/union
     *
     * @return  number of fields
     * @exception  none
     */
    public int getNumFields() {
        if (getType() != this)
            return getType().numFields;
        return numFields;
    }

    /**
     * Return type of a field specified by field index
     *
     * @param   index    field index
     * @return  CType of specified field
     * @exception  none
     */
    public CType getFieldType( int index ) {
        if (getType() != this)
            return getType().getFieldType(index);
        if (index < numFields && index >= 0)
            return fields[index].type.getType();
        else
            return null;
    }

    /**
     * Return byte offset of a field specified by field index
     *
     * @param   index    field index
     * @return  Byte offset of specified field
     * @exception  none
     */
    public int getFieldOffset( int index ) {
        if (getType() != this)
            return getType().getFieldOffset(index);
        if (index < numFields && index >= 0)
            return fields[index].offset/8;
        else
            return -1;
    }

    /**
     * Return byte size of a field specified by field index
     *
     * @param   index    field index
     * @return  Size in bytes of specified field
     * @exception  none
     */
    public int getFieldSize( int index ) {
        if (getType() != this)
            return getType().getFieldSize(index);
        if (index < numFields && index >= 0)
            return fields[index].size/8;
        else
            return -1;
    }

    /**
     * Return name of a field specified by field index
     *
     * @param   index    field index
     * @return  Name of specified field
     * @exception  none
     */
    public String getFieldName( int index ) {
        if (getType() != this)
            return getType().getFieldName(index);
        if (index < numFields && index >= 0)
            return fields[index].name;
        else
            return null;
    }

    /**
     * Return type of a field specified by field name
     *
     * @param   name   field name
     * @return  CType of specified field
     * @exception  none
     */
    public CType getFieldType( String name ) {
        return getFieldType(getFieldIndex(name));

    }

    /**
     * Return byte offset of a field specified by field name
     *
     * @param   name   field name
     * @return  Byte offset of specified field
     * @exception  none
     */
    public int getFieldOffset( String name ) {
        return getFieldOffset(getFieldIndex(name));
    }

    /**
     * Return byte size of a field specified by field name
     *
     * @param   name   field name
     * @return  Size in bytes of specified field
     * @exception  none
     */
    public int getFieldSize( String name ) {
        return getFieldSize(getFieldIndex(name));
    }

    /**
     * Return index of a field specified by field name
     *
     * @param   name   field name
     * @return  Field index of specified field
     * @exception  none
     */
    public int getFieldIndex( String name ) {
        if (getType() != this)
            return getType().getFieldIndex(name);
        for ( int i = 0; i < numFields; i++ )
            if (name.equals(fields[i].name))
                return i;
        return -1;
    }


    /**
     * Return value of enumeration constant
     *
     * @param   name   name of enumeration constant
     * @return  integer value of enumeration constant
     * @exception  none
     */
    public static int getConstant(String name) {
        for ( int i = 0; i < enumValues.length; i++) {
            if (name.equals(enumValues[i].name))
                 return enumValues[i].value;
        }
        return 0;
    }

    static void list() {
        for (Iterator iter = declaredTypes.iterator(); iter.hasNext(); ) {
            CType t = (CType) iter.next();
            System.out.println(t);
        }
    }

    class FieldInfo {
        CType type;
        int offset;
        int size;
        String name;
    }
    class ArrayInfo {
        int dimensions;
        int start[];
        int end[];
        CType type;
    }

    class EnumValues {
        String name;
        int value;
    }
        

    /**
     * Convert to readable string
     *
     * @param   none
     * @return  Multi-line string output
     * @exception  none
     */
    public String toString() {
        String str = "Typenum: "+typeNum+"\n";
        str = str + "name = " + name+" " +"baseType = "+baseType+"\n";
        if (baseType == TYPE_STRUCT || baseType == TYPE_UNION) {
            for ( int i = 0; i < numFields; i++ ) {
                CType fieldType = fields[i].type;
                int indirectionLevel = 0;
                while (fieldType.derefType != null) {
                    indirectionLevel++;
                    fieldType = fieldType.derefType;
                }
                str += "\t"+fieldType.name +" ";
                for (int j = 0; j < indirectionLevel; j++)
                    str += "*";
                str += "\t"+fields[i].name+" : "+fields[i].offset/8+"\n";
                //str += "\t\t"+fields[i].type.toString()+"\n";
            }
        }
        return str;
    }

    /**
     * Delete all known types. This function should be called before
     * loading a new type file, to ensure that old types are not used.
     *
     * @param   none
     * @return  none
     * @exception  none
     */
    public void deleteAll() {
        declaredTypes = null;
    }

}
