/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/CTypeFile.java, ras, dumpviewer 1.4
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 2002
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */
/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION: Support to read in C typedefinitions from typedef file.
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.lang.*;
import java.util.*;
import java.io.*;

public class CTypeFile {


    /* Read a NULL terminated string from byte array
     */
    private String getString(byte strings[], int offset, String encoding) {
        int len = 0;
        
        for (int i = offset; strings[i] != '\0'; i++, len++) ;
        if (len == 0)
            return null;
        if (strings[offset+len-1] == ';')
            len--; /* skip trailing ; from description */
        try {
            String str = new String(strings, offset, len, encoding);
            return str;
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }


    /**
     * Process type definitions from typedef file specified by <filename>
     * Note that loading of multiple typedef files can lead to conflicts.
     * CTypes corresponding to all types read from the file can be accessed
     * using CType functions after the file has been processed.
     *
     * @param   fileName   Name of file to load
     * @return  none
     * @exception  none
     */
    public void processTypes(String fileName) {
        try {
            byte magic[] = new byte[8];
            int numSyms;
            int symOffset;
            int symSize;
            int strOffset;
            int strSize;
            int version, enc;
            String encoding;

            RandomAccessFile in = new RandomAccessFile(fileName, "r");
            in.read(magic, 0, 8);
            version = in.readInt();
            enc= in.readInt();
            encoding = enc == 1?"Cp037":"ascii";
            String magicStr = new String(magic, 0, 7, encoding);
            if (!magicStr.equals("JVMSYMS")) {
                System.err.println("Unrecognized symbol file " + fileName);
                return;
            }
            numSyms = in.readInt();
            symOffset = in.readInt();
            symSize = in.readInt();
            strOffset = in.readInt();
            strSize = in.readInt();
            long cur = in.getFilePointer();
            in.seek(strOffset);
            byte strings[] = new byte[strSize];
            in.read(strings, 0, strSize);
            in.seek(cur);
            for (int i = 0; i < numSyms; i++ ) {
                int typeNum = in.readInt();
                int baseType = in.readInt();
                int size = in.readInt();
                int nameOffset = in.readInt();
                int descriptionOffset = in.readInt();
                String name, description;
                if (descriptionOffset == 0) {
                    if (typeNum < 0)
                        description = "";
                    else
                        continue;
                } else
                    description = getString(strings, descriptionOffset, encoding);
                name = getString(strings, nameOffset, encoding);
                if (nameOffset == 0)
                    name = "";
                CType type = CType.create(typeNum, baseType, size, name, description);
                
            }

        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
