/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/Dumpviewer.java, ras, dumpviewer 1.18
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date  Origin  Description
 * ------  ----  ------   ----------------------------------------------------
 * 36524  020502 PHILR:   New for feature 36524
 * 36524  130502 philr    Add in DvConsole hookup and use of
 *                        command plugins and commandPluginResponse
 * 36524  200502 philr    Now implements observer so that it can look
 *                        for responses from launched CommandPlugin methods
 * 36524  300502 philr    Tidy up for coding standards
 * 36524  140602 philr    try catch around readBytes
 * 36524  270602 philr    memoryscroller upgrade (CMVC GLITCH?)
 * 36524  180702 mpartrid Added findInMemory()
 * 36524  190702 mpartrid Changed call to displayMemory().
 * 36524  220702 mpartrid Added file history.
 *
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.awt.*;
import java.awt.event.*;
import java.lang.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.tree.*;

import javax.swing.event.*;
import java.util.*;
import java.lang.reflect.Method;
import java.io.*;



/**
 *  Description: Whats the purpose of this class
 *
 *
 *
 *
 *
 */
public class Dumpviewer extends JFrame implements Observer, DvFileHistory.IFileHistory {

    static boolean waitingOnConsoleCommand=false;
    static Object  savedConsoleOutput=null;

    static boolean Trace_is_on = true;
    static Dvifm theIFM;
    static JDesktopPane desktop;
    static JCheckBoxMenuItem mbiTc,wm;
    static DVExitListener exitListener;
    static JInternalFrame wm_frame;
    static JInternalFrame tc_frame;
    static String theFileName;
    static JMenuBar mainmenubar;
    static JMenu mbFile,mbMemory,mbWindows,mbEdit,mbSubmenu;
    static JLabel status;   // part of Status line on bottom of screen
    static JInternalFrame mMapDisplayed = null;
    static int desktopWidth;
    static boolean is32BitDump = false;
    static boolean is64BitDump = false;
    static boolean isBigEndDump = false;
    static boolean isLittleEndDump = false;
    static int currentAS = 0;

    static DvConsole dvConsole;


    String[] dumpPlugins;


    static DvDump theDump;

    static boolean file_chosen = false; // gets set to true when user has
                                        // chosen a file

    static DvMemoryFind dvMemoryFind;   // Retains search parameters from any previous search. 
    private DvFileHistory fileHistory;  // File History object.


    /**
     *   <p>Description:
     *   </p>
     */
    public   Dumpviewer()   {

        super("Dumpviewer");

        // First of all lets get the property file going
        String dp = DvUtils.getDvProperty("Version");
        this.setTitle(dp);

        String conTrace = DvUtils.getDvProperty("ConsoleTrace");
        System.out.println("\n"+conTrace);
        if (null != conTrace && true == conTrace.equals("Y")) {
            DvUtils.enableConsoleTrace(true);
        }
        String earlyTrace =  DvUtils.getDvProperty("EarlyTrace");

        desktop = new JDesktopPane();
        setupDesk(desktop);       //  menu, and decent size/position
        desktop.putClientProperty("JDesktopPane.dragMode","outline");
        setContentPane(desktop);
        desktopWidth = getWidth();

        fileHistory = new DvFileHistory(this); // init FileHistory
        fileHistory.initFileMenuHistory();

        exitListener = new DVExitListener(fileHistory);
        addWindowListener(exitListener);

        // we use DvConsole support to get the output for many
        // pieces that don't require special handling, we do this
        // by creating a new DvConsole and then directing commands to it
        // using the execValidCommand method
        dvConsole = new DvConsole(true,this); // true means we are dumpviewer

        // Lets start up Internal Frame Manager
        theIFM =  new Dvifm(64); // we now have potentially 64 internal windows.

        // Construct a dvMemoryFind object, this will retain the parameters from
        // any previous search memory search.

    dvMemoryFind = new DvMemoryFind();

        // We should have some command plugins available from instantiation of
        // DvConsole, so we use these to set the Plugins menu item up....
        theIFM.addPluginMenuItem(dvConsole);

        DvUtils.setTheIFM(theIFM); // tell DvUtil where we are

        if (null != earlyTrace && true == earlyTrace.equals("Y")) {
            mbiTc.setSelected(true); //toggle the check button status.
                                  // and let the listener take the strain
        }

        // lets add a status line (internal frame!)

        addStatusLine();

        status.setText("Awaiting user input....");


       // After this point the work is being driven primarily
       // from assorted listeners assocuiated with the menu bar
       // and all the internal frames.
       //
       // There is a GenericMMBActionListener that is listening
       // for actions associated with the menu bar.
       //
       // There is a  CheckBoxListener watching for Window Manager and
       // Trace Console being toggled....
       //
       // There is an ExitListener watching for "quit" of DumpViewer
       //
       // There is a WMMouseListener that looks primarily for double
       // clicks on Windows in the WM so that you can then jump to them

    }

    /**
     *   <p>Description:
     *   </p>
     */
    private void addStatusLine()  {
        JInternalFrame sl = theIFM.createNewFrame("Status Line",0,
                             DvConstants.FRAMETYPE_IS_STATUSLINE) ;





        // Not happy with the numbers below!
        // May need some smarts to do a status line this way!
        addIFrameToDesktop(sl,
                          30,getWidth()-10,
                          0,getHeight()-80);


        status = new JLabel("Initializing....");
        sl.setBackground(Color.lightGray);
        sl.setBorder( BorderFactory.createLoweredBevelBorder());
        sl.getContentPane().add("West",status);
    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static void main(String[] args) {


        Dumpviewer dv = new Dumpviewer();

        dv.setVisible(true); // This should show it....

    }

    /**
     *   <p>Description:
     *   </p>
     */
    public JInternalFrame  setupASIDframe() {
        // So lets set up the first frame..... the asid/process/#threads frame
        // Coupled with general information
        // Note that this one is NON CLOSEABLE

        JInternalFrame jif  = theIFM.createNewFrame("Asid-Process-Threads",
                                       theIFM.getDefaultProperties()-
                                       DvConstants.FRAME_IS_CLOSEABLE,
                                       DvConstants.FRAMETYPE_IS_NORMAL );

        // Not happy with the numbers below!
        // May need some smarts to do a status line this way!
        addIFrameToDesktop(jif,
                          120,desktopWidth-10,
                          0,0);
        //addIFrameToDesktop(theIFM.getJInternalFrame(i));

        // Now lets create a table and fill it with asid and process
        JTable table = fillASIDtable(jif);

        // Now we issue a console type command and get system info
        // in order to fill the text area. Note that the data
        // comes back to our update method as we are an Observer of
        // the command response object in the plugin command!

        dvConsole.execValidCommand("DIS SYS" );

        // we need to wait until the info is delivered....
        waitingOnConsoleCommand = true;
        while (true == waitingOnConsoleCommand) {

            try {
                Thread.sleep(100);
            }
            catch  (java.lang.InterruptedException e) {
                System.err.println("\nInterruptedException");
            }
        }

        Vector v = (Vector) savedConsoleOutput;
        String sysInfoAsString="";
        for (int i=0;i<v.size();i++) {
            sysInfoAsString+=(v.get(i)+"\n");
        }
        theIFM.addContentToFrame(jif,table,sysInfoAsString);

        JMenuItem pmenuItem = new JMenuItem("Test");
        theIFM.addPopupMenuItemToFrame(jif,pmenuItem);
        JMenuItem pmenuItem1 = new JMenuItem("Test1");
        theIFM.addPopupMenuItemToFrame(jif,pmenuItem1);
        return jif;
    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static void setupWMFrame() {
        // This creates and fills a Window Manager Frame
        wm_frame  = theIFM.createNewFrame("Window Manager",
                                       theIFM.getDefaultProperties()-
                                       DvConstants.FRAME_IS_CLOSEABLE,
                                       DvConstants.FRAMETYPE_IS_WM);
        //System.out.println("\nframe is "+wm_id);
        addIFrameToDesktop(wm_frame);
        // now we need to create the jtree and add this to the
        // frame and find out what window hierarchy we have
        JTree tree = fillWMTree(wm_frame);
        theIFM.addContentToFrame(wm_frame,tree);

    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static void setupTraceFrame() {
        // This creates and fills a Trace Frame
        tc_frame  = theIFM.createNewFrame("Trace Console",
                                       theIFM.getDefaultProperties()-
                                       DvConstants.FRAME_IS_CLOSEABLE,
                                       DvConstants.FRAMETYPE_IS_TRACE);


        addIFrameToDesktop(tc_frame);
        JTextArea jtext =  new JTextArea("",8,25);
        jtext.setEditable(false);   // This window is just for info.

        String ss = jtext.getText();


        theIFM.addContentToFrame(tc_frame,jtext);
        Calendar cal = Calendar.getInstance();
        DvUtils.setTc_text(jtext);
        DvUtils.enableConsoleTrace(true);
        DvUtils.writetoTrace("Trace Console started at " +
                                 cal.getTime().toString());



    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static JTree  fillWMTree(JInternalFrame iframe) {

        // Create the nodes
        DefaultMutableTreeNode top = new DefaultMutableTreeNode("Windows");

        boolean finished=false;
        int where=-1;
        int count=0;

        Enumeration enum = theIFM.getIFrameElements();


        while (enum.hasMoreElements()) {

            DefaultMutableTreeNode next =
              new DefaultMutableTreeNode(((DvifmFrame) enum.nextElement()).iframeTitle);

            top.insert(next,count);
            count++;

        }


        JTree tree = new JTree(top);
        tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        Icon nullIcon = null;
        DefaultTreeCellRenderer renderer =  new DefaultTreeCellRenderer();
        renderer.setLeafIcon(nullIcon);
        renderer.setClosedIcon(nullIcon);
        renderer.setOpenIcon(nullIcon);

        tree.setCellRenderer(renderer);

        // Now shove in a tree selection listener against this tree....
        tree.addTreeSelectionListener(new TreeSelectionListener() {
            public void valueChanged(TreeSelectionEvent e) {}
        });

        // Now shove in a mouse event listener against this tree....
        WMMouseListener gml = new WMMouseListener();
        gml.whatIframe = iframe;
        tree.addMouseListener(gml);

        return tree;
    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static JTable fillASIDtable(JInternalFrame mf ) {

        String[] sColumnHeadings= {
            "ASID",
            "Process",
            "# Threads"
        };
        JTable table=null;

        // Now figure out how many rows we want.......
        // Ask for address space count, n will then be use in the
        // dataTableModel setup
        int rows_required = 0;

        DvAddressSpace[] as_array = theDump.getAddressSpaces();

        if (null == as_array[0]) {
            System.err.println("\nAs_array[0] is null!!!!");
        }

        as_array[0].getMemRanges();   // just for now

        int nAsid = as_array.length; // figure out number of elements

        DvProcess[] pr_array=null;

        for (int x=0;x<nAsid;x++) {
            pr_array = as_array[x].getProcesses();
            rows_required = rows_required + pr_array.length;

        }

        System.out.println("\nRows is "+rows_required);

        long[] asid_array = new long[rows_required];
        long[] pid_array  = new long[rows_required];
        long[] threadCount_array = new long[rows_required];



        for (long i=0;i<nAsid;i++) {
            long asid = 0;
            asid = Long.valueOf(as_array[(int) i].id()).longValue();

            pr_array = as_array[(int)i].getProcesses();
            long m = pr_array.length;
            for (long j=0;j<m;j++) {

                asid_array[(int)(i+j)] = asid;
                pid_array[(int)(i+j)] =
                Long.valueOf(pr_array[(int) j].id()).longValue();
                int numThreads = (pr_array[(int) j].getThreads()).length;

                threadCount_array[(int)(i+j)] = numThreads;
            }


        }

        AsidTableModel dataModel =
            new AsidTableModel(asid_array,pid_array,
                               threadCount_array,
                               rows_required,sColumnHeadings);

        table = new  JTable(dataModel);
        table.setAutoResizeMode(table.AUTO_RESIZE_OFF);

        return table;

    }



    /**
     *   <p>Description:
     *   </p>
     */
    public static void addIFrameToDesktop(JInternalFrame f) {

        addIFrameToDesktop(f,300,300,10,10);

    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static void addIFrameToDesktop(JInternalFrame f,
                                          int height,
                                          int width,
                                          int xpos,
                                          int ypos) {

        f.setVisible(true);
        f.setLocation(xpos,ypos);
        f.setSize(width, height);
        f.setBackground(Color.white);
        desktop.add(f); // add this internal frame to the desktop
        try {
            f.setSelected(true);
        }
        catch(java.beans.PropertyVetoException e) {
        }

    }


    /**
     *   <p>Description:
     *   </p>
     */
    public  void setupStatusLine(JDesktopPane f ) {
        status = new JLabel("Status Line");
        status.setBorder(
        BorderFactory.createLoweredBevelBorder());
        status.setVisible(true) ;
        f.add(status,BorderLayout.SOUTH);


    }

    /**
     *   <p>Description:
     *   </p>
     */
    public void setupDesk(JDesktopPane f) {


        setupMenuBar(f);    // we should have a menu bar

        // lets make most of the stuff disabled until they choose a file
        mbMemory.setEnabled(false);
        mbWindows.setEnabled(false);
        mbEdit.setEnabled(false);
        //mbSubmenu.setEnabled(false);


        Dimension d = Toolkit.getDefaultToolkit().getScreenSize();

        Dimension d1 = new Dimension((int) (d.getWidth() - 40),
                                     (int) (d.getHeight() - 40));

        setBounds(50,50,((int) (d.getWidth()-100)),((int) (d.getHeight()-100)));


     }

    /**
     *   <p>Description:
     *   </p>
     */
    public   void setupMenuBar(JDesktopPane f) {

        mbFile = new JMenu("File");
        mbFile.setMnemonic(KeyEvent.VK_F);
        mbEdit = new JMenu("Edit");
        mbEdit.setMnemonic(KeyEvent.VK_E);
        mbWindows = new JMenu("Windows");
        mbWindows.setMnemonic(KeyEvent.VK_W);
        mbMemory = new JMenu("Memory");
        mbMemory.setMnemonic(KeyEvent.VK_M);
        // It is intended that
        // submenu takes its context depending on which
        // Internal frame is activated and where in the
        //  internal frame we are.
        // So as we activate internal frames we change the nature of the submenu!
        JMenu mbSubmenu = new JMenu("Submenu");
        // Add Test item to the Sub menu
        JMenuItem testItem = new JMenuItem("Test");
        mbSubmenu.setMenuLocation(200,200);
        mbSubmenu.add (testItem);

        // Add Display Memory Map to Memory menu
        // Add Display Memory to Memory menu
        JMenuItem mmapItem = new JMenuItem("Map");
        JMenuItem memdisplay = new JMenuItem("Display");
        JMenuItem memoryFind = new JMenuItem("Find");

        mbMemory.add ( mmapItem);
        mbMemory.add (memdisplay);
        mbMemory.add(memoryFind);
        
        GenericMMBActionListener mmapAl = new GenericMMBActionListener();
        mmapAl.setOwner("mmap");
        mmapItem.addActionListener(mmapAl);

        memoryFind.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e ){
                findInMemory();
            }
        } );

        // Add Open to the File menu
        JMenuItem openItem = new JMenuItem("Open");
        openItem.setMnemonic(KeyEvent.VK_O);
        openItem.setAccelerator(
           KeyStroke.getKeyStroke(KeyEvent.VK_O,Event.CTRL_MASK));
        GenericMMBActionListener openAl = new GenericMMBActionListener();
        openAl.setOwner("open");
        openItem.addActionListener(openAl);



        mbFile.add (openItem);


        // Add Save WorkSpace to the file menu

        // Add Quit to the File menu
        JMenuItem quitItem = new JMenuItem("Quit");
        quitItem.setMnemonic(KeyEvent.VK_Q);
        quitItem.setAccelerator(
           KeyStroke.getKeyStroke(KeyEvent.VK_Q,Event.CTRL_MASK));
        quitItem.addActionListener(new ActionListener( ) {
            public void actionPerformed(ActionEvent e) {
                exitListener.doTerminationCode();
                //System.exit(0);
            }
        });
        mbFile.add (quitItem);

        // Add window manager to the window menu
        wm = new JCheckBoxMenuItem("Window Manager");
        mbWindows.add(wm);


        // Add trace console to the Window menu
        mbiTc = new JCheckBoxMenuItem("Trace Console");
        mbWindows.add(mbiTc);
        CheckBoxListener traceCheckBoxListener = new CheckBoxListener();
        // Share the checkboxlistener
        mbiTc.addItemListener(traceCheckBoxListener);
        wm.addItemListener(traceCheckBoxListener);

        mainmenubar = new JMenuBar();
        mainmenubar.add(mbFile);
        mainmenubar.add(mbEdit);
        mainmenubar.add(mbWindows);
        mainmenubar.add(mbMemory);
        mainmenubar.add(mbSubmenu);

        setJMenuBar(mainmenubar);
    }

    /**
     *   <p>Description:
     *   </p>
     */
    static class AsidTableModel extends AbstractTableModel {

        boolean valsgiven = false;
        long[] col1_vals;
        long[] col2_vals;
        long[] col3_vals;
        String[] columnNames = {"C1","C2","C3","C4","C5","C6"} ;
        int nRows;
        int nColumns;

        public int getColumnCount() {
            return nColumns;
        }

        public String getColumnName(int col) {
            return columnNames[col];
        }

        public int getRowCount()  {
             return nRows;
        }

        public Object getValueAt(int row,int col) {
            if (valsgiven == false) {
               return new Integer(0);
            }
            else {
                if (col == 0)   {
                    return new Long(col1_vals[row]);
                }
                else {
                    if (col == 1) {
                        return new Long(col2_vals[row]);
                    }
                    else {
                        if (col == 2) {
                            return new Long(col3_vals[row]);
                        }
                    }
                }

            }
            return new Long(0);
        }

        public void setValueAt(Object value,int row,int col) {

        }

        public AsidTableModel(int rows,int columns,String[] sa) {
            nRows = rows;
            nColumns = columns;
            try {
                columnNames[0] = sa[0];
                columnNames[1] = sa[1];
                columnNames[2] = sa[2];
                columnNames[3] = sa[3];
                columnNames[4] = sa[4];
                columnNames[5] = sa[5];
            }
            catch (ArrayIndexOutOfBoundsException e)  {

            }
        }

        public AsidTableModel(long[] col1,
                              long[] col2,
                              long[] col3,
                              int rows,
                              String[] sa) {
            valsgiven = true;
            nRows = rows;
            nColumns = sa.length;
            col1_vals = col1;
            col2_vals = col2;
            col3_vals = col3;


            try {
                columnNames[0] = sa[0];
                columnNames[1] = sa[1];
                columnNames[2] = sa[2];
                columnNames[3] = sa[3];
                columnNames[4] = sa[4];
                columnNames[5] = sa[5];
            }
            catch (ArrayIndexOutOfBoundsException e)  {

            }
        }


    }

    /**
     *   <p>Description:
     *   </p>
     */
    static class MmapTableModel extends AbstractTableModel {


        boolean valsgiven = false;
        String[] col1_vals;
        String[] col2_vals;
        long[] col3_vals;
        String[] columnNames = {"C1","C2","C3","C4","C5","C6"} ;
        int nRows;
        int nColumns;

        public int getColumnCount() {
            return nColumns;
        }

        public String getColumnName(int col) {
            return columnNames[col];
        }

        public int getRowCount() {
            return nRows;
        }

        public Object getValueAt(int row,int col) {
            if (valsgiven == false) {
                return new Integer(0);
            }
            else {
                if (col == 0) {
                     return new String(col1_vals[row]);
                }
                else {
                    if (col == 1) {
                        return new String(col2_vals[row]);
                    }
                    else {
                        if (col == 2) {
                             return new Long(col3_vals[row]);
                        }
                        else {
                            return new Long(999);
                        }
                    }
                }
            }
        }

        public void setValueAt(Object value,int row,int col) {

        }

        public MmapTableModel(int rows,int columns,String[] sa) {
            nRows = rows;
            nColumns = columns;
            try {
                columnNames[0] = sa[0];
                columnNames[1] = sa[1];
                columnNames[2] = sa[2];
                columnNames[3] = sa[3];
                columnNames[4] = sa[4];
                columnNames[5] = sa[5];
            }
            catch (ArrayIndexOutOfBoundsException e)  {

            }
        }



        public MmapTableModel(String[] col1,
                             String[] col2,
                             long[] col3,
                             int rows,
                             String[] sa) {
            valsgiven = true;
            nRows = rows;
            nColumns = sa.length; // number of columns is set by number of headings
            col1_vals = col1;
            col2_vals = col2;
            col3_vals = col3;
            try {
                columnNames[0] = sa[0];
                columnNames[1] = sa[1];
                columnNames[2] = sa[2];
                columnNames[3] = sa[3];
                columnNames[4] = sa[4];
                columnNames[5] = sa[5];
            }
            catch (ArrayIndexOutOfBoundsException e)  {

            }
        }

    }


    /**
     *   <p>Description:
     *   </p>
     */
    static class CheckBoxListener implements ItemListener {

        public void itemStateChanged(ItemEvent e) {
            // aha they are playing with a checkbox
            // lets find out which one.....
            boolean bSelecting = true;
            if (e.getStateChange() == ItemEvent.DESELECTED) {
                bSelecting = false;
            }
            Object source = e.getItemSelectable();

            if (source ==  mbiTc) {
            //System.out.println("\nplaying with tc");
                if (!bSelecting) {
                    // Deactivating the trace window
                    // Get rid of the iframe and close it
                    //removeIframe(traceIframe,true);
                    Calendar cal = Calendar.getInstance();
                    DvUtils.writetoTrace("Trace Console stopped at " +
                                            cal.getTime().toString());
                    theIFM.removeIFrame(tc_frame,true);
                    DvUtils.enableConsoleTrace(false); // disable trace

                }
                else {
                    setupTraceFrame();
                }

            }
            else {
                if (source == wm) {
                    // playing with window manager
                    //System.out.println("\nplaying with wm");
                    if (!bSelecting) {
                        // Deactivating the trace window
                        // Get rid of the iframe and close it
                        //removeIframe(traceIframe,true);
                        theIFM.removeIFrame(wm_frame,true);

                    }
                    else {
                        setupWMFrame();
                    }

                }
                else {
                    // playing with some unknown checkbox
                }
            }
        }

    }

    /**
     *   <p>Description:
     *   </p>
     */
    class GenericMMBActionListener implements ActionListener {
        String whatcmd=null;

        public void setOwner(String cmd) {
            whatcmd = cmd;
        }

        public void actionPerformed(ActionEvent e) {

            DvUtils.writetoTrace("Menu action: "+whatcmd);
            if (true == whatcmd.equals("open")) {
                status.setText("Identifying Dump...");
                boolean b = doOpenFile();
                if (true == b) {
                    String bitness="unknown";
                    String endianess="unknown";
                    String systemType = "unknown";
                    int numProcessors = 0;
                    String subSystemType = "unknown";
                    String processorType = "unknown";
                    // Cor blimey - we actually have something that
                    // looks recognised - so lets put up the ASID frame
                    // and then the user can take it from there......
                    status.setText("Dump format identified as....");
                    JInternalFrame mf = setupASIDframe();  // Lets put up the top frame......

                    // Now we will add the information from sdffhead segment into the
                    // text frame of the dump
                    int arch = theDump.getArchitecture();
                    //arch =10; // 10 is LITTLE_ENDIAN and 64 bit
                    if (DvConstants.ANSWER_UNKNOWN == arch) {
                        status.setText("Unknown bitness");
                    }
                    else {
                        status.setText("arch= "+convertArchitecture(arch));
                        // now fill in the text within the asid/process window...
                        if (true == is32BitDump) {
                            bitness = "32-Bit";
                        }
                        else {
                            if (true == is64BitDump) {
                                bitness = "64-Bit";
                            }
                        }
                        if (true == isBigEndDump) {
                            endianess = "Big Endian";
                        }
                        else {
                            if (true == isLittleEndDump) {
                                endianess = "Little Endian";
                            }
                        }

                //?????theIFM.jtext[mf].setEditable(false); // This window is just for info.
                //?????theIFM.jtext[mf].setText("Bitness      = " +
                //   bitness + "         Endianess         = "+ endianess +
                //   "\nSystem      = " + systemType + "    SubSystem        = " +
                //   subSystemType + "\nProcessor = " + processorType +
                //   "    # of processors = "+ numProcessors);
                // ;

                    }
                }
                else {
                    status.setText("Dump format not recognised");
                }
            }

            if (true == whatcmd.equals("mmap")) {
                status.setText("Displaying Memory Map...");
                displayMemoryMap();
            }


        }


    }

    /**
     *   <p>Description:
     *   </p>
     */
    private static void displayMemoryMap() {

        // We'll only do the memory mmap once and then hide or display it
        // So the close action actualy becomes a hide option unbeknown to
        // user - makes performance good....

        if (null != mMapDisplayed) {
            // Find out where and display it.....
        }
        else
        {
            // Lets create one,load it and display it.....
            // So first get an internal frame
            mMapDisplayed = theIFM.createNewFrame("Memory Mapping",
                                       theIFM.getDefaultProperties(),
                                       DvConstants.FRAMETYPE_IS_MMAP);

            // for the size here 300,250 we should probably replace the
            // width (250) with something greater if its a 64 bit system
            // should probably run off defaults in properties table and attach
            // this window to edge of frame
            addIFrameToDesktop(mMapDisplayed,
                               300,250,100,100 );
        }
        // The memory map best suits a table
        // First of all we need to obtain the MemRanges object
        // lets assume only 1 process/asid for now.

        DvMemRanges[] mr =(theDump.getAddressSpaces())[0].getMemRanges();
        //System.out.println("\nAt point B "+mr.length+ " " + mr[10].length());
        //Now we set up a table with mr.length rows and 3 columns
        // start address, end address and length


        String[] sColumnHeadings= {
            "Start Address",
            "End Address  ",
            "Length   "
        };
        JTable table=null;

        // Now figure out how many rows we want.......
        int rows_required = mr.length;

        String[] sa_array = new String[rows_required];
        String[] ea_array = new String[rows_required];
        long[] len_array  = new long[rows_required];

        SDFFAddress ea = new SDFFAddress();

        for (int i=0;i<rows_required;i++) {
            sa_array[i] = mr[i].addressStart().convertToHex();
            len_array[i] =mr[i].length();
            ea.setAddress((mr[i].addressStart().getAddressAsLong())+len_array[i]);
            ea_array[i] = ea.convertToHex();

        }

        MmapTableModel dataModel = new MmapTableModel(  sa_array,
                                                        ea_array,
                                                        len_array,
                                                        rows_required,
                                                        sColumnHeadings);


        table = new  JTable(dataModel);
        table.setAutoResizeMode(table.AUTO_RESIZE_OFF);

        theIFM.addContentToFrame(mMapDisplayed,table);

        MmapMouseListener mml = new MmapMouseListener();
        mml.setOwner(mMapDisplayed);
        table.addMouseListener(mml);

    }

    /**
     *   <p>Description:
     *   </p>
     */
    private boolean doOpenFile() {

        boolean bret=false;
        DvUtils.writetoTrace("Dumpviewer: doOpenFile entry");

        JFileChooser chooser =
            new JFileChooser(System.getProperty("user.dir"));
        chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);

        chooser.setFileFilter(new javax.swing.filechooser.FileFilter() {

            public boolean accept(File file) {
                return true;
            }

            public String getDescription() {
                return "*.*";
            }
        });

        int result = chooser.showOpenDialog(desktop);
        if (result == JFileChooser.APPROVE_OPTION) {
            theFileName = chooser.getSelectedFile().getPath();

            // having accepted and validated
            // we can now enable all those disabled menus
            theDump =   DvUtils.findDumpPlugin(theFileName);
            if (null != theDump) {

                // and importantly tell the console what it is!!!!!
                dvConsole.setTheDump(theDump,null,theFileName );
                fileHistory.insertPathname(theFileName); // hook into FileHistory class

                mbMemory.setEnabled(true);
                mbWindows.setEnabled(true);
                mbEdit.setEnabled(true);
                file_chosen = true;
                bret = true;
            }
        }

        DvUtils.writetoTrace("Dumpviewer: doOpenFile exit - " +
                              (bret?"true":"false"));
        return bret;
    }

    /**
     *   <p>Description:
     *   </p>
     */
    static class WMMouseListener implements MouseListener {

        private JInternalFrame whatIframe; // who owns me!


        public void setOwner(JInternalFrame i) {
            whatIframe = i;
        }

        public void mouseClicked(MouseEvent e) {
            // If it was the right mouse button and there is a selection active on
            // the component then we display the sub menu and then wait for either
            // a mouse click off or on the submenu. If its on then we activate the
            // chosen action...... if its off then we remove the submenu.
            // If it was a left mouse button then we check to see if theres another
            // window to launch - depends on the type of display
            //System.out.println("\nwow we have detected a mouse click in the Window Manager");
            //System.out.println("\n and i is "+whatIframe);

            int x= e.getClickCount();
            DvUtils.writetoTrace("Mouse click in the Window Manager - count "+x);

            if (x>1) {
               // A double mouse click lets find what window is selected
               // and make it the front one!
               // Can do this by getting title and searching for id
                DefaultMutableTreeNode node =
                    (DefaultMutableTreeNode)
                     (theIFM.findTree(whatIframe)).getLastSelectedPathComponent();
                if (node != null) {
                    Object nodeInfo = node.getUserObject();
                    DvifmFrame frame = theIFM.findDvifmFrame((String)  nodeInfo);

                    theIFM.setIFrameSelected(frame.iframe);

               }

            }
        }

        public void mousePressed(MouseEvent e) {
        }

        public void mouseReleased(MouseEvent e) {
        }

        public void mouseEntered(MouseEvent e) {
        }

        public void mouseExited(MouseEvent e) {
        }
  }

    /**
     *   <p>Description:
     *   </p>
     */
    static class MmapMouseListener implements MouseListener {

        private JInternalFrame whatIframe; // who owns me!

        public void setOwner(JInternalFrame jif) {
            whatIframe = jif;
        }

        public void mouseClicked(MouseEvent e) {


            int x= e.getClickCount();
            DvUtils.writetoTrace("Mouse click in the Memory Manager - count "+x);

            if ((e.getModifiers() & InputEvent.BUTTON3_MASK) ==
                     InputEvent.BUTTON3_MASK)
            {
                DvUtils.writetoTrace("its the right mouse button");
            }
            else
            {
                if (x>1) {
                    // its a double click - lets go do the biz......
                    // here we determine what row is selected and
                    JTable table = theIFM.findTable(whatIframe);

                    if (null != table) {
                        int row =  table.getSelectedRow();

                        if (-1 != row) {
                            DvAddressSpace[] as_array = theDump.getAddressSpaces();
                            DvAddress addr =
                             (as_array[currentAS].getMemRanges())[row].addressStart();
                            long len = (as_array[currentAS].getMemRanges())[row].length();

                            // Its at this point that we may want to do some fiddling with
                            // len so that we can do some clever cacheing stuff......
                                                        // we are in business can now a launch a
                            // dumpviewer window based
                            // on this array data......
                            displayMemory(addr,len);
                        }
                    }
                }
            } // end of else
        }

        public void mousePressed(MouseEvent e) {
        }

        public void mouseReleased(MouseEvent e) {
        }

        public void mouseEntered(MouseEvent e) {
        }

        public void mouseExited(MouseEvent e) {
        }
    }

    /**
     *   <p>Description:
     *   </p>
     */
    public static String convertArchitecture(int arch) {

        int[] pow2={1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384};
        int work = arch;
        boolean found1=false;
        String wReturn = "";

        if (work < 2*16384) {
            for (int i=pow2.length-1; i>-1; i--) {
                if (work>=pow2[i]) {
                    try {
                        wReturn = wReturn+DvConstants.ARCH_DESCRIPTIONS[i];
                        // Set globals so we never need ask again
                        found1=true;
                        if (pow2[i] == DvConstants.ARCH_IS_32BIT)
                             is32BitDump = true;
                        if (pow2[i] == DvConstants.ARCH_IS_64BIT)
                             is64BitDump = true;
                        if (pow2[i] == DvConstants.ARCH_IS_BIGENDIAN)
                             isBigEndDump = true;
                        if (pow2[i] == DvConstants.ARCH_IS_LITTLEENDIAN)
                             isLittleEndDump = true;
                    }
                    catch (ArrayIndexOutOfBoundsException e) {
                    }
                    work = work - pow2[i];
                }

            } // end of for loop
        }

        if (true == found1) return wReturn;
        else return "Architecture unknown";
    }

    /**
     *   <p>Description: Creates an Internal frame and then creates a memory
     *                   scroller object.
     *   </p>
     */

    public static void displayMemory(DvAddress addr,long len) {

        JInternalFrame jif = theIFM.createNewFrame("Memory Scroller "+addr.convertToHex(),
                                       theIFM.getDefaultProperties(),
                                       DvConstants.FRAMETYPE_IS_MS);
        addIFrameToDesktop(jif,270,500,250,50);

        DvUtils.setTheIFM(theIFM);

        DvMemoryScroller MemoryScroller = new DvMemoryScroller( jif, dvConsole, addr, len );
    }

    /**
     *   <p>Description: Creates an Internal frame and then creates a find in
     *                   memory object.
     *   </p>
     */

    public static void findInMemory() {

        JInternalFrame findJif = theIFM.createNewFrame("Memory Find",
                                       theIFM.getDefaultProperties(),
                                       DvConstants.FRAMETYPE_IS_MS);
        addIFrameToDesktop(findJif,270,320,350,200);

        DvUtils.setTheIFM(theIFM);

        DvFindInMemory finder = new DvFindInMemory( findJif, dvConsole, dvMemoryFind ); 
    }

    /**
     *   <p>Description: This is where we see returned output from
     *                   launched command plugins
     *   </p>
     */
    public void update(Observable o,Object arg) {

        // now we consume the data from the CommandPlugin....
        // it might be that we have some more data later
        synchronized (o) {

            waitingOnConsoleCommand = false;
            savedConsoleOutput = arg;
        }


    }

    // --- Implementation of dvFileHistory.IFileHistory interface ----------------

    /**
     *   <p>Description:    Get the application name to identify the configuration
     *                      file in the the USER_HOME directory. This name should
     *                      be unique in this directory.
     *   </p>
     *   <p>Input      :    None.
     *   </p>
     *   <p>Returned   :    The application name.
     *   </p>
     */

    public String getApplicationName() {
        return "DUMPVIEWER";
    }

    /**
     *   <p>Description:    Get a handle to the frame's file menu.
     *   </p>
     *   <p>Input      :    None.
     *   </p>
     *   <p>Returned   :    The frame's file menu   
     *   </p>
     */

    public JMenu getFileMenu() {
        return mainmenubar.getMenu(0);
    }

    /**
     *   <p>Description:    Simulate a load file activity. 
     *   </p>
     *   <p>Input      :    None.
     *   </p>
     *   <p>Returned   :    the pathname of the loaded file 
     *   </p>
     */

    public void loadFile(String theFileName) {

        // having accepted and validated
        // we can now enable all those disabled menus

        theDump =   DvUtils.findDumpPlugin(theFileName);

        if (null != theDump) {

            // and importantly tell the console what it is!!!!!

            dvConsole.setTheDump(theDump, null, theFileName);

            mbMemory.setEnabled(true);
            mbWindows.setEnabled(true);
            mbEdit.setEnabled(true);
            file_chosen = true;
            DvUtils.writetoTrace("Dumpviewer: loadFile success - " + theFileName );
        }
        else {
            DvUtils.writetoTrace("Dumpviewer: loadFile failure - " + theFileName );
        }
    }
}
