/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvAddressIndexFile.java, ras, dumpviewer 1.4
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  190702  philr   New for feature 36524
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;
import java.io.*;
import java.util.*;


 /**
  *  This file supports a database of object pointers, linked to
  *  reference (to and from) his object.
  */
public class DvAddressIndexFile extends RandomAccessFile {

    private static String file_name;
    private static IndexHeader IH;
    private static RandomAccessFile self;
    private static long [] indexArray;
    private static long next_record_pos=0;
    private static int index_table_size=0;
    private long saved_prev_position=0;



    /**
     *  Constructor
     */
    public DvAddressIndexFile (String file_name,
                        String mode,
                        int index_table_size,
                        int record_size,
                        int key_size,
                        long magic)
            throws java.io.IOException{

        super(file_name,mode);
        self = this;




        if (mode.equals("r")) {
            // read existing control information from the file
            IH = IndexHeader.readExistingHeader();
            //checkIntegrity();
            if (IH.magic == magic) {
                // System.err.println("Magic numbers check out");
                // Need to rebuild indexArray
                seek(IH.index_ptr);
                indexArray = new long[IH.index_table_size];
                for (int i=0;i<IH.index_table_size;i++) {
                    indexArray[i]=readLong();
                }
            }
            else {
                throw new IOException("Magic numbers mismatch");
            }


        }
        else {

            initTheIndex(index_table_size,
                         record_size, key_size,
                         magic);
            //this.index_table_size = index_table_size;
        }


    } //end Constructor


    /**
     *  Write a record to this file
     */
    public boolean write(long key,short heapId) {

        // we know we need a number between 0 and index_table_size-2
        // so we get this by taking the key byte array and summing the
        // individual bytes and then getting remainder when we divide
        // by (index_table_size-1) - we further "complicate it" (in order
        // to give bigger numbers by multiplying the first four bytes
        // in the key together (unless the value is 0). We reserve
        // array position (index_table_size-1) for those records where
        // the key is such that its size is small!!!
        long key_total=hashTheKey(key);
        int  array_index=0;

        //System.out.println("key_total is :"+key_total);
        // System.out.println("index_table_size"+IH.index_table_size);
        array_index = (int) (key_total%IH.index_table_size);
        //if (array_index < 0) {
           // System.out.println("array index is "+array_index);
        //}
        // Now convert  this into an offset into the index

        long pos = IH.index_ptr+8*array_index;
        //if (pos < 0) {
            //System.out.println("pos is "+pos);
        //}
        try {
            seek(pos);
            long x = readLong();
            //if (x < -1) {
            //    System.out.println("x is "+x +"@"+pos+":"+IH.index_ptr+":"+array_index);
            //}


            if (-1 == x) {
                // this is a new one - Don't forget to copy it to indexArray in memory
                //indexArray[array_index] =

                long whereItwent = writeNewKeyRecord(pos,key,x);
                seek(pos);
                writeLong(whereItwent);
                // Don't forget to copy it to indexArray in memory

            }
            else {
                long whereItwent=writeNewKeyRecord(pos,key,x);
                seek(pos);
                writeLong(whereItwent);
            }


        }
        catch (IOException e) {
            //String s = new String(key);
            System.out.println("IOException "+pos+" "+key);
        }

        //String s = new String(key);
        // System.out.println("Index is :"+array_index+ " "+s);

        return true;
    }

    /**
     *  write a new key record
     */
    private long writeNewKeyRecord(long pos,long key,long keypos) {
        // keypos is the contents of the  index record -
        // if it is -1 then we are starting a new chain and need to
        // update the index record - if it is not -1 then we
        // will be using the pos to update the next in the previous
        // record
        //System.err.println("wnr:"+pos+":"+key+":"+keypos);
        long retVal=0;
        try {
            seek(length()); // go to end of file
            long saved_pos = getFilePointer();
            writeByte((byte)'K');
            writeByte((byte)':');

            writeLong(key); //The key (the object address)
            if (-1 != keypos) {
                // This is for chaining, basically if this is not the
                // first record in the chain we update the previous one with
                // this records offset in the file
                writeLong(keypos);
            }
            else {
                writeLong(0); // previous one with this index in array index
            }

            writeLong(0); // pointer to Rto: data record associated with this key

            writeLong(0); // lastRto offset will be stored here
            writeLong(0); // pointer to Rfr: data record associated with this key

            writeLong(0); // lastRfr offset will be stored here

            retVal = saved_pos;  // caller will update index

       }
       catch (IOException e) {
        System.err.println("IOException in writeNewKeyRecord");
       }

       return retVal;
    }

    /**
     *  write a new data record
     */
    private long writeNewDataRecord(long instanceAddr,long lastDRAptr,
                                    long chainPtrAddress) {

        long retLong = 0;
        try {
            seek(length()); // go to end of file
            retLong = getFilePointer();
            writeByte((byte)'O');
            writeByte((byte)'b');
            writeByte((byte)'j');
            writeByte((byte)':');
            writeInt(1); // we will fill 0'th one here
            writeLong(instanceAddr);
            for (int i=0;i<8;i++) {
                writeLong(0);
            }
            seek(lastDRAptr);
            writeLong(retLong);
            if (0 != chainPtrAddress) {
                // if its not the first one then we are chaining
                seek(chainPtrAddress);
                writeLong(retLong);
            }
         }
         catch (IOException e) {
            System.err.println("IOException in writeNewDataRecord");
         }
         return retLong;
    }

    /**
     *  Initialise the index file
     */
    private static boolean initTheIndex (int index_table_size,
                                        int record_size,int key_size,
                                        long magic)
                                throws IOException  {

        // Need to create a new IndexHeader;
        IH = new IndexHeader(index_table_size,
                             record_size,
                             magic);
        indexArray = new long[index_table_size]; // keep it in core to avoid
                                                 // excessive IO
        for (int i=0;i<index_table_size;i++) {
            indexArray[i] = -1; // -1 means no record.
        }
        return true;
    }

    /**
     *  Get a long hash value from a key using our own algorithm
     */
    private long hashTheKey(long key) {

        long retLong=0;

        retLong = key >>> 3;
        retLong = retLong%IH.index_table_size;


        return retLong;
     }

    /**
     *  Internal class dealing with the index header
     */
    private static class IndexHeader  {

        static long magic;
        static int  index_table_size;
        static int  record_size;
        static int  key_size;
        static long index_ptr;

        private IndexHeader() {
        }

        private IndexHeader(int index_table_size,
                            int record_size,
                            long magic) throws IOException  {

            this.magic = magic;
            // we make index table size an interation of 256
            //System.out.println(" "+index_table_size);
            this.index_table_size = (index_table_size/256)*256;
            //System.out.println(" "+index_table_size);
            this.record_size = record_size;
            this.key_size = key_size;
            writeIH();
        }

        private  int loadIH() {

            return 0;
        }

        private static IndexHeader readExistingHeader() throws IOException {

            IndexHeader IH = new IndexHeader();
            self.seek(0); // go to start of file
            magic = self.readLong();
            index_table_size = self.readInt();
            //System.out.println("At point B "+index_table_size);
            record_size = self.readInt();
            key_size = self.readInt();
            self.readInt();
            index_ptr = self.readLong();
            //System.out.println("index_ptr: " + index_ptr);
            next_record_pos = self.length();

            return IH;
        }

        private boolean writeIH() throws IOException {
            self.writeLong(magic);
            self.writeInt(index_table_size);
            System.out.println("index_table_size:" + index_table_size);
            self.writeInt(record_size);
            self.writeInt(key_size);


            self.writeInt(0); // for extension purposes
            long ptr_to_index_ptr=self.getFilePointer();
            self.writeLong(0); // index_ptr goes here

            self.writeLong(0); // start of data records goes here
            self.writeLong(0); // for extension purposes
            self.writeLong(0); // for extension purposes
            self.writeLong(0); // for extension purposes

            for (int i=0;i<4;i++) {
                self.writeByte((byte)'i');
                self.writeByte((byte)'n');
                self.writeByte((byte)'d');
                self.writeByte((byte)'x');
            }
            index_ptr = self.getFilePointer();
            // Now we write a whole series of longs corresponding to the
            //index_table_size - all 0's
            for (int i=0;i<index_table_size;i++) {
                self.writeLong(-1); // for extension purposes

            }
            for (int i=0;i<4;i++) {
                self.writeByte((byte)'d');
                self.writeByte((byte)'a');
                self.writeByte((byte)'t');
                self.writeByte((byte)'a');
            }
            long start_of_data_records = self.getFilePointer();
            self.seek(ptr_to_index_ptr);
            self.writeLong(index_ptr); // index_ptr goes here
            self.writeLong(start_of_data_records);
            next_record_pos = start_of_data_records;



            return true;
        }
    }


    /**
     *  Internal class dealing with the data records
     */
    private static  class dataRecord {


           byte[] eye = new byte[2]; // should be K:
           long key;
           long prevPtr;
           long rtoPtr;
           long lastRtoDRA;
           long rfrPtr;
           long lastRfroDRA;





           private dataRecord(long where)   {

                try {
                    self.seek(where);
                    self.read(eye);

                    key =  self.readLong();
                    prevPtr = self.readLong();
                    rtoPtr = self.readLong();
                    lastRtoDRA  = self.readLong();
                    rfrPtr = self.readLong();
                    lastRfroDRA  = self.readLong();
                    //name = new byte[length];
                    //self.read(name);

                }
                catch (IOException e) {
                    System.err.println("IOexception in dataRecord constructor");
                }

            }
     }



    /**
     *
     */
    boolean checkForKnownObject(long theAddress) {


        // So we go through the address index file chains looking for one that matches

        long key_total=hashTheKey(theAddress);

        int array_index = (int) (key_total%IH.index_table_size);


        long pos = IH.index_ptr+8*array_index;
        boolean isFound = false;
        try {
            seek(pos);
            long x = readLong(); // so we read where the last one is and go up through
                                 // the chain
            while (x != 0 && x != -1 && false == isFound ) {
                dataRecord record = new dataRecord(x);
                //System.err.println("theAddress:0x" +
                //                    Long.toHexString(theAddress) +
                //                    "   record.key:0x" +
                //                    Long.toHexString(record.key) +
                //                    "   x:"+x);

                if (record.key == theAddress) {
                   isFound = true;
                }
                else {
                    x = record.prevPtr;
                }

            }
        }
        catch (IOException e) {
            System.err.println("IOError in checkForKnownObject");
        }

        return isFound;
    }

    /**
     *
     */
    public int sizeOfIndex() {

        return IH.index_table_size;

    }

    public Vector getAllObjectsAtIndex(int index) {

        Vector v = new Vector();

        // so we go to the index and read all the objects up the chain adding
        // the position adding a Long to the vector

        long pos = IH.index_ptr+8*index;
        byte[] eye = new byte[2]; // should be K:

        try {

            long x = pos; // so we read where the last one is and go up through
                                 // the chain
            seek(x);
            x = readLong(); // get to last one in chain
            while (x != 0 && x != -1) {

                seek(x);
                read(eye);
                x = readLong(); // the key (address of object)
                v.add(new Long(x));
                x = readLong(); // prev pointer

             }

        }
        catch (IOException e) {
            System.err.println("IOError in getAllObjectsAtIndex");
        }

        return v;
    }

}


