/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvBaseFmtCmds.java, ras, dumpviewer 1.19
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */
/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502  PHILR:  New for feature 36524
 * 36524  270502  philr   Plugin methods need to be non static.
 * 36524  270502  philr   Extend format command to take a hint as input as
 *                        well as an address.
 * 36524  260602  philr   Support for reset of command plugins plus tidyup
 * 36524  110702  philr   Add in trace points to aid debugging
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.util.*;
import java.io.*;

 /**
  *  Description: This class handles the Format commands supplied
  *               as part of the base.
  *
  *
  *
  *
  */
public class DvBaseFmtCmds extends CommandPlugin  {

    private static String[] commandsSupported =
            {  "FORMAT,FOR, , ,showFormatInfo",
               "FORMAT,FOR,execenv,ee,formatEE",
               "FORMAT,FOR,jvm,jvm,formatJVM",
               "FORMAT,FOR,stglobal,stg,formatSTGlobal",
               "FORMAT,FOR,*,*,doFormat",
               "DISPLAY,DIS,hints,hints,displayHints",
               "HELP,?, , ,displayHelp",
               "HELP,?,*, ,displayHelp"

              };



    /**
     *   <p>Description:
     *   </p>
     */
    public  String  pluginName() {
        return "Base format commands (DvBaseFmtCmds)";
    }



    /**
     *   <p>Description:
     *   </p>
     */
    public String[] getSyntax() {
        return commandsSupported;
    }

    /**
     *   Static method for the use of the Help command as
     *   implemented in DvBaseCommands
     */
    public static Object getGenericHelp() {
        //DvUtils.trace(" Entry DvClassCommands:getGenericHelp",0,false);
        Vector v = DvUtils.getMultiplePropertyValues("GeneralHelp",
                        "DvBaseFmtCmds.properties");

        return v;

    }


    /**
     *  Generic help for the commands supported in this plugin
     *
     */
    public String   help() {return null;}

    /**
     *   <p>Description:
     *   </p>
     */
    public String   help(String what) {return null;}

    /**
     *   <p>Description:
     *   </p>
     */
    public String[] guiMenuSyntax() {return null;}

    /**
     *   <p>Description:
     *   </p>
     */
    public String[] guiPopupSyntax(){return null;}


    // All the CommandPlugin requirements are above
    // Now for the real methods...............


    /**
     *   <p>Description: Display what the formatting base is and
     *                   what sort of areas it can format
     *   </p>
     */
    public void showFormatInfo() {

        Vector output=new Vector();
        String infoFileName = "DvFormat.txt";

        output.add("  Format information");
        output.add("  ==================\n");

        File infoFile = DvUtils.searchForFile(infoFileName);


        if (null == infoFile || false == infoFile.canRead()) {
            output.add("  "+ infoFileName+ " not found!!!!");
        }
        else {

            //we are in business
            // cheat a little assume never going to be a big file!!!!
            try {
                FileInputStream fis = new FileInputStream(infoFile);
                byte[] buffer= new byte[5000];
                fis.read(buffer);
                String s = new String(buffer);
                s = s.trim();
                output.add(s);
            }
            catch (java.io.FileNotFoundException e) {}
            catch (java.io.IOException e) {
                output.add(" IOError for "+infoFileName);
            }

        }
        output.add("\n");


        cpr.output(output);
        forcedEnd = true; // so DvConsole knows we've finished
    }

    /**
     *   <p>Description: The catcher format command - driven by input such as
     *                   FORMAT addr as execenv
     *   </p>
     */
    public void doFormat() {

        Vector output = new Vector();


        if (null != theDump) {
            output.add(doFormatEx(paramString,theDump,verbModifier));

        }
        else {
            output.add(" Unable to comply as no dump accessed yet");
        }



        cpr.output(output);
        forcedEnd = true; // so DvConsole knows we've finished

    }


    /**
     *   <p>Description: Formats an area of memory and returns the result
     *                   - static as it is used from outside this plugin
     *   </p>
     */
    public static Vector doFormatEx(String paramString,DvDump theDump,String verbModifier) {


        DvUtils.trace(" Entry to DvBaseFmtCmds:doFormatEx ("+ verbModifier + " " +
                         paramString + ")",0,false);
        Vector output = new Vector();

        // lets analyze the modifier and the parameters
        if (null == paramString) {
            // its format xxx, so we check that xxx is one of our recognised
            // "standard" control blocks.... such as ?
            // this means we will format those blocks for current a/p/t combo....
            output.add(" paramString is null !!!!");
        }
        else {

            Vector individualParms = DvUtils.splitIntoPieces(paramString," ");
            String p1 = (String) individualParms.get(0); // get first part of params

            // lets see if its format xxx as yyyyy, in which case we
            // assume that xxx is a hex address - if it isn't we have a glitch
            if ((p1.toUpperCase()).equals("AS")) {
                // so lets check modifier is a valid address
                // lets check that we have another parameter

                if (individualParms.size() != 2) {
                    output.add(" Invalid request");
                }
                else {
                    long a = 0;
                    try {
                        String asidVal = DvUtils.getValue("ASID");
                        DvAddressSpace as = DvUtils.findAsid(theDump,asidVal);
                        // we allow the specification of Hints instead of addresses
                        // so lets try and convert it as though it is a hint
                        //String hint = DvUtils.getJVMHint(as,verbModifier);
                        // ---- REMOVED --------------
                        //if (null != hint) {
                        //    verbModifier = hint;
                        //}
                        // we allow 0x form and non0x forms.
                        verbModifier = DvUtils.remove0x(verbModifier);
                        a = DvUtils.hexToLong(verbModifier);
                        // we also allow the specification of Hints instead of addresses
                        // so lets try and convert it as though it is a hint
                        // so now get what we want to format it as
                        String type =  (String) individualParms.get(1);
                        //output.add("type is "+type);
                        // so now we can get address space
                        // so we can create an address object and verify it.
                        if (null != CType.find(type)) {
                            if (null != as) {
                                DvAddress addr = as.createAddress(a);
                                if (DvUtils.checkAddressInDump(addr,as)) {
                                //if (false == DvUtils.isTypeFileLoaded()) loadFormatFile(output);
                                    CTypeObject cTO = DvUtils.formatAddrAsType(as,addr,type);
                                    if (null != cTO) {
                                        output.add(cTO);
                                        output.add("\n "+paramString+ " finished");
                                    }
                                    else {
                                        output.add(" Format failed!!!!");
                                    }
                                }
                                else {
                                    output.add(" Address "+ verbModifier + " not in dump");
                                }
                            }
                            else {
                                output.add(" Unable to find address space: "+asidVal);
                            }
                        }
                        else {
                            output.add(" Unsupported structure \"" + type+
                                        "\" (case is relevent).");
                        }


                    }  // end of try
                    catch (NumberFormatException e) {
                        // Extended to treat an invalid hex as a possible hint and therefor
                        // redo it with the hint substituting for the invalid piece.....
                        output.add(" Invalid hex: "+verbModifier);

                    }
                }
            }

        }

        DvUtils.trace(" Exit from DvBaseFmtCmds:doFormatEx",0,false);
        return output;
    }


    /**
     *   <p>Description: Formats execenv for given thread!
     *                   assumes current asid/process
     *   </p>
     */
    public  void formatEE() {

        Vector output = new Vector();

        if (null != theDump) {
            DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
            DvProcess proc  = (DvProcess) DvUtils.getAPT(theDump,"P");
            DvThread thread = (DvThread) DvUtils.getAPT(theDump,"T");

            String threadVal;

            if (null == paramString) {
                threadVal = DvUtils.getValue("THREAD");
            }
            else {
                threadVal = paramString;
                if (as != null && proc != null) {
                    thread = DvUtils.findThread(theDump,threadVal,as,proc);
                }
            }

            output.add("\n execenv for thread "+threadVal+
                        "\n ==========================\n");

            if (null != as && null != proc && null != thread) {
                DvAddress addr= as.createAddress(thread.eeAddress().getAddressAsLong());
                // Check that the address is actually in memory!!!!!
                if (DvUtils.checkAddressInDump(addr,as)) {
                    //if (false == DvUtils.isTypeFileLoaded()) loadFormatFile(output);

                    CTypeObject cTO = DvUtils.formatAddrAsType(as,addr,"execenv");
                    if (null != cTO) {
                        output.add(cTO);
                        output.add("\n Format of execenv finished");
                    }
                    else {
                        output.add(" Format of execenv failed!!!!");
                    }
                }
                else {
                    output.add(" Address " + thread.eeAddress().convertToHex() +
                             " is not available in the dump" );
                }
            }
            else {
                output.add(" ASID/Process/Thread invalid");
            }

        }
        else {
            output.add(" Unable to comply as no dump accessed yet");
        }
        cpr.output(output);
        forcedEnd = true; // so DvConsole knows we've finished

    } // end of method formatEE()


    /**
     *   Formats jvm for given thread!
     *   assumes current asid/process
     *   goes via ee->jvmP
     *
     */
    public  void formatJVM() {


        DvUtils.trace(" Entry to DvBaseFmtCmds:formatJVM",0,false);
        // so we need to format the execenv first and extract the jvmP!
        Vector output = new Vector();

        if (null != theDump) {
            DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
            DvProcess proc  = (DvProcess) DvUtils.getAPT(theDump,"P");
            DvThread thread = (DvThread) DvUtils.getAPT(theDump,"T");

            String threadVal;

            if (null == paramString) {
                threadVal = DvUtils.getValue("THREAD");
            }
            else {
                threadVal = paramString;
                if (as != null && proc != null) {
                    thread = DvUtils.findThread(theDump,threadVal,as,proc);
                }
            }

            output.add("\n Jvm via thread "+threadVal+
                        "\n ==========================\n");

            if (null != as && null != proc && null != thread) {
                DvAddress ee_addr= as.createAddress(thread.eeAddress().getAddressAsLong());
                // Check that the address is actually in memory!!!!!
                if (DvUtils.checkAddressInDump(ee_addr,as)) {
                    //if (false == DvUtils.isTypeFileLoaded()) loadFormatFile(output);

                    CTypeObject ee_cTO = DvUtils.formatAddrAsType(as,ee_addr,"execenv");

                    if (null != ee_cTO) {
                        // so now we have an execenv - we can use getField on jvmP
                        // to get to the jvm control block
                        CTypeObject jvmP_cTO = ee_cTO.getField("jvmP");
                        if (null != jvmP_cTO) {
                            String s= jvmP_cTO.toString();
                            //System.out.println("jvmP:"+s);
                            if (s.startsWith("null ")) {
                                s = s.substring(5);
                            }
                            if (s.startsWith("(null) ")) {
                                s = s.substring(7);
                            }
                            s = DvUtils.removeAtSign(s);
                            s = DvUtils.remove0x(s);
                            //System.out.println("jvmP:"+s);
                            DvAddress a2 = null;
                            DvAddress a1=
                            as.createAddress(DvUtils.hexToLong(s));
                            try {
                                a2 = as.readPointer(a1);
                            }
                            catch (DvAddressException e) {
                                output.add("\n jvmP addressing failure");
                            }
                            CTypeObject jvm_cTO = null;
                            if (null != a2) {
                                jvm_cTO = DvUtils.formatAddrAsType(as,a2,"Jvm");
                            }

                            if (null != jvm_cTO) {
                                DvUtils.setupJVMHints(jvm_cTO,as);
                                output.add(jvm_cTO);
                                output.add("\n Format of Jvm finished");
                            }
                            else {
                                output.add(" Unable to format Jvm!");
                            }
                        }
                        else {
                            output.add(" Unable to get jvmP!");
                        }
                    }
                    else {
                        output.add(" obtaining execenv failure!!!!");
                    }
                }
                else {
                    output.add(" Address " + thread.eeAddress().convertToHex() +
                         " is not available in the dump" );
                }
            }
            else {
                output.add(" ASID/Process/Thread invalid");
            }

        }
        else {
            output.add(" Unable to comply as no dump accessed yet");
        }
        cpr.output(output);
        forcedEnd = true; // so DvConsole knows we've finished
        DvUtils.trace(" Exit from DvBaseFmtCmds:formatJVM",0,false);

    } // end of method formatJVM()

    /**
     *   <p>Description: Formats STGlobal for current jvm!
     *                   ee->jvmP->Jvm->st:dataP which should
     *                   format nicely as a STGlobal. we cheat on this one
     *                   having allowed the format jvm to set up some
     *                   hints.
     *   </p>
     */
    public  void formatSTGlobal() {

        Vector output = new Vector();

        if (null != theDump) {

            // lets look for the hint

            DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
            if (null != as) {
                String hint = DvUtils.getJVMHint(as,"STGLOBALPTR");
                if (null != hint) {
                    verbModifier = hint;
                    paramString = "as STGlobal";
                    doFormat();
                    return; // do format will have set forced end
                }
                else {
                    output.add(" Unable to comply - no hint found for STGLOBALPTR");
                }
            }
            else {
                output.add(" Unable to comply as no valid address space found");
            }
        }
        else {
            output.add(" Unable to comply as no dump accessed");
        }

        cpr.output(output);
        forcedEnd = true; // so DvConsole knows we've finished

    } // end of method formatSTGlobal()



    /**
     *   <p>Description: Display all the hints stored away for this addressspace
     *
     *   </p>
     */
    public  void displayHints() {

        Vector output = new Vector();

        if (null != theDump) {
            DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

            if (null != as) {
                output.add( "\n Jvm Hints for AddressSpace "+as.id()+
                            "\n ==============================\n");
                output.add(DvUtils.getAllJVMHints(as));
            }
            else {
                output.add("No current addressSpace establishable");
            }

        }
        else {
            output.add(" No dump established yet");
        }
        cpr.output(output);
        forcedEnd = true; // so DvConsole knows we've finished

    } // end of method


    /**
     *   <p>Description: reset saved data when new dump opened
     *
     *   </p>
     */
    public void reset() {

        // Not a lot to do.....

    }

}
