/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvClassCommands.java, ras, dumpviewer 1.26
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  120602  PHILR:  New for feature 36524
 * 36524  260602  philr:  Fix DIS CLASS not ending problem
 * 36524  260602  philr   reset method added
 * 36524  070702  philr   add fld enhancer
 * 36524  090702  philr   fix statics not printing problem )(signature not being passed)
 * 36524  090702  philr   added code to print java\Lang\String fields for objects
 * 36524  180702  philr   amend getInstanceValues such that we look at superclasses
 * 36524  300702  philr   fixed 64 bit bugginess
 * 36524  060802  philr   expanded interpretValue
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.util.*;
import java.io.*;

 /**
  *  Description: Whats the purpose of this class
  *
  *
  *
  *
  *
  */
public class DvClassCommands extends CommandPlugin  {


    private static String prevVerb=null;
    private static String prevModifier=null;
    private static String prevParms=null;
    private static String prevSeperator=null;
    private static int cc1=0,cc2=0,cc3=0,totalClassesCount=0;
    private static Vector theClasses = new Vector();
    private static Hashtable classHashByAddr = new Hashtable();
    private static Hashtable classHashByName = new Hashtable();
    private static Hashtable fieldsHashByName = new Hashtable();
    private static Vector theFields = new Vector();
    private static int methodBlockSize;
    private static int fieldBlockSize;
    private static ClassDetails javaLangString = null;
    private static long stringValueOffset = 0;


    private static final String badString = "Unable to get Instance Values";

    private boolean bSummary;




    //private static Properties myProps=null;
    //private static Enumeration myPropsEnum;
    //private static Properties setValues=null;
    private static String[] commandsSupported =
            {
               "DISPLAY,DIS,CLASSSUMMARY,CLS,displayCS",
               "DISPLAY,DIS,CLASS,CL,displayClasses",
               "DISPLAY,DIS,METHODS,ME,displayMethods"
            };

    /**
     *                   Constructor used to allow Threading on top of
     *                   runnable
     */
    public DvClassCommands() {
        super();
    }

    /**
     *
     */
    public  String  pluginName() {
        return "Class commands (DvClassCommands)";
    }



    /**
     *
     *
     */
    public String[] getSyntax() {

        return commandsSupported;
    }


    /**
     *   Static method for the use of the Help command as
     *   implemented in DvBaseCommands
     */
    public static Object getGenericHelp() {
        //DvUtils.trace(" Entry DvClassCommands:getGenericHelp",0,false);
        Vector v = DvUtils.getMultiplePropertyValues("GeneralHelp",
                        "DvClassCommands.properties");

        return v;

    }


    /**
     *  Generic help for the commands supported in this plugin
     *
     */
    public String   help() {
        //getGenericHelp(null);
        return null;
    }

    /**
     *
     *
     */
    public String   help(String what) {return null;}



    /**
     *
     *
     */
    public String[] guiPopupSyntax(){return null;}


    /**
     *
     *
     */
    public void displayHelp() {


        String key;
        if (null == verbModifier) {
            key="GeneralHelp";
        }
        else {
            key = verbModifier;
        }


        cpr.output(DvUtils.getMultiplePropertyValues(key,"DvBaseCommands.properties"));


        forcedEnd = true; // so DvConsole knows we've finished


    }

    /**
     *                   Display a summary of the classes loaded in memory
     *
     */
    public void displayCS() {

        bSummary = true;

        Vector theOutput = new Vector();

        if (null != theDump) {
            theOutput.add("\n Classes Summary"+
                          "\n ===============\n" );

            // So lets go through the classes,system classes and ACS classes

            getClasses(theOutput);

            theOutput.add(" Number of classes found via loadedClasses       = "+
                         cc1);
            theOutput.add(" Number of classes found via loadedACSClasses    = "+
                         cc2);
            theOutput.add(" Number of classes found via loadedSystemClasses = "+
                         cc3+"\n\n");

            cpr.output(theOutput);
            theOutput.clear();

            displayClasses();
       }
       else {
            //System.out.println("Sure enough theDump was null!!!!");
            theOutput.add("Unable to comply - no dump accessed");
       }
       bSummary = false;
       cpr.output(theOutput);
       forcedEnd = true;
    }

    /**
     *      Display the methods for a class or set of
     *      f classes ... we can then filter them if necessary
     *      paramString to see what level of detail is required.
     *
     *        dis me   (no parameters) just shows help for the dis me
     *                - if you truly want all methods for each class
     *                - then dis me(*) will do it
     *
     */
    public void displayMethods() {

        Vector theOutput = new Vector();
        Vector theFilters = new Vector();
        boolean bWildCards = false;
        boolean bSummary = false;

        if (null != theDump) {
            // first we analyse the parameter list
            System.out.println("paramString:"+paramString);
            if (null != paramString) {
                StringTokenizer st = new StringTokenizer(paramString,",");
                while (st.hasMoreTokens()) {
                  // Note we assume that the filters make sense
                  String token = st.nextToken().trim();
                  if (token.startsWith("*") ||
                      token.endsWith("*") ||
                      token.startsWith("!")) {
                     bWildCards = true;
                  }
                  theFilters.add(token);
                }

                getClasses(theOutput); // make sure classes vector has been
                                       // filled

            // now for each class - see whether its name matches a filter
            // assuming there are any filters
            if (theFilters.size() >0 && false == bWildCards) {
                for (int i=0;i<theFilters.size();i++) {
                    String f = (String) theFilters.get(i);
                    Integer n = (Integer) classHashByName.get(f.toUpperCase());
                    if (null != n) {
                        // we have found it
                        //System.out.println("In here");
                        ClassDetails c = (ClassDetails) theClasses.get(n.intValue());
                        showMethods(c,theOutput,bSummary);
                    }
                }
            }
            else {
                // OK they have wildcards ... got to scan through the lot

                for (int i=0;i<theClasses.size();i++) {
                    // go and find thename of the class
                    ClassDetails c = (ClassDetails) theClasses.get(i);
                    boolean bShowit = true; // assume this one is to be shown
                    if (0 !=theFilters.size()) {
                        // if there are any filters, then lets see if this name passes
                        bShowit = DvUtils.filtrate(c.name,theFilters);
                    }

                    if (true == bShowit) {
                        showMethods(c,theOutput,bSummary);
                    }
                }
            }



            }
            else {
                // we do this to stop dis mel throwing out huge amounts of data
                // if they really want huge amounts of data then dis me(*) does it
                // display help for dis me
                doMethodHelp(theOutput);

            }
        }
        else {
            theOutput.add("Unable to comply - no dump accessed");
        }

        cpr.output(theOutput);
        forcedEnd = true;
    }

    /**
     *                   Display the classes loaded in memory. Evaluates
     *      // of classes ... we can then filter them if necessary
     *        paramString to see what level of detail is required.
     *
     *        dis cl   (no parameters) shows all classes in summary mode !!!!
     *
     *        dis cl(a,b,c,d.......) uses the a,b,c,d as class_specs to
     *                               filter the classes displayed based on
     *                               classname.....
     *
     *
     */
    public void displayClasses() {

        boolean bWildCards = false;
        Vector theOutput = new Vector();

        Vector theFilters = new Vector();
        boolean bDisplayMethods = false;
        boolean bDisplayStatics = false;
        boolean bDisplayFields = false;


        if (null != theDump) {
            // first we analyse the parameter list
            //System.out.println("paramString:"+paramString);
            if (null != paramString) {
                StringTokenizer st = new StringTokenizer(paramString,",");
                while (st.hasMoreTokens()) {
                  // Note we assume that the filters make sense
                  String token = st.nextToken().trim();
                  if (token.startsWith("*") ||
                      token.endsWith("*") ||
                      token.startsWith("!")) {
                     bWildCards = true;
                  }
                  theFilters.add(token);
                }

            }
            else {
                // we do this to stop dis cl throwing out huge amounts of data
                // if they really want huge amounts of data then dis cl(*) does it
                bSummary=true;
            }

            // We go and look if there are any "enhancers" and if there are then check that
            // they match up with what we expect
            if (null != enhancers) {
                String[] validEnhancers = {"me","st","fld"}; // what ones we allow
                if (true == enhancersCheck(validEnhancers)) {
                    // Lets look for each one individually and set the
                    // flags.....
                    bDisplayMethods = enhancerIsPresent("me");
                    bDisplayStatics = enhancerIsPresent("st");
                    bDisplayFields =  enhancerIsPresent("fld");
                }
                else {
                    theOutput.add("Invalid enhancer detected: "+
                        enhancersCheck(validEnhancers));
                    // but we still continue....
                }
            }

            // Go off to DvUtils to do the scan (if necessary) and fill theClasses vector
            getClasses(theOutput);

            // now for each class - see whether its name matches a filter
            // assuming there are any filters
            if (theFilters.size() >0 && false == bWildCards) {
                for (int i=0;i<theFilters.size();i++) {
                    String f = (String) theFilters.get(i);
                    Integer n = (Integer) classHashByName.get(f.toUpperCase());
                    if (null != n) {
                        // we have found it
                        //System.out.println("In here");
                        ClassDetails c = (ClassDetails) theClasses.get(n.intValue());
                        showClass(c,theOutput,bSummary,theDump);
                        if (true == bDisplayMethods) {
                            showMethods(c,theOutput,bSummary);
                        }
                        if (true == bDisplayStatics) {
                            showStatics(c,theOutput,bSummary);
                        }
                        if (true == bDisplayFields) {
                            showFields(c,theOutput,bSummary,theDump);
                        }
                    }
                }
            }
            else {
                // OK they have wildcards ... got to scan through the lot

                for (int i=0;i<theClasses.size();i++) {
                    // go and find thename of the class
                    ClassDetails c = (ClassDetails) theClasses.get(i);
                    boolean bShowit = true; // assume this one is to be shown
                    if (0 !=theFilters.size()) {
                        // if there are any filters, then lets see if this name passes
                        bShowit = DvUtils.filtrate(c.name,theFilters);
                    }

                    if (true == bShowit) {
                        showClass(c,theOutput,bSummary,theDump);
                        if (true == bDisplayMethods) {
                            showMethods(c,theOutput,bSummary);
                        }
                        if (true == bDisplayStatics) {
                            showStatics(c,theOutput,bSummary);
                        }
                        if (true == bDisplayFields) {
                            showFields(c,theOutput,bSummary,theDump);
                        }
                    }
                }
            }
        }
        else {
                theOutput.add("Unable to comply - no dump accessed");
        }


        cpr.output(theOutput);


       //System.out.println("!!!"+verbModifier);
       if ( false == bSummary ||
            (null == paramString && verbModifier.toUpperCase().equals("CL")) ||
            (null == paramString && verbModifier.toUpperCase().equals("CLASS")) ) {
            forcedEnd = true;
       }

    }



    /**
     *                   Goes through the assorted types of loaded classes and
     *                   stores them away........
     *
     */
    private void getClasses(Vector v) {

        // only do it once per dump.....
        if (0 == theClasses.size()) {
            cc1=getClassesForCategory("LOADEDCLASSES",v);
            cc3=getClassesForCategory("LOADEDSYSTEMCLASSES",v);
            cc2=getClassesForCategory("LOADEDACSCLASSES",v);

        }
        //also figure out method block size used when scanning
        // method table and FieldBlock size to be used when
        // scanning fields
        methodBlockSize = CType.find("methodblock").getSize();
        fieldBlockSize = CType.find("fieldblock").getSize();
        DvUtils.trace("methodblock size is x"+
                    Integer.toHexString(methodBlockSize),0,false);
        DvUtils.trace("fieldblock size is x"+
                    Integer.toHexString(fieldBlockSize),0,false);
    }

    /**
     *  Goes through the assorted types of loaded classes and
     *  stores them away........
     *
     */
    private int getClassesForCategory (String category,Vector output) {
        DvUtils.trace("Entry: DvClassCommands.getClassesForCategory -"+
                        category,0,false );
        int count = 0;
        DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        long obj = 0;
        //boolean detailedDebug = false;
        //if (detailedDebug) output.add(" !!dd - getClassesForCategory: detailed debug started");


        // LoadedClasses is our starting point hint
        // we loop through using memory_table out of mirrored_data
        // as all the classes are chained together......
        String hint = DvUtils.getJVMHint(as,category);
        if (null != hint) {
            long current_start = DvUtils.hexToLong(hint);
            try {
                //if (detailedDebug) output.add(" !!dd - A");
                DvUtils.trace("  current_start = "+Long.toHexString(current_start),0,false );
                long next_one = 0;
                if (0 != current_start) {
                    next_one =  DvUtils.getFieldAddress(current_start,
                                new String[] {"Hjava_lang_Class","obj"},
                                as);
                }

                while (0 != next_one) {

                    obj = next_one;
                    //if (detailedDebug) output.add(" !!dd - B"+Long.toHexString(obj));
                    long ptrToNameAddr = DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","name"},
                                        as);
                    // actually want the field not the field address
                    //if (detailedDebug) output.add(" !!dd - C");
                    long next_one_ptr = DvUtils.getFieldAddress( obj,
                                        new String[] {"MirroredClassData","memory_table"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - D");
                    long static_variable_values = DvUtils.getFieldAddress( obj,
                                        new String[] {"MirroredClassData","static_variable_values"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - E"+
                    //        Long.toHexString(static_variable_values) +
                    //        Long.toHexString(obj));
                    static_variable_values = DvUtils.getPtrFromPtr(theDump,as,
                                                            static_variable_values);
                    //if (detailedDebug) output.add(" !!dd - F");
                    long version_ptr = DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","major_version"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - G");
                    long methods_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","methods"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - H");
                    long methods = DvUtils.getPtrFromPtr(theDump,as,methods_ptr);

                    //if (detailedDebug) output.add(" !!dd - I");
                    long fields_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","fields"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - J");
                    long fields = DvUtils.getPtrFromPtr(theDump,as,fields_ptr);

                    //if (detailedDebug) output.add(" !!dd - K");
                    long methods_count_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","methods_count"},
                                        as);

                    //if (detailedDebug) output.add(" !!dd - L");
                    int methods_count = (int) as.readShort(as.createAddress(methods_count_ptr));


                    //    methods_count = (methods_count << 16) >>>16;
                    //if (detailedDebug) output.add(" !!dd - M");
                    long fields_count_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","fields_count"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - N");
                    int fields_count = (int) as.readShort(as.createAddress(fields_count_ptr));


                    //if (detailedDebug) output.add(" !!dd - M");
                    long instance_size_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","instance_size"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - N");
                    int instance_size = (int) as.readShort(as.createAddress(instance_size_ptr));

                    //if (detailedDebug) output.add(" !!dd - O");
                    long static_var_count_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class"
                                                      ,"static_variable_count"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - P");
                    int static_variable_count = (int) as.readShort(as.createAddress(static_var_count_ptr));


                    // Now figure out whether the classloader is initialized we do this
                    //if (detailedDebug) output.add(" !!dd - Q");
                    byte[] bVersion =as.readBytes(as.createAddress(version_ptr),4);


                    // y is now the address of the name as a null terminated string in memory
                    // so lets go and get it
                    //if (detailedDebug) output.add(" !!dd - R");
                    long nameAddr = DvUtils.getPtrFromPtr(theDump,as,ptrToNameAddr);

                    //System.out.println("nameAddr: "+nameAddr);
                    //System.out.println("ptrToNameAddr: "+ptrToNameAddr);

                    String name;
                    // Now we are finally able to get the name.......
                    try {
                        //if (detailedDebug) output.add(" !!dd - S");
                        name =
                            DvUtils.getStringFromBytes(as.readBytes(
                                    as.createAddress(nameAddr),256));
                    }
                    catch (DvAddressException e) {
                        //if (detailedDebug) output.add(" !!dd - T");
                        name = " --- unknown --- @ 0x"+Long.toHexString(nameAddr);
                    }

                    //if (detailedDebug) output.add(" !!dd - U");
                    long sa_ptr =  DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","superclass"},
                                        as);
                    //if (detailedDebug) output.add(" !!dd - V");
                    long superAddress = (as.readPointer(as.createAddress(sa_ptr))).getAddressAsLong();

                    String superName=" --- none ---";
                    //if (detailedDebug) output.add(" !!dd - W");
                    if (superAddress != 0) {
                        //if (detailedDebug) output.add(" !!dd - X");
                        ptrToNameAddr = DvUtils.getFieldAddress( obj,
                                        new String[] {"Classjava_lang_Class","super_name"},
                                        as);
                        //if (detailedDebug) output.add(" !!dd - Y");
                        nameAddr = DvUtils.getPtrFromPtr(theDump,as,ptrToNameAddr);

                        //if (detailedDebug) output.add(" !!dd - Z");
                        // Now we are finally able to get the name.......
                        try {
                            //if (detailedDebug) output.add(" !!dd - AA");
                            superName =
                                DvUtils.getStringFromBytes(as.readBytes(
                                        as.createAddress(nameAddr),256));
                        }
                        catch (DvAddressException e) {
                            //if (detailedDebug) output.add(" !!dd - BB");
                            superName = " --- unknown --- @ 0x"+Long.toHexString(nameAddr);
                        }


                    }

                    // loader is either 0 (which means it is initialized) or
                    // an address of initialized field offset off obj
                    // (Classjava_lang_ClassLoader structure)
                    //if (detailedDebug) output.add(" !!dd - CC");
                    long loaderPtr = DvUtils.getFieldAddress( obj,
                                        new String[] {"MirroredClassData","loader"},
                                        as);

                    //System.err.println(name + " loader:"+loaderPtr+":"+Long.toHexString(loader));
                    long loader = 0;
                    boolean classloaderInited = false;
                    try {
                        //if (detailedDebug) output.add(" !!dd - DD");
                        loader =
                            as.readPointer(as.createAddress(loaderPtr)).getAddressAsLong();

                        if (0 == loader) {
                            classloaderInited = true;
                            //System.err.println("class loader initialised");
                        }
                        else {
                            //System.err.println("loader:" + loader + ":" + Long.toHexString(loader));
                            // need to increment loader by the offest of the obj field in jvmobject
                            // and the offset of the initialized field within this - basically the obj
                            // is address of Class
                            //loader = loader +
                            //    as.readPointer(as.createAddress(loader)).getAddressAsLong();

                            CType cl = CType.find("Classjava_lang_ClassLoader");
                            if (null != cl) {

                                long objOffset = next_one-current_start;
                                DvAddress  tAddr = as.createAddress(loader+objOffset);



                                CTypeObject cTO = DvUtils.formatAddrAsType(as,tAddr,
                                                    "Classjava_lang_ClassLoader");
                                CTypeObject cTO1 = cTO.getField("initialized");

                                tAddr = as.createAddress(cTO1.getAddr());
                                int ini = as.readInt(tAddr);
                                if (0 != ini)  {
                                    classloaderInited = true;
                                }

                            }
                            else {
                                // leave classLoader assumed as un-inited
                            }
                        }
                    }
                    catch (DvAddressException e) {
                        //if (detailedDebug) output.add(" !!dd - EE");
                        // just in case the address isn't in the dump
                        classloaderInited = true; // we assume that the fact that there is
                                                  // an address means that the loader is assumed
                                                  // to be inited.
                    }


                    // major version is a short at bVersion
                    // minor version is a short at bVersion+2
                    // but need to remember endianess

                    //if (detailedDebug) output.add(" !!dd - FF");
                    short maj_ver = DvUtils.bytesToShort(bVersion,0,theDump.getArchitecture());
                    short min_ver = DvUtils.bytesToShort(bVersion,2,theDump.getArchitecture());
                    //    System.out.println("maj_ver:" + maj_ver + " min_ver:" + min_ver);



                    //System.out.println(  " name: "+name );
                    // Now add this class to our maintained table of classes
                    // but do a circularity check first
                    if (totalClassesCount > 0) {

                        Integer n = (Integer) classHashByAddr.get(new Long(current_start));
                        if (n != null) {
                            ClassDetails c = (ClassDetails) theClasses.get(n.intValue());
                            if (current_start == c.address) {
                                output.add(" !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                                output.add(" !!! Circularity detected creating class table !!!");
                                output.add(" !!!  Class "+ c.name + " #"+n.intValue()+
                                       " has matching address to proposed #"+
                                       totalClassesCount+1);
                                output.add(" !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                                next_one = 0; // lets get out!!!!
                            }
                        }

                    }

                    if (0 != next_one) {
                        ClassDetails c = new ClassDetails();
                        c.name = name;
                        c.superName = superName;
                        c.superAddress = superAddress;
                        c.address = current_start;
                        c.major_version = maj_ver;
                        c.minor_version = min_ver;
                        c.methods = methods;
                        c.fields = fields;
                        c.methods_count = methods_count;
                        c.fields_count = fields_count;
                        c.loaderInitd = classloaderInited;
                        c.instance_size = instance_size;
                        c.static_variable_count = static_variable_count;
                        c.static_variable_values = static_variable_values;


                        theClasses.add(c);
                        if (name.equals("java/lang/String")) {
                            javaLangString = c;
                        }
                        // Hash Table allows us to figure out if we already have this one
                        // and do circularity check....
                        classHashByAddr.put(new Long(current_start),new Integer(totalClassesCount));
                        classHashByName.put(c.name.toUpperCase(),new Integer(totalClassesCount));
                        count++;
                        totalClassesCount ++;

                        next_one = DvUtils.getPtrFromPtr(theDump,as,next_one_ptr);
                        current_start = next_one;
                        if (0 != next_one) {
                            next_one =  DvUtils.getFieldAddress(next_one,
                                        new String[] {"Hjava_lang_Class","obj"},
                                        as);
                        }


                    }

                } // end of while
            }
            catch (DvAddressException e) {

                output.add(" !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                output.add(" !!! DvAddressException whilst accessing class");
                output.add(" !!! data from file:\n" + e.getMessage()   );
                output.add(" !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                // e.printStackTrace();

                //throw new Error(" Unexpected failure - 0x" + Long.toHexString(obj) );
            }
        }
        else {
            // Ok so we didn't find the hint - means that the STGlobalPtr was
            // probably adrift....
            output.add(" !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
            output.add(" !!! Hint missing for category "+category);
            output.add(" !!! Probably due to mismatch in types file.");
            output.add(" !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        }
        DvUtils.trace("Exit: DvClassCommands.getClassesForCategory -"+
                        category + "  count="+count,0,false );
        return count;
    }


    /**
     *   A data class for storing class info for later use
     *   Extended to include superclass name (java/Lang/Object ignored).
     */
    private static class ClassDetails{

        protected String name; // whats the name of this class
        protected long address;   // where the class is in memory
        protected short major_version;
        protected short minor_version;
        protected long methods;
        protected long methods_count;
        protected boolean loaderInitd; // whether loader is initialized
                                       // ans therefore whether statics
                                       // are valid
        protected long static_variable_count; // How many statics
        protected long static_variable_values; // values for the statics
        protected long fields_count; // How many fields
        protected long fields; // address of field blocks
        protected long instance_size; // How big an instance is
        protected String superName;
        protected long superAddress;   // where the super class is in memory



       public String toString() {

           return ("\nClass:"+ name + "\n  @"+Long.toHexString(address)+
            "  version:" + major_version + "." +  minor_version +"\n" +
            "  super:"+superName+ "  @"+Long.toHexString(superAddress) +"\n" +
            "  methods:@" + Long.toHexString(methods) +
            "   method count:" + methods_count + "\n" +
            "  Class Loader Initialized: " + loaderInitd + "\n" +
            "  fields:@" + Long.toHexString(fields) +
            "   fields count:" + fields_count +
            "   instance size:" + instance_size + "\n" +
            "  statics:@" + Long.toHexString(static_variable_values) +
            "   statics count:" + static_variable_count ) ;
       }

    }

    /**
     *
     *
     *
     */
    private  void showMethods( ClassDetails c,Vector theOutput,boolean bSummary) {

        // for class c go through the Method Blocks which all follow each
        // other sequentially. Can check for errors in that they should
        // all point back to same class.....

        DvUtils.trace("Entry: DvClassCommands.showMethods",0,false );
        DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        boolean bRet = true;
        int arch = theDump.getArchitecture();
        boolean is64Bit = DvUtils.is64BitSystem(arch);
        // first print out class name
        theOutput.add(  "\n Method Block details follow....." +
                        "\n ===========================\n");


        theOutput.add(c); // will show summary data
        theOutput.add("\n");

        if (0<c.methods_count) {
            long where=c.methods;
            // now go through the method table for requisite count
            for (int i=0;i<c.methods_count;i++,where+=methodBlockSize) {
                // Now go and get the Method Block and output it
                // but first check that where is really in memory
                DvAddress addr = as.createAddress(where);
                if (true ==DvUtils.checkAddressInDump(addr,as)) {
                    CTypeObject mb = DvUtils.formatAddrAsType(as,addr,"methodblock");
                    // Check that the clazz is equal to c.address
                    // clazz is first field of method block
                    if (false == clazzCheck(i,addr,c.address,as,is64Bit)) {
                        theOutput.add("    *** Consistency check failed - " +
                                        "clazz field not matching (showMethods)");
                    }

                    theOutput.add(  "\n Method Block entry #"+(i+1)+
                                    "\n ======================");
                    theOutput.add(mb);
                }
                else {
                    theOutput.add("   *** Method block #"+i+" address (" +
                     Long.toHexString(where)+") not valid***");
                }

            }
        }
        else {
            // No methods
            theOutput.add("   *** Class has NO METHODS ***");
        }

        DvUtils.trace("Exit: DvClassCommands.showMethods",0,false);

    }

    /**
     *                   Show the statics for a class
     *
     *
     */
    private  void showStatics( ClassDetails c,Vector theOutput,boolean bSummary) {

        // for class c go through the Method Blocks which all follow each
        // other sequentially. Can check for errors in that they should
        // all point back to same class.....

        DvUtils.trace("Entry: DvClassCommands.showStatics",0,false);

        boolean bRet = true;
        int arch = theDump.getArchitecture();
        boolean is64Bit = DvUtils.is64BitSystem(arch);
        DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // first print out class name
        theOutput.add(  "\n Statics for " + c.name +
                        "\n ===========================\n");
        //theOutput.add(c);

        if (true == c.loaderInitd) {
            //theOutput.add("Field count is:" + c.fields_count );
            // for each field see if it is a static one and then
            if (0<c.fields_count) {
                long where=c.fields;
            // now go through the method table for requisite count
                for (int i=0;i<c.fields_count;i++,where+=fieldBlockSize) {
                    DvAddress addr = as.createAddress(where);
                    if (true ==DvUtils.checkAddressInDump(addr,as)) {
                        // Lets see if its a static (ACC_STATIC)

                        //long access_ptr = 0;
                        short access_flags = 0;
                        try {
                           long obj =  DvUtils.getFieldAddress(where,
                                   new String[] {"MemberBlock","access"},
                                   as);

                           access_flags = as.readShort(as.createAddress(obj));



                        }
                        catch (DvAddressException e) {
                        }
                        // remember that the top bit is propagated when shifting right.

                        if (DvConstants.ACCESS_STATIC == (DvConstants.ACCESS_STATIC & access_flags)) {

                        //access_flags = (access_flags << 28)>>31;
                        //theOutput.add("access_flags is x"+Integer.toHexString(access_flags));
                        //if (-1 == access_flags) {
                              //theOutput.add("It looks to be static" );

                            CTypeObject fb = DvUtils.formatAddrAsType(as,addr,"fieldblock");
                            // Check that the clazz is equal to c.address
                            // clazz is first field of field block
                            if (false == clazzCheck(i,addr,c.address,as,is64Bit)) {
                                theOutput.add("    *** Consistency check failed - " +
                                        "clazz field not matching (showStatics)");
                            }


                            //theOutput.add(  "\n Static Field Block entry #"+(i+1)+
                            //                "\n =============================");
                            //theOutput.add(fb);
                            // Now add value details
                            long valueAddress = 0;
                            long sigAddress = 0;
                            long nameAddress = 0;
                            String sig=null;
                            try {
                                sigAddress = DvUtils.getFieldAddress(where,
                                   new String[] {"MemberBlock","signature"},
                                   as);



                                nameAddress = DvUtils.getFieldAddress(where,
                                   new String[] {"MemberBlock","name"},
                                   as);

                                valueAddress =  DvUtils.getFieldAddress(where,
                                   new String[] {"fieldblock","value"},
                                   as);


                                if (is64Bit) {

                                    nameAddress  = as.readLong(as.createAddress(nameAddress));
                                    valueAddress = as.readLong(as.createAddress(valueAddress));
                                    sigAddress =  as.readLong(as.createAddress(sigAddress));
                                }
                                else {
                                    nameAddress  = as.readInt(as.createAddress(nameAddress));
                                    valueAddress = as.readInt(as.createAddress(valueAddress));
                                    sigAddress =  as.readInt(as.createAddress(sigAddress));
                                }
                                sig = DvUtils.stringFromAddress(sigAddress, as);
                                //theOutput.add(" sigAddress is:"+Long.toHexString(sigAddress));

                            }
                            catch (DvAddressException e) {
                                // The fact that value = 0 should handle this exception
                            }

                            if (nameAddress !=0) {
                                String nameString = " Field #" + (i+1) + " @0x" +
                                        Long.toHexString(where)+ " : "+
                                        DvUtils.stringFromAddress(nameAddress, as);
                                theOutput.add(nameString);
                            }

                            if (valueAddress != 0 && sigAddress != 0) {

                                String valueString = "     "+fieldValue(valueAddress, sig,
                                                     as,false);
                                theOutput.add(valueString);
                            }
                            else {
                                theOutput.add("valueAddress:" + valueAddress +
                                              "  sigAddress:" + sigAddress);
                                theOutput.add(DvUtils.stringFromAddress(sigAddress, as));
                                theOutput.add(" !!! Value unobtainable for this field");
                            }

                            theOutput.add("      Access flags:"+ DvConstants.interpretAccess(access_flags));

                        }

                    }
                    else {
                        theOutput.add("   *** Field block #"+i+" address (" +
                            Long.toHexString(where)+") not valid***");
                    }
                }
            }

            else {
                theOutput.add("   *** Class has NO FIELDS ***");
            }
        }
        else {
            // Loader not initialized - statics not valid
            theOutput.add("   *** Loader not initialized - statics invalid ***");
        }

        DvUtils.trace("Exit: DvClassCommands.showFields",0,false);

    }

    /**
     *   Show the fields for a class - non-static ones
     *
     *
     */
    private static void showFields( ClassDetails c,Vector theOutput,
                                    boolean bSummary,DvDump theDump) {

        // for class c go through the Method Blocks which all follow each
        // other sequentially. Can check for errors in that they should
        // all point back to same class.....

        DvUtils.trace("Entry: DvClassCommands.showFields",0,false);

        boolean bRet = true;
        int arch = theDump.getArchitecture();
        boolean is64Bit = DvUtils.is64BitSystem(arch);
        DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // first print out class name
        theOutput.add(  "\n Fields (non-static) for " + c.name +
                        "\n =====================================\n");

            FieldInfo fInfo = (FieldInfo) fieldsHashByName.get(c.name);
            boolean newFInfo = false;

            if (null == fInfo) {
                newFInfo = true;
                fInfo = new FieldInfo();
            }

            if (0<c.fields_count) {
                long where=c.fields;
            // now go through the field table for requisite count
                for (int i=0;i<c.fields_count;i++,where+=fieldBlockSize) {
                    DvAddress addr = as.createAddress(where);
                    if (true ==DvUtils.checkAddressInDump(addr,as)) {
                        // Lets see if its a non - static (ACC_STATIC)

                        //long access_ptr = 0;
                        short access_flags = 0;
                        try {
                           long obj =  DvUtils.getFieldAddress(where,
                                   new String[] {"MemberBlock","access"},
                                   as);

                           access_flags = as.readShort(as.createAddress(obj));

                        }
                        catch (DvAddressException e) {
                        }

                        // only want non-static ones
                        if (DvConstants.ACCESS_STATIC != (DvConstants.ACCESS_STATIC & access_flags)) {



                            CTypeObject fb = DvUtils.formatAddrAsType(as,addr,"fieldblock");
                            // Check that the clazz is equal to c.address
                            // clazz is first field of field block
                            if (false == clazzCheck(i,addr,c.address,as,is64Bit)) {
                                theOutput.add("    *** Consistency check failed - " +
                                        "clazz field not matching (showFields)");
                            }


                            // Now add value details
                            long valueAddress = 0;
                            long sigAddress = 0;
                            long nameAddress = 0;
                            String sig = null;
                            try {
                                sigAddress = DvUtils.getFieldAddress(where,
                                   new String[] {"MemberBlock","signature"},
                                   as);


                                nameAddress = DvUtils.getFieldAddress(where,
                                   new String[] {"MemberBlock","name"},
                                   as);

                                valueAddress =  DvUtils.getFieldAddress(where,
                                   new String[] {"fieldblock","value"},
                                   as);

                                if (is64Bit) {

                                    nameAddress  = as.readLong(as.createAddress(nameAddress));
                                    valueAddress = as.readLong(as.createAddress(valueAddress));
                                    sigAddress =  as.readLong(as.createAddress(sigAddress));
                                }
                                else {
                                    nameAddress  = as.readInt(as.createAddress(nameAddress));
                                    valueAddress = as.readInt(as.createAddress(valueAddress));
                                    sigAddress =  as.readInt(as.createAddress(sigAddress));
                                }
                                sig = DvUtils.stringFromAddress(sigAddress, as);
                                //theOutput.add(" sigAddress is:"+Long.toHexString(sigAddress));

                            }
                            catch (DvAddressException e) {
                                // The fact that value = 0 should handle this exception
                            }
                            String nameString = "";

                            if (nameAddress !=0) {
                                nameString = DvUtils.stringFromAddress(nameAddress, as);
                                String tS = " Field #" + (i+1) + " @0x" +
                                        Long.toHexString(where)+ " : "+
                                        nameString;
                                theOutput.add(tS);
                                // Now for some special processing to aid quick
                                // finding of value field in java/lang/String
                                //System.out.println("!!!"+nameString+"!!!"+
                                //    c.name+"!!!");
                                if (nameString.equals("value") &&
                                   (c.name.equals("java/lang/String"))) {
                                    stringValueOffset = valueAddress;

                                }
                            }

                            if (sigAddress != 0) {


                                String valueString = "     "+fieldValue(valueAddress, sig,
                                                     as,true);
                                theOutput.add(valueString);
                            }
                            else {
                                theOutput.add("valueAddress:" + valueAddress +
                                              "  sigAddress:" + sigAddress);
                                theOutput.add(DvUtils.stringFromAddress(sigAddress, as));
                                theOutput.add(" !!! Value unobtainable for this field");
                            }

                            theOutput.add("      Access flags:"+ DvConstants.interpretAccess(access_flags));
                            if (true == newFInfo) {
                                fInfo.addField( DvUtils.stringFromAddress(nameAddress, as),
                                                DvUtils.stringFromAddress(sigAddress, as),
                                                valueAddress,access_flags);
                            }
                        }

                    }
                    else {
                        theOutput.add("   *** Field block #"+i+" address (" +
                            Long.toHexString(where)+") not valid***");
                    }
                }
            }

            else {
                theOutput.add("   *** Class has NO FIELDS ***");
            }

            // Now hash fInfo so we can find it quickly later....
            // only if we don't already have one
            if (true == newFInfo) {
                fieldsHashByName.put(c.name,fInfo);
            }

            DvUtils.trace("Exit: DvClassCommands.showFields",0,false);

    }


    /**
     *
     *
     *
     */
    private static void showClass( ClassDetails c,Vector theOutput,boolean bSummary,
                                    DvDump theDump) {
        // and format it .....
        // put together a FORMAT c.address as ClassClass and go for it
        if (false == bSummary) {
            theOutput.add("\n Class details for " + c.name +
                          "\n ========================================\n");
            theOutput.add(DvBaseFmtCmds.doFormatEx("as Hjava_lang_Class",
                                                   theDump,
                                                   Long.toHexString(c.address)));
        }
        else {
            theOutput.add(c);
        }
    }


    /**
     *   getClassName from theClasses vector
     *
     *
     */
    public static String getClassName(long classDescriptor) {

        String s = null;

        Integer n = (Integer) classHashByAddr.get(new Long(classDescriptor));
        if (n != null) {

             ClassDetails c =  (ClassDetails) theClasses.get(n.intValue());
             s = new String(c.name);

        }


        return s;
    }

    /**
     *  Method help
     *
     *
     */
    private void doMethodHelp(Vector v) {

        Vector helpLines = DvUtils.getMultiplePropertyValues("MethodHelp",
                                                "DvClassCommands.properties");
        if (helpLines.size() != 0) {
            for (int i=0;i< helpLines.size();i++) {
                v.add(helpLines.elementAt(i));
            }
        }

    }

    /**
     *  converts a value and signature into something printable
     *
     *
     */
     public static String fieldValue(long valueAddr, String signature,
                            DvAddressSpace as,boolean isOffset) {

        // first we look at the start of the signature to see what type
        // of value we are dealing with .....
        //char firstChar = signature.charAt(0);
        // Then depending on first char we form a print line
        String badMemory = "!!! unable to read value from memory at "+
                                Long.toHexString(valueAddr);
        String iOffset = " instance offset is "+ valueAddr +
                        " (0x"+Long.toHexString(valueAddr)+")" ;
        boolean finished = false;
        String retString = " !!!! Field Value could not be established";
        byte b= 0;
        int offset = 0;

        //String signature = DvUtils.stringFromAddress(sigAddr,as);
        //System.out.println("signature is :"+signature+"  valueAddr:" + Long.toHexString(valueAddr));
        char firstChar = signature.charAt(0);
        switch(firstChar) {
            case 'I' :  {
                // Its an integer
                if (true == isOffset) {
                    retString = " integer:" + iOffset;
                }
                else {
                    try {
                        int i = as.readInt(as.createAddress(valueAddr));
                        retString = " integer: "+ i+" (0x"+ Integer.toHexString(i)+")";
                    }
                    catch (DvAddressException e) {
                        retString = " integer: " + badMemory;
                    }
                }
                break;
            }
            case 'J' :  {
                // Its a long
                if (true  == isOffset) {
                    retString = " long:" + iOffset;
                }
                else {
                    try {
                        long l = as.readLong(as.createAddress(valueAddr));
                        retString = " long: "+ l + " (0x"+Long.toHexString(l)+")";
                    }
                    catch (DvAddressException e) {
                        retString = " long: " + badMemory;
                    }
                }
                break;
            }
            case 'C' :  {
                // Its a char- which is a 16 bit unsigned integer obeying
                // unicode standards..........
                if (true == isOffset) {
                retString = " char:" + iOffset;
                }
                else {
                    try {
                        int i = as.readInt(as.createAddress(valueAddr));
                        Character c = new Character((char) i);
                        retString = " char: "+ c + " (0x"+Long.toHexString(i)+")";
                    }
                    catch (DvAddressException e) {
                        retString = " char: " + badMemory;
                    }
                }
                break;
            }
            case '[' :  {
                // Its an array of some sort
                // lets look at 2nd char to establish an array of whats

                char secondChar = signature.charAt(1);
                String type = "";
                switch (secondChar) {
                    case 'I' : type = "integer"; break;
                    case 'J' : type = "long"; break;
                    case 'L' : type = "objects"; break;
                    case 'C' : type = "char"; break;
                    case 'B' : type = "byte"; break;
                    case 'E' : type = "enum"; break;
                    case 'F' : type = "float"; break;
                    case 'D' : type = "double"; break;
                    case 'S' : type = "short"; break;
                    case 'V' : type = "void"; break;
                    case 'Z' : type = "boolean"; break;
                    case 'A' : type = "any"; break;
                    case '[' : type = "array";break;
                    default  : type = " !!! Not known ("+secondChar+")"; break;
                }
                // go and read the pointer at valueAddr
                if (true == isOffset) {
                    retString = " array of "+ type + " :"+iOffset;
                }
                else {
                    try {
                        DvAddress p =  as.readPointer(as.createAddress(valueAddr));
                        retString = " Array of "+type + " @ 0x" +
                            Long.toHexString(p.getAddressAsLong())  ;
                        if (type.equals("char")) {
                                // special processing for array of char
                                retString += "\n\t<<"+interpretValue(signature,valueAddr, as)+
                                ">>";
                        }
                        else {
                            retString += "\t"+interpretValue(signature,valueAddr, as);

                        }
                    }
                    catch (DvAddressException e) {
                        retString = " Array of "+type + " @ !!! unable to read memory from "+
                               Long.toHexString(valueAddr) ;
                    }
                }
                break;
            }
            case 'L' :  {
                // Its a class instance
                if (true == isOffset) {
                    retString = " class:" + iOffset;
                }
                else {
                    retString = " Class instance of " + signature + " @ 0x";

                    try {
                        DvAddress p =  as.readPointer(as.createAddress(valueAddr));
                        retString += Long.toHexString(p.getAddressAsLong()) ;
                        // Only do Special code for java/lang/String instances.....
                        if (signature.equals("Ljava/lang/String;")) {
                            // we will have stored away (during object summary run)
                            // the offset of the field named "value" so we do some
                            // quick arithmetic and get the position in memory
                            // of thia array of chars and then we do the business
                            try {
                                long x = p.getAddressAsLong();
                                //System.out.println(Long.toHexString(x));
                                if (x != 0) {

                                    long obj =  DvUtils.getFieldAddress(x,
                                           new String[] {"Hjava_lang_Object","obj"},as);


                                    retString += "\n\t\t  <<"+interpretValue("[C",
                                                obj,
                                                as)+">>";

                                }
                                else {
                                    // its a null string
                                    retString+= "\n\t\t  Null String <<>>";
                                }
                            }
                            catch (DvAddressException e) {
                                retString+= "\n\t\t  Unable to evaluate String value";
                            }

                        }
                    }
                    catch (DvAddressException e) {
                        retString += " !!! unable to read memory from "+
                               Long.toHexString(valueAddr) ;
                    }

                }

                break;
            }
            case 'F' :  {
                // Its a float
                if (true == isOffset) {
                    retString = " float:" + iOffset;
                }
                else {
                    try {
                        float f = (float) as.readLong(as.createAddress(valueAddr));
                        retString = " float: "+ f + " (0x"+Long.toHexString((long)f)+")";
                    }
                    catch (DvAddressException e) {
                        retString = " float: " + badMemory;
                    }
                }
                break;
            }
            case 'D' :  {
                // Its a double which actually takes up same space as a long
                // so (I think) we can get it by using readLong and casting
                // it as a double .....
                if (true == isOffset) {
                    retString = " double:" + iOffset;
                }
                else {
                    try {
                        double d = (double) as.readLong(as.createAddress(valueAddr));
                        retString = " double: "+ d + " (0x"+Long.toHexString((long)d)+")";
                    }
                    catch (DvAddressException e) {
                        retString = " double: " + badMemory;
                    }
                }
                break;
            }
            case 'Z' :  {
                // Its a boolean which is an integer where anything other than 0
                // is true..........
                if (true == isOffset) {
                    retString = " boolean:" + iOffset;
                }
                else {
                    try {
                        int i = as.readInt(as.createAddress(valueAddr));
                        if (0 == i) {
                            retString =  " boolean: false"+
                                         " (0x"+Integer.toHexString(i)+")";
                        }
                        else {
                            retString = " boolean: true"+
                                        " (0x"+Integer.toHexString(i)+")";
                        }
                    }
                    catch (DvAddressException e) {
                        retString = " boolean: " + badMemory;
                    }
                }
                break;
            }
            case 'B' :  {
                // Its a byte
                if (true == isOffset) {
                    retString = " byte:" + iOffset;
                }
                else {
                    try {
                        Byte bite = new Byte(as.readByte(as.createAddress(valueAddr)));
                        retString = " byte: 0x"+ Integer.toHexString(bite.intValue())  ;

                    }
                    catch (DvAddressException e) {
                        retString = " byte: " + badMemory;
                    }
                }
                break;
            }
            case 'S' :  {
                // Its a short - 2 bytes

                if (true == isOffset) {
                    retString = " short:" + iOffset;
                }
                else {
                    try {
                        short s = as.readShort(as.createAddress(valueAddr));
                        retString = " short: 0x"+ Integer.toHexString((int) s)  ;

                    }
                    catch (DvAddressException e) {
                        retString = " short: " + badMemory;
                    }  ;
                }
                break;
            }
            default: {
                retString = retString + "( signature: "+signature+")";
            }
        }



        return retString;
     }



    /**
     * reset saved data when new dump opened
     *
     *
     */
    public void reset() {

        prevVerb=null;
        prevModifier=null;
        prevParms=null;
        prevSeperator=null;
        cc1=0;
        cc2=0;
        cc3=0;
        totalClassesCount=0;
        theClasses.clear();
        classHashByAddr.clear();
        classHashByName.clear();
        fieldsHashByName.clear();
        methodBlockSize=0;

    }


    /**
     * given a JVMObject then the instance values for the object are returned
     * from the matching class....
     * 18/07/02 Extended this to go up the chain of super classes
     *
     */
    public static String getInstanceValues(long addr,String className,DvDump theDump,Vector o) {

        DvUtils.trace(" Entry: getInstanceValues  - addr="+Long.toHexString(addr)+
                        "  className="+ className,0,false);
        Vector v = new Vector();
        String retString="";
        boolean goodOutput = false;
        boolean classHasASuperClass = false;
        String superName = "";
        DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        // First we make sure that the classes vector has been filled
        //getClasses(v); // Not too bothered with the output here

        // Now we see if there are any classes in the vector
        if (0 == theClasses.size()) {
            o.add( "\nRunning imbedded displayCS to establish class formatting!!!\n");
            //String retString = "\nRunning imbedded displayCS !\n";
            // need to run dummy dis cls first
            DvClassCommands x = new DvClassCommands();
            x.verb = "dis";
            x.verbModifier = "cls";
            x.theDump = theDump;
            x.displayCS();

        }



         // so now we take the incoming className and find it in theClasses vector
         // using the hashing
        if (0 != theClasses.size()) {
           // System.out.println("!!!"+className+"!!!");
            if (null == classHashByName) {
                System.out.println("classHashByName is null!!!");
            }
            Integer n = (Integer) classHashByName.get(className.toUpperCase());
            Integer m = (Integer) classHashByAddr.get(new Long(addr));
            if ( m == null) {
                // ok its a class instance and not the class definition
                // So now we can get the field names and offsets and form
                // a retString..... do some special processing for Strings
                // which are arrays of chars ......
                ClassDetails c;
                boolean bItsAString = false;
                goodOutput = true;
                if (className.equals("java/Lang/String")) {
                    c = javaLangString;
                    bItsAString = true;
                }
                else {

                    if (null != n) {

                        c = (ClassDetails) theClasses.get(n.intValue());
                        if ((0 != c.superAddress) &&
                             (false == c.superName.equals("java/lang/Object"))) {

                            classHasASuperClass = true;
                            superName = c.superName;
                        }
                    }
                    else {

                        goodOutput = false;
                    }
                }

                if (true == goodOutput) {

                    // Now we go and figure out the names and offsets
                    // Look in the fieldsHashByName for remembered ones

                    FieldInfo fInfo = (FieldInfo) fieldsHashByName.get(className);
                    if (null == fInfo) {
                        // ok it wasn't in the hash table - so lets see if we can find it and
                        // - do this by calling Showfields
                        Integer num = (Integer) classHashByName.get(className.toUpperCase());
                        if (null != num) {
                            // we have found it
                            //System.out.println("In here");
                            Vector vBin = new Vector(); // bin for output
                            ClassDetails cl = (ClassDetails) theClasses.get(num.intValue());
                            showFields(cl,vBin,true,theDump);
                            fInfo = (FieldInfo) fieldsHashByName.get(className);
                        }

                    }
                    if (null == fInfo) {
                        // if fInfo still null then it looks as though we cannot
                        // format this one...............................
                        goodOutput = false;
                    }
                    else {
                        // Do the format
                        int arch = theDump.getArchitecture();
                        int count = fInfo.name.size();


                        try {
                            // if its an array then number of elements is
                            // contained in methods field
                            //long sizeAddr =  DvUtils.getFieldAddress(addr,
                            //        new String[] {"Hjava_lang_Object","methods"},as);
                            //int size = as.readInt(as.createAddress(sizeAddr));
                            long obj =  DvUtils.getFieldAddress(addr,
                                    new String[] {"Hjava_lang_Object","obj"},as);
                            //DvAddress a1 = as.readPointer(as.createAddress(obj));
                            //obj = a1.getAddressAsLong();
                            for (int i=0;i<count;i++) {
                                String sig =  (String) fInfo.signature.get(i);
                                //retString += ("\n" + (String) fInfo.name.get(i) + " ("+ sig + ") \t:");
                                String name = (String) fInfo.name.get(i);
                                retString += ("\n" + name); // + "\t\t:");
                                if (name.length() > 7) {
                                    retString += "\t";
                                }
                                else {
                                    retString += "\t\t";
                                }
                                //retString += interpretValue(sig, ((Long) fInfo.offset.get(i)).longValue(),
                                //                        obj,as,arch);
                                long x = obj+((Long)fInfo.offset.get(i)).longValue();

                                retString += fieldValue(x, sig, as,false);

                            }

                            if (true == classHasASuperClass) {
                               //retString += "\n    ===== Super Class expansion for: "+superName ;
                               String temp = getInstanceValues(addr,superName,
                                                                theDump,o);
                               if (temp != null && false == temp.equals("")) {
                                  retString += "\n    ===== Super Class expansion for: "+
                                                superName + temp;
                               }
                            }
                        }
                        catch (DvAddressException e) {
                            retString = " DvAddressException in DvClassCommands:getInstanceValues";

                        }
                    }

                }
                else {
                    // At this point we do special processing for arrays
                    if  (className.startsWith("array ")) {
                        long lenAddr = 0;
                        int len = -1;
                        try {
                            lenAddr = DvUtils.getFieldAddress(addr,
                                    new String[] {"Hjava_lang_Object","methods"},as);

                            len = as.readInt(as.createAddress(lenAddr));



                            if (className.equals("array char")) {
                            //DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
                                // Cheat slightly......
                                if ( len > 0) {
                                    byte[] ba;
                                    ba = as.readBytes(as.createAddress(lenAddr+8),len*2);
                                    byte[] ba1 = new byte[ba.length/2] ;
                                    for (int i=0;i<ba1.length;i++) {
                                        ba1[i] = ba[2*i+1];
                                    }
                                    String s = new String(ba1);
                                    if (s.length()<61) {
                                        retString="\t<<"+s+">>";
                                    }
                                    else {
                                        retString="\t<<"+s.substring(0,50)+"...>> " +
                                        (s.length()-50) + " char not shown";
                                    }
                                }
                                else {
                                    retString= "\tIts a null array of char (length 0)";
                                }

                            }
                            else {
                                retString=" No array expansion available";
                            }

                            retString = "Array Size:"+len+"\n"+retString;
                        }
                        catch (DvAddressException e) {
                                retString="  !!!exception in array processing in getInstanceValues";
                        }

                    }
                    else {
                    }

                    goodOutput = true;
                }

            }
            else {
                retString=" No instance values: ClassClass rather than instance of class";
                goodOutput = true;
            }
        }
        else {
            // we are in trouble as we haven't any classes extablishable
            // tell them
            o.add( "\n !!! After displayCS there are still no classes available !!!\n");
        }


        if (true == goodOutput) {
            return retString;
        }
        else {
            return retString + badString;
        }
    }

    /**
     * Change an array of char into something printable
     *
     *
     */
    private static String interpretValue(String sig,long where, DvAddressSpace as) {
        // ok we need to look at sig and do the business
        // also check that addr is actually available in dump
        // Note that the address incoming is the address of the object
        // (the obj field from for xxx as JVMObject).
        String retString=null;
        boolean bHandled = false;
        //System.out.println(" addr is:"+ addr + "  offset is:" + off);
        //System.out.println( "  where:" + Long.toHexString(where));
        if (sig.startsWith("[")) {
            char ch = sig.charAt(1);
            bHandled = true;
            String s = null;
            int len = 0;
            try {
                        DvAddress a1 = as.readPointer(as.createAddress(where));
                        // so now a1 points to an array of chars!!! length is in first int
                        // and locknflags in second int so lets read an array of bytes of
                        // length 8 plus length...... if for 64 bit they increase size of
                        // length field or locknflags then we may need to adjust code
                        // also need to read the bytes as shorts in order to get the endianess
                        // considered!!!!!
                        if (0 != a1.getAddressAsLong()) {
                             len = as.readInt(a1);
                        }
            }
            catch (DvAddressException e) {
            }

            try {
                DvAddress a1=null;
                a1 = as.readPointer(as.createAddress(where));
                switch(ch) {
                    case 'C' :


                        // so now a1 points to an array of chars!!! length is in first int
                        // and locknflags in second int so lets read an array of bytes of
                        // length 8 plus length...... if for 64 bit they increase size of
                        // length field or locknflags then we may need to adjust code
                        // also need to read the bytes as shorts in order to get the endianess
                        // considered!!!!!


                        // String s = null;
                        int outputLen = len;
                        if (len > 91) {
                            outputLen = 80;
                        }


                        if (len > 0) {
                            //System.out.println("len is:"+len);
                            char[] ca = new char[outputLen];

                            a1.offsetBy(8);
                            for (int n=0;n<outputLen;n++) {
                                ca[n] = (char) as.readShort(a1);
                                    a1.offsetBy(2);

                            }

                            s=new String(ca);
                            if (len > 91) {
                                s = s + " (" + (len-80) + " bytes not displayed.)";
                            }
                        }
                        else {
                            s = "... Null String (length 0) ...";
                        }

                        return s;

                    case 'B':
                        // Its an array of bytes......
                         if (len > 0) {
                            s = "  Array Size:" + len + " \n\t{";
                            a1.offsetBy(8);
                            int cutoff = 10;
                            for (int i=0;i<len && i < 10;i++) {

                                    byte b = as.readByte(a1);
                                    int ival = (int) b;
                                    if (ival < 0) {
                                        ival=ival+256;
                                    }
                                    s += (DvUtils.byteToHex).substring(2*ival,
                                                        2*ival+2);

                                    a1.offsetBy(1);


                                if ((cutoff-1) != i && i != (len-1)) {
                                    s+= ",";
                                }

                            }
                            if (len <= cutoff) {
                                s += "}" ;
                            }
                            else {
                                s += ", ........";
                            }
                            if (len > cutoff) {
                                s = s+ "\n\t"+ (len-cutoff)+ " values not shown";
                            }
                            return s;

                        }
                        else {
                            s = "... Null Array (length 0) ...";
                            return s;
                        }


                    case 'I' :
                    case 'S' :
                    case 'J' :
                    case 'F' :
                    case 'D' :


                        // so now a1 points to an array of integers


                        // Its an array of integers,shorts,longs,floats or doubles
                        if (len > 0) {
                            s = "  Array Size:" + len + " \n\t{";
                            a1.offsetBy(8);
                            int cutoff = 10;
                            for (int i=0;i<len && i < 10;i++) {
                                if ('I' == ch) {
                                    s = s+ as.readInt(a1);
                                    a1.offsetBy(4);
                                }
                                if ('J' == ch || 'D' == ch || 'F' == ch) {
                                    // Floats and doubles also take up
                                    // 8 bytes..........
                                    if ('J' == ch) {
                                        s = s+ as.readLong(a1);
                                    }
                                    if ('D' == ch) {
                                        cutoff = 4;
                                        s = s+ (double) as.readLong(a1);
                                    }
                                    if ('F' == ch) {
                                        cutoff = 4;
                                        s = s+ (float) as.readLong(a1);
                                    }
                                    a1.offsetBy(8);
                                }
                                if ('S' == ch) {
                                    s = s+ as.readShort(a1);
                                    a1.offsetBy(2);
                                }
                                if ((cutoff-1) != i && i != (len-1)) {
                                    s+= ",";
                                }

                            }
                            if (len <= cutoff) {
                                s += "}" ;
                            }
                            else {
                                s += ", ........";
                            }
                            if (len > cutoff) {
                                s = s+ "\n\t"+ (len-cutoff)+ " values not shown";
                            }
                            return s;

                        }
                        else {
                            s = "... Null Array (length 0) ...";
                            return s;
                        }


                    default  : {
                        s = " Array Size:"+len;
                        return s;

                    }
                } // end of switch
            }
            catch (DvAddressException e) {
                            return ("\n !!!!! DvAddressException in DvClassCommands:interpretValue"+
                            " sig:"+sig + "  where:" + Long.toHexString(where) + "  !!!!!\n");
            }


        }
        else {

        }



        if (false == bHandled) {
            // catch all case
            retString = "Unable to interpret field (sig="+sig+", where="+ where+")";
        }

        return retString;
    }

    /**
     *   A class for keeping field definitions within classes under control
     *
     */
    static class FieldInfo {

        public Vector name = new Vector();
        public Vector signature = new Vector();
        public Vector offset = new Vector();
        public Vector access = new Vector();

        public void addField(String n, String sig, long o, short acc) {
            name.add(n); // name of this field
            signature.add(sig);  // signature for this field
            offset.add(new Long(o)); // offset into instance
            access.add(new Short(acc));

        }


        public String toString() {

            String retString="";
            int count = name.size();
            retString+="Field count = "+count;
            for (int i=0;i<count;i++) {
                retString += "\n name: "+ (String) name.get(i);
                retString += "\n signature: "+ (String) signature.get(i);
                retString += "\n offset: "+ (Long) offset.get(i);
                retString += "\n access: "+ (Short) access.get(i);
            }

            return retString;

        }
    }


    /**
     *   Consistency check that the clazz value matches owning class
     *
     */
    private static boolean clazzCheck(int i,DvAddress addr,long cAddr,
                                DvAddressSpace as,boolean is64Bit) {


        boolean bRet=true;

        try {
            // Cover 32 and 64 bit check.....
            if (true ==  is64Bit) {
                if (cAddr != as.readLong(addr)) {
                    bRet = false;

                    //theOutput.add(Long.toHexString(as.readLong(addr)));
                    //                theOutput.add(Long.toHexString(cAddr));
                }
            }
            else {
                if (cAddr != (long) as.readInt(addr)) {
                    bRet = false;
                }
            }
        }
        catch (DvAddressException e) {
            DvUtils.trace  ( "\n !!!!!!!!!!!!!!!!!!!!!!" +
                            "\n Method Block entry #"+(i+1)+
                            "\n produced DvAddressException" + Long.toHexString(cAddr) +
                            "\n Address: " +
                            "\n !!!!!!!!!!!!!!!!!!!!!!",0,true);
            bRet = false;
        }

        return bRet;
    }

    /**
     *   return the instance size for a class name
     *
     *
     */
    public static long getInstanceSize(String className,Vector o) {
        //System.err.println("className="+className+"!");
        long retLong = 0;

        if (0 == theClasses.size()) {

            o.add( "\nRunning imbedded displayCS to establish class formatting!!!\n");
            //String retString = "\nRunning imbedded displayCS !\n";
            // need to run dummy dis cls first
            DvClassCommands x = new DvClassCommands();
            x.verb = "dis";
            x.verbModifier = "cls";
            x.theDump = DvConsole.getCurrentDump();
            x.displayCS();

        }

        Integer num = (Integer) classHashByName.get(className.toUpperCase());
        if (null != num) {
                // we have found it

                ClassDetails cl = (ClassDetails) theClasses.get(num.intValue());
                //System.out.println("In here "+ cl.instance_size);
                retLong = cl.instance_size;

        }

        return retLong;
    }


}

