/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvConsole.java, ras, dumpviewer 1.40
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502  PHILR:  New for feature 36524
 * 36524  200502  philr   Changed to support new CommandPluginResponse mode
 *                        ... lots of changes
 * 36524  230502  philr   Now support space as a seperator so that commands
 *                         such as FORMAT aaa as bbb are supportable
 * 36524  270502  philr   Plugin methods need to be non static.
 * 36524  270502  philr   plusminus support.
 * 36524  300502  philr   Search complete command table before looking for
 *                        generic action on second search
 * 36524  300502  philr   Special useage of ~ and ^ to enable command preprocessor
 *                        (see method preProcessInput)
 * 36524  040602  philr   Added support for verbModifierForFind
 * 36524  160602  philr   Added support for enhancers beyond the end bracket
 * 36524  260702  philr   Support for redirecting output and error streams
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
//import com.ibm.jvm.Trace;

/**
  *  Description: This class is both the "main" class for the console version
  *               of the dump formatter and also a class instantiated by the
  *               GUI version of the dumpviewer. When driven by the GUI version
  *               no input loop is driven. The constructor for this class uses
  *               a boolean (stored as runningUnderGUI) to determine whether
  *               DvConsole has been instantiated from the GUI.
  *
  */
public class DvConsole implements Observer{


    static PrintStream original_output = System.out; // so we can reset output to console
                                                     // after setting it to a file;
    static PrintStream original_error = System.err; // so we can reset output to console
                                                    // after setting it to a file;
    BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
    boolean traceOn = false;
    String  traceFileName=null; // null= console
    String  output=null;  // null = console
    String  inputFileName=null; // null = console
    static Properties inParams;
    static DvDump theDump;
    static CommandPlugin currentCP; // for interrupting purposes with !
    static Vector cmdList = new Vector();
    static Vector commandPlugins = new Vector();
    static Vector fastPath=new Vector();   // (*1,*2,*3 etc support)
    static String currDump=null;
    private static Properties myHelpProps=null; // used for help
    private static Enumeration myPropsEnum;
    static int  currASIndex=0;       // what index in the as array we should use
    static int   currProcessIndex=0;  // what index in the process array we should use
    static int   currThreadIndex=0;   // what index in the thread array we should use
    static Vector currentClassSpecs = new Vector();
    static Vector currentCBSpecs = new Vector();
    static int    currWidth = 80; // assume 80 to start with
    static char   currOffset = 'N';

    static boolean echoOn = false; // input echoed to output.
    static boolean ipDump=false;
    static boolean ipInput=false;
    static boolean ipOutput=false;
    static int traceHandle;
    static String[] templates;
    static boolean runningUnderGUI=false;
    static Dumpviewer theGUI=null;
    static String lastMemCommandPosition; // used by +/- processing
    static int    lastMemCommandLength; // used by +/- processing

    // variables required to support execValidCmdToArray
    private static Vector outputVector=null; 
    private static boolean useOutputVector=false;
    private static DvConsole that=null;
    private static CommandPlugin childCP; 


    //**********************************************************
    //*
    //*   Now for some code
    //**********************************************************

    /**
     *   <p>Description: Construct a new DvConsole, either driven by the main
     *                   method within DvConsole class or from Dumpviewer the
     *                   GUI version of the dump formatter.
     *   </p>
     *   <p>Input      : boolean indicating whether we are gui driven or not
     *                   (true is gui driven)
     *   </p>
     */
    public  DvConsole(boolean guiDriven,Dumpviewer gui) {

        // save away an object reference for use in a static context
        that=this;

        DvUtils.trace("DvConsole:DvConsole("+guiDriven+")",0,false);
        // If the gui is driving us then we don't expect input
        // via command line;
        runningUnderGUI = guiDriven;
        theGUI = gui;
        DvUtils.setConsoleOutput(!guiDriven);


        // Lets look for plugins and format their command set

        commandPlugins = DvUtils.findCommandPlugins();

        int numPlugins= commandPlugins.size();

        // Now for each plugin ask it what it supports....
        String[] syntax=null;
        DvCommand command = null;

        DvUtils.trace("Asking each command plugin what it supports ",0,false);
        for (int i=0;i<numPlugins;i++) {

            CommandPlugin plugin = (CommandPlugin) commandPlugins.get(i);
            DvUtils.trace(" Plugin "+ plugin.pluginName() + " supports: ",0,false);
            syntax=plugin.getSyntax();
            for (int j=0;j<syntax.length;j++) {
                command = new DvCommand(syntax[j],i);
                DvUtils.trace("    "+syntax[j],0,false);
                cmdList.add(command);
            }

        }

        // lets setup ability to set/get useful properties
        DvUtils.initSetCommand();


        // first we do any cmmands implied by input parameters
        // these are all set commands and part of DvBaseCommands
        if (true == ipOutput) {
            String cmd= "Set Output="+ inParams.getProperty("o");

            execCmd(new ValidCmd(cmd));
        }

        if (true == ipDump) {
            String cmd= "Set Dump="+ inParams.getProperty("d");
            execCmd(new ValidCmd(cmd));
        }

        if (true == ipInput) {
        // we have a load of input from file
            String cmd= "Exec "+ inParams.getProperty("i");
            execCmd(new ValidCmd(cmd));
        }


        String inputLine = "";


        // Now a do loop until quit is entered
        // unless we are running under GUI, in which case input will be
        // programatic instead of manual.
        DvUtils.trace("Entering master loop - terminated with quit command",0,false);
        while (false == runningUnderGUI &&
             true != inputLine.equals("quit")) {

            output("\nReady......");

            try {
                inputLine = in.readLine();
            }
            catch (IOException e) {
                // Should never happen !!!!!
                DvUtils.errorMsg(" IOException in DvConsole constructor input loop",0);
            }
            if (echoOn) {
                output(inputLine);   // echo the input line

            }

            inputLine = inputLine.trim(); // get rid of leading and trailing spaces
            String savedInputLine = new String(inputLine);
            inputLine = preProcessInput(inputLine);
            if (false == inputLine.equals(savedInputLine)) {
                //transformation has occurred
                output("Transformed command:"+inputLine+"\n");
            }


            // now validate input line
            // n gives position where error detected so we can echo
            // it back with a mark under it.....


            ValidCmd validCmd = validate(inputLine);
            if (validCmd.errorPos == 0) {
                output("\n ......command executing");
                execCmd(validCmd);
                // wait for command to finish
                int singleChar;
                while (null != currentCP &&
                       true == currentCP.getInProgress()) {
                    try {
                        Thread.sleep(100);
                        //try {
                        //  singleChar = in.read();
                        //  if (singleChar == '!') {
                        //   System.out.println("An Interrupt!!");
                        //  currentCP.setForcedEnd(true);
                        //
                        //  }
                        // }
                        // catch (IOException e1) {
                        //   System.out.println("An IOException!!");
                        // }
                    }
                    catch  (java.lang.InterruptedException e) {
                        System.err.println("\nInterruptedException");
                    }
                } // end of while

            }
            else
            {
                String s=new String();
                // must be an easier way to do this......
                for (int i=0;i<validCmd.errorPos;i++) {
                    s = s+" ";
                }
                output("\n !!!Invalid Command: " +
                    validCmd.savedCmd+ "\n" +
                    "                    " + s + "^\n");
                if (null != currentCP) {
                    currentCP.cpInProgress = false;
                }
            }

        } // end of while

    }


    /**
     *   <p>Description:
     *   </p>
     */
    public class DvCommand {

        String verb=null;
        String verbShort=null;
        String modifier=null;
        String modifierShort=null;
        String method=null;
        //String[] parameters=null;
        int     index; // index into commandPlugins vector


        public DvCommand(String syntax, int i) {


            // Have to split the command into pieces
            //
            index = i;
            StringTokenizer st = new StringTokenizer(syntax,",");
            if (5 != st.countTokens()) {
                System.err.println("!!!Bad command definition!!!");
            }
            else {
                verb = st.nextToken().toUpperCase();
                verbShort = st.nextToken().toUpperCase();
                modifier = st.nextToken().toUpperCase();
                if (true == modifier.equals(" ")) {
                    modifier = null;
                }
                modifierShort = st.nextToken().toUpperCase();
                if (true == modifierShort.equals(" ")) {
                    modifierShort = null;
                }
                method=st.nextToken();
            } // end of else

      }  // end of DvCommand constructor
    }

    /**
     *   <p>Description:
     *   </p>
     */
    public   static  void main(String[] args) {


        templates=new String[1];
        templates[0]="0 DvConsole: %s";
        //traceHandle = Trace.registerApplication("DvConsole",templates);

        //Trace.trace(traceHandle,0,args);



        boolean finis = false;
        boolean bgood = true; // assume parameters are good

        if (args.length>0) {
            inParams  = new Properties();
            for (int i=0;i<args.length && finis == false;i++) {
                // Set various things depending on input arguments
                // all arguments need to start with a -x, where x is
                // -i (input from file name following)
                // -o (output to file name following)
                // -t trace to console (messy!!!)
                // -t.... trace to filename
                // -d (dump identifier)
                // -? (give me some help)
                // -h (give me some help)
                // invalid input will result in a syntax help

                if (args[i].charAt(0) == '-') {
                    // good its looking like a parameter
                    char ch = args[i].charAt(1);
                    switch(ch) {
                        case 'd' :
                        case 'D' : {
                              bgood = true;
                              ipDump = true;
                              inParams.setProperty("d",args[i].substring(2));
                              break;
                        }
                        case 'i' :
                        case 'I' : {
                              bgood = true;
                              ipInput = true;
                              inParams.setProperty("i",args[i].substring(2));
                              break;
                        }
                        case 'o' :
                        case 'O' : {
                              bgood = true;
                              ipOutput = true;
                              inParams.setProperty("o",args[i].substring(2));
                              break;
                        }
                        case 't' :
                        case 'T' : {
                              //trace is to be set
                              //output("caught a t "+i);
                              DvUtils.setVerbose(true);
                              inParams.setProperty("t",args[i].substring(2));
                              bgood = true;
                              break;
                        }
                        case '?' :
                        case 'H' :
                        case 'h' : {
                              bgood = false;
                              break;
                        }
                        default  : {
                              output("in default "+i);
                              bgood = false;
                              finis = true ;
                        }
                    } // end of switch
                    if (bgood == false) {
                        showHelp("InvokeHelp");
                        finis = true;
                    }

                }

            }


        }
        else {
            showHelp("InvokeHelp"); // If no args then show help
        }

        DvUtils.trace("DvConsole:main - instantiating DvConsole instance",0,false);

        DvConsole dv = new DvConsole (false,null); // false = console driven

    } // end of main

    /**
     *   <p>Description: Function used by GUI interface to tell us what
     *                   the dump is and also used by the SET DUMP command
     *                   when executed as a console command.
     *   </p>
     */
    protected static void setTheDump(DvDump dump,Vector output,String params) {

        DvUtils.trace(" Entry to DvConsole:setTheDump",0,false);
        theDump= dump;

        // Now we will set ASID,PID and THREAD values to
        // 1st ASID,1st PID,  current (if known) or 1st Java Thread
        // (one with an execenv)



        // This will need changing to use currentASID, currentProcess
        // and currentThread!!!!!!!
        DvAddressSpace as[] = dump.getAddressSpaces();
        DvUtils.setValue("ASID",as[0].id());

        DvProcess proc[] = as[0].getProcesses();
        DvProcess currProc = proc[0];
        DvUtils.setValue("PID",currProc.id());
        DvThread currThread = proc[0].getCurrentThread();

            // we get memranges so that dis mem() works without
            // needing to do this first
        as[0].getMemRanges();

        if (null !=currThread && true == currThread.isJava()) {
            DvUtils.setValue("THREAD",currThread.id());
        }
        else {
            DvThread threads[] = proc[0].getThreads();
            boolean bSuitableThread=false; // will set to true when good thread found
            if (threads.length != 0) {
                for (int i=0;i<threads.length;i++) {
                    if (true == threads[i].isJava() &&
                        threads[i].eeAddress().getAddressAsLong() != 0) {
                        DvUtils.trace(" Found a suitable 1st thread:"+threads[i].id()
                                        ,0,false);
                        bSuitableThread=true;
                        DvUtils.setValue("THREAD",threads[i].id());
                        currThread = threads[i];
                        i = threads.length;
                    } // end of if
                 } // end of for
            } // end of if
        } // end of else

        DvUtils.setValue("DUMP",params);

        // Lets see if we have some EnvData - so as we can establish the
        // correct types file to use.
        DvEnvData envData = currProc.getEnvData();
        if (null != envData) {
                String versionInfo = envData.javaVersion() ;

                // Now go through all the CTypeFile stanzas in Dv.properties to
                // get a file name to set for FORMATFILE
                // Note that we pass the dump as if the ctype info is embedded in
                // the dump we will use that in preference to a file....
                // Note that if ff=name comes back as * then we are using
                // ctype info from the dump dump!!!!!!

                String ffname = DvUtils.getCtypeFile(theDump,versionInfo);
                if (ffname != null) {
                    // To be replaced
                    //DvUtils.setValue("FORMATFILE",ffname);
                }



        }
        else {
                DvUtils.trace(" EnvData is null",0,false);
                //System.err.println(" EnvData is null");
                String ffname = DvUtils.getCtypeFile(theDump,"abcd");
                if (ffname != null) {
                    // To be replaced
                    //DvUtils.setValue("FORMATFILE",ffname);
                }
        }


        if (false == DvUtils.isTypeFileLoaded()) {
                DvUtils.loadFormatFile(output);
        }

        // We also set the hints into action provided the THREAD value is not
        // -1 (i.e. we have found a java thread!)

        DvAddress ee_addr= as[0].createAddress(currThread.eeAddress().getAddressAsLong());
        // Check that the address is actually in memory!!!!!
        if (DvUtils.checkAddressInDump(ee_addr,as[0])) {


            CTypeObject ee_cTO = DvUtils.formatAddrAsType(as[0],ee_addr,"execenv");

                if (null != ee_cTO) {
                    // so now we have an execenv - we can use getField on jvmP
                    // to get to the jvm control block
                    CTypeObject jvmP_cTO = ee_cTO.getField("jvmP");

                    if (null != jvmP_cTO) {
                        String s= jvmP_cTO.toString();

                        if (s.startsWith("null ")) {
                            s = s.substring(5);
                        }
                        if (s.startsWith("(null) ")) {
                            s = s.substring(7);
                        }
                        s = DvUtils.removeAtSign(s);
                        s = DvUtils.remove0x(s);
                        //System.out.println("jvmP:"+s);
                        DvAddress a2 = null;
                        DvAddress a1 = null;
                        boolean goodJvmP = true;
                        try {
                            a1= as[0].createAddress(DvUtils.hexToLong(s));
                        }
                        catch (NumberFormatException e) {
                            // if s is not a hex string then we bomb out - probably due to
                            // jvmP being crap!!!
                            goodJvmP = false;
                            DvUtils.trace(" DvConsole:setTheDump -String returned for jvmP_cTO is:<<"+s+">>",
                                        0,true);
                            output.add("\n jvmP addressing failure");
                        }

                        CTypeObject jvm_cTO = null;
                        if (true == goodJvmP) {
                            try {
                                a2 = as[0].readPointer(a1);
                            }
                            catch (DvAddressException e) {
                                // Display an error?
                                if (null != output) {
                                    output.add("\n jvmP addressing failure");
                                }
                            }

                            if (null != a2) {
                                jvm_cTO = DvUtils.formatAddrAsType(as[0],a2,"Jvm");
                            }
                        }


                        if (null != jvm_cTO) {
                            DvUtils.setupJVMHints(jvm_cTO,as[0]);
                        }
                        else {
                            if (null != output) {
                                output.add(" Unable to format Jvm!");
                            }
                        }
                    }
                    else {
                        if (null != output) {
                            output.add(" Unable to get jvmP!");
                        }
                    }
                }

        }
        else {
            DvUtils.trace(" execenv was 0",0,false);
            output.add(" execenv was 0!");
        }

        DvUtils.trace(" Exit from DvConsole:setTheDump",0,false);

    }

    /**
     *   <p>Description:
     *   </p>
     */
    private static void showHelp(String key) {

        // This is the generic help, we get it from the
        // DvGenericHelp.properties file
        DvUtils.trace("DvConsole:showHelp entry",0,false);
        String key1 = null;
        String value ="";
        String value1=null;
        if (null == myHelpProps) {
            try {
                FileInputStream myPropsFIS =
                     new FileInputStream(".\\DvGenericHelp.properties");
                myHelpProps  = new Properties();
                myHelpProps.load(myPropsFIS);

            }
            catch (IOException e) {
                // didn't find it on current directory
                String sep = System.getProperty("file.separator");
                String path=  "com"+sep+"ibm"+sep+"jvm"+sep+
                    "dump"+sep+"DvGenericHelp.properties";
                myHelpProps =  DvUtils.searchClassPathForPropFile(path);
                if (null == myHelpProps) {
                    path=  "DvGenericHelp.properties";
                    myHelpProps = DvUtils.searchClassPathForPropFile(path);
                }

            }
        }

        if (null != myHelpProps) {

            myPropsEnum = myHelpProps.propertyNames();


            for (int i=1; i<999;i++) {
                key1 = key+i;
                value1 = myHelpProps.getProperty(key1);
                if (null != value1) {
                    value += value1;
                }
                else {
                    i = 999;
                }

            }
        }
        output(value);
        DvUtils.trace("DvConsole:showHelp exit",0,false);

    }

    /**
     *   <p>Description: Create a Valid Command given known good input
     *                   and execute it. Return the CommandPlugin instance
     *                   for guito be able to check for completion
     */
    protected void execValidCommand(String input) {

        ValidCmd cmd = new ValidCmd(input);

        execCmd(cmd);

    }



    /**
     *   <p>Description:
     *   </p>
     */
    private static ValidCmd validate(String input) {

        // this will return a ValidCmd block
        // with the main verb available and
        // the parameters all tokenised
        ValidCmd x = new ValidCmd(input);

        return x;

    }

    /**
     *   <p>Description:
     *   </p>
     */
    private static void output(String o) 
    {
        if(useOutputVector) {
            if (outputVector==null  || o==null) {
                String s="***ERROR*** Either output vector or object does not exist!\n"+o;
                DvUtils.output(s);
            } else {
                outputVector.add(o);
            }
        } else {
            // Here we'll redirect the output if required
            // and also do formatting to medium.......
            // ..... but for now
            DvUtils.output(o);
        }
    }

    /**
     *   <p>Description:
     *   </p>
     */
    static class ValidCmd {

        public int errorPos; // 0 = no error
        public String mainVerb;
        public int pluginID=-1;
        public boolean isNonPlugin= false;
        public String seperator=null;
        public String savedCmd;
        public char    mainVerbOne;
        public String restOfInput=null;
        public String params = null ;
        public String modifier=null;
        public String method=null;
        public String enhancersString=null;


        public ValidCmd(String theCmd) {
        //System.out.println("Cmd:"+theCmd);
        errorPos = 1;  //assume command is bad
        savedCmd = theCmd.trim();   // saveit
        // Now tokenise the input - first word/letter is assumed to be the
        // command verb - convert everything to uppercase....
        String mainVerb_oc=null;
        StringTokenizer st = new StringTokenizer(savedCmd);
        try {
            mainVerb_oc = st.nextToken();
            mainVerb =  mainVerb_oc.toUpperCase();
        }
        catch (java.util.NoSuchElementException e) {
           mainVerb=null;
        }


        // Note don't convert to upper case as case can be important
        // particularly on Unix systems

        if (mainVerb != null) {
            mainVerbOne=mainVerb.charAt(0);
            // We now set verbModifer to rest of input string
            seperator = null;
            if (st.hasMoreTokens()) {
                restOfInput=theCmd.substring(mainVerb.length()).trim();
                // now seperate out the verbModifier and u/case it
                StringTokenizer st1 = new StringTokenizer(restOfInput,"=( ");


                modifier = st1.nextToken(); //.toUpperCase() ;
                modifier = modifier.trim();
                restOfInput=restOfInput.substring(modifier.length()).trim();
                seperator = " ";
                if (true == restOfInput.startsWith("=")) {
                    seperator = "=";
                    restOfInput = restOfInput.substring(1);
                }
                if (true == restOfInput.startsWith("(")) {
                    seperator = "(";
                    restOfInput = restOfInput.substring(1);
                }


                if (st1.hasMoreTokens()) {

                    params = restOfInput;


                    if (seperator.equals("(")) {
                        // look for end bracket and figure out whether we have any
                        // enhancers
                        if (params.endsWith(")")) {
                            params = params.substring(0,params.length()-1);
                        }
                        else {
                            int n = params.indexOf(')');
                            if (-1 == n) {
                                errorPos = params.length()+mainVerb.length()+modifier.length()+2;
                                return;
                            }
                            else {
                                // we have some "enhancers"
                                enhancersString =  params.substring(n+1);
                                params = params.substring(0,n);
                                //System.out.println("enhancersString:" + enhancersString);
                                //System.out.println("params:" + params);
                            }
                        }
                    }
                    //System.out.println("\nParams are: "+params);
                }
                //modifier = modifier.toUpperCase();

            }

            switch(mainVerbOne) {

                case '!':
                    // The breakout command
                    isNonPlugin = true;
                    errorPos=0 ;
                    break;


                case '+':
                case '-':
                    // Plus and minus are special cases - used to
                    // redo last dis mem command with updated address
                    // and (in case of plus when the output is
                    // non streamed to get next output block......
                    isNonPlugin = true;
                    errorPos=0 ;
                    break;


                case '*':
                    // The echo character....
                    isNonPlugin = true;
                    errorPos=0 ;
                    break;


                case 'Q':
                    if (true == mainVerb.equals("Q") ||
                        true == mainVerb.equals("QUIT") ) {
                        isNonPlugin = true;
                        errorPos=0;
                        //output("Its valid");
                     }

                    break;

                case 'E':
                    if (true == mainVerb.equals("E") ||
                        true == mainVerb.equals("EXEC") ||
                        true == mainVerb.equals("EXECUTE") ) {

                        isNonPlugin = true;
                        errorPos=0;
                        //output("Its valid");
                    }

                    break;


                default: errorPos = 1; // main verb is not in list
                                 // above - section of cod ebelow will
                                 // now look through command plugins
            }  // end of switch

            //System.out.println("\nAt point xxx "+ errorPos);
            if (0 != errorPos ) {


                // scan the cmdList to see if this verb and verb modifier are
                // supported and by which plugin
                isNonPlugin = false;
                boolean foundVerb = false;
                boolean foundVerbModifier = false;
                int matchFound = -1;
                DvCommand cmd = null;

                for (int i=0;i<cmdList.size() && matchFound == -1;i++) {

                    cmd = (DvCommand) cmdList.get(i);
                    if (mainVerb.equals(cmd.verb) ||
                        mainVerb.equals(cmd.verbShort)) {

                        foundVerb = true; // verb is good

                        // do two searches leaving the generic modifier search till
                        // after we have done all non generics.....
                        if (((modifier == null) && (cmd.modifier == null)) ||
                            (modifier != null) && (modifier.toUpperCase()).equals(cmd.modifier) ||
                            (modifier != null) && (modifier.toUpperCase()).equals(cmd.modifierShort)) {
                            //System.out.println("\n Found verbModifier !! match for "+
                            //                       cmd.modifier +
                            //                  " " + cmd.index);
                            matchFound = cmd.index;
                            pluginID = cmd.index;
                            method = cmd.method;
                            errorPos = 0;
                            foundVerbModifier = true;

                        }
                    }

                } // end of for loop

                // now for second search for '*' case.....
                for (int i=0;i<cmdList.size() && matchFound == -1;i++) {

                    cmd = (DvCommand) cmdList.get(i);
                    if (mainVerb.equals(cmd.verb) ||
                        mainVerb.equals(cmd.verbShort)) {
                        foundVerb = true; // verb is good

                        if (((modifier == null) && (cmd.modifier == null)) ||
                            (modifier != null) && (modifier.toUpperCase()).equals(cmd.modifier) ||
                            (modifier != null) && (modifier.toUpperCase()).equals(cmd.modifierShort) ||
                            (null != cmd.modifier) && (cmd.modifier.equals("*"))) {
                            //System.out.println("\n Found verbModifier match for "+
                            //                      cmd.modifier +
                            //                 " " + cmd.index);
                            matchFound = cmd.index;
                            pluginID = cmd.index;
                            method = cmd.method;
                            errorPos = 0;
                            foundVerbModifier = true;

                        }
                   }

                } // end of for loop

                if (true == foundVerb) {
                    if (false == foundVerbModifier) {
                        if (null != modifier)
                            errorPos = savedCmd.indexOf(modifier.trim())+1;

                        }
                    }
                    else {
                      errorPos=1;
                    }
                }

            }
         }
    }


    /**
     *   <p>Description:
     *   </p>
     */
    protected void execCmd( ValidCmd theCmd )  {

        boolean result = false; // assume failure

        if (true == theCmd.isNonPlugin) {
            switch (theCmd.mainVerbOne) {
                case '!':
                    // ! is our interrupt that tells the currently executing thread
                    // that we'd like it to stop.....
                    if (null != currentCP) {
                        currentCP.setForcedEnd(true);
                    }
                    break;

                case '+':
                    doPlusMinus(true);
                    break;

                case '-':
                    doPlusMinus(false);
                    break;

                case '*':
                    // * is echo - used primarily by scripts
                    output(theCmd.savedCmd);
                    break;
                case 'Q':
                    System.exit(0);
                    break;
                case 'E':
                    execExecuteCmd(theCmd.modifier);
                    break;
                //case 'L':
                // List - fill in the *n vector
                //execListCmd(theCmd.verbModifier);
                //result = execListCmd(theCmd.verbModifier);
                //  break;

                default:
                    if (null != currentCP &&
                        true == currentCP.cpInProgress) {
                        doSingleOutput("\n Unable to execute "+theCmd.savedCmd+
                             "\n .. last command still in progress\n");
                        return ;
                    }


            } // end of switch
        }
        else {
            if (null != currentCP && true == currentCP.cpInProgress)
            {
                doSingleOutput("\n Unable to execute "+theCmd.savedCmd+
                             "\n .. last command still in progress\n");
                return  ;
            }


            Class clz=null;
            Method method=null;
            CommandPlugin cp = (CommandPlugin) commandPlugins.get(theCmd.pluginID);;

            try {
                clz = cp.getClass();
                //   Class clz = ClassLoader.getSystemClassLoader().loadClass("   ");

                // Now look for the method that they have told us thay want to
                // support this command with and check that it exists in the
                // class instance
                method = clz.getMethod(theCmd.method, new Class[]{});
            }
            catch (java.lang.NoSuchMethodException e) {

                // ok so they have a specified a method that dont exist!!!


                if (false == runningUnderGUI)
                    doSingleOutput("\n ***** Error *****"+
                                   "\n Expected method (" +
                                   theCmd.method +") not found in "+
                                   clz.getName()+
                                   "\n ***** Error *****\n");

                return ;
            }

            try {
                currentCP =  (CommandPlugin) clz.newInstance();
            }
            catch (java.lang.InstantiationException e) {
                System.err.println(" InstantiationException");
            }
            catch (java.lang.IllegalAccessException e) {
                System.err.println(" IllegalAccessException");
            }

            // set the method variable in the command plugin class instance
            currentCP.setMethod(method);
            // set the verb, modifier and parameter strings
            currentCP.setVerb(theCmd.mainVerb);
            // Note - must set paramString before verb Modifier!!!!!
            // so that setVerbModifier correctly processes verbModifierForFind
            currentCP.setParamString(theCmd.params);
            if (null != theCmd.enhancersString) {
                //System.out.println("enhancersString:"+theCmd.enhancersString);
                currentCP.setEnhancers(theCmd.enhancersString);

            }

            currentCP.setVerbModifier(theCmd.modifier);

            currentCP.setSeperator(theCmd.seperator);
            currentCP.setForcedEnd(false);
            currentCP.setTheDump(theDump);

            // if we are running under GUI then its update method gets the
            //output not ours!!!!!
            if  (false == runningUnderGUI) {
                currentCP.addObserverToResponse(this);
            }
            else {
                currentCP.addObserverToResponse(theGUI);
            }

            try {
                currentCP.start(); // its running as a seperate thread
                                 //  so we can interrupt it with !
            }
            catch (Exception e) {
                System.err.println("\n!!! Exception !!!");

            }

       }

    }


    /**
     *   <p>Description:
     *   </p>
     *   <p>Input      :
     *   </p>
     *   <p>Returned   :
     *   </p>
     */
    private boolean execExecuteCmd( String execFile)  {
        boolean result = false ; // assume failure
        String inputLine=null;
        BufferedReader in=null;
        int lineCount = 0;

        execFile= execFile.trim();
        output("Will try and execute "+ execFile+"!");
        // so we read through the input given and execute it one line at
        // a time....................
        try {
            in = new BufferedReader(new FileReader(execFile));
        }
        catch (java.io.FileNotFoundException e) {
            // we can't do owt - so we ought to put out an error line
            output("File " + execFile + " not found");
        }

        if (null != in) {

            boolean finis = false;
            while (false == finis) {
            try {
                inputLine = in.readLine();

                // Some things in the input line are translated into other things!!!
                inputLine = preProcessInput(inputLine);

                lineCount++;
                output("\n Execute(" + execFile + ":"+lineCount+
                        "): " +inputLine +"\n");
                if (null==inputLine) {
                    finis=true;
                }
                else {

                    ValidCmd validCmd = validate(inputLine);

                    if (validCmd.errorPos == 0) {

                        //System.out.println("\nExecute: "+inputLine+"\n");
                        execCmd(validCmd);

                        // Now need to wait until the last command has finished
                        // before going round the loop and trying next one
                        //
                        while (true == currentCP.cpInProgress) {
                            try {
                            Thread.sleep(500);
                            }
                            catch (java.lang.InterruptedException e) {
                            }
                        }
                    }
                    else  {
                        System.err.println("Error pos is"+validCmd.errorPos);
                        String invalidErr="Invalid Command - possible reasons:";
                        invalidErr+= "\n An expected plugin was not found";
                        invalidErr+= "\n .... what else ?????";

                        System.err.println(invalidErr);
                        // echo input with error position shown
                    }

                }
            } // end of try
            catch (IOException e) {
                finis=true; }
            }

        }

        return result;

    }
    
    /** Run a command and return the output as a String Array.
     *  <p>Starts up a child Command Plugin to run the command supplied.</p> 
     *  <p>NB - as suggested in the function name, the command is assumed to 
     *          be valid.  It is NOT checked.</p>
     *
     * @param  input String containing the command as you would type in at the 
     *         command prompt.
     * @return String Array sized to length of output.  Each line is a 
     *         line of text as would appear on the console from running 
     *         the command.
     * @since   JDK1.3.1
     */
    protected static String[] execValidCmdToArray( String input ) 
    {
        String[] s=that.doExecValidCmdToArray(input);
        return s;
    }

    /** Routine that does the real work for execValidCmdToArray().
     *  
     *  <p>This is an instance function called by the static execValidCmdToArray().
     *     It runs the command and returns output as described above.</p>
     */
    private String[] doExecValidCmdToArray( String input ) 
    {
        outputVector=new Vector(); 
        useOutputVector=true;
        Class clz=null;
        Method method=null;

        ValidCmd theCmd = new ValidCmd(input);
        
        CommandPlugin cp = (CommandPlugin) commandPlugins.get(theCmd.pluginID);
        
        try {
            clz = cp.getClass();
            method = clz.getMethod(theCmd.method, new Class[]{});
        }
        catch (java.lang.NoSuchMethodException e) {
            // ok so they have a specified a method that don't exist!!!
            System.out.println("***** Error *****\n"+
                               "Expected method "+ theCmd.method + " not found in "+clz.getName()+ "\n"+
                               "***** Error *****\n");
            String[] rubbish={"***ERROR***"};
            return rubbish;
        }

        try {
            childCP =  (CommandPlugin) clz.newInstance();
        } catch (java.lang.InstantiationException e) {
            System.err.println(" InstantiationException");
        } catch (java.lang.IllegalAccessException e) {
            System.err.println(" IllegalAccessException");
        }

        // set the method verb and param string
        childCP.setMethod(method);
        childCP.setVerb(theCmd.mainVerb);
        childCP.setParamString(theCmd.params);

        if (null != theCmd.enhancersString) {
            childCP.setEnhancers(theCmd.enhancersString);
        }

        childCP.setVerbModifier(theCmd.modifier);
        childCP.setSeperator(theCmd.seperator);
        childCP.setForcedEnd(false);
        childCP.setTheDump(theDump);
   
        // our update method gets the output
        childCP.addObserverToResponse(this);

        try {
            childCP.start(); // its running as a seperate thread so we can interrupt it with !
        }
        catch (Exception e) {
            System.err.println("\n!!! Exception !!!");
        }

        // wait for the command to complete...
        while (childCP.getInProgress()) {
            try {
                Thread.sleep(100);
            }
            catch  (java.lang.InterruptedException e) {
                System.err.println("\nInterruptedException");
            }
        } // end of while

        // stop sending output to vector
        useOutputVector=false; 

        // put the vector's contents into a String array
        String s[] = new String[outputVector.size()];
        Enumeration e=outputVector.elements();
        for (int i=0; e.hasMoreElements(); i++) {
            s[i]=(String) e.nextElement();
        }

        outputVector=null;
        return s;
    }



    /**
     *   <p>Description:
     *   </p>
     */
    public void update(Observable o,Object arg) {

        // now we consume the data from the CommandPlugin....
        // it might be that we have some more data later
        DvUtils.trace("Entry DvConsole:output",0,false);
        synchronized (o) {
            doSingleOutput(arg);
        }

        DvUtils.trace("Exit DvConsole:output",0,false);
    }

    /**
     *   <p>Description:
     *   </p>
     */
    private void doSingleOutput(Object arg) {
        // it could be that we have a vector of vectors
        // or mixed vector or just a single object  - for
        // vectors we get iterative or an array of objects
        if (arg instanceof Vector) {
            Vector v = (Vector) arg;
            for (int i=0;i<v.size();i++) {
                doSingleOutput(v.get(i));
            }
        }
        else {
            if (arg instanceof Array) {
                Array a = (Array) arg;
                for (int i=0;i<a.getLength(a);i++) {
                    doSingleOutput(a.get(a,i));
                }
            }

            else {
                if (arg instanceof String) {
                    String s = (String) arg;
                    output(s);
                }
                else {
                    // we try the toString method for the object
                    // - if what they have passed has bothered to do
                    // a nice one then we will get nice output
                    // otherwise.......
                    if (null != arg) {
                        output(arg.toString());
                    }
                    else {
                        output(" --- null --- ");
                    }
                }
            }
        }
    } // end of method;

    /**
     *   <p>Description:
     *   </p>
     */
    private void doPlusMinus(boolean plus) {

        if (null != lastMemCommandPosition) {
            String command;
            long x = DvUtils.hexToLong(lastMemCommandPosition);

            if (plus) {
                x = x +  lastMemCommandLength;
            }
            else {
                x = x - lastMemCommandLength;
            }


            command = "dis mem("+ Long.toHexString(x) + ","+lastMemCommandLength+")";
            doSingleOutput(command); // echo the command

            //System.out.println("!!!:"+command) ;
            //now issue the command
            ValidCmd validCmd = validate(command);
            execCmd(validCmd);
        }
        else {
            doSingleOutput("\n No previous dis mem command found\n");
        }



    } // end of method



    /**
     *   <p>Description: Takes a input line and processes it for tokens and
     *                   also handles 'special' forms. These special forms
     *                   are loaded from an override file (DvOverides.txt).
     *                   Special forms stored in this file take the
     *                   format aaa == bbb, the space==space piece is crucial as
     *                   it allows for differentiation between the aaa and bbb parts.
     *                   For example if the input line was jvmP = 0xabcdef00 then an
     *                   entry in the overrides file of :
     *                        jvmP = ^ == format ^ as Jvm
     *                   would convert the input line to, (* in this case means "rest
     *                   of line").
     *                        format 0xabcdef00 as Jvm.
     *                   Another example
     *                        ^1 @ ^2 == format ^2 as ^1
     *                   would convert StThreadLocalData @ 0x806bef4 to
     *                         format 0x806bef4 as StThreadLocalData
     *
     *                   The main useage for these abilities is to allow easy
     *                   navigation for format commands.....
     *
     *                   On a simpler level
     *                        1 == dis proc
     *                   would allow 1 to be used as a short form for dis proc.
     *
     *                   Token handling is used when the input line contains tokens
     *                   (things preceded by a ^) thus
     *                   "text ^1 text2" would if (^1 was set to "text 3"
     *                   become text1 text3 text2. The tokens are stored as
     *                   set values.
     *   </p>
     */
    private String preProcessInput(String input) {

        String output = " ";

        String workingString = new String(input);
        workingString = " " + workingString + " ";

        // We do a scan of the line looking for ~. Note that we scan backwards
        // from end as this gives us ability to handle ~~~ type notation for free.
        // Having found a ~ we then scan foward looking for a delimiter which is
        // space,comma,(,) or =. Then we have a substring that represents input
        // in looking for a SET value.... which we then replace the identified
        // piece with and start again.....

        boolean error = false;
        boolean finished=false;
        while (false == finished) {
            int len = workingString.length();
            int i = 0;
            boolean iDone = false;
            for (  i=len;i>0 && false == iDone ;i--) {

                if (workingString.charAt(i-1) == '~') {
                    // found a tilda,now search forward for delimiter
                    boolean jDone = false;
                    int j = 0;
                    for (j=i;j<len&& false == jDone;j++) {
                        char x = workingString.charAt(j);
                        if (x == ' ' || x == ',' || x == '(' ||
                            x == ')' || x == '=' || x == '^') {
                          jDone = true;
                        }
                    }
                    if (j==i) {
                        //its a command with tilda followed immediately by rubbish
                        // bad boy
                        System.err.println("bad bad bad");
                        finished = true;
                        error = true;
                        iDone = true;
                    }
                    else {
                        // ok lets work out what follows tilda
                        String key = workingString.substring(i,j-1);
                        // so now we find value for s (if there is one)
                        String replacer = DvUtils.getValue(key);
                        //System.err.println("Replacer:"+replacer);
                        if (null == replacer) {
                            //System.err.println(key + " not found");
                            finished = true;
                            error = true;
                            iDone = true;
                        }
                        else {
                          String top = workingString.substring(0,i-1);
                          String tail = workingString.substring(i+key.length());
                          //System.err.println("top:"+top);
                          //System.err.println("tail:"+tail);
                          workingString = top +  "^" + replacer + tail;
                          //System.err.println("Updated:"+workingString);
                        }

                        //System.err.println("key is:"+key);

                        iDone = true; // so we go round with new workingString

                    }
                }

            }
            if (i == 0) {
                finished = true;
            }
        }

        if (true == error) {
            return input;
        }
        else {

            //we now remove any ^ chars and finally trim the command
            for (int i=0;i<workingString.length();i++) {
              if (workingString.charAt(i) == '^') {
              }
              else output = output + workingString.charAt(i);
            }
            output = output.trim();
            //System.err.println("output is:"+output);
            return output;
        }



    }   // end of method



    // Assorted "helper" functions
    protected static CommandPlugin getCurrentCP() {return currentCP;}
    protected static DvDump getCurrentDump() {return theDump;}
    protected static int getCurrentAsIndex() {return  currASIndex;}
    protected static int getCurrentProcessIndex() {return  currProcessIndex;}
    protected static int getCurrentThreadIndex() {return  currThreadIndex;}

    protected static void setLastDisMem(String position,int length) {
    //System.out.println("p:"+position+" l:"+length);
        lastMemCommandPosition = position;
        lastMemCommandLength = length;
    }

    public static PrintStream getOriginal_output() {
        return original_output;
    }

    public static PrintStream getOriginal_error() {
        return original_error;
    }

}
