/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvDump.java, ras, dumpviewer 1.11
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502  PHILR:  New for feature 36524
 * 36524  210502  philr   getCurrentAddressSpace added
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION: Abstract class
 * ===========================================================================
 */

/**
 * Abstract class to support indirection to a dump file in assorted
 * formats and access mechanisms
 *
 * @author Phil Rosenthal
 *
 */

package com.ibm.jvm.dump;

 /**
  *  Description: Whats the purpose of this class
  *
  *
  *
  *
  *
  */
public abstract class DvDump {


        // System values as used in SdffHead
        public static final int SYSTEM_IS_UNKNOWN = -1;
        public static final int SYSTEM_IS_WINDOWS = 1;
        public static final int SYSTEM_IS_AIX     = 2;
        public static final int SYSTEM_IS_LINUX   = 3;
        public static final int SYSTEM_IS_OS390    = 4;

        public static final String[ ] SYSTEM_DESCRIPTION =
        {
           "!!!Unknown System!!!",
           "Windows",
           "AIX",
           "Linux",
           "OS390"
        } ;

        // SubSystem values as used in SdffHead
        // should be kept in line with string array
        public static final int SUBSYS_IS_UNKNOWN = -1;
        public static final int SUBSYS_WINNT = 1;
        public static final int SUBSYS_WIN2000 = 2;
        public static final int SUBSYS_WIN95 = 3;
        public static final int SUBSYS_WIN98 = 4;
        public static final int SUBSYS_WINME = 5;
        public static final int SUBSYS_WINXP = 6;

        public static final int SUBSYS_AIX51   = 7;
        public static final int SUBSYS_AIX52   = 8;
        public static final int SUBSYS_REDHAT  = 9;
        public static final int SUBSYS_SUSE  = 10;
        public static final int SUBSYS_OS390    = 11;
        // add additional ones as required but don't
        // forget to update string array below....

        public static final String[ ] SUBSYS_DESCRIPTION =
        {
           "!!!Unknown SubSystem!!!",
           "Windows NT",
           "Windows 2000",
           "Windows 95",
           "Windows 98",
           "Windows ME",
           "Windows XP",
           "AIX 5.1",
           "AIX 5.2",
           "REDHAT",
           "SUSE",
           "OS 390"
        }  ;

        // Architecture values as used in SdffHead

        public static final int ARCHITECTURE_IS_UNKNOWN = -1;
        public static final int ARCHITECTURE_IS_32BIT        =  1;
        public static final int ARCHITECTURE_IS_64BIT        =  2;
        public static final int ARCHITECTURE_IS_BIGENDIAN    =  4;
        public static final int ARCHITECTURE_IS_LITTLEENDIAN =  8;

        public static final String[ ] ARCHITECTURE_DESCRIPTION =
        {
           "Unknown or invalid Architecture",
           "32 bit - Unknown endianess",
           "64 bit - Unknown endianess",
           "Ambiguous architecture value",
           "Big Endian - Unknown bitness",
           "32 bit - Big Endian",
           "64 bit - Big Endian",
           "Big Endian - ambiguous bitness",
           "Little Endian - Unknown bitness",
           "32 bit - Little Endian",
           "64 bit - Little Endian"
        };

        // Processor values as used in SdffHead
        // Used in conjunction with processor subtype information
        public static final int PROCESSOR_IS_UNKNOWN = -1;
        public static final int PROCESSOR_IS_INTEL    = 1;
        public static final int PROCESSOR_IS_PPC      = 2;
        // add more as and when required.....


        public static final String[] PROCESSOR_DESCRIPTION =
        {
          "Unknown processor",
          "Intel processor",
          "Motorola Power PC"
        };

        // Processor subtype  values as used in SdffHead
        // Used in conjunction with processor subtype information
        public static final int PROC_SUBTYPE_UNKNOWN = -1;
        public static final int PROC_SUBTYPE_IS_PENTIUM = 1;
        public static final int PROC_SUBTYPE_IS_PENTIUM2 = 2;
        public static final int PROC_SUBTYPE_IS_PENTIUM3 = 3;
        public static final int PROC_SUBTYPE_IS_PENTIUM4 = 4;
        // add more as and when required.....


        public static final String[] PROC_SUBTYPE_DESCRIPTION =
        {
          "Unknown processor subtype",
          "Pentium",
          "Pentium II",
          "Pentium III",
          "Pentium IV"
        };
 /**
  *   <p>Description: Determines whether the input dump source is compatible
  *                   with the concreate dump plugin.
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public static Boolean isSupportedSource(String theDumpSource) {

/**
  * This static method allows determination of whether the dump source
  * is compatible with the concrete classes extending this class. Because
  * its static no instantiation of the concrete class will be necessary, just
  * a load of it. The default one returns false - meaning that this dump source
  * isn't suported by this class.
  *
  * @param theDumpSource - where the dump is
  *
  */
 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
     return (new Boolean(false));
 }


 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract DvAddressSpace[] getAddressSpaces();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract DvAddressSpace getCurrentAddressSpace();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getArchitecture();   // returns the architecture for the creator
                                          // of the dump
 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getSystemType();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getSubSystemType();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getProcessor();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getProcessorSubType();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getCurrentProcessor();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 public abstract int getProcessorCount();

 /**
  *   <p>Description:
  *   </p>
  *   <p>Input      :
  *   </p>
  *   <p>Returned   :
  *   </p>
  */
 byte[] getMetaData() {

 /**
  * getMetaData can be optionally implemented for extenders of this abstract class
  * to return a byte array representing the metadata
  *
  */

  return null;
 }




}
