/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvFileReader.java, ras, dumpviewer 1.6    
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 41452  210502 nmasson  New
 * 41452  240602 nmasson  Added readString and setPos
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.io.*;

/**
 *  DvFileReader is a class to implement random access to data in the 
 *  address space for a process where the data is held in a file.  
 */

class DvFileReader extends DvReader {
   /**
    * The file to read from
    */
    RandomAccessFile core;			// Source of data
    /**
     * How to map addresses to file offsets
     */
    Mapper mapper;				// Map addresses to file pos
    /**
     * Constructs a DvFileReader to read data of endian type and word length
     * size from the file file
     */
    DvFileReader(RandomAccessFile file, WordType type, WordLength size) {
	core = file;
	endType = type;
	wordLength = size;
    }
    /**
     * Set up the address map
     */
    void setMapper(Mapper mapper) {
	this.mapper = mapper;
    }
    /**
     * Read an array of bytes from the file
     */
    void read(byte[] b) throws IOException {
	core.readFully(b);
    }
    /**
     * reads a byte array from the given offset and moves to the next
     * position
     */
    void read(byte[] b, int off, int len) throws IOException {
	core.read(b, off, len);
    }
    /**
     * reads a len byte value from the file.  len should be less than or
     * or equal to eight.
     */
    long readBytes(int len)  throws IOException {
	byte[] b = new byte[len];
	core.read(b);
	return(endType.toLong(b));
    }
    /**
     * reads a null-terminated string pointed to from the file
     */
    String readStringPtr() throws IOException {
	long addr = readAddress(); 
	long pos = core.getFilePointer();	// Save the new offset
	DvUtils.trace("Ptr = " + Long.toHexString(addr) + " filepos = " + 
	    Long.toHexString(pos), 2, false);
	setAddress(addr);

	int i = 0;
	int size = 512;
	byte[] b = new byte[size];
	while(i < size) {
	    b[i] = core.readByte();
	    if(b[i] == 0) {
		break;
	    }
	    i++;
	}
	DvUtils.trace("String = " + new String(b, 0, i), 2, false);
	setPos(pos);		// Restore the offset
	return(new String(b, 0, i));
    }
    /**
     * reads a null-terminated string from the file
     */
    String readString() throws IOException {
	int i = 0;
	int size = 512;
	byte[] b = new byte[size];
	while(i < size) {
	    b[i] = core.readByte();
	    if(b[i] == 0) {
		break;
	    }
	    i++;
	}
	return(new String(b, 0, i));
    }	
    /**
     * reads a string from the file
     */
    String readString(int len) throws IOException {
	byte[] b = new byte[len];
	core.read(b);
	return(new String(b));
    }	
    /**
     * reads a one-byte value from the file
     */
    int readBuffer(byte[] buffer) throws IOException {
	long len = core.read(buffer);
	return((int)len);
    }
    /**
     * reads a one-byte value from the file
     */
    long readByte() throws IOException {
	return(readBytes(1));
    }
    /**
     * reads a two-byte value from the file
     */
    long readHalf() throws IOException {
	return(readBytes(2));
    }
    /**
     * reads a four-byte value from the file
     */
    long readWord() throws IOException {
	return(readBytes(4));
    }
    /**
     * reads a eight-byte value from the file
     */
    long readLong() throws IOException {
        return(readBytes(8));
    }
    /**
     * gets the current file position to newOffset
     */
    long getPos() throws IOException {
	return core.getFilePointer();
    }
    /**
     * sets the current file position to newOffset
     */
    void setPos(long newOffset) throws IOException {
	core.seek(newOffset);
    }
    /**
     * sets the current file position to the position corresponding to the
     * process Address
     */
    void setAddress(long Address) throws IOException {
	long newOffset = mapper.getOffset(Address);
	setPos(newOffset);
    }
    /**
     * read an address offset from the current offset
     */
    long readAddressOffset() throws IOException {
	/* The difficulty is that the offset could be negative so
	 * we need to take account of the word length
	 */
	long offset = readBytes(wordLength.size);
	if(wordLength.size == 4) {
	    int off = (int)offset;
	    DvUtils.trace("Offset is l:" + offset + " i:" + off, 2, false);
	    offset = (long)off;
	    DvUtils.trace("Offset is l:" + offset, 2, false);
	}
	return(offset);
    }
    /**
    /**
     * read an address value from the current offset
     */
    long readAddress() throws IOException {
	return(readBytes(wordLength.size));
    }
    /**
     * reads an address value from the offset in the buffer
     */
    long getAddr(byte[] b, int offset) {
	byte[] b2 = new byte[wordLength.size];
	for(int i=0; i<wordLength.size; i++) {
	    b2[i] = b[offset+i];
	}
	return(endType.toLong(b2));
    }
}
