/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvFindInMemory.java, ras, dumpviewer 1.1
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin   Description
 * ------  ----   ------   ----------------------------------------------------
 * 36524  180502  mpartrid New for feature 36524

 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION: DvFindInMemory class
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.event.*;
import javax.swing.*;

 /**
  *  Description: This class creates a window and allows the user to enter
  *               the text they wish to search for in memory. This window
  *               contains a JTabbed pane. The first panel allows the search
  *               string to be entered and the second panel allows the the user
  *               to specify aditional options for the search. If the search is
  *               successful a new window is launched to display the results.
  *
  */

public class DvFindInMemory implements ActionListener, KeyListener  {

    Dvifm theIFM;              // the internal frame manager object
    JInternalFrame jif;        // the internal frame we are working in.
    DvConsole dvConsole;       // the dvconsole object to get the memory.
    DvMemoryFind dvMemoryFind; // the parameters for Find.

    JTextField findText;        // Field containing search text.
    JTextField startField;      // Field contaning start address.
    JTextField finishField;     // Field containing end address.
    JTextField boundaryField;   // Field containing byte boundary.
    JTextField bytesField;      // Field containing number of bytes to display
    JTextField occurenceField;  // Field containing the number of occurrences to display

    JButton findButton = new JButton("Find");
    JButton cancelButton = new JButton("Cancel");

    /**
     *   <p>Description: Creates a DvFindInMemory Object
     *   </p>
     *   <p>Input      : Internal frame to use, a DvConsole object
     *                   and a DvMemoryFind Object.
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    public DvFindInMemory(JInternalFrame thejif, DvConsole theDvConsole, DvMemoryFind theDvMemoryFind)
    {
        theIFM=DvUtils.getTheIFM();

        // Save the parameters away in there appropriate class variables.

        jif = thejif;
        dvConsole = theDvConsole;
        dvMemoryFind = theDvMemoryFind;

        constructWindow();
    }

    /**
     *   <p>Description: Constructs the FindInMemory Window. which is made up
     *                   up of a tabbed pane consisting of two panels, Search
     *                   and Options. This is thne added to a panel in the
     *                   centre of the window and some buttons are added to
     *                   a panel at the bottom of the window.
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    private void constructWindow()
    {
        JPanel mainPanel = new JPanel();
        JPanel buttonPanel = new JPanel();
        JTabbedPane findPane = new JTabbedPane();
        JPanel findPanel;
        JPanel optionsPanel;

        findButton = new JButton("Find");
        cancelButton = new JButton("Cancel");

        findButton.addActionListener(this);
        cancelButton.addActionListener(this);

        mainPanel.setLayout( new GridLayout(0,1) );

        findPanel = constructFindPanel();
        findPane.addTab("Find", findPanel);

        optionsPanel = constructOptionsPanel();
        findPane.addTab("Options", optionsPanel);

        findPane.setSelectedIndex(0);       
 
        buttonPanel.add(findButton);
        buttonPanel.add(cancelButton);
        
        mainPanel.add(findPane);
        theIFM.addContentToFrame(jif, mainPanel, buttonPanel);

        // Request that the intitial focus is set to the find field.

        findText.requestFocus();
        
    }

    /**
     *   <p>Description: Constructs the find panel, which consists of a label
     *                   and text field. Thes fields are added to there own panels
     *                   with a  FLowlayout manager,s o tha the size can be controlled.
     *                   They are then added to the main findPanel, which uses a
     *                   BorderLayout.
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : The completed find panel.
     *   </p>
     
     */
    private JPanel constructFindPanel()
    {
        JPanel findPanel = new JPanel();
        
        findPanel.setLayout( new BorderLayout() );
        findPanel.setBorder(BorderFactory.createEmptyBorder(50, 30, 50, 30));

        JLabel labelField = new JLabel("Search For:");
        JPanel labelPanel= new JPanel();
        labelPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 10));

        labelPanel.add(labelField);
        findPanel.add(labelPanel, BorderLayout.WEST);

        
        findText = new JTextField(dvMemoryFind.findText, 10);

        findText.addKeyListener(this);

        JPanel fieldPanel = new JPanel();
        fieldPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 10));

        fieldPanel.add(findText);

        findPanel.add(fieldPanel, BorderLayout.EAST);

        return findPanel;
    }

    /**
     *   <p>Description: Constructs the options panel. It consists of two panels
     *                   with grid layouts to hold the labels and fields. These
     *                   are then added to another panel which uses a
     *                   BorderLayout as Flow manager.
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : The constructed options panel.
     *   </p>
     */

    private JPanel constructOptionsPanel()
    {
        JPanel optionsPanel = new JPanel();

        optionsPanel.setLayout(new BorderLayout());
        optionsPanel.setBorder(BorderFactory.createEmptyBorder(25, 30, 25, 30));

        // Labels identifying the text fields.

        JLabel startLabel = new JLabel( "Start Address" );
        JLabel finishLabel = new JLabel( "Finish Address");
        JLabel boundaryLabel = new JLabel( "Byte Boundary" );
        JLabel bytesLabel = new JLabel( "Display Bytes" );
        JLabel occurenceLabel = new JLabel( "Occurences" );

        JPanel labelPanel= new JPanel();
        labelPanel.setLayout(new GridLayout(0, 1, 3, 3));

        labelPanel.add(startLabel);
        labelPanel.add(finishLabel);
        labelPanel.add(boundaryLabel);
        labelPanel.add(bytesLabel);
        labelPanel.add(occurenceLabel);

        optionsPanel.add(labelPanel, BorderLayout.WEST);

        // Text fields

        startField= new JTextField(dvMemoryFind.startAddress, 10);
        finishField = new JTextField(dvMemoryFind.finishAddress, 10);
        boundaryField = new JTextField(Integer.toString(dvMemoryFind.byteBoundary), 10);
        bytesField = new JTextField(Integer.toString(dvMemoryFind.displayBytes), 10);
        occurenceField = new JTextField(Integer.toString(dvMemoryFind.occurences), 10);

        // Add key listener to text fields.

        startField.addKeyListener(this);
        finishField.addKeyListener(this);
        boundaryField.addKeyListener(this);
        bytesField.addKeyListener(this);
        occurenceField.addKeyListener(this);

        JPanel fieldPanel = new JPanel();
        fieldPanel.setLayout(new GridLayout(0, 1, 3, 3));

        fieldPanel.add(startField);
        fieldPanel.add(finishField);
        fieldPanel.add(boundaryField);
        fieldPanel.add(bytesField);
        fieldPanel.add(occurenceField);

        optionsPanel.add(fieldPanel, BorderLayout.EAST);

        return optionsPanel;
    }

    /**
     *   <p>Description: In response to the user requesting a find to be started 
     *                   this routine intialises the dvMemoryfind object and then
     *                   creates a new internal frame for the results and generates
     *                   the results via a DvMemoryFindResults object.
     *                   pressing one of the scroll buttons.
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    public void doFindInMemory()
    {
        // Setup the values of dvMemoryFindClass.

        dvMemoryFind.findText      = findText.getText();
        dvMemoryFind.startAddress  = startField.getText();
        dvMemoryFind.finishAddress = finishField.getText();
        dvMemoryFind.displayBytes  = Integer.valueOf(bytesField.getText()).intValue();
        dvMemoryFind.byteBoundary  = Integer.valueOf(boundaryField.getText()).intValue();
        dvMemoryFind.occurences    = Integer.valueOf(occurenceField.getText()).intValue();

        // Create a new interanl frame for the results and add it to our desktop.

        Dvifm theIFM = DvUtils.getTheIFM();
        JInternalFrame resultJif = theIFM.createNewFrame(
                                           "Memory Find - " + dvMemoryFind.findText,
                                           theIFM.getDefaultProperties(),
                                           DvConstants.FRAMETYPE_IS_MS);
        Dumpviewer.addIFrameToDesktop(resultJif,325,470,300,75);

        // Now generate and display the results.

        DvMemoryFindResults MemoryFindResults = new DvMemoryFindResults( 
                                                        resultJif, dvConsole, dvMemoryFind ); 

        MemoryFindResults.DoFindResults();
    }

    /**
     *   <p>Description:  This routine checks that all the data entered
     *                    for the 'find' by the user is valid.
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : True if data valid False otherwise
     *   </p>
     */

    public boolean checkValidFields()
    {
        boolean retval = true;  // Value returned by the function.
        String findFieldText   = findText.getText();;
        String startFieldText  = startField.getText();
        String finishFieldText = finishField.getText();
        int    displayBytesField = Integer.valueOf(bytesField.getText()).intValue();
        int    bytesBoundaryField = Integer.valueOf(boundaryField.getText()).intValue();
        int    occurencesField = Integer.valueOf(occurenceField.getText()).intValue();
        long   startAddress = -1;
        long   endAddress = -1;


        DvUtils.writetoTrace( "Find " + findText.getText() );

        // Check to make sure start field is not blank, before converting address to a long.

        if ( !startFieldText.equals("") ) {

            try {
                startAddress = DvUtils.hexToLong(startFieldText);
            }
            catch ( NumberFormatException e ) {
                retval = false;
                JOptionPane.showMessageDialog(null, "Invalid Start Address" );
                startField.requestFocus();
                return retval;
            }
            DvUtils.writetoTrace( "StartAddress = " + startAddress );
        }
        
        // Check to make sure finish field is not blank, before converting address to a long.

        if ( !finishFieldText.equals("") ) {

            try {
                endAddress = DvUtils.hexToLong(finishFieldText);
            }
            catch ( NumberFormatException e ) {
                retval = false;
                JOptionPane.showMessageDialog(null, "Invalid finish Address" );
                finishField.requestFocus();
                return retval;
            }
            DvUtils.writetoTrace( "endAddress = " + endAddress );
        }

        if ( findText.getText().equals("") ) {
            retval = false;
            JOptionPane.showMessageDialog(null, "A search string must be entered." );
            findText.requestFocus();
        } else if ( ( startAddress != -1 ) && ( endAddress != -1 ) ) {
            if ( startAddress > endAddress ){
                retval = false;
                JOptionPane.showMessageDialog(null, "Start Address must be less than finish Address" );
                startField.requestFocus();
            }
        } else if ( ( bytesBoundaryField < 1 ) || ( bytesBoundaryField > 100 ) ) {
           retval = false;
           JOptionPane.showMessageDialog(null, "Boundary must be in the range 1 to 100" );
           boundaryField.requestFocus();
        } else if ( displayBytesField < 1 ) {
           retval = false;
           JOptionPane.showMessageDialog(null, "Display bytes must be greater than 1" );
           bytesField.requestFocus();
        } else if (( occurencesField < 1 ) || ( occurencesField > 32767 )) {
            retval = false;
            JOptionPane.showMessageDialog(null, "Occurrences must be in the range 0 to 32767 " );
            occurenceField.requestFocus();
        }

        return retval;
    }

    /**
     *   <p>Description: Handles one of the buttons pressed
     *      by the user
     *   </p>
     *   <p>Input      : Action event
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    public void actionPerformed(ActionEvent e) {

        if( e.getSource() == findButton ){

            DvUtils.writetoTrace( "Find Button action in FindInMemory" );

            // Check that valid data has been entered for the search,
            // before doing the find.

            if ( checkValidFields() == true ) {
                doFindInMemory();
            }


        } else if ( e.getSource() == cancelButton) {

            DvUtils.writetoTrace( "Cancel Button action in FindInMemory" );

            // Close the 'find' window.

            jif.setVisible(false);
        }
    }

    /**
     *   <p>Description: Handle the key pressed event from the text field.
     *   </p>
     *   <p>Input      : Key event
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */
    
    public void keyTyped(KeyEvent e) {
    }

    /**
     *   <p>Description: Handle the key pressed event from the text field.
     *                   Treats it the same as the 'Find' button being
     *                   pressed by the user
     *   </p>
     *   <p>Input      : Key event
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */
    
    public void keyPressed(KeyEvent e) {
        int keyCode = e.getKeyCode();

        if (keyCode == KeyEvent.VK_ENTER ) {
            DvUtils.writetoTrace( "Enter Key pressed in findText field" );
            if ( checkValidFields() == true ) {
                doFindInMemory();
            }
        }
    }

    /**
     *   <p>Description: Handle the key released event from a text field.
     *   </p>
     *   <p>Input      : Key event
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    public void keyReleased(KeyEvent e) {
    }
}

