/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvMemoryFindResults.java, ras, dumpviewer 1.1
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin   Description
 * ------  ----   ------   ----------------------------------------------------
 * 36524  020502  mpartrid New for feature 36524
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.JDialog;

 /**
  *  Description: This class creates a window and displays the results
  *               in it for the current find settings. It also allows the
  *               user to issue a find next command,based on these settings
  *
  */
public class DvMemoryFindResults implements ActionListener  {

    Dvifm theIFM;          // the internal frame manager object
    JInternalFrame jif;    // the internal frame we are working in.
    DvConsole dvConsole;   // the dvconsole object to get the memory.
    DvMemoryFind dvMemoryFind; // The parameters to Find results for.
    int occurrencesToFind;      // The number of occurrences of the string to find.
                                // Initially this will be the number specified, but
                                // if 'find next' is selected it will be reset to 1.

    String findString;      // String containing find text for DvConsole.
    JTextArea jtext;        // the windows text area.
    String resultsAsString; // Find results to display, stored as a string.

    static String waitPrompt = "\nSearching....results will appear shortly....";

    Font thisFont;
    JPanel panel;          // Main panel to hold other panels

    JButton findNextBtn;
    JButton cancelBtn;

    /* Constant definitions. */


    /**
     *   <p>Description: Creates a DvMemoryFindResults Object
     *   </p>
     *   <p>Input      : Internal frame to use, a DvConsole object
     *                 and a DvMemoryFind Object.
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    public DvMemoryFindResults(JInternalFrame thejif, DvConsole theDvConsole, DvMemoryFind theDvMemoryFind)
    {
        theIFM=DvUtils.getTheIFM();

        // Save the parameters away in there appropriate class variables

        jif = thejif;
        dvConsole = theDvConsole;
        dvMemoryFind = theDvMemoryFind;

        constructWindow();

        // Construct the find command.


        // Determine wether we are looking for a string or a pointer,
        // by trying to format the 'search' string and if it is an
        // invalid number format, then assume we are searching for
        // a string.

        findString = "FINDPTR ";

        try {
             Long.decode( dvMemoryFind.findText );
        }
        catch ( NumberFormatException e ) {
            findString = "FIND ";
        }
        

        findString += dvMemoryFind.findText + ",";
        findString += dvMemoryFind.startAddress + ",";
        findString += dvMemoryFind.finishAddress + ",";
        findString += dvMemoryFind.byteBoundary + ",";
        findString += dvMemoryFind.displayBytes + ",";
        findString += dvMemoryFind.occurences;

        occurrencesToFind = dvMemoryFind.occurences;
    }

    public void DoFindResults()
    {
        final SwingWorker worker = new SwingWorker() {

            boolean resultFound = false;
        
            public Object construct() {
            
                Object dummy = null;    // dummy object used as return value.

                resultFound = getFindResults(findString);

                return dummy;           // return value not used by this program.
            }
        
            // Runs on the event-dispatching thread.

            public void finished() {

                displayResults( resultFound );
            }
        };

        worker.start();  // starts the seperate worker thread.

    }

    /**
     *   <p>Description: Sets up all the window components
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    private void constructWindow()
    {
        JPanel buttonPanel = new JPanel();
        thisFont = new Font("Monospaced", Font.PLAIN, 12);

        jtext =  new JTextArea("",8,25);
        jtext.setEditable(false); // This window is just for info.
        jtext.setFont(thisFont);

        theIFM.addContentToFrame(jif, jtext, buttonPanel);

        findNextBtn     = new JButton( "Find Next" );
        cancelBtn       = new JButton( "Cancel" );

        buttonPanel.add(findNextBtn);
        buttonPanel.add(cancelBtn);

        // Add tooltips for buttons.

        findNextBtn.setToolTipText("Click this button to find Next occurence.");
        cancelBtn.setToolTipText("Click this button to cancel Find.");

        // Add action listener for all buttons.

        findNextBtn.addActionListener(this);
        cancelBtn.addActionListener(this);

        jtext.setText(waitPrompt);
    }

    /**
     *   <p>Description: Gets the results for the find options from the DvConsole.
     *   </p>
     *   <p>Input      : String used to find the results, can either be a valid find
     *                 command or a find next command.
     *   </p>
     *   <p>Returned   : True if text found, false otherwise
     *   </p>
     */

    private boolean getFindResults(String findString)
    {
        resultsAsString = "";     // Reset results string.
        StringBuffer tempBuf;     // Temporary string buffer.
        boolean resultFound = false;

        DvUtils.writetoTrace( "Find string = " + findString );

        dvConsole.execValidCommand( findString );

        // We need to wait until the info is delivered....

        Dumpviewer.waitingOnConsoleCommand = true;

        while (true == Dumpviewer.waitingOnConsoleCommand) {

            try {
                Thread.sleep(100);
            }
            catch  (java.lang.InterruptedException e) {
                System.out.println("\nInterruptedException");
            }
        }

        // Now we need to make sure it is a Vector, because it could be an error string.

        if (Dumpviewer.savedConsoleOutput instanceof Vector) {

            Vector consoleOutput = (Vector) Dumpviewer.savedConsoleOutput;

            for (int i=0; i< consoleOutput.size(); i++) {
                System.out.println("\nLine  " + i + " " + consoleOutput.get(i));
            }

            // Check to make sure the find was successfull. If we have less than
            // 2 lines/2 vector entries, we know we the search has failed

            if ( consoleOutput.size() > 2 ) {

                resultFound = true;

                // One of the vector entries stored in savedConsoleOutput, contains
                // another vector which contains the memory vector. It's location will depend
                // upon the number of occurrences of the 'string' we have asked to find.
                // If only one occurrence is requested it will be entry '1' in the vector,
                // otherwise it will be stored entry occurences + 1.

                Vector memory = (Vector)consoleOutput.get(( occurrencesToFind > 1 ?
                                        occurrencesToFind + 1 : occurrencesToFind ));

                // Format the memory vector, with a '\n on the end of each line.

                for (int i=0; i< memory.size(); i++) {
                    System.out.println("\nLine  " + i + " " + memory.get(i));
                    resultsAsString+=(memory.get(i)+"\n");
                }

                // Use a string Buffer to remove the first and last characters which
                // are carriage returns.

                tempBuf = new StringBuffer( resultsAsString );
                tempBuf.deleteCharAt( tempBuf.length() - 1);
                tempBuf.deleteCharAt( 0 );
                resultsAsString = tempBuf.toString();

                // If the number of occurences to find are more than 1, we want to add a
                // blank line and then write out all the addresses of these occurrences.

                if (occurrencesToFind > 1) {

                    resultsAsString+="\n";

                    for (int i=1; i<= occurrencesToFind; i++) {
                        System.out.println("\nLine  " + i + " " + consoleOutput.get(i));
                        resultsAsString+=("\n" + consoleOutput.get(i));
                    }
                }
            }
            else {
                // The 'find' failed, so we should display the first vector entry to 
                // inform the user of this.

                resultFound = false;
                resultsAsString = (consoleOutput.get(0)+"\n");
            }
        } else if ( Dumpviewer.savedConsoleOutput instanceof String ) {
            // The 'find' failed, so we should display the first vector entry to 
            // inform the user of this.

            resultFound = false;
            resultsAsString = Dumpviewer.savedConsoleOutput +"\n";

        } else {
            // Unexpected object returned by Console.

            resultFound = false;
            resultsAsString = "Unexpected response from console\n";
        }

        return resultFound;
    }

    /**
     *   <p>Description: Displays the results for the current find command
     *                   held in currentAddr.
     *   </p>
     *   <p>Input      : None
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    private void displayResults(boolean resultFound)
    {
        // Display the current 'find' results.

        jtext.setText( resultsAsString );

        //  Disable the Find next button if result not found

        if ( resultFound == false ) {
            findNextBtn.setEnabled(false);
        }
    }

    /**
     *   <p>Description: Handles the find next and cancel commands.
     *   </p>
     *   <p>Input      : Action event
     *   </p>
     *   <p>Returned   : None
     *   </p>
     */

    public void actionPerformed(ActionEvent e) {

        // Check which button has been pressed.

        if( e.getSource() == findNextBtn ){
            DvUtils.writetoTrace( "findNextBtn pressed in FindResults" );

            // Reset the occurrences to find back to 1 for 'find next'

            jtext.setText(waitPrompt);
            occurrencesToFind = 1;

            // Issue the find next command and display the results.

            findString = "FN";
            DoFindResults();
        }
        else if ( e.getSource() == cancelBtn) {
            DvUtils.writetoTrace( "cancelBtn pressed in FindResults" );

            // Close the 'find results' window.

            jif.setVisible(false);
        }
    }
}



