/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvMemoryScroller.java, ras, dumpviewer 1.6
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502 PHILR:  New for feature 36524

 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

 /**
  *  Description: This class creates a window and displays memory
  *               details in it. It also allows the user to scroll
  *               up and down through a memory range via line up,
  *               half page up, page up, line down, half page down
  *               and page down buttons
  *
  */
public class DvMemoryScroller implements ActionListener  {

  long current; // the currently loaded piece (which is between top and top + len)

  Dvifm theIFM;          // the internal frame manager object
  JInternalFrame jif;    // the internal frame we are working in.
  DvConsole dvConsole;   // the dvconsole object to get the memory.
  DvAddress currentAddr; // Current start address of memory.

  long startMemoryAddr;  // The start of memory.
  long MemorySize;       // Size of the memory block.

  JTextArea jtext;       // the windows text area.
  String memoryAsString; // Memory to display, storeda s a string.

  Font thisFont;
  JPanel panel;          // Main panel to hold other panels

  JPanel upButtonPanel;  // Panel to hold up buttons.       
  JPanel downButtonPanel;// Panel to hold down buttons

  JButton oneLineUpBtn;
  JButton halfPageUpBtn;
  JButton onePageUpBtn;
  JButton oneLineDownBtn;
  JButton halfPageDownBtn;
  JButton onePageDownBtn;

  /* Constant definitions. */

  public static final int ONE_LINE = 16;
  public static final int HALF_PAGE = 256;
  public static final int ONE_PAGE = 512;

  /**
   *   <p>Description: Creates a DvMemoryScroller Object
   *   </p>
   *   <p>Input      : Internal frame to use, a DvConsole, DvAddress for memory block
   *                   length of memory block
   *   </p>
   *   <p>Returned   : None
   *   </p>
   */

  public DvMemoryScroller(JInternalFrame thejif, DvConsole theDvConsole, DvAddress addr, long length)
  {
    theIFM=DvUtils.getTheIFM();

    // Save the parameters away in there appropriate class vari

    jif = thejif;
    dvConsole = theDvConsole;
    currentAddr = addr;

    // Save the start memory address and length, so we can check any scrolling done
    // is in the correct range.

    startMemoryAddr = addr.getAddressAsLong();

    MemorySize = length;

    constructWindow();

    getMemoryDetails(currentAddr);

    displayMemory();
  }

  /**
   *   <p>Description: Sets up all the window components
   *   </p>
   *   <p>Input      : None
   *   </p>
   *   <p>Returned   : None
   *   </p>
   */

  private void constructWindow()
  {
      thisFont = new Font("Monospaced", Font.PLAIN, 12);

      jtext =  new JTextArea("",8,25);
      jtext.setEditable(false); // This window is just for info.
      jtext.setFont(thisFont);

      panel = new JPanel();
      panel.setLayout( new BorderLayout() );
      theIFM.addContentToFrame(jif, jtext, panel);

      upButtonPanel = new JPanel();
      downButtonPanel = new JPanel();

      oneLineUpBtn    = new JButton( ">" );
      halfPageUpBtn   = new JButton( ">>" );
      onePageUpBtn    = new JButton( ">>>" );
      oneLineDownBtn  = new JButton( "<" );
      halfPageDownBtn = new JButton( "<<" );
      onePageDownBtn  = new JButton( "<<<" );

      // Add tooltips for all buttons.

      oneLineUpBtn.setToolTipText("Click this button to move one line up.");
      halfPageUpBtn.setToolTipText("Click this button to move half a page up.");
      onePageUpBtn.setToolTipText("Click this button to move one page up.");
      oneLineDownBtn.setToolTipText("Click this button to move one line down.");
      halfPageDownBtn.setToolTipText("Click this button to move half a page down.");
      onePageDownBtn.setToolTipText("Click this button to move one a page down.");

      // Add action listener for all buttons.

      oneLineUpBtn.addActionListener(this);
      halfPageUpBtn.addActionListener(this);
      onePageUpBtn.addActionListener(this);
      oneLineDownBtn.addActionListener(this);
      halfPageDownBtn.addActionListener(this);
      onePageDownBtn.addActionListener(this);

      upButtonPanel.add(oneLineUpBtn);
      upButtonPanel.add(halfPageUpBtn);
      upButtonPanel.add(onePageUpBtn);

      downButtonPanel.add(onePageDownBtn);
      downButtonPanel.add(halfPageDownBtn);
      downButtonPanel.add(oneLineDownBtn);

      panel.add( upButtonPanel, BorderLayout.EAST );
      panel.add( downButtonPanel, BorderLayout.WEST );
  }

  /**
   *   <p>Description: Gets the memory details from the DvConsole.
   *   </p>
   *   <p>Input      : Dvaddress for the required memory address
   *   </p>
   *   <p>Returned   : None
   *   </p>
   */

  private void getMemoryDetails( DvAddress addr )
  {
      memoryAsString = "";      // Reset memory string.
      StringBuffer tempBuf;     // Temporary string buffer.

      dvConsole.execValidCommand("DIS MEM(" + addr.convertToHex() + "," + ONE_PAGE + ")" );

      // we need to wait until the info is delivered....

      Dumpviewer.waitingOnConsoleCommand = true;

      while (true == Dumpviewer.waitingOnConsoleCommand) {

        try {
         Thread.sleep(100);
        }
        catch  (java.lang.InterruptedException e) {
         System.out.println("\nInterruptedException");
        }
      }

      Vector v = (Vector) Dumpviewer.savedConsoleOutput;

      for (int i=0; i< v.size(); i++) {
        memoryAsString+=(v.get(i)+"\n");
      }

      // Use a string Buffer to remove the first and last characters which
      // are carriage returns.

      tempBuf = new StringBuffer( memoryAsString );
      tempBuf.deleteCharAt( tempBuf.length() - 1);
      tempBuf.deleteCharAt( 0 );
      memoryAsString = tempBuf.toString();

  }

  /**
   *   <p>Description: Displays the memory contents for the current address
   *                   held in currentAddr.
   *   </p>
   *   <p>Input      : None
   *   </p>
   *   <p>Returned   : None
   *   </p>
   */

  private void displayMemory()
  {
      // Set the title, to include the current address start.

      jif.setTitle("Memory Scroller " + currentAddr.convertToHex() );

      // Display the current memory range.

      jtext.setText( memoryAsString );
  }

  /**
   *   <p>Description: Handles the scrolling of memory in response to the user
   *                   pressing one of the scroll buttons.
   *   </p>
   *   <p>Input      : Action event
   *   </p>
   *   <p>Returned   : None
   *   </p>
   */

  public void actionPerformed(ActionEvent e) {

      long memoryOffset = 0;    // Amount of memory to scroll by.
      long memoryAddress;       // Start address of current memory.
      long newAddress;          // Start of the new address.

      // Check which button has been pressed.

      if( e.getSource() == oneLineUpBtn ){
          DvUtils.writetoTrace( "OneLineUpBtn pressed in Memory Scroller" );
          memoryOffset -= ONE_LINE;
      }
      else if ( e.getSource() == halfPageUpBtn) {
          DvUtils.writetoTrace( "halfPageUpBtn pressed in Memory Scroller" );
          memoryOffset -= HALF_PAGE;
      }
      else if ( e.getSource() == onePageUpBtn) {
          DvUtils.writetoTrace( "onePageUpBtn pressed in Memory Scroller" );
          memoryOffset -= ONE_PAGE;
      }
      else if ( e.getSource() == oneLineDownBtn) {
          DvUtils.writetoTrace( "oneLineDownBtn pressed in Memory Scroller" );
          memoryOffset += ONE_LINE;
      }
      else if ( e.getSource() == halfPageDownBtn) {
          DvUtils.writetoTrace( "halfPageDownBtn pressed in Memory Scroller" );
          memoryOffset += HALF_PAGE;
      }
      else if ( e.getSource() == onePageDownBtn) {
          DvUtils.writetoTrace( "onePageDownBtn pressed in Memory Scroller" );
          memoryOffset += ONE_PAGE;
      }
      else {
          DvUtils.writetoTrace( "Non Button action in Memory Scroller" );
          return;
      }

      memoryAddress = currentAddr.getAddressAsLong();
      newAddress = memoryAddress + memoryOffset;

      // Check to make sure that the new offset is greater than the start of the
      // memory block and less than the end of the memory block.

      if ( newAddress < startMemoryAddr ) {
          DvUtils.writetoTrace( "New Address before memory start address" );
      }
      else if ( newAddress > ( startMemoryAddr + MemorySize ) ) {
          DvUtils.writetoTrace( "New Address beyond memory block" );
      } else {

          // We have a valid address, so we can update currentAddr

          try {
             currentAddr.offsetBy( memoryOffset );
          } 
          catch(DvAddressException exception ) {
             DvUtils.writetoTrace( "Address out of Range" );
          }

          DvUtils.writetoTrace( "Memory Start = " + currentAddr.convertToHex() );

          /* Retrieve the new memory range and display it. */

          getMemoryDetails(currentAddr);
          displayMemory();
      }

   }
}


