/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvNamesIndexFile.java, ras, dumpviewer 1.5
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  190702  philr   New for feature 36524
 * 36524  060802  philr   Sort the summary information
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;
import java.io.*;
import java.util.*;




 /**
  *  This file supports a database of names, linked to longs (which are object
  *  references in the case of SDFF.
  *
  *
  *
  *
  */
public class DvNamesIndexFile extends RandomAccessFile {


    private static int ptrLength;  // set to 4/8 depending on bitness
    private static String file_name;
    private static IndexHeader IH;
    private static RandomAccessFile self;
    private static long [] indexArray;
    private static long next_record_pos=0;
    private static int index_table_size=0;
    private long saved_next_position=0;


    /**
     *  Constructor
     */
    public DvNamesIndexFile (String file_name,
                        String mode,
                        int index_table_size,
                        int record_size,
                        int key_size,
                        long magic)
            throws java.io.IOException{

        super(file_name,mode);
        self = this;
        initPtrLength();


        if (mode.equals("r")) {
            // read existing control information from the file
            IH = IndexHeader.readExistingHeader();
            //checkIntegrity();
            if (IH.magic == magic) {
                // System.err.println("Magic numbers check out");
                // Need to rebuild indexArray
                seek(IH.index_ptr);
                indexArray = new long[IH.index_table_size];
                for (int i=0;i<IH.index_table_size;i++) {
                    indexArray[i]=readLong();
                }
            }
            else {
                throw new IOException("Magic numbers mismatch");
            }


        }
        else {

            initTheIndex(index_table_size,
                         record_size, key_size,
                         magic);
            //this.index_table_size = index_table_size;
        }


    } //end Constructor

   /**
     *  Write a record to this file
     */
    public boolean write(String key,long data,short heapId,long objLen) {

        return(write(key.getBytes(),data,heapId,objLen));
    }

    /**
     *  Write a record to this file
     */
    public boolean write(byte[] key,long data,short heapId,long objLen) {
        // we know we need a number between 0 and index_table_size-2
        // so we get this by taking the key byte array and summing the
        // individual bytes and then getting remainder when we divide
        // by (index_table_size-1) - we further "complicate it" (in order
        // to give bigger numbers by multiplying the first four bytes
        // in the key together (unless the value is 0). We reserve
        // array position (index_table_size-1) for those records where
        // the key is such that its size is small!!!
        long key_total=hashTheKey(key);
        int  array_index=0;

        //System.out.println("key_total is :"+key_total);
        // System.out.println("index_table_size"+IH.index_table_size);
        array_index = (int) (key_total%IH.index_table_size);
        if (array_index < 0) {
            System.out.println("array index is "+array_index);
        }
        // Now convert  this into an offset into the index

        long pos = IH.index_ptr+8*array_index;
        if (pos < 0) {
            System.out.println("pos is "+pos);
        }
        try {
            seek(pos);
            long x = readLong();
            if (x < -1) {
                System.out.println("x is "+x +"@"+pos+":"+IH.index_ptr+":"+array_index);
            }


            if (-1 == x) {
                // this is a new one - Don't forget to copy it to indexArray in memory

                indexArray[array_index] = writeNewKeyRecord(pos,key,data,objLen); // 0 means new chain start
                // Don't forget to copy it to indexArray in memory

            }
            else {
                updateExisting(pos,key,x,data,objLen);
            }


        }
        catch (IOException e) {
            String s = new String(key);
            System.out.println("IOException "+pos+" "+s);
        }

        //String s = new String(key);
        // System.out.println("Index is :"+array_index+ " "+s);

        return true;
    }

    /**
     *  write a new key record
     */
    private long writeNewKeyRecord(long pos,byte[] key,long data,
                                    long objLen) {
        long retVal=0;
        try {
            seek(length()); // go to end of file
            long saved_pos = getFilePointer();
            writeByte((byte)'K');
            writeByte((byte)'e');
            writeByte((byte)'y');
            writeByte((byte)':');
            writeByte((byte)'<');
            writeInt(key.length);
            writeInt(1); // count of items with this key
            //String s = new String(key);


            writeLong(objLen); // size of all objects for this name
            writeLong(0); // next one with this index in array index

            long saved_data_position = getFilePointer();
            writeLong(0); // pointer to data record associated with this key
            long saved_lastDRAptr_position = getFilePointer();
            writeLong(0); // lastDRA will be stored here
            write(key);
            writeByte((byte)'>');
            if (0 != pos) {
                // This is for chaining, basically if this is not the
                // first record in the chain we update the previous one with
                // this records offset in the file
                seek(pos);
                writeLong(saved_pos);
            }
            retVal = saved_pos;
            long datapos = addInstanceAddress(0,data,saved_lastDRAptr_position);

            seek(saved_data_position);
            writeLong(datapos);



       }
       catch (IOException e) {
        System.err.println("IOException in writeNewKeyRecord");
       }

       return retVal;
    }

    /**
     *  Shows all the classes and ther counts
     */
    public Vector summary() {

        Vector rV = new Vector();
        int total_count = 0;
        int total_bytes = 0;
        for (int i=0;i<IH.index_table_size;i++) {
            if (-1 == indexArray[i]) {
               // No one there
            }
            else {
                // go and process the names under here
                //System.out.println("Its "+ indexArray[i] + " (" + i +")");

                long ongoing = indexArray[i];
                while (ongoing != 0) {
                   // seek(ongoing);
                   // int instanceCount =
                   dataRecord record = new dataRecord(ongoing);
                   String sName = new String(record.name);
                   //long numBytes = 0;
                   //if (sName.startsWith("array ")) {
                        //numBytes = 0;
                        //  for arrays we have to go off to look at size of each
                        //  one individually......

                   //}
                   //else {
                   //     numBytes = DvClassCommands.getInstanceSize(sName,rV);
                   //     numBytes = numBytes*record.count;
                   //}
                   //rV.add(record.count+ " instances " +
                   //" (="+record.totalsize+" bytes) of " +sName + " found" );
                   rV.add(sName + "  : "+ record.count+ " instances " +
                            " (="+record.totalsize+" bytes)");
                   total_count += record.count;
                   total_bytes += record.totalsize;
                   ongoing = record.nextPtr;
                }

            }

        }

        // now sort thsi into an array of Strings

        String [] rString = new String[rV.size()+2];
        for (int i=0; i<rV.size();i++) {
            rString[i] = (String) rV.get(i);
        }

        rString[rV.size()] =  "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz";
        rString[rV.size()+1] = "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz";

        Arrays.sort(rString);

        rString[rV.size()] =  ("\n Total number of objects = "+  total_count);
        rString[rV.size()+1] =  ("\n Total byte count = "+ total_bytes);
        //rV.add("\n Total number of objects = "+  total_count);
        //rV.add("\n Total byte count = "+ total_bytes);

        rV.clear();

        for (int i=0;i<rString.length;i++) {
            rV.add(rString[i]);
        }

        return rV;

    }

    /**
     *  get an array of all the class names
     */
    public Vector getAllClassNames( ) {

        Vector rV =  new Vector();



        for (int i=0;i<IH.index_table_size;i++) {
            if (-1 == indexArray[i]) {
               // No one there
            }
            else {
                // go and process the names under here
                //System.out.println("Its "+ indexArray[i] + " (" + i +")");

                long ongoing = indexArray[i];
                while (ongoing != 0) {
                   // seek(ongoing);
                   // int instanceCount =
                   dataRecord record = new dataRecord(ongoing);
                   String sName = new String(record.name);
                   rV.add(sName);

                   ongoing = record.nextPtr;
                }

            }

        }
        return rV;

    }



    /**
     *  get a Vector of all the instances for a particular class
     */
    public long[]  getInstancesForClass(String className) {


        DvUtils.trace("Entry to getInstancesForClass: "+className,0,false);

         long[] rLongArray = null;
         byte[] key = className.getBytes();

         long key_total = hashTheKey(key);
         long ongoing = indexArray[(int) (key_total%IH.index_table_size)];
         boolean done = false;
         while (false == done && ongoing != 0 && -1 != ongoing) {

            // seek(ongoing);
            // int instanceCount =
            dataRecord record = new dataRecord(ongoing);
            String sName = new String(record.name);
            if (sName.equals(className)) {
                //System.err.println("Point A");
                // ok we are in business
                // get the first
                done = true; // so we can escape the loop
                rLongArray = new long[record.count];
                int foundSoFar = 0;
                long IRAddr = record.dataPtr;
                while (0 != IRAddr) {
                    instanceRecord IR = new instanceRecord(IRAddr);
                    for (int j=0;j<IR.count;j++) {
                        rLongArray[foundSoFar] = IR.instance[j];
                        foundSoFar++;
                    }
                    IRAddr = IR.nextIRptr;
                }

            }

            ongoing = record.nextPtr;
        }

        if (null != rLongArray) {
            DvUtils.trace("Exit from getInstancesForClass: "+rLongArray.length,0,false);
        }
        else {
            DvUtils.trace("Exit from getInstancesForClass: no instances found",0,false);
        }

        return rLongArray;
    }




    /**
     *  Update an existing key chain
     */
    public boolean updateExisting(long pos,byte[] key,long keypos,
                                    long instanceAddr,long objLen) {
        try {
            //System.out.println("In updateExisting");
            boolean foundit = false;
            boolean done = false;
            int checkitOutCnt = 0;

            // we loop down the chain until we find one that matches
            // our key - if we don't find one then we chain a new one
            while (false == done) {
                long new_keypos = checkItOut(keypos,key);
                checkitOutCnt++;
                //System.err.println("saved_next_position is:"+saved_next_position);
                if (new_keypos == keypos) {
                    done=true;
                    foundit= true;
                }
                if (0 == new_keypos) {
                    done=true;
                }
                if (false == done) {
                    keypos = new_keypos;
                }

            }

            if (true== foundit) {
                //if (checkitOutCnt > 1) {
                //    System.err.println("checkitOutCnt: "+checkitOutCnt+":"+ keypos);
                //}
                seek(keypos);
                byte[] eye = new byte[4];
                read(eye); // Should say Key: - sanity check
                //String s=new String(eye);
                //if (s.equals("Key:")) {
                    // looks good;
                readByte(); // ">"
                int keylen = key.length;

                int storedLen = readInt();

                //if (storedLen == keylen) {
                    // possibly a good one
                long saved_count_position = getFilePointer();
                int instances_count=readInt();
                long saved_totsize_position = getFilePointer();
                long total_size = readLong(); // totalsize
                seek(saved_totsize_position);
                writeLong(total_size + objLen);

                saved_next_position = getFilePointer();

                readLong();
                long saved_data_position = getFilePointer();
                long dataPos = readLong();   // this is pointer to a storage area for obj
                              // addresses - stored in blocks of 8.

                long saved_lastDRAptr_position = getFilePointer();


                seek(saved_count_position);
                        //System.out.println("At B ");
                int x = readInt()+1;
                seek(saved_count_position);
                        //System.out.println("At B ");
                writeInt(x);
                        //System.out.println("At C ");
                // saved_data_position will have a value since this is an update

                addInstanceAddress(saved_data_position,instanceAddr,
                                    saved_lastDRAptr_position);

            }
            else {
                // must be new one....
                //System.out.println("New one - !!!" );
                long newPos = writeNewKeyRecord(saved_next_position,key,
                                            instanceAddr,objLen); // 1 means new chain is ongoing
                                                             // rather than new

            }


        }
        catch (IOException e) {
            System.out.println(" IOError in updateExisting");
        }




        return true;
    }


    private long checkItOut(long keypos,byte[] key) {
        long retLong = 0;
        try {
            seek(keypos);
            byte[] eye = new byte[4];
            read(eye); // Should say Key: - sanity check
            String s=new String(eye);
            if (s.equals("Key:")) {
                // looks good;
                readByte(); // ">"
                int storedLen = readInt();

                int instances_count=readInt();
                readLong();  // this is total byte count
                            // only important for arrays
                saved_next_position = getFilePointer();
                long chain=readLong();

                readLong();
                readLong(); // lastDRAptr

                byte[] ba=new byte[storedLen];
                read(ba);
                String s1 = new String(key);
                String s2 = new String(ba);
                if (s1.equals(s2)) {
                    retLong = keypos;
                }
                else {
                    retLong = chain;
                    //System.err.println("Going down chain "+s1+":"+s2+":"+retLong+":"+
                    //                        saved_next_position);

                }
            }
            else {
                System.err.println("Bad format detected in checkItOut");
            }
        }
        catch (IOException e) {
            System.err.println("IOException in checkItOut");
        }

        return retLong ;
    }

    /**
     *  Add an instance Address to the list for this named class
     */
    private long addInstanceAddress(long dataRecord,long instanceAddr,
                                        long lastDRAptr) {
        // Here we have a block prefixed with Obj: followed by a
        // next free count, followed by 8 longs (which will hold object
        // addresses, followed by a continuation block address
        // Note lastDRA is maintained in order that adding a new data object
        // performs well....
        //System.out.println("Adding obj "+instanceAddr);
        long retLong = dataRecord;
        if (0 == dataRecord) {
            // this is first one for this class name
            //System.err.println("At point A");
            retLong = writeNewDataRecord( instanceAddr,lastDRAptr,0);


        }
        else {
            //System.err.println("At point B");
            try {
                seek(lastDRAptr);
                seek(readLong());
                long saved_pos =  getFilePointer();
                readInt(); // gets us past 4 byte eyecatcher
                long next_free_pos = getFilePointer();
                int nextFree = readInt();
                //if (lastDRAptr == 10030)
                //System.err.println("10030:" + nextFree+":"+next_free_pos+
                //                    ":"+saved_pos);

                if (nextFree<8) {
                    //System.err.println("10030:" + nextFree+":"+next_free_pos+
                    //                    ":"+saved_pos);
                    // have to simply figure out where to put next one
                    seek(saved_pos + 8*(nextFree+1));
                    writeLong(instanceAddr);
                    seek(next_free_pos);
                    nextFree++;
                    writeInt(nextFree);
                    //System.err.println("Wrote next free:"+next_free_pos+":"+
                    //                    nextFree);

                }
                else {
                    // else we have to create a new one.....
                    long chainPointerAddr = saved_pos + 72;
                    retLong = writeNewDataRecord( instanceAddr,lastDRAptr,
                                                  chainPointerAddr  );

                }
            }
            catch (IOException e) {
                System.err.println("IOException in addInstanceAddress");
            }

        }
        return retLong;
    }

    private long writeNewDataRecord(long instanceAddr,long lastDRAptr,
                                    long chainPtrAddress) {

        long retLong = 0;
        try {
            seek(length()); // go to end of file
            retLong = getFilePointer();
            writeByte((byte)'O');
            writeByte((byte)'b');
            writeByte((byte)'j');
            writeByte((byte)':');
            writeInt(1); // we will fill 0'th one here
            writeLong(instanceAddr);
            for (int i=0;i<8;i++) {
                writeLong(0);
            }
            seek(lastDRAptr);
            writeLong(retLong);
            if (0 != chainPtrAddress) {
                // if its not the first one then we are chaining
                seek(chainPtrAddress);
                writeLong(retLong);
            }
         }
         catch (IOException e) {
            System.err.println("IOException in writeNewDataRecord");
         }
         return retLong;
    }

    /**
     *  Initialise the index file
     */
    private static boolean initTheIndex (int index_table_size,
                                        int record_size,int key_size,
                                        long magic)
                                throws IOException  {

        // Need to create a new IndexHeader;
        IH = new IndexHeader(index_table_size,
                             record_size,
                             magic);
        indexArray = new long[index_table_size]; // keep it in core to avoid
                                                 // excessive IO
        for (int i=0;i<index_table_size;i++) {
            indexArray[i] = -1; // -1 means no record.
        }
        return true;
    }

    /**
     *  Get a long hash value from a key using our own algorithm
     */
    private long hashTheKey(byte[] key) {

        long key_total=0;
        if (key.length > 4) {

            for (int i=0;i<key.length;i++) {
                long temp = (long) key[i];
                 //System.out.println("temp is :"+temp);
                if (i>0 && i<4) {
                    key_total = key_total*temp;
                }
                else {
                    key_total = key_total+temp*i;

                }
            }
        }
        return key_total;
     }

    /**
     *  Internal class dealing with the index header
     */
    private static class IndexHeader  {

        static long magic;
        static int  index_table_size;
        static int  record_size;
        static int  key_size;
        static long index_ptr;

        private IndexHeader() {
        }

        private IndexHeader(int index_table_size,
                            int record_size,
                            long magic) throws IOException  {

            this.magic = magic;
            // we make index table size an interation of 256
            //System.out.println(" "+index_table_size);
            this.index_table_size = (index_table_size/256)*256;
            //System.out.println(" "+index_table_size);
            this.record_size = record_size;
            this.key_size = key_size;
            writeIH();
        }

        private  int loadIH() {

            return 0;
        }

        private static IndexHeader readExistingHeader() throws IOException {



            IndexHeader IH = new IndexHeader();
            self.seek(0); // go to start of file
            magic = self.readLong();
            index_table_size = self.readInt();

            record_size = self.readInt();
            key_size = self.readInt();
            self.readInt();
            index_ptr = self.readLong();
            //System.out.println("index_ptr: " + index_ptr);
            next_record_pos = self.length();

            return IH;
        }

        private boolean writeIH() throws IOException {
            self.writeLong(magic);
            self.writeInt(index_table_size);
            System.err.println("index_table_size:" + index_table_size);
            self.writeInt(record_size);
            self.writeInt(key_size);


            self.writeInt(0); // for extension purposes
            long ptr_to_index_ptr=self.getFilePointer();
            self.writeLong(0); // index_ptr goes here

            self.writeLong(0); // start of data records goes here
            self.writeLong(0); // for extension purposes
            self.writeLong(0); // for extension purposes
            self.writeLong(0); // for extension purposes

            for (int i=0;i<4;i++) {
                self.writeByte((byte)'i');
                self.writeByte((byte)'n');
                self.writeByte((byte)'d');
                self.writeByte((byte)'x');
            }
            index_ptr = self.getFilePointer();
            // Now we write a whole series of longs corresponding to the
            //index_table_size - all 0's
            for (int i=0;i<index_table_size;i++) {
                self.writeLong(-1); // for extension purposes

            }
            for (int i=0;i<4;i++) {
                self.writeByte((byte)'d');
                self.writeByte((byte)'a');
                self.writeByte((byte)'t');
                self.writeByte((byte)'a');
            }
            long start_of_data_records = self.getFilePointer();
            self.seek(ptr_to_index_ptr);
            self.writeLong(index_ptr); // index_ptr goes here
            self.writeLong(start_of_data_records);
            next_record_pos = start_of_data_records;



            return true;
        }
    }


    /**
     *  Internal class dealing with the data records
     */
    private static  class dataRecord {


           byte[] eye = new byte[5]; // should be Key:<
           int  length;
           int  count;
           long totalsize; // How many bytes does represented by all the objects!
           long nextPtr;
           long dataPtr;
           long lastDRA;
           byte[] name;




           private dataRecord(long where)   {

                try {
                    self.seek(where);
                    self.read(eye);
                    length=self.readInt();
                    count=self.readInt();
                    totalsize = self.readLong();

                    nextPtr = self.readLong();
                    dataPtr = self.readLong();
                    lastDRA  = self.readLong();
                    name = new byte[length];
                    self.read(name);

                }
                catch (IOException e) {
                    System.err.println("IOexception in dataRecord constructor");
                }

            }
     }


    /**
     *  Internal class dealing with the instanceRecords
     */
    private static  class instanceRecord {


           byte[] eye = new byte[4]; // should be obj:
           int  count; // how many instances recorded here (max 8)
           long instance[] = new long[8];
           long nextIRptr; // where the next IR is





           private instanceRecord(long where)   {

                try {
                    self.seek(where);
                    self.read(eye);
                    count=self.readInt();
                    for (int i=0;i<8;i++) {
                        instance[i]=self.readLong();
                    }

                    nextIRptr  = self.readLong();


                }
                catch (IOException e) {
                    System.err.println("IOexception in instanceRecord constructor");
                }

            }
    }

    private void initPtrLength() {
        // pointer length is used in several places -
        // we establish it here so that everyone can get at it
        DvDump theDump = DvConsole.getCurrentDump();
        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            ptrLength=8;

        }
        else
        {
            ptrLength=4;
        }
    }


}


