/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvObjectsCommands.java, ras, dumpviewer 1.61
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 *
 * (C) Copyright IBM Corp. 2002.
 *
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  270502  hdrc    New for feature 36524
 * 36524  160702  philr   Upgrade to object storage for performance
 * 36524  170702  hdrc    Inflated Monitor code added
 * 36524  230702  philr   Upgraded to use namesIndexFile
 * 36524  260702  jkbar3  Added Thread Local Heap Traversal
 * 36524  260702  philr   Scaffolding for object analysis and memory
 *                        requirements for large dumps
 * 36524  290702  philr   Addded in workdir support for where index files go
 * 36524  300702  philr   Fixed 64 bit support
 * 36524  300702  philr   TLH cache block address is 0 fix  (Courtesy jkbar3)
 * 36524  310702  hdrc    Added lkDumpRoutine for javacore plugin
 * 36524  310702  jkbar3  handle errors caused by null TLH cache blocks
 * 36524  050802  philr   remove need for address Index file (performance partly)
 *                        and use an allocbits model - also fill in heap name
 *                        by using range array.
 * 36524  060802  jkbar3  Check for execenv ZERO before processing TLH
 * 36524  060802  philr   Better display in dis lo and time on dis os
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.util.*;
import java.io.*;

/**
 *  Command Pluging, containing functions to manipulate Object and Locks data
 *
 *  @author hdrc rich.cole@uk.ibm.com
 */
public class DvObjectsCommands extends CommandPlugin  {
    /* buffer for data waiting to be output to the console */
    private static Vector theOutput = new Vector();
    private final static String oSeperator = "\n============================================";



    //boolean doOnce = true;
    /* counts of the number of Objects founds in the different heaps */
    private static boolean heapsTraversed = false;
    private static int totalObjectCount;
    private static int totalLockedObjectCount;

    // The array below are is to record
    // the ranges associated with the System and
    // the TLH's, thus enabling us to determine (by object address)
    // which heap an object is in. The first one is start, second is
    // end and third is type....  note that this array is sorted by
    // start position.....
    //  -1,-2,-3 etc is system heap...
    // 0,1,2,3,4,5 is index into threads for this process (1=0,2=1 etc)

    private static long[][] rangeArray;




    private static int middlewareH;
    private static int transientH;
    private static int systemH;
    private static int acsH;
    static int threadLocalH;
    private static int ptrLength; // gets set to 4 or 8 depending on whether its a 64 bit
                                  // system (done during objectsummary)

    private static boolean is64bit;
    private static boolean isBigEndian;
    private static long allocBitsAddress = 0;
    private static final byte[] allocBitsTestBytes={(byte)0x80,(byte)0x40,(byte)0x20,(byte)0x10,
                                           (byte)0x08,(byte)0x04,(byte)0x02,(byte)0x01};

    private static boolean debug=true;
    private static boolean verboseTLH=false;  // set true for verbose output during TLH traversal

    private static DvAddressIndexFile  addressIndexFile;
    private static DvNamesIndexFile    namesIndexFile;  // Index file
    private static boolean address_index_built = false;
    private static boolean names_index_built = false;

    // Count of threads with and without Thread local Heap (TLH) cache_blocks
    private int javaThreadsWithoutTLH;
    private int javaThreadsWithTLH;
    private int javaThreadsWithoutTHTLH;
    private int javaThreadsWithTHTLH;

    private static byte[] extendedAllocBits=null;
    private static byte[] lockedAllocBits=null;
    private static long   startOfMHeap = 0;
    private static long   endOfMHeap = 0;

    private static long startOfTransHeap = 0;
    private static long endOfTransHeap = 0;


    // extendedAllocBits  is a large array of bytes that
    // largely mimics the allocbits in the JVM
    // with the additional feature that objects
    // within TLH's will be recorded here. As we
    // find objects during our scan then the
    // allocbits will be set - thus we will
    // easily be able to tell if an address is
    // an object. This area will be the size of
    // the heap divided by 64 with 1 bit =
    // 8 bytes....


    private static HashSet systemHeapHashByAddr = new HashSet();
    private static HashSet acsHeapHashByAddr = new HashSet();



    // because the system heap falls outside of the "normal" heap then we need to keep a
    // check on the ranges - so that when we try and look for an object outside our normal
    // EAB array we can check in systemHeapHashByAddr for it .....

    private static Vector lockedObjects = new Vector();
    // There is a comparitively small number of locked objects compared to the total number
    // of objects - so we create a vector of all those that we discover, note that this vector
    // is added to the namesIndexFile at the end of the traversal



    /**
     *  The commands supported by DvObjectsCommands
     */
    private static String[] commandsSupported =
    {
        "DISPLAY,DIS,OBJECT,obj,displayObjects",
        "DISPLAY,DIS,LOCK,loc,displayLocks",
        "DISPLAY,DIS,LOCKSUMMARY,ls,displayLockSummary",
        "DISPLAY,DIS,LOCKEDOBJECTS,lo,displayLockedObjects",
        "DISPLAY,DIS,OBJECTSUMMARY,os,displayObjectSummary"


    };

    /**
     *
     */
    public DvObjectsCommands() {
        super();

    }

    /**
     *
     */
    public String[] getSyntax() {
        return commandsSupported;
    }

    /**
     *
     */
    public  String  pluginName() {return "Object commands Plugin";};


    /**  AddressSpace  */
    private static DvAddressSpace as;

    /**
     *  Clear the Object and Lock information
     *  This is called by the set dump routine to clear old info
     *
     */
    public void reset() {
        // output any data left in buffer before clearing
        cpr.output(theOutput);
        theOutput.clear();

        //theObjArray = null;  111
        //theObjects.clear();
        systemHeapHashByAddr.clear();
        acsHeapHashByAddr.clear();
        //objectHashByAddr.clear();
        //objectHashByName.clear();
        totalObjectCount = 0;
        totalLockedObjectCount=0;
        middlewareH = 0;
        transientH = 0;
        systemH = 0;
        acsH = 0;
        threadLocalH = 0;
        ptrLength=0;

    }

    /**
     *  Displays INFORMATION about the objects
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Filters and wildcards can be used
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @console param      filters and wildcards based on names of objects
     *  @console return     displays information on the Objects in the heap
     *                      based on what filters and wildcards are used
     *
     */
    public void displayObjects() {

        boolean bWildCards = false;
        Vector theFilters = new Vector();

        // check that dump has been opened
        if (null == theDump) {
            theOutput.add("Unable to comply as no Dump opened yet\n");
            returnControl();
            return;
        }
        initPtrLength();
        is64bit = DvUtils.is64BitSystem(theDump.getArchitecture());
        isBigEndian = DvUtils.isBigEndian(theDump.getArchitecture());
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        if (null == paramString) {
            displayObjectSummary();
            returnControl();
            return;
        }


        // if parameter after command then tokenize
        if (null != paramString) {
            try{
                // process parameter if it is a hex address, must start 0x
                if ((paramString.charAt(0)=='0') && ((paramString.charAt(1)=='x')||(paramString.charAt(1)=='X')))
                {
                    long theAddress=Long.parseLong(paramString.substring(2),16);
                    if (!DvUtils.checkAddressInDump(as.createAddress(theAddress),as))
                    {
                        theOutput.add("Not a valid address");
                        returnControl();
                        return;
                    }
                    // If heaps have not been travered do that
                    //if (null == theObjArray || 0==theObjArray.length) {
                    if (false == heapsTraversed) {
                        traverseAllHeaps();
                    }



                    if (true == checkForKnownObject(theAddress-ptrLength)) {
                        //System.err.println("Known object");
                        ObjectDetails o = (ObjectDetails) getObjectDetails(theAddress-ptrLength);
                        showObjects(o);
                    }
                    else {
                         //System.out.println("At point b");
                         theOutput.add("@ 0x" + Long.toHexString(theAddress)+
                            " not the start of an object in the heap");
                    }

                    returnControl();
                    return;

                }
            }
            catch (NumberFormatException e)
            {
                if (debug) theOutput.add("The address is not valid hex");
                returnControl();
                return;
            }
            catch (StringIndexOutOfBoundsException e) {
                if (debug) theOutput.add("The address is not valid          StringIndexOutOfBoundsException:");
                returnControl();
                return;
            }
            if (Character.isDigit(paramString.charAt(0))) {
                if (debug) theOutput.add("Object names must start with letter");
                returnControl();
                return;
            }
            StringTokenizer st = new StringTokenizer(paramString,",");

            while (st.hasMoreTokens()) {

                // Note we assume that the filters make sense

                String token = st.nextToken().trim();

                if (token.startsWith("*") ||

                    token.endsWith("*") ||

                    token.startsWith("!")) {

                    bWildCards = true;

                }

                theFilters.add(token);

            }

        }

        // If heaps have not been travered do that
        //if (null == theObjArray || 0==theObjArray.length) {
        if (false == heapsTraversed) {
            traverseAllHeaps();
        }
        //if (0==theObjects.size()) {traverseAllHeaps();}
        // now for each class - see whether its name matches a filter
        // assuming there are any filters



        if (theFilters.size() >0 && false == bWildCards) {

            for (int i=0;i<theFilters.size();i++) {

                String f = (String) theFilters.get(i);


                long[] objectList = namesIndexFile.getInstancesForClass(f);

                if (null != objectList) {
                    // we have found it
                    for (int j=0;j<objectList.length;j++) {
                        ObjectDetails o =
                                (ObjectDetails) getObjectDetails(objectList[j]);

                        showObjects(o);
                            //  (ObjectDetails) getObjectDetails(n.intValue());

                    }

                }

            }

        }

        else {

            // OK they have wildcards ... got to scan through the lot

            //for (int i=0;i<theObjArray.length;i++) {

                // get all the class names and go through them 1 at a time
                // to see if they satisay one of the filters....
                Vector v= namesIndexFile.getAllClassNames();

                for (int j=0; j<v.size();j++) {
                    String name = (String) v.get(j);
                    boolean bShowit = true; // assume this one is to be shown

                    if (0 !=theFilters.size()) {
                        bShowit = DvUtils.filtrate(name,theFilters);
                    }


                    if (bShowit) {

                        long[] objectList = namesIndexFile.getInstancesForClass(name);

                        if (null != objectList) {
                            // we have found it
                            for (int k=0;k<objectList.length;k++) {
                                ObjectDetails o =
                                    (ObjectDetails) getObjectDetails(objectList[k]);

                                showObjects(o);
                                //  (ObjectDetails) getObjectDetails(n.intValue());

                            }
                        }
                    }
                }

            //}

        }

        returnControl();
    }

    /**
     *  Displays a SUMMARY of information about the objects
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @console return     displays a summary of the information on the
     *                      Objects in the heap
     *
     */
    public void displayObjectSummary() {
        boolean bWildCards = false;
        Vector theFilters = new Vector();

        // check that dump has been opened
        if (null == theDump) {
            theOutput.add("Unable to comply as no Dump opened yet\n");
            returnControl();
            return;
        }

        initPtrLength();
        is64bit = DvUtils.is64BitSystem(theDump.getArchitecture());
        isBigEndian = DvUtils.isBigEndian(theDump.getArchitecture());


        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // If heaps have not been travered do that
        //if (null == theObjArray || 0==theObjArray.length) {
        if (false == heapsTraversed) {
            traverseAllHeaps();
        }

        theOutput.add("\n objects Summary"+
                      "\n ===============\n" );

        theOutput.add(" Number of objects found via Thread Local Heap   = "+
                      threadLocalH);
        theOutput.add(" Number of objects found via Middleware Heap     = "+
                      middlewareH);
        theOutput.add(" Number of objects found via Transient Heap      = "+
                      transientH);
        theOutput.add(" Number of objects found via System Heap         = "+
                      systemH);
        theOutput.add(" Number of objects found via ACS Classes Heap    = "+
                      acsH+"\n\n");

        theOutput.add(" Total number of objects found via all Heaps     = "+
                      totalObjectCount+"\n\n");



        // Now go through the index file and find all class names with
        // their counts.....
        theOutput.add(namesIndexFile.summary()) ;

        theOutput.add("\n Number of Locked objects  = "+
                      totalLockedObjectCount+"\n");

        returnControl();

    }


    /**
     *  Displays Information about the objects LOCKS
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Filters and wildcards can be used
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @console param      filters and wildcards based on names of objects
     *  @console return     displays information on the object Locks in the
     *                      heap based on what filters and wildcards are used
     *
     */
    public void displayLocks() {
        boolean bWildCards = false;
        Vector theFilters = new Vector();

        // check that dump has been opened
        if (null == theDump) {
            theOutput.add("Unable to comply as no Dump opened yet\n");
            returnControl();
            return;
        }
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // if parameter after command then tokenize
        if (null != paramString) {
            try{
                // process parameter if it is a hex address, must start 0x
                if ((paramString.charAt(0)=='0') && ((paramString.charAt(1)=='x')||(paramString.charAt(1)=='X')))
                {
                    long theAddress=Long.parseLong(paramString.substring(2),16);
                    if (!DvUtils.checkAddressInDump(as.createAddress(theAddress),as))
                    {
                        theOutput.add("Not a valid address");
                        returnControl();
                        return;
                    }
                    // If heaps have not been travered do that
                    //if (null == theObjArray || 0==theObjArray.length) {
                    if (false == heapsTraversed) {
                        traverseAllHeaps();
                    }

                    if (true == checkForKnownObject(theAddress-ptrLength)) {
                        //System.err.println("Known object");
                        ObjectDetails o = (ObjectDetails) getObjectDetails(theAddress-ptrLength);
                        showObjects(o);
                    }
                    else {
                         //System.out.println("At point C");
                         theOutput.add("@ 0x" + Long.toHexString(theAddress)+
                            " not the start of an object in the heap");
                    }
                    returnControl();
                    return;

                }
            }
            catch (NumberFormatException e)
            {
                if (debug) theOutput.add("The address is not valid hex");
                returnControl();
                return;
            }
            catch (StringIndexOutOfBoundsException e) {
                if (debug) theOutput.add("The address is not valid          StringIndexOutOfBoundsException:");
                returnControl();
                return;
            }
            if (Character.isDigit(paramString.charAt(0))) {
                if (debug) theOutput.add("Object names must start with letter");
                returnControl();
                return;
            }
            StringTokenizer st = new StringTokenizer(paramString,",");
            while (st.hasMoreTokens()) {
                // Note we assume that the filters make sense
                String token = st.nextToken().trim();
                if (token.startsWith("*") ||
                    token.endsWith("*") ||
                    token.startsWith("!")) {
                    bWildCards = true;
                }

                theFilters.add(token);
            }
        }
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // If heaps have not been travered do that
        //if (null == theObjArray || 0==theObjArray.length) {
        if (false == heapsTraversed) {
            traverseAllHeaps();
        }

        // now for each class - see whether its name matches a filter
        // assuming there are any filters
        if (theFilters.size() >0 && false == bWildCards) {
            for (int i=0;i<theFilters.size();i++) {
                String f = (String) theFilters.get(i);

                long[] objectList = namesIndexFile.getInstancesForClass(f);

                if (null != objectList) {
                    // we have found it
                    for (int j=0;j<objectList.length;j++) {
                        ObjectDetails o =
                                (ObjectDetails) getObjectDetails(objectList[j]);

                        showLocks(o);

                    }


                }

            }
        }
        else {
            // OK they have wildcards ... got to scan through the lot
            Vector v= namesIndexFile.getAllClassNames();

            for (int j=0; j<v.size();j++) {
                String name = (String) v.get(j);
                boolean bShowit = true; // assume this one is to be shown

                if (0 !=theFilters.size()) {
                    bShowit = DvUtils.filtrate(name,theFilters);
                }

                if (bShowit) {

                    long[] objectList = namesIndexFile.getInstancesForClass(name);
                    if (null != objectList) {
                        // we have found it
                        for (int k=0;k<objectList.length;k++) {
                            ObjectDetails o =
                                (ObjectDetails) getObjectDetails(objectList[k]);

                            showLocks(o);
                            //  (ObjectDetails) getObjectDetails(n.intValue());

                        }
                    }
                 }
             }

        }

        returnControl();
    }


    /**
     *  Displays Information about the objects LOCKS
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Filters and wildcards can be used
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @console param      filters and wildcards based on names of objects
     *  @console return     displays information on the object Locks in the
     *                      heap based on what filters and wildcards are used
     *
     */
    public void displayLockSummary() {
        // check that dump has been opened
        if (null == theDump) {
            theOutput.add("Unable to comply as no Dump opened yet\n");
            returnControl();
            return;
        }
        initPtrLength();
        is64bit = DvUtils.is64BitSystem(theDump.getArchitecture());
        isBigEndian = DvUtils.isBigEndian(theDump.getArchitecture());
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        long lkGlPtrAddr=Long.parseLong(DvUtils.getJVMHint(as,"LKGLOBALPTR"),16);
        long dgGlPtrAddr=Long.parseLong(DvUtils.getJVMHint(as,"DGGLOBALPTR"),16);
        if (lkGlPtrAddr==dgGlPtrAddr) {
            theOutput.add("               This will not work without the modified JVM ie. needs lk: dataP");
            theOutput.add("               The \"LKGLOBALPTR\" will appear in \"dis hints\" but it is actually \"DGGLOBALPTR\"");
            theOutput.add("               You can check to see if lk dataP exist by \"for Jvm\" and looking under lk");

            returnControl();
            return;
        }

        // If heaps have not been travered do that
        if (false == heapsTraversed) {
            traverseAllHeaps();
        }

        // At this point the lockedObjects vector will contain
        // the addresses of all locked objects!!!!  Also
        // totalLockedObjects will contain a count
//--------------------------------------------------
        theOutput.add("Monitor pool info:");
        theOutput.add("");
        theOutput.add("");
/*        theOutput.add("Initial monitor count: xx");
        theOutput.add("Minimum number of free monitors before expansion: xx");
        theOutput.add("Pool will next be expanded by: xx");
        theOutput.add("Current total number of monitors: xx");
        theOutput.add("Current number of free monitors: xx");
*/        theOutput.add("");

//---------------------------------------------------
        theOutput.add("Monitor Pool Dump (inflated object-monitors):");
        try{
            long micb = DvUtils.getFieldValue(lkGlPtrAddr,new String[] {"lkglData","micb"}, as, theDump);
            theOutput.add("table of inflated monitor ");
            theOutput.add("0x"+ Long.toHexString(micb) + " monitor_index_cb_t" );
            theOutput.add("");
            long table = DvUtils.getFieldValue(micb,new String[] {"monitor_index_cb_t","table"}, as, theDump);
            long count = DvUtils.getFieldValue(micb,new String[] {"monitor_index_cb_t","count"}, as, theDump);
            long inflMonT;
            int index=1;
            if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
                inflMonT = (long)as.readLong(as.createAddress(table));
                table += 8;
            }
            else {
                inflMonT = (long)as.readInt(as.createAddress(table));
                table += 4;
            }

            //  The first one should be blank, lets test this
            if (0!=inflMonT) {
                theOutput.add("the first element in the infl_mon table is not 0x00000000");
            }
            else { // passed check continue with Monitor Table processing
                while (count-index>0) {//DvUtils.checkAddressInDump(as.createAddress(table),as)) {

                    if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
                        inflMonT = (long)as.readLong(as.createAddress(table));
                        index++;
                        table += 8;
                    }
                    else {
                        inflMonT = (long)as.readInt(as.createAddress(table));
                        index++;
                        table += 4;
                    }
                    long SysMonT = DvUtils.getFieldValue(inflMonT,new String[] {"infl_mon","_sysmon"}, as, theDump);
                    theOutput.add("sys_mon_t:0x" + Long.toHexString(SysMonT) + " infl_mon_t: 0x"+ Long.toHexString(inflMonT));
                    //theOutput.add("<name> <owner>");
                    //theOutput.add("Waiting to be notified:");
                    //theOutput.add("<    > (0x00000000) ");
                }

            }
        }
        catch (DvAddressException e) {
            theOutput.add("DvAddressException in displayLockSummary");
            returnControl();
        }
        theOutput.add("");
//-------------------------------------------------
        //JVM System Monitor Dump (registered monitors)
        theOutput.add("");

        long ptrToReg;
        long reg;
        String name = "";
        try {
            ptrToReg = DvUtils.getFieldValue(lkGlPtrAddr, new String[] {"lkglData", "reg"}, as, theDump);
            if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
                reg = (long)as.readLong(as.createAddress(ptrToReg));
            }
            else {
                reg = (long)as.readInt(as.createAddress(ptrToReg));
            }
        }
        catch (DvAddressException e) {
            theOutput.add("DvAddressException in displayLockSummary");
            returnControl();
            return;
        }
        theOutput.add("JVM System Monitor Dump (registered monitors): ");
        theOutput.add("0x"+Long.toHexString(reg)+" reg_mon");
        theOutput.add("");

        while (DvUtils.checkAddressInDump(as.createAddress(reg),as)) {
//            theOutput.add("2LKREGMON          Method trace lock: <unowned>");
//            theOutput.add("2LKREGMON          Heap lock: owner \"Signal dispatcher\" (0x21226C70), entry count 1");

            try {
                long nameAddr = DvUtils.getFieldValue(reg, new String[] {"reg_mon", "name"}, as, theDump);
                name = DvUtils.getStringFromBytes(as.readBytes(as.createAddress(nameAddr),256));

                theOutput.add("0x" +Long.toHexString(reg) +" "+ name +":");
                reg = DvUtils.getFieldValue(reg, new String[] {"reg_mon", "next"}, as, theDump);
            }
            catch (DvAddressException e) {
            theOutput.add("DvAddressException in displayLockSummary");
                returnControl();
                return;
            }

        }
        theOutput.add("");
//-------------------------------------------------
        theOutput.add("Thread identifiers (as used in flat monitors");
        theOutput.add("");
        theOutput.add("");

//---------------------------------------------------
        theOutput.add("Java Object Monitor Dump (flat & inflated object-monitors):");
        theOutput.add("for further information is \"dis lo(0x00000000)\"");
        theOutput.add("");


        // go through all the objects and see which ones are locked
        int count = lockedObjects.size();
        for (int i=0;i<count;i++) {
            long objAddress =  ((Long) lockedObjects.get(i)).longValue();
            ObjectDetails o = (ObjectDetails) getObjectDetails(objAddress);
            if (0==o.shapeBit) {
                        // either flat
                        theOutput.add(o.objectName + " 0x"+Long.toHexString(o.address));
                        theOutput.add("    locknflags 0x" + Long.toHexString(o.locknflags) +
                                 " flat locked by thread ident " + o.threadId +
                                 ", entry count "+ o.entryCount);
            }
            else {
                        // or inflated
                        theOutput.add(o.objectName + " 0x"+Long.toHexString(o.address));
                        theOutput.add("    locknflags 0x" + Long.toHexString(o.locknflags) +
                                 " Monitor inflated infl_mon 0x"+Long.toHexString(o.inflMon));
            }

        }

        //count= addressIndexFile.sizeOfIndex();
        //System.out.println("Index table size is: "+ count);

        //for (int i=0;i<count;i++) {
        //    Vector v = addressIndexFile.getAllObjectsAtIndex(i);
        //    for (int j=0;j<v.size();j++) {
        //        long objAddress =  ((Long) v.get(j)).longValue();
        //        ObjectDetails o = (ObjectDetails) getObjectDetails(objAddress);
        //        if (o.lock) {
        //            if (0==o.shapeBit) {
                        // either flat
        //                theOutput.add(o.objectName + " 0x"+Long.toHexString(o.address));
        //                theOutput.add("    locknflags 0x" + Long.toHexString(o.locknflags) +
        //                         " flat locked by thread ident " + o.threadId +
        //                         ", entry count "+ o.entryCount);
        //           }
        //            else {
                        // or inflated
        //                theOutput.add(o.objectName + " 0x"+Long.toHexString(o.address));
        //                theOutput.add("    locknflags 0x" + Long.toHexString(o.locknflags) +
        //                         " Monitor inflated infl_mon 0x"+Long.toHexString(o.inflMon));
        //            }
        //        }

        //    }
        //}

        returnControl();
    }

    /**
     *  Likly to change in the future may become part of obj
     *
     *  Test to see if is an object
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Filters and wildcards can be used
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @param theAddress       the address to test
     *  @return                 <code>true</code> if the address is the begining of an
     *                          Object in the Heap
     *  @throws                 DvAddressException
     *
     */
    public boolean displayObject(long theAddress) throws DvAddressException {
        if (null == theDump) {
            returnControl();
            throw new DvAddressException("Unable to comply as no Dump opened yet");
        }

        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        //if (0==theObjects.size()) {
        //if (null == theObjArray || 0==theObjArray.length) {
        if (false == heapsTraversed) {
            traverseAllHeaps();
        }

        if (!DvUtils.checkAddressInDump(as.createAddress(theAddress),as))
        {
            returnControl();
            throw new DvAddressException("Unable to comply as not a valid address");
        }

        if (true == checkForKnownObject(theAddress-ptrLength)) {
            //System.err.println("Known object");
            ObjectDetails o = (ObjectDetails) getObjectDetails(theAddress-ptrLength);
            showObjects(o);
        }
        else {
            //System.out.println("At point D");
            theOutput.add("@ 0x" + Long.toHexString(theAddress)+
                " not the start of an object in the heap");
        }
        returnControl();
        return false;
    }

    /**
     *  Likly to change in the future may become part of obj
     *
     *  Test to see if is an object
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Filters and wildcards can be used
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @param theName          the Name to test
     *  @return                 <code>true</code> if the Name is of an
     *                          Object in the Heap
     *  @throws                 DvAddressException
     *
     */
    public boolean displayObject(String theName) throws DvAddressException {
        // check that dump has been opened
        if (null == theDump) {
            returnControl();
            throw new DvAddressException("Unable to comply as no Dump opened yet");
        }

        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // If heaps have not been travered do that
        //if (null == theObjArray || 0==theObjArray.length) {
        if (false == heapsTraversed) {
            //if (0==theObjects.size()) {
            traverseAllHeaps();
        }

        long[] objectList = namesIndexFile.getInstancesForClass(theName);

        if (null != objectList) {
                    // we have found it
            for (int j=0;j<objectList.length;j++) {
                ObjectDetails o =
                     (ObjectDetails) getObjectDetails(objectList[j]);

                showObjects(o);
                returnControl();
                return true;
            }

        }


        returnControl();
        return false;
    }


    /**
     *  Displays Information about the objects which are LOCKED
     *  in the Middleware Heap, Transient Heap, System Heaps, ACS Heap.
     *  Filters and wildcards can be used
     *  Traverse the heaps.
     *  Requires a dump to be set.
     *
     *  @console param      filters and wildcards based on names of objects
     *  @console return     displays information on the object Locks in the
     *                      heap based on what filters and wildcards are used
     *
     */
    public void displayLockedObjects() {
        boolean bWildCards = false;
        Vector theFilters = new Vector();

        // check that dump has been opened
        if (null == theDump) {
            theOutput.add("Unable to comply as no Dump opened yet\n");
            returnControl();
            return;
        }

        initPtrLength();
        is64bit = DvUtils.is64BitSystem(theDump.getArchitecture());
        isBigEndian = DvUtils.isBigEndian(theDump.getArchitecture());
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        // if parameter after command then tokenize
        if (null != paramString) {
            try{
                // process parameter if it is a hex address, must start 0x
                if ((paramString.charAt(0)=='0') && ((paramString.charAt(1)=='x')
                    ||(paramString.charAt(1)=='X')))
                {
                    long theAddress=Long.parseLong(paramString.substring(2),16);
                    if (!DvUtils.checkAddressInDump(as.createAddress(theAddress),as))
                    {
                        theOutput.add("Not a valid address");
                        returnControl();
                        return;
                    }
                    // If heaps have not been travered do that

                    if (false == heapsTraversed) {
                        traverseAllHeaps();
                    }


                    //System.out.println("theAddress:" + Long.toHexString(theAddress)+
                    //            "ptrLength:" +ptrLength ) ;
                    if (true == checkForKnownObject(theAddress-ptrLength)) {
                        //System.err.println("Known object");
                        ObjectDetails o = (ObjectDetails) getObjectDetails(theAddress-ptrLength);
                        if (o.lock) {
                            showLocks(o);
                        }
                        else theOutput.add("Object is not locked");
                    }
                    else {
                         //System.out.println("At point E");
                         theOutput.add("@ 0x" + Long.toHexString(theAddress)+
                            " not the start of an object in the heap");
                    }

                    returnControl();
                    return;

                }
            }
            catch (NumberFormatException e)
            {
                if (debug) theOutput.add("The address is not valid hex");
                returnControl();
                return;
            }
            catch (StringIndexOutOfBoundsException e) {
                if (debug) theOutput.add("The address is not valid          StringIndexOutOfBoundsException:");
                returnControl();
                return;
            }

            if (Character.isDigit(paramString.charAt(0))) {
                if (debug) theOutput.add("Object names must start with letter");
                returnControl();
                return;
            }
            StringTokenizer st = new StringTokenizer(paramString,",");
            while (st.hasMoreTokens()) {
                // Note we assume that the filters make sense
                String token = st.nextToken().trim();
                if (token.startsWith("*") ||
                    token.endsWith("*") ||
                    token.startsWith("!")) {
                    bWildCards = true;
                }
                theFilters.add(token);
            }
        }


        // enhancers
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        // If heaps have not been travered do that
        if (false == heapsTraversed) {
            traverseAllHeaps();
        }
        // now for each class - see whether its name matches a filter
        // assuming there are any filters
        if (theFilters.size() >0 && false == bWildCards) {
            for (int i=0;i<theFilters.size();i++) {
                String f = (String) theFilters.get(i);

                //long[] objectList = namesIndexFile.getInstancesForClass(f);

                //if (null != objectList) {
                    // we have found it
                for (int j=0;j<lockedObjects.size();j++) {
                    long objAddress =
                        ((Long) lockedObjects.get(j)).longValue();
                    ObjectDetails o =
                        (ObjectDetails) getObjectDetails(objAddress);

                    if (f.equals(o.objectName)) {
                        showLocks(o);
                    }

                }

                //}

            }
        }
        else {
            // OK they have wildcards ... got to scan through the lot
            Vector v= namesIndexFile.getAllClassNames();

            for (int j=0; j<v.size();j++) {
                String name = (String) v.get(j);
                boolean bShowit = true; // assume this one is to be shown

                if (0 !=theFilters.size()) {
                        bShowit = DvUtils.filtrate(name,theFilters);
                }


                if (bShowit) {

                    for (int k=0;k<lockedObjects.size();k++) {
                        long objAddress =
                            ((Long) lockedObjects.get(k)).longValue();
                        ObjectDetails o =
                            (ObjectDetails) getObjectDetails(objAddress);

                        if (name.equals(o.objectName)) {
                            showLocks(o);
                        }

                    }
                }
            }
        }

        returnControl();
    }




    /**
     *  lkDumpRoutine
     *  returns Vector of Strings of javacore locking component information
     *
     *  @return Vector of Strings
     *
     */
    public static Vector lkDumpRoutine()
    {
        Vector theLkInfo = new Vector();
        theLkInfo.addElement("NULL");
        theLkInfo.addElement("");
        theLkInfo.addElement("1LKPOOLINFO");
        theLkInfo.addElement("Monitor pool info:");
        theLkInfo.addElement("2LKPOOLINIT");
        theLkInfo.addElement("Initial monitor count: <n\\a>");
        theLkInfo.addElement("2LKPOOLEXPNUM");
        theLkInfo.addElement("Minimum number of free monitors before expansion: <n\\a>");
        theLkInfo.addElement("2LKPOOLEXPBY");
        theLkInfo.addElement("Pool will next be expanded by: <n\\a>");
        theLkInfo.addElement("2LKPOOLTOTAL");
        theLkInfo.addElement("Current total number of monitors: <n\\a>");
        theLkInfo.addElement("2LKPOOLFREE");
        theLkInfo.addElement("Current number of free monitors: <n\\a>");
        theLkInfo.addElement("NULL");
        theLkInfo.addElement("");
        theLkInfo.addElement("1LKMONPOOLDUMP");
        theLkInfo.addElement("Monitor Pool Dump (flat & inflated object-monitors):");
        theLkInfo.addElement("NULL");
        theLkInfo.addElement("");
        theLkInfo.addElement("1LKREGMONDUMP");
        theLkInfo.addElement("JVM System Monitor Dump (registered monitors):");
        theLkInfo.addElement("NULL");
        theLkInfo.addElement("");
        theLkInfo.addElement("1LKFLATMONDUMP");
        theLkInfo.addElement("Thread identifiers (as used in flat monitors):");
        theLkInfo.addElement("NULL");
        theLkInfo.addElement("");
        theLkInfo.addElement("1LKOBJMONDUMP");
        theLkInfo.addElement("Java Object Monitor Dump (flat & inflated object-monitors):");
        return theLkInfo;
    }

    /**
     *  Display the object info from the object
     *
     *  @param o         Object to be displayed on console
     *
     */
    private void showObjects(ObjectDetails o) {
           theOutput.add(oSeperator);
           theOutput.add(o);
           if (null != o) {
                theOutput.add(DvClassCommands.getInstanceValues(o.address,
                                                                o.objectName,
                                                                theDump,
                                                                theOutput));
           }
           else {
            theOutput.add(" ---- Null or invalid object ---- " );
           }


           cpr.output(theOutput);
           theOutput.clear();

    }

    /**
     *  Display the lock info from the object, exact behavour depends on the
     *  value of the locks
     *
     *  @param o        Object to have locks displayed
     */
    private void showLocks(ObjectDetails o) {
         String cn = "";
         if (null != o.className) {
             cn = o.className;
         }

         if (o.lock){
             if (0==o.shapeBit) {
             // flat
                 theOutput.add( "\n"+o.objectName+" @ 0x"+
                                Long.toHexString(o.address) +
                                " \n" + cn +
                                "          type:" + o.type +
                                " (Flat) Entry Count: "+ o.entryCount +
                                " Thread Id: " + o.threadId );

           }
           else {
           // inflated
               theOutput.add("\n"+o.objectName+" @ 0x"+
                                Long.toHexString(o.address) +
                                " \n" + cn +
                                "          type:" + o.type +
                                " (Inflated) Monitor Index: " + o.monitorIndex );
               showMonitor(o);
           }
       }
       else {
           // no locks.
           theOutput.add("@ 0x"+Long.toHexString(o.address) +
                        " (" + o.objectName  + ") " + cn +
                        "          type:" + o.type );

       }
   }


    /**
     *  @param o        Object to have monitor displayed
     */
    private void showMonitor(ObjectDetails o){
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        long lkGlPtrAddr=Long.parseLong(DvUtils.getJVMHint(as,"LKGLOBALPTR"),16);
        long dgGlPtrAddr=Long.parseLong(DvUtils.getJVMHint(as,"DGGLOBALPTR"),16);
        if (lkGlPtrAddr==dgGlPtrAddr) {
            //theOutput.add("               This will not work without the modified JVM ie. needs lk: dataP");
            returnControl();
            return;
        }

        try {
            CTypeObject inflMon = DvUtils.formatAddrAsType(as,as.createAddress(o.inflMon),"infl_mon");
            theOutput.add(inflMon);
            long sysMonAddr = DvUtils.getFieldValue(o.inflMon, new String[] {"infl_mon", "_sysmon"},as, theDump);
            CTypeObject sysMon = DvUtils.formatAddrAsType(as,as.createAddress(sysMonAddr),"sys_mon");
            theOutput.add("_sysmon");
            theOutput.add(sysMon);
        }
        catch (DvAddressException e) {
            theOutput.add("DvAddressException in showMonitors");
            returnControl();
            return;
        }
        catch (NumberFormatException e) {
            theOutput.add("DvNumberFormatException in showMonitors");
            theOutput.add("Possible cause by not having hints setup properly");
            returnControl();
            return;
        }

   }

    /**
     *  This method traverses all the heap actegories in the dump
     *  including the Thread Local Heaps. If Name and Address Index
     *  files do not exist then they are created during the scan (which
     *  means that an initial scan can take a long time for a big dump)
     *  as these need to be filled with data.....
     *
     *  On subsequent entry to DvConsole (when name and address index
     *  files exist then the scan is simply to count up all the
     *  individual classes in each category
     *
     */

   private void traverseAllHeaps() {

        //


        // right lets see if we have Index files yet
        if (null == namesIndexFile) {
            openNamesIndexFile();
        }

        // right lets see if we have Index files yet
        //if (null == addressIndexFile  ) {
        //    openAddrIndexFile();
        //}

        // Changed so there is only 1 pass
        //for (int i=1;i<2;i++) {
            totalObjectCount = 0;
            totalLockedObjectCount=0;

            // look through the Thread Local Heap (TLH) objects
            traverseEveryTLH();

            //So lets go through the Middleware Heap, Transient Heap, System Heap and ACS classes
            //start the Middleware Heap objects
            traverseMiddlewareHeap();

            // start Transient Heap objects
            traverseTransientHeap();

            // start System Heap objects
            traverseSystemHeaps();

            // start ACS Heap -it only contains classes
            traverseACSHeap();
            //if (0 == i) {
                // if its pass 0 then allocate the array
                // which could be a very big array
                //theObjArray = new long[totalObjectCount];
            //}
        //} changed so there is only 1 pass
        heapsTraversed = true;
        names_index_built = true;
        address_index_built = true;
    }

    /**
     *  Traverses MiddlewareHeap
     *  Requires the global DvDump theDump
     *
     */
    private void traverseMiddlewareHeap() {
        int objectCount=0;
        theOutput.add("Traversing the Middleware heap\n");
        cpr.output(theOutput);
        theOutput.clear();
        startOfMHeap = Long.parseLong(DvUtils.getJVMHint(as,"CURHEAPMIN"),16);
        endOfMHeap = Long.parseLong(DvUtils.getJVMHint(as,"CURHEAPMAX"),16);
        objectCount = traverseHeap(startOfMHeap,endOfMHeap,
                               Long.parseLong(DvUtils.getJVMHint(as,"MH_HEAPBASE"),16),
                              "Mid",DvConstants.HEAP_MID);
        middlewareH = objectCount;
    }

    /**
     *  Traverses TransientHeap
     *  Requires the global DvDump theDump
     *
     */

    private void traverseTransientHeap() {
        int objectCount = 0;
        theOutput.add("Traversing the transient heap \n");
        cpr.output(theOutput);
        theOutput.clear();
        startOfTransHeap = DvUtils.hexToLong(DvUtils.getJVMHint(as,"TH_HEAPBASE"));
        endOfTransHeap = DvUtils.hexToLong(DvUtils.getJVMHint(as,"TH_HEAPLIMIT"));
        objectCount = traverseHeap(startOfTransHeap,endOfTransHeap,
                              DvUtils.hexToLong(DvUtils.getJVMHint(as,"MH_HEAPBASE")),
                              "Tra",DvConstants.HEAP_TRANS);

        transientH = objectCount;
    }

    /**
     *  Traverses SystemHeaps
     *  Requires the global DvDump theDump
     *
     */
    private void traverseSystemHeaps() {

        int objectCount = 0;

        int heapCount = 0;
        theOutput.add("Traversing the system heaps \n");
        cpr.output(theOutput);
        theOutput.clear();
        Vector ranges = new Vector();

        // traverseSystemHeaps happens after going through TLH's hence there may be
        // some stuff already recorded in ranges lets transfer it back it our
        // cnew vector

        int raLength = 0;

        if (null != rangeArray) {
            raLength = rangeArray.length;




            for (int i =0; i< rangeArray.length;i++) {
                //System.out.println("qqq:"+rangeArray[i][0]);
                ranges.add(new Range(rangeArray[i][0],
                                    rangeArray[i][1],rangeArray[i][2]));



            }
        }

       // A small vector to keep Range objects in, detailing the ranges of the System and
       // thread local heap address ranges - used for fillinr rangeArray;


        DvAddressSpace theAs = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        DvAddress STGlobalAddr = theAs.createAddress(DvUtils.hexToLong(DvUtils.getJVMHint(theAs,"STGLOBALPTR")));

        DvAddress SH_extentAddr = theAs.createAddress(0);

        long mh_heapbase = DvUtils.hexToLong(DvUtils.getJVMHint(theAs,"MH_HEAPBASE"));
        long heapBegin;
        long heapEnd;
        try{



            long st_jab = DvUtils.getFieldValue(STGlobalAddr.getAddressAsLong(), new String[] {"STGlobal", "st_jab"},theAs, theDump);
            SH_extentAddr = theAs.createAddress(DvUtils.getFieldValue(st_jab,
                new String[] {"StorageJab", "system_heap_this_extent"}, theAs,theDump));


            if (DvUtils.checkAddressInDump(SH_extentAddr,theAs)) {

                heapBegin = DvUtils.getFieldValue(SH_extentAddr.getAddressAsLong(),
                        new String[] {"heap_extent_info","base"}, theAs, theDump);
                heapEnd = DvUtils.getFieldValue(st_jab,
                        new String[] {"StorageJab", "system_heap_next"} , theAs, theDump);


                Range r = new Range(heapBegin,heapEnd,(long) -1);
                ranges.add(r);


                objectCount += traverseHeap(heapBegin,
                                            heapEnd,
                                            mh_heapbase,
                                            ("Sys "+heapCount),
                                            (short) (DvConstants.HEAP_SYS-heapCount));


                heapCount++;
            }

            SH_extentAddr = theAs.createAddress(DvUtils.getFieldValue(SH_extentAddr.getAddressAsLong()
                     ,new String[] {"heap_extent_info","prev"},theAs,theDump));

            while(DvUtils.checkAddressInDump(SH_extentAddr,theAs)) {

                heapBegin = DvUtils.getFieldValue(SH_extentAddr.getAddressAsLong(),
                    new String[] {"heap_extent_info","base"}, theAs,theDump);
                heapEnd = DvUtils.getFieldValue(SH_extentAddr.getAddressAsLong(),
                    new String[] {"heap_extent_info","limit"}, theAs,theDump);

                Range r = new Range(heapBegin,heapEnd,(long) -1);
                ranges.add(r);


                objectCount += traverseHeap(heapBegin, heapEnd,
                                            mh_heapbase, ("Sys "+heapCount),
                                            (short) (DvConstants.HEAP_SYS-heapCount));

                heapCount++;

                SH_extentAddr = theAs.createAddress(DvUtils.getFieldValue(SH_extentAddr.getAddressAsLong() ,new String[] {"heap_extent_info","prev"}, theAs,theDump));

            }

        } catch (DvAddressException e) {
             if (debug) theOutput.add("DvAddressException in traverseSystemHeaps");
             systemH = objectCount;
             return;
        }

        // Now we send the ranges vector off and produce a sorted rangeArray
        int[] sortList = sortRanges(ranges);

        if (null != sortList) {
            int rSize = ranges.size();
            rangeArray =  new long[rSize][3];
            for (int i =0;i<rSize;i++) {
                 Range r = (Range) ranges.get(sortList[i]);
                 rangeArray[i][0] = r.start;
                 rangeArray[i][1] = r.end;
                 rangeArray[i][2] = r.index;
                 //System.out.println("s:"+rangeArray[i][0]+
                 //                   "e:"+rangeArray[i][1]+
                 //                   "i:"+rangeArray[i][2]);
            }
        }


        systemH = objectCount;

    }


    /**
     *  Traverses ACSHeap
     *  Requires the global DvDump theDump
     *
     *   THIS IS INCOMPLETE
     */
    private void traverseACSHeap() {
        DvAddressSpace theAs = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        DvThread thread = (DvThread) DvUtils.getAPT(theDump,"T");

        int savedObjectCount = 0;
        int theObjectCount = 0;
        theOutput.add("Traversing the ACS heaps \n");
        cpr.output(theOutput);
        theOutput.clear();
        DvAddress loadedACSClassesAddr = theAs.createAddress(
                                        DvUtils.hexToLong(DvUtils.getJVMHint(theAs, "LOADEDACSCLASSES")));
        if (0==loadedACSClassesAddr.getAddressAsLong()){
            theOutput.add("No ACSClasses");
            returnControl();
            return;
        }
        if (debug) theOutput.add("The ACS Classes Traverser has not been fully tested and may break after this point");
        DvAddress cbAddr = loadedACSClassesAddr;
        while (DvUtils.checkAddressInDump(cbAddr,theAs)){
//      This section is very sinular to the code inside the for loop

            // Lets get bitness and endianess
            int arch = theDump.getArchitecture();
            DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

//            int theObjectCount=0;
            long objAddress = 0;
            long objLength=0;
            long methods=0;
            long locknflags=0;
            long obj=0;

            // Go get the memory at running pos
            // We have (hopefully) a length (int on 32 bit
            // long on 64 bit)
            long workingPos = cbAddr.getAddressAsLong();

            try {
                if (is64bit) {
                    objLength = as.readLong(as.createAddress(workingPos));
                    objLength = objLength & 0x3ffffffffffffff8L;

                }
                else { // 32bit

                        objLength =(long) as.readInt(as.createAddress(workingPos));
                        objLength = objLength & 0x3ffffff8 ;
                }

            }
            catch (DvAddressException e) {
                if (debug) theOutput.add("DvAddressException in traverseHeap");
                returnControl();
                return;// (int)theObjectCount;
            }

            // so we've found an object - lets store it away
            // for posterity
            // ACS Objects live in a world of their own
            // not conected to the real heap


            processFoundObject( workingPos,false,
                                    DvConstants.HEAP_ACS, objLength);

            theObjectCount++;

            if (theObjectCount >= (savedObjectCount + 5000)) {
                cpr.output("\t"+ DvUtils.getTime()+ " "+theObjectCount +
                        " objects processed...");
                savedObjectCount = theObjectCount;
            }

            //Get next loaded class.
            //  This code mimics ObjectEnumerate
            //  cb = (cb->obj->shared_class_id != 0) ? (ee->cl_thread_local.class_mirrors)[cb->obj->shared_class_id)]->obj->mirrored_data
            //                                       : cb->obj->mirrored_data->memory_table;

            try{
                long objAddr = DvUtils.getFieldAddress(cbAddr.getAddressAsLong(),new String[] {"Hjava_lang_Class","obj"},theAs);

                long shared_class_id = DvUtils.getFieldValue((objAddr) ,new String[] {"Classjava_lang_Class", "shared_class_id"},theAs,theDump);
                if (0==shared_class_id) {
                    long mirrored_data = DvUtils.getFieldValue((objAddr) ,new String[] {"Classjava_lang_Class", "mirrored_data"},theAs,theDump);
                    long memory_table = DvUtils.getFieldValue((mirrored_data) ,new String[] {"Classjava_lang_Class", "memory_table"},theAs,theDump);
                    cbAddr = (DvAddress) as.createAddress(memory_table);
                }
                else { // is shared class
                    if (null != thread) {
                        long cl_thread_localAddr = DvUtils.getFieldAddress(thread.eeAddress().getAddressAsLong(),new String[] {"ExecEnv","cl_thread_local"},theAs);
                        long class_mirrorsAddr = DvUtils.getFieldValue(cl_thread_localAddr,new String[] {"ClThreadLocalData","class_mirrors"},theAs,theDump);
                        long class_mirrorsTableIndex;
                        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
                            class_mirrorsTableIndex=class_mirrorsAddr + (shared_class_id * 8);
                        }
                        else {
                            class_mirrorsTableIndex=class_mirrorsAddr + (shared_class_id * 4);
                        }
                        objAddr = DvUtils.getFieldAddress(class_mirrorsTableIndex,new String[] {"Hjava_lang_Class","obj"},theAs);
                        long mirrored_data = DvUtils.getFieldValue((objAddr) ,new String[] {"Classjava_lang_Class", "mirrored_data"},theAs,theDump);
                        long memory_table = DvUtils.getFieldValue((mirrored_data) ,new String[] {"Classjava_lang_Class", "memory_table"},theAs,theDump);
                        cbAddr = (DvAddress) as.createAddress(memory_table);
                    }
                    else {theOutput.add("thread is null");}
                }
             } catch (DvAddressException e) {
                  if (debug) theOutput.add("DvAddressException in traverseACSHeap");
             }
         }
         acsH = theObjectCount;
    }

    /**
     *  Traverses heaps, based on start and finish values
     *  used for MiddlewareHeap, TransientHeap, SystemHeaps
     *  Requires the global DvDump theDump
     *   NOT used for ACS Heap
     *
     *  @param min              start of section of heap interested in
     *  @param max              end of section of heap interested in
     *  @param startofHeap      start of the whole heap
     *  @param theHeapName      the name of the heap so it can be stored in object,
     *                          also used to decide if to allocbits need to be
     *  @param heapId           Heap identifier
     *
     *  @return int
     *
     */
    private int traverseHeap(long min,long max,long startofHeap,
                                String theHeapName,
                                short heapId){

        int theObjectCount=0;
        int savedObjectCount = 0;
        // if allocCheck is true DOES NOT do alloc check
        boolean allocCheck = false;
        if (null  == theDump) {
            theOutput.add("Unable to comply - no dump accessed");
            returnControl();
            return (int)theObjectCount;
        }
        // Lets get bitness and endianess
        int arch = theDump.getArchitecture();
        DvAddressSpace as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        if (!DvUtils.checkAddressInDump(as.createAddress(min),as) ||
            !DvUtils.checkAddressInDump(as.createAddress(max),as) ||
            min==max)
        {
            theOutput.add("Event traverseHeap min or max value not in dump");
            returnControl();
            return (int)theObjectCount;
        }
        //is64bit = DvUtils.is64BitSystem(arch);
        //boolean isBigEndian = DvUtils.isBigEndian(arch);

        long runningPos = min;
        long objAddress = 0;
        long objLength=0;
        long methods=0;
        long locknflags=0;
        long obj=0;

        // check allocbits..
        // set alloc bit if middleware or transient heap
        if ((0==theHeapName.compareToIgnoreCase("MIDDLEWARE"))||
            (0==theHeapName.compareToIgnoreCase("TRANSIENT"))||
            (0==theHeapName.compareToIgnoreCase("MID"))||
            (0==theHeapName.compareToIgnoreCase("TRA")))
         {
            // if allocCheck is true DOES NOT do alloc check
            allocCheck=true;
         }

        for (runningPos=min; runningPos<max; runningPos += objLength) {
            // Go get the memory at running pos
            // We have (hopefully) a length (int on 32 bit
            // long on 64 bit)
            long workingPos = runningPos;

            try {
                if (is64bit) {
                    objLength = as.readLong(as.createAddress(workingPos));
                    objLength = objLength & 0x3ffffffffffffff8L;
                }
                else { // 32bit
                    objLength =(long) as.readInt(as.createAddress(workingPos));
                    objLength = objLength & 0x3ffffff8 ;
                }

            }
            catch (DvAddressException e) {
                if (debug) theOutput.add("DvAddressException in traverseHeap");
                returnControl();
                return (int)theObjectCount;
            }

            if (!allocCheck || (checkAllocbit(runningPos-startofHeap,as))) {
                //theObjArray[totalObjectCount] = runningPos;

                // transfer the object to the list of objects
                processFoundObject(runningPos, allocCheck,
                                    heapId,objLength);


                theObjectCount++;

            }


            if (theObjectCount >= (savedObjectCount + 5000)) {
                cpr.output("\t"+  DvUtils.getTime()+ " " +
                     theObjectCount + " objects processed...");
                savedObjectCount = theObjectCount;
            }
        } // end of for loop traversing heap

        cpr.output(theOutput);
        theOutput.clear();

        cpr.output("\t"+theHeapName+ " traversal finished - "
                    + theObjectCount + " objects");


        returnControl();
        return (int)theObjectCount;
    }

    /**
     *  Process all Thread Local Heaps (TLH).
     *
     *  Find all the processes in the dump, find all their threads and
     *  for all the Java ones, run traverseOneTLH on it to find its objects.
     */
    private void traverseEveryTLH()
    {
        int objectCount=0;
        javaThreadsWithoutTLH=0;
        javaThreadsWithTLH=0;
        javaThreadsWithoutTHTLH=0;
        javaThreadsWithTHTLH=0;

        if (theDump == null) {
            theOutput.add("Unable to comply as no Dump opened yet\n");
            returnControl();
            return;
        }
        DvDump      theDump               = DvConsole.getCurrentDump();
        as                                = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        theOutput.add("\nTraversing Thread Local Heaps \n");
        cpr.output(theOutput);
        theOutput.clear();

        Vector ranges = new Vector();

        try {
            // get all processes in the address space
            DvProcess[] allProcesses=as.getProcesses();
            for (int i=0; i<allProcesses.length; i++) {
                // get all threads in the process
                DvThread[] allThreads=allProcesses[i].getThreads();
                for (int j=0; j<allThreads.length; j++) {
                    if(allThreads[j].isJava()){
                        // for Java threads only, get its TLH objects
                        objectCount += traverseOneTLH(allThreads[j],j,ranges);

                    } else {
                        if(verboseTLH) theOutput.add("        "+allThreads[j]+" non-Java thread");
                    }
                }
            }
        } catch(Exception e) {
            theOutput.add("*** Exception received while traversing TLH - "+e.getMessage());
            cpr.output(theOutput);
            theOutput.clear();
            e.printStackTrace();
        }

        // At the end of above process all the range vectors should have been
        // filled.....
        int rSize = ranges.size();
        if (rSize != 0) {
            rangeArray =  new long[rSize][3];
            for (int i =0;i<rSize;i++) {
                 Range r = (Range) ranges.get(i);
                 rangeArray[i][0] = r.start;
                 rangeArray[i][1] = r.end;
                 rangeArray[i][2] = r.index;

            }
        }

        // final status and set global var
        theOutput.add("        "+
                      javaThreadsWithoutTLH+ " of "+
                      (javaThreadsWithTLH+javaThreadsWithoutTLH)+
                      " java threads did not have TLH cache blocks");
        theOutput.add("        "+
                      javaThreadsWithoutTHTLH+ " of "+
                      (javaThreadsWithTHTLH+javaThreadsWithoutTHTLH)+
                      " java threads did not have TH-TLH cache blocks");
        theOutput.add("        TLH finished - "+objectCount+" objects\n");
        cpr.output(theOutput);
        theOutput.clear();

        threadLocalH=objectCount;
    }

/**
     *  Go through the TLH for specified thread, finding its TLH objects
     */
    private int traverseOneTLH(DvThread aThread, int thread_index,
                               Vector ranges)
    {
        int objectCount=0;
        DvAddress Dv_temp=null;
        long cache_block_addr=0, cache_size=0, cache_orig_size=0;
        int MH=0;
        int TH=1;

        // locate alloc cache (TLH)
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        //if (as == null) System.out.println("as is null");
        CTypeObject ee                    = new CTypeObject("execenv", aThread.eeAddress(), as);
        if((ee==null) || (as==null) || (aThread==null) ) {
            theOutput.add("        "+"*** Warning: Unable to process TLH for thread - unexpected NULL variable");
            cpr.output(theOutput);
            theOutput.clear();
            return 0;
        }
        if (aThread.eeAddress().getAddressAsLong()==0) {
            System.out.println("        "+"*** Warning: ee address is ZERO - thread ignored.");
            return 0;
        }

        // Now loop twice, alloc_cahe on first pass, TH_alloc_cache on second
        for (int heap=MH; heap<=TH; heap++) {
            // select which cache structure we're looking at
            CTypeObject CTOtc = null;
            String THString="";
            if (heap==MH) {
                CTOtc = ee.getField("alloc_cache.c.tc");
                THString="";
            } else {
                CTOtc = ee.getField("TH_alloc_cache.c.tc");
                THString="TH-";
            }
            // if tc is NULL, continue
            if(CTOtc==null) {
                theOutput.add("*** Error: Unable to get tc in DvObjectsCommands.traverseOneTLH()");
                cpr.output(theOutput);
                theOutput.clear();
                continue;
            }

            // get cache_block, size and orig_size from cahce structure
            CTypeObject CTO_cache_block     = CTOtc.getField("cache_block");
            CTypeObject CTO_cache_size      = CTOtc.getField("cache_size");
            CTypeObject CTO_cache_orig_size = CTOtc.getField("cache_orig_size");

            // if any of these values are NULL, we're wasting our time
            if((CTO_cache_block==null) ||
               (CTO_cache_size==null) ||
               (CTO_cache_orig_size==null) ) {
                theOutput.add("*** Error: Unexpected NULL variable in DvObjectsCommands.traverseOneTLH()");
                cpr.output(theOutput);
                theOutput.clear();
                continue;
            }

            try {
                // get the address and size of the cache block
                Dv_temp = as.readPointer(as.createAddress(CTO_cache_block.getAddr()));
                cache_block_addr = Dv_temp.getAddressAsLong();
                if(cache_block_addr==0){
                    if(heap==MH)
                        javaThreadsWithoutTLH++;
                    else
                        javaThreadsWithoutTHTLH++;
                    continue;
                } else {
                    theOutput.add("        "+THString+
                                  "TLH cache block located for thread with execenv=0x"+
                                  Long.toHexString(ee.getAddr()) );
                    if(heap==MH)
                        javaThreadsWithTLH++;
                    else
                        javaThreadsWithTHTLH++;
                }
                cache_size = as.readInt( as.createAddress(CTO_cache_size.getAddr()) );
                cache_orig_size = as.readInt( as.createAddress(CTO_cache_orig_size.getAddr()) );
                Range r = new Range(cache_block_addr,cache_block_addr+cache_orig_size,
                                    (long) thread_index);
                ranges.add(r);
            } catch(DvAddressException DvAE) {
                theOutput.add("*** Unexpected exception while getting cache block for "+THString+"TLH");
                cpr.output(theOutput);
                theOutput.clear();
                DvAE.printStackTrace();
            } catch(Exception e) {
                theOutput.add("*** Unexpected other exception while getting cache block for "+THString+"TLH");
                theOutput.add("    Exception type = "+e.getMessage());
                cpr.output(theOutput);
                theOutput.clear();
                e.printStackTrace();
            }

            //starts.add(new Long(cache_block_addr));
            //ends.add(new Long(cache_block_addr+cache_orig_size);
            //threadIndex.add(new Long(threadIndex));
            // Has any of the TLH been used?
            if (cache_orig_size == cache_size){
                if(verboseTLH) theOutput.add("        Nothing in "+THString+"TLH for Thread ("+aThread+")");
            } else {
                if(verboseTLH) theOutput.add("        Objects in "+THString+"TLH for Thread ("+aThread+")");

                // set address obj_ptr to first object on TLH
                // which incidentally begins at the end of the piece
                // of memory that thread has reserved on the heap
                long objPointer= cache_block_addr+cache_size;
                if(verboseTLH) theOutput.add("        "+THString+"TLH objects begin at : 0x"+Long.toHexString(objPointer));

                //starts.add(new Long(cache_block_addr));
                //ends.add(new Long(cache_block_addr+cache_orig_size);
                //threadIndex.add(new Long(threadIndex));
                // loop through all the objects in this thread's TLH
                objectCount += traverseOneTLHCacheBlock(aThread, objPointer,
                                                        (cache_block_addr+cache_orig_size),
                                                        THString);
            }
        }

        return objectCount;
    }

    /**
     *  Go through the cache block representing the TLH, finding its TLH objects
     */
    private int traverseOneTLHCacheBlock(DvThread aThread, long objPointer,
                                         long endOfCacheBlock, String THString)
    {
        int objectCount=0;
        DvAddress Dv_temp=null;
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");

        //int ptrLength; // made static and class wide - wpr
        long objLenMask;
        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            //ptrLength=8;
            objLenMask = 0x3ffffffffffffff8L;
        } else {
            //ptrLength=4;
            objLenMask = 0x3ffffff8L;
        }

        try{
            // loop through all the objects in this threads TLH
            while(objPointer < endOfCacheBlock) {
                long locknFlags=0, methods=0, obj=0, objLen=0;
                CTypeObject object=null;

                try {    /*try catch for DvAddressExceptions in object setup only*/

                    // get the length field & mask off flags
                    Dv_temp=as.readPointer(as.createAddress(objPointer));
                    objLen=Dv_temp.getAddressAsLong();
                    objLen = objLen & objLenMask;

                    // CTypeObject for object
                    object=new CTypeObject("Hjava_lang_Object",
                                           (as.createAddress(objPointer+ptrLength)),
                                            as);

                    // locknflags
                    CTypeObject CTOLocknFlags=object.getField("locknflags");
                    Dv_temp=as.readPointer(as.createAddress(CTOLocknFlags.getAddr()));
                    locknFlags=Dv_temp.getAddressAsLong();

                    // methods
                    CTypeObject CTOMethods=object.getField("methods");
                    Dv_temp=as.readPointer(as.createAddress(CTOMethods.getAddr()));
                    methods=Dv_temp.getAddressAsLong();

                    // obj
                    CTypeObject CTOObj=object.getField("obj");
                    Dv_temp=as.readPointer(as.createAddress(CTOObj.getAddr()));
                    obj=Dv_temp.getAddressAsLong();

                } catch(DvAddressException DvAE) {
                    theOutput.add("*** Unexpected exception processing an object in "+THString+"TLH");
                    theOutput.add("    objPointer="+objPointer);
                    cpr.output(theOutput);
                    theOutput.clear();
                    DvAE.printStackTrace();
                }


                // analyse and write them to file

                // set a flag to say whether the object is an array
                boolean isAnArray=false;
                if ((locknFlags & 0x2L) != 0) isAnArray=true;

                ObjectDetails o = new ObjectDetails();
                o.methods = methods;
                o.obj = obj;
                o.heapName = (THString+"TLH:0x"+Long.toHexString(aThread.eeAddress().getAddressAsLong()));
                o.length = objLen;
                o.address = object.getAddr();
                o.locknflags = locknFlags ;
                o = processData(o); // process locknflags bits
                if(verboseTLH)
                {
                    if(isAnArray) {
                        theOutput.add("            Array  @ 0x"+
                                      Long.toHexString(objPointer + ptrLength)+
                                      " (length=0x"+Long.toHexString(objLen)+") - "+
                                      o.type+"["+methods+"]");
                    }
                    else {
                            theOutput.add("            Object @ 0x"+
                                          Long.toHexString(objPointer + ptrLength)+
                                          " (length=0x"+Long.toHexString(objLen)+") - "+
                                          o.type);
                    }
                 }

                    //objectHashByAddr.put(new Long(objPointer+ptrLength),new Integer(totalObjectCount));
                 if(verboseTLH) theOutput.add("            *** putting: "+
                                                 Long.toHexString(objPointer+ptrLength)+
                                                 ", "+totalObjectCount);

                 processFoundObject(objPointer,true,
                                        (short) 0,objLen);
                 objectCount++;

                 // step down to the next object
                 objPointer += objLen;
            }
        } catch(Exception e){
            theOutput.add("*** Unexpected exception recieved while examining items on "+THString+"TLH");
            theOutput.add("  Thread="+aThread+", objectPointer="+objPointer+", endOfCacheBlock="+endOfCacheBlock);
            cpr.output(theOutput);
            theOutput.clear();
            e.printStackTrace();
        }
        return objectCount;
    }

    /**
     *
     */
    private ObjectDetails processData(ObjectDetails o){

        // o.type is a string set up with the object typre
        // such as "array of x" or "normal object"
        o.type = processTypes(o);


        o.lock = false;  // assume not locked first
        // processShapeBit determines whether we
        // are looking at a normal lock or inflated lock, however
        // we can bypass all this processing if the locknflags is
        // 0 as this means that we are definitely not locked - so
        // we leave all the values at 0.
        if (0 != o.locknflags) {
            o.shapeBit=processShapeBit(o);
            if (0==o.shapeBit) {
                o.entryCount = processEntryCount(o);
                o.threadId = processThreadId(o);
            }
            else{
                o.monitorIndex=processMonitorIndex(o);
                o.inflMon=processInflMon(o);
            }

            if ((0!=o.threadId)||(0!=o.entryCount)||(0!=o.monitorIndex)) {
                o.lock = true;
            }
        }
        else {
            o.entryCount = 0;
            o.threadId = 0;
            o.monitorIndex = 0;
            o.inflMon = 0;
            o.lock = false;
        }


        o.objectName = processObjectName(o);
        o.className = processClassName(o);
        return o;
    }

    /**
     *
     */
    private long processShapeBit(ObjectDetails o){

        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            long retValue = (o.locknflags>>62L)&2L;
            return retValue;
        }
        else { // 32bit
            long retValue = (o.locknflags>>31)&1;
            return retValue;
        }
    }

    /**
     *
     */
    private long processEntryCount(ObjectDetails o){

        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            long retValue = (o.locknflags>>16L)&0x7FFFL;
            return retValue;
        }
        else { // 32bit
            long retValue = (o.locknflags>>8)&0xFF;
            return retValue;
        }
    }

    /**
     *
     */
    private long processThreadId(ObjectDetails o){
        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            long retValue = (o.locknflags>>32L)&0x7FFFFFFFL;
            return retValue;
        }
        else { // 32bit
            long retValue = (o.locknflags>>16)&0x7FFF;
            return retValue;
        }
    }

    /**
     *
     */
    private long processMonitorIndex(ObjectDetails o){
        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            long retValue = (o.locknflags>>16L)&0x7FFFFFFFFFFFL;
            return retValue;
        }
        else { // 32bit
            long retValue = (o.locknflags>>8)&0x7FFFFF;
            return retValue;
        }
    }

    /**
     *  @param o        Object to have monitor displayed
     */
    private long processInflMon(ObjectDetails o){
        as = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        long lkGlPtrAddr = 0;
        long dgGlPtrAddr = 0;
        //try {
             lkGlPtrAddr=Long.parseLong(DvUtils.getJVMHint(as,"LKGLOBALPTR"),16);
             dgGlPtrAddr=Long.parseLong(DvUtils.getJVMHint(as,"DGGLOBALPTR"),16);
        //}
        //catch (NumberFormatException e) {
           // theOutput.add("               NumberFormatException caught in processInflMon");
           // theOutput.add("               This will not work without the modified JVM ie. needs lk: dataP");
           // returnControl();
           // return 0;
        //}
        if (lkGlPtrAddr==dgGlPtrAddr) {
            //theOutput.add("               This will not work without the modified JVM ie. needs lk: dataP");
            returnControl();
            return 0;
        }
        long objAddr=o.address;
        long inflMon=0;
        // format lkGlPtr as lkglData  or +4 ptrfromptr
        long micbAddr;
        long inflCount;
        long inflSize;
        long inflMonTable;
        long inflMonAddr;
        try{
            if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
                lkGlPtrAddr += 8;
                micbAddr = as.readPointer(as.createAddress(lkGlPtrAddr)).getAddressAsLong();
                inflCount=as.readLong(as.createAddress(micbAddr));
                micbAddr += 8;
                inflSize=as.readLong(as.createAddress(micbAddr));
                micbAddr += 8;
                inflMonTable =
as.readPointer(as.createAddress(micbAddr)).getAddressAsLong();
                inflMonAddr = inflMonTable + (4*o.monitorIndex);

             }
             else { // 32bit
                lkGlPtrAddr += 4;
                micbAddr = as.readPointer(as.createAddress(lkGlPtrAddr)).getAddressAsLong();
                inflCount=(long)as.readInt(as.createAddress(micbAddr));
                micbAddr += 4;
                inflSize=(long)as.readInt(as.createAddress(micbAddr));
                micbAddr += 4;
                inflMonTable =
as.readPointer(as.createAddress(micbAddr)).getAddressAsLong();
                inflMonAddr = inflMonTable + (4*o.monitorIndex);

             }
             inflMon =
as.readPointer(as.createAddress(inflMonAddr)).getAddressAsLong();
        }
        catch (DvAddressException e) {
            theOutput.add("DvAddressException in processInflMon");
            returnControl();
        }
        catch (NumberFormatException e) {
            theOutput.add("DvNumberFormatException in processInflMon");
            theOutput.add("Possible cause by not having hints setup properly");
            returnControl();
        }
        return inflMon;
    }

    /**
     *
     */
    private boolean processLock(ObjectDetails o){
         // processShapeBit, processThreadId, processEntryCount, processMonitorIndex must be done first
         if ((0!=o.threadId)||(0!=o.entryCount)||(0!=o.monitorIndex)) {return true;}
         else return false;
    }

    /**
     *  Gets name of object..
     *  Requires the global DvDump theDump
     *  @param
     *  @return String
     */
    private String processObjectName(ObjectDetails o){
        String name="";
        long theObjAddress = o.address;
        return processObjectName(theObjAddress, (DvAddressSpace) DvUtils.getAPT(theDump,"A"));
    }

    /**
     *  Gets name of object..
     *  Requires the AddressSpace and objectAddress
     *  @param
     *  @return String
     */
    private String processObjectName(long theObjAddress,DvAddressSpace theAs) {
        long retLong;
        String name="";
        // test to see if not object
        //System.out.println("OA:"+Long.toHexString(theObjAddress));
        boolean object=true;
        long locknflags=0;
        try {
            if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
                locknflags =  (long)as.readLong(as.createAddress(theObjAddress+8));
            }
            else {
                 locknflags =  (long)as.readInt(as.createAddress(theObjAddress+4));
            }
        }
        catch (DvAddressException e) {
            theOutput.add("DvAddressException in processObjectName ");
            returnControl();
            return name;
        }
        //System.out.println("OA:"+Long.toHexString(theObjAddress)+":"+locknflags);
        if ((((locknflags>>DvTypeCodes.OBJ_TYPECODE_SHIFT)&DvTypeCodes.OBJ_TYPECODE_MASK)==0)&&
            ((locknflags>>1)&1)==0){
            try{
                long methodsAddr = DvUtils.getFieldValue(theObjAddress,new String[] {"Hjava_lang_Object", "methods"},theAs,theDump);
                retLong = methodsAddr;
                long classdescriptorAddr = DvUtils.getFieldValue(methodsAddr,new String[] {"methodtable", "classdescriptor"},theAs,theDump);
                long objAddr = DvUtils.getFieldAddress(classdescriptorAddr,new String[] {"Hjava_lang_Class","obj"},theAs);
                long nameAddr = DvUtils.getFieldValue(objAddr,new String[] {"Classjava_lang_Class","name"},theAs,theDump);
                name = DvUtils.getStringFromBytes(theAs.readBytes(theAs.createAddress(nameAddr),256));

            }
            catch (DvAddressException e) {
                if (debug) theOutput.add("DvAddressException in processObjectName(long theObjAddress,DvAddressSpace theAs)");
                if (debug) theOutput.add("Possible cause by address not being in dump");
                object=false;
            }
        }
        else {
//            theOutput.add("not object");
            object=false;
        }

        if (!object) {
            retLong = (locknflags>>DvTypeCodes.OBJ_TYPECODE_SHIFT)&DvTypeCodes.OBJ_TYPECODE_MASK;
            name = processTypes(locknflags);
        }
        // if class class get name
        return name;
    }

    /**
     *  Gets name of class when object is class
     *  Requires the global DvDump theDump
     *  @param theObjAddress    The start address of the object
     *  @return                 The name of the class
     */
    private String processClassName(ObjectDetails o){
        if (0!=o.objectName.compareToIgnoreCase("java/lang/Class")) {return null;}
        long theObjAddress = o.address;
        String name=null;
        // test to see if not object
        DvAddressSpace theAs = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
        try{
            long objAddr = DvUtils.getFieldAddress(theObjAddress,new String[] {"Hjava_lang_Class","obj"},theAs);
            long nameAddr = DvUtils.getFieldValue(objAddr,new String[] {"Classjava_lang_Class","name"},theAs,theDump);
            name = DvUtils.getStringFromBytes(theAs.readBytes(theAs.createAddress(nameAddr),256));
        }
        catch (DvAddressException e) {
            if (debug) theOutput.add("DvAddressException in processClassName");
        }
        return name;
    }

    /**
     *  Processes locknflags and returns the value of the type
     *  Looks up values from DvTypeCodes
     *
     *  @param theLockNFlags    locknflags value
     *  @return                 value of type only.
     */
     private String processTypes(ObjectDetails o){
        long theLockNFlags = o.locknflags;
        return processTypes(theLockNFlags);
    }

    /**
     *  Processes locknflags and returns the value of the type
     *  Looks up values from DvTypeCodes
     *
     *  @param theLockNFlags    locknflags value
     *  @return                 value of type only.
     */
     private String processTypes(long theLockNFlags){
        String theType = "";
        switch((int)theLockNFlags&7) {
            case (int)DvTypeCodes.OBJ_HashedAndMoved : theType += "hashedAndMoved "; break;
            case (int)DvTypeCodes.OBJ_IsArray        : theType += "array "; break;
            case (int)DvTypeCodes.OBJ_Hashed         : theType += "hashed "; break;
        }
        switch((int)(theLockNFlags>>DvTypeCodes.OBJ_TYPECODE_SHIFT)&DvTypeCodes.OBJ_TYPECODE_MASK) {
            case (int)DvTypeCodes.T_CHAR             : theType += "char"; break;
            case (int)DvTypeCodes.T_BOOLEAN          : theType += "boolean"; break;
            case (int)DvTypeCodes.T_FLOAT            : theType += "float"; break;
            case (int)DvTypeCodes.T_DOUBLE           : theType += "double"; break;
            case (int)DvTypeCodes.T_BYTE             : theType += "byte"; break;
            case (int)DvTypeCodes.T_SHORT            : theType += "short"; break;
            case (int)DvTypeCodes.T_INT              : theType += "int"; break;
            case (int)DvTypeCodes.T_UINTEGER         : theType += "uintpointer"; break;
            case (int)DvTypeCodes.T_LONG             : theType += "long"; break;
            case (int)DvTypeCodes.T_CLASS            : theType += "class"; break;
            case (int)DvTypeCodes.T_NORMAL_OBJECT    : theType += "normal_object"; break;
        }
        return theType;


    }



    private boolean checkAllocbit(long distInHeap,DvAddressSpace theAs){

        DvUtils.trace("Entry:checkAllocbit  (0x"+Long.toHexString(distInHeap)+")",0,false);

        // The allocbits memory is effectively an array of java ints -
        // but with the added complication that they aren't really ints
        // (big endian and little endian meaning comes into this) they are
        // really a collections of bytes - thus 0x80000000 on a bigendian system
        // (mvs) coreesponds to 0x01000000 on a little endian system - this
        // stuff will blow your brain if you let it.

        // from our perspective the simplest thing to do is treat the allocbits
        // as a virtual byte array - so first we figure out which byte
        // corresponds to our address in the heap (1 bit in alocbits = 8 bytes in heap,
        // and 1 byte = 64 bytes) so we figure out the offset of the byte we are
        // interested in by taking the distance from the heap start (which is always
        // a multiple of 8!!!) as a long and dividing this by 64
        // thus 0 means we are interested in byte 0,64 in byte 1, 128 in byte 2 etc...
        // The remainder from the division is itself divided by 8 and tells us which
        // of the 8 bits we are interested in, with the proviso that we use the 8
        // complement on little endian systems!!!!
        boolean bRet = false; // assume bit isn't set.
       // DvAddressSpace theAs = (DvAddressSpace) DvUtils.getAPT(theDump,"A");
       // long startofheap = DvUtils.hexToLong(DvUtils.getJVMHint(theAs,"MH_HEAPBASE"));
       // long distInHeap = theRunningPos - startofheap;
        int allocBitIndex = 0;
        long testValue = 0;
        byte allocTableByte=0x00;
        long allocBitsByteOffset = distInHeap/64;
        // Now read this byte
        if (0 == allocBitsAddress) {
            allocBitsAddress = DvUtils.hexToLong(DvUtils.getJVMHint(as,"ALLOCBITS"));
        }
        if (0 != allocBitsAddress) {
            try {
                allocTableByte = as.readByte(theAs.createAddress(
                                     allocBitsAddress+allocBitsByteOffset));

            }
            catch (DvAddressException e) {
                // Should never happen!!!!
                DvUtils.trace  ( "\n !!!!!!!!!!!!!!!!!!!!!!" +
                                "\n DvAddressException trying to read AllocBits" +
                                "\n Address = 0x" + Long.toHexString(allocBitsAddress+allocBitsByteOffset) +
                                "\n Address: " +
                                "\n !!!!!!!!!!!!!!!!!!!!!!",0,true);
            }
        }
        else {
            // Should never happen
            DvUtils.trace  ( "\n !!!!!!!!!!!!!!!!!!!!!!" +
                             "\n ALLOCBITS hint was 0 - suggests an error in control"+
                             "\n block structure or wrong types file information"+
                             "\n !!!!!!!!!!!!!!!!!!!!!!",0,true);
        }

        // Now figure out which bit of our byte is the interesting one
        allocBitIndex = (int) (distInHeap%64)/8;


        if (false == DvUtils.isBigEndian(theDump.getArchitecture())) {
              allocBitIndex = 7-allocBitIndex;
        }

        //DvUtils.trace("/tallocBitIndex="+allocBitIndex+ "allocTableByte=0x" +
        //              Long.toHexString((long)allocTableByte),0,false);
        if (0 != (allocTableByte & allocBitsTestBytes[allocBitIndex])) {
            bRet = true;
        }


        DvUtils.trace("Exit:checkAllocbit ("+(bRet?"true)":"false)"),0,false);

        return bRet;
    }



    /**
     *  Outputs anything waiting to be but to the console
     *  and resets "theOutput" buffer
     */
    private void returnControl(){
        cpr.output(theOutput);
        theOutput.clear();
        forcedEnd = true;
    }

    /**
     *  A data class for storing range information for later use.
     */
    private static class Range{
        protected long start;
        protected long end;
        protected long index; // -1 = ACS 0 onwards is thread index


        protected Range(long start,long end, long index) {
            this.start=start;
            this.end = end;
            this.index = index;
        }



    }

    /**
     *  A data class for storing object information for later use.
     *  Allows us to store information when object has been processed
     *  so can retreive quickly
     */
    private static class ObjectDetails{

        // information stored in Control Block (JVMObject)
        protected long address;   // where the object is in memory
        protected long methods;
        protected long locknflags;
        protected long obj;

        // information gained by processing.
        protected String objectName; // whats the name of this object
        protected String className; // if this object is  a class then name of class
        protected String heapName;
        protected String type;  // from locknflags
        protected boolean lock;  // from locknflags
        protected long shapeBit;  // from locknflags
        protected long entryCount;  // from locknflags
        protected long threadId;  // from locknflags
        protected long monitorIndex;  // from locknflags
//        protected long inflMonAddr; // from monitorIndex
        protected long inflMon; // from monitorIndex

        protected long length;

        public String toString() {
            if (null == className) {
                return ("\n @ 0x"+Long.toHexString(address) +
                   " (" + objectName  + ") " +  "  (heap: " + heapName  + ")");
            }
            else {

                return ("\n @ 0x"+Long.toHexString(address) +
                   " (" + objectName  + ") " + className + "  (heap: " + heapName  + ")");
            }
        }

    }
    /**
     *  guiPopupSyntax
     *
     *  @return (String[]) null
     */
    public String[] guiPopupSyntax(){return null;}

    /**
     *   Static method for the use of the Help command as
     *   implemented in DvBaseCommands
     */
    public static Object getGenericHelp() {
        //DvUtils.trace(" Entry DvClassCommands:getGenericHelp",0,false);
        Vector v = DvUtils.getMultiplePropertyValues("GeneralHelp",
                        "DvObjectsCommands.properties");

        return v;

    }

   //private ObjectDetails getObjectDetails(int i) {
        //long workingPos = theObjArray[i];
   //     return getObjectDetails( workingPos);
   //}


   /**
     *   Expand the object into something more presentable
     *
     */
    private ObjectDetails getObjectDetails(long workingPos) {
        // So we use the ith entry out of the array to create an
        // ObjectDetails structure
        //long workingPos = theObjArray[i];
        long objLength = 0;
        long methods = 0;
        long locknflags = 0;
        long obj = 0;
        long objAddress;
        ObjectDetails o = new ObjectDetails();
        try {
            if (is64bit) {
                objLength = as.readLong(as.createAddress(workingPos));
                objLength = objLength & 0x3ffffffffffffff8L;
                //move on to next field
                workingPos+=8;
                objAddress = workingPos;
                // now read the methods
                // the locknflags
                // and the obj (JVMObject fields)

                methods = as.readLong(as.createAddress(workingPos));
                //move on to next
                workingPos+=8;
                locknflags =  as.readLong(as.createAddress(workingPos));
                //move on to next
                workingPos+=8;
                obj =  as.readLong(as.createAddress(workingPos));
             }
             else {
                 // Its 32 bit
                 objLength =(long) as.readInt(as.createAddress(workingPos));
                 objLength = objLength & 0x3ffffff8 ;
                 //move on to next
                 workingPos+=4;
                 objAddress = workingPos;
                 // now read the methods
                 // the locknflags
                 // and the obj (JVMObject fields)
                 methods = (long) as.readInt(as.createAddress(workingPos));
                 //move on to next
                 workingPos+=4;
                 locknflags =  (long)as.readInt(as.createAddress(workingPos));
                 // anding need because of casting when read int
                 locknflags = locknflags & 0x00000000ffffffffL;

                 //move on to next
                 workingPos+=4;
                 obj =  as.readInt(as.createAddress(workingPos));


             }

             o.methods = methods;
             o.obj = obj;
             // we figure out the heapName by first looking in rangeArray to see if its
             // in there - we then know its in the syatem or TLH - if its in TLH then we
             // figure out which thread by gettin the thread array
             // If its outside the above ranges then if its within the start/end of the
             // middleware or transient heap we assume that the object is one of those
             // and report so.
             o.heapName = establishHeapName(objAddress);
             if (null == o.heapName) {
                o.heapName = "??????";
             }

             o.length = objLength;
             o.address = objAddress;
             o.locknflags = locknflags;
            // process locknflags bits
             o = processData(o);
        }
        catch (DvAddressException e) {
            if (debug) theOutput.add("DvAddressException in getObjectDetails");
            // what to do?
        }

        return o;
    }

   /**
     *   open existing names index or create new one
     *
     */
    private void openNamesIndexFile() {


            String namefile = DvUtils.getValue("Dump")+".nidx";
            String workingDir = DvUtils.getValue("WorkDir");
            if (null != workingDir) {
                namefile = DvUtils.stripToFileName(namefile);
                namefile = workingDir +  namefile;

            }



            try {
                System.out.println(" Looking for existing Index file "+namefile);
                namesIndexFile = new DvNamesIndexFile(namefile,"r",0,0,0,123456);
                names_index_built = true;

            }
            catch (java.io.IOException e) {
                theOutput.add("Invalid or non-existant class names index file"+
                    "\nDump will be scanned and index created");
                cpr.output(theOutput);
                theOutput.clear();
                // So lets create one
                try {
                    namesIndexFile = new DvNamesIndexFile(namefile,"rw",1024,100,100,123456);
                    System.out.println(" Created a nameIndex file");
                }
                catch(java.io.IOException e1) {
                    System.out.println(" Couldn't create a nameIndex file" +
                                        " '"+namefile+"'");
                }
            }


    }

    /**
     *   open existing object addr index or create new one
     *
     */
    private void openAddrIndexFile() {



           String addressfile = DvUtils.getValue("Dump")+".aidx";
           String workingDir = DvUtils.getValue("WorkDir");
           if (null != workingDir) {
                addressfile = DvUtils.stripToFileName(addressfile);
                addressfile = workingDir +  addressfile;

           }



           try {
                System.out.println(" Looking for existing addressIndex file");
                addressIndexFile = new DvAddressIndexFile(addressfile,"r",0,0,0,123456);
                address_index_built = true;

            }
            catch (java.io.IOException e) {
                theOutput.add("Invalid or non-existant addressindex file"+
                    "\nDump will be scanned and index created");
                cpr.output(theOutput);
                theOutput.clear();
                // So lets create one
                try {
                    addressIndexFile = new DvAddressIndexFile(addressfile,"rw",32768,100,100,123456);
                    System.out.println(" Created an addressIndex file");
                }
                catch(java.io.IOException e1) {
                    System.out.println(" Couldn't create a addressIndex file" +
                                        " '"+addressfile+"'");
                }
            }
    }

    private void initPtrLength() {
        // pointer length is used in several places -
        // we establish it here so that everyone can get at it
        if (DvUtils.is64BitSystem(theDump.getArchitecture())) {
            ptrLength=8;

        }
        else
        {
            ptrLength=4;
        }
    }


    /**
     *   open existing object addr index or create new one
     *
     */
    public void objectAnalyze() {

    }


    /**
     *   open existing object addr index or create new one
     *
     */
    public void createEAB(DvAddressSpace as) {
        // figure out the size of the heap!!!
         String base = DvUtils.getJVMHint(as,"MH_HEAPBASE");
         String limit = DvUtils.getJVMHint(as,"MH_HEAPLIMIT");
         startOfMHeap = Long.parseLong(base,16); // need this to figure out which bit to flip
         endOfMHeap  = Long.parseLong(limit,16);
         int howbig = (int) ((endOfMHeap - startOfMHeap)/64);
         DvUtils.trace("Allocating "+ howbig + " bytes for extendedAllocBits",
                        0,false);

         extendedAllocBits = new byte[howbig];

    }

    /**
     *   flip the allocated bit in the EAB
     *
     */
    public void flipEAB(long addr) {
        if (extendedAllocBits == null) {
            createEAB(as); // go and init the EAB
        }
        long flip_byte = addr;
        int  flip_position;
        long distInHeap = flip_byte - startOfMHeap;
        flip_byte = distInHeap/64;
        flip_position = (int) (distInHeap%64)/8;
        // now flip the bit - but first check that it ain't already
        // flipped !!! If it is then record a glitch.....
        // This keeps the other heap traversers honest
        try {
            if ( 0 == (extendedAllocBits[(int)flip_byte] &
                allocBitsTestBytes[flip_position])) {
                // flip the bit as it isn't already
                //System.out.println("flipping bit for "+Long.toHexString(addr));
                extendedAllocBits[(int)flip_byte] |=
                    allocBitsTestBytes[flip_position];

                //if ( 1 == (extendedAllocBits[(int)flip_byte] &
                //        allocBitsTestBytes[flip_position])) {
                //    if (doOnce)  {
                     //System.out.println("flipped bit for "+Long.toHexString(addr));
                     //System.out.println("flip_byte: "+flip_byte+ " flip_position: "+flip_position);
                     int ival =  extendedAllocBits[(int)flip_byte];
                     if (ival < 0) {
                        ival=ival+256;
                     }

                     //System.out.println("flip_byte: " + (DvUtils.byteToHex).substring(2*ival,2*ival+2));
                //     doOnce = false;
                //    }
                //}

            }
            else {
                DvUtils.trace(" !!! Found extendedAllocBits inconsistency "+
                          "\n !!! for: 0x" + Long.toHexString(addr)
                          + " bit already on !", 0,true);
            }
        }
        catch (ArrayIndexOutOfBoundsException  e) {
            DvUtils.trace(" !!! Trying to flip bit outside of array for 0x"+
                    Long.toHexString(addr),0,true);
        }
    }


    /**
     *   Check if an object is Known!!!
     *
     */
    public boolean checkForKnownObject( long addr)  {

        //System.out.println("startOfMHeap:" + startOfMHeap +"  endOfMHeap:"+ endOfMHeap +
        //            "  addr:"+addr);
        boolean bRet=false; // assume they have given us a bad one


        // first we check that its on an 8 byte boundary (otherwise we refuse to play)
        // but mustn't forget that we stored addr of length rather than object
        if (0 == ((addr+ptrLength)%8)) {
            // now we check up on the EAB (assuming address is within range)

            if (addr >= startOfMHeap  && addr <= endOfMHeap) {

                long test_byte = addr;
                int  test_position;
                long distInHeap = test_byte - startOfMHeap;
                test_byte = distInHeap/64;
                test_position = (int) (distInHeap%64)/8;
                //System.out.println("here I am addr ="+ addr);
                //System.out.println("test_byte: "+test_byte+ " test_position: "+test_position);
                //int ival =  extendedAllocBits[(int)test_byte];
                //     if (ival < 0) {
                //        ival=ival+256;
                //     }
                //System.out.println("flip_byte: " + (DvUtils.byteToHex).substring(2*ival,2*ival+2));
                if ( 0 != (extendedAllocBits[(int)test_byte] &
                    allocBitsTestBytes[test_position])) {

                    bRet = true;
                }


            }
            else {

                // we check whether it was a system one (not on main heap!!!)
                // we check whether it was an ACS one (not on main heap!!!);
                Long lAddr = new Long(addr);
                if (systemHeapHashByAddr.contains(lAddr) ||
                    acsHeapHashByAddr.contains(lAddr)) {

                    bRet=true;
                }



            }
        }

        return bRet;

    }


    /**
     *   One of the treverses has found an object
     *   - so we add it to the record
     *
     */
    private void processFoundObject(long runningPos,boolean isEAB,
                                    short heapId,long objLength) {

        String name;
        if (false ==names_index_built) {
                 name = processObjectName(runningPos+ptrLength,as);
                 namesIndexFile.write(name, runningPos,heapId,objLength);
        }


        if (true == isEAB) {
            flipEAB(runningPos);
        }
        else {
            // For the system and ACS Heaps we record the existance of stuff by using a
            // hash set coupled to an array of start/end points - thus we can
            // tell when an object is in a system heap (and which one its in).
            // This is not a performance hit as there aren't a huge number of
            // objects on the system heap .....
            if (heapId == DvConstants.HEAP_SYS) {
                systemHeapHashByAddr.add(new Long(runningPos));
            }
            else {
                acsHeapHashByAddr.add(new Long(runningPos));
            }
            //System.out.println(" Adding Sys 0x"+ Long.toHexString(runningPos));

        }

        // Now we want to look see whether this is a locked object
        // so lets read in the locknflags - if they are non zero then we
        // need to figure out whether this is a locked object and
        // store it away for later use... thus avoiding scanning all
        // the objects again
        long locknflags=0;
        try {
            if (is64bit) {
                locknflags = as.readLong(as.createAddress(runningPos+16));
            }
            else {
                locknflags = as.readInt(as.createAddress(runningPos+8));
            }
        }
        catch (DvAddressException e) {
            System.err.println("DvAddressException in processFoundObject");
        }

        // Ones with locknflags other than 0 might have locks
        // but then again they might just be arrays so lets truy and
        // seperate out the pure arrays from the locked objects
        if (locknflags != 0 ) {
            //locknflags &= 0xfffffffffffffff8L;
            locknflags &= 0xfffffffffffff800L;
           // theLockNFlags>>DvTypeCodes.OBJ_TYPECODE_SHIFT)&DvTypeCodes.OBJ_TYPECODE_MASK
            if (locknflags != 0 ) {
                // We have a locked object - lets save it away
                // for later analysis by "dis lo"
                lockedObjects.add(new Long(runningPos));
                totalLockedObjectCount++;
                //System.out.println("locknflags = "+
                //locknflags + " pos: 0x"+Long.toHexString(runningPos));
            }

        }

        totalObjectCount++;
    }




    /**
     *   One of the traverses has found an object
     *   - so we add it to the record
     *
     */

    private int[] sortRanges(Vector ranges) {

        int[] sortList = null;
        if (0 != ranges.size()) {
            // now get a sort of the ranges vector as an array of ints
            sortList =  new int[ranges.size()];
            long[] startPoints = new long[ranges.size()];
            for (int i=0;i<sortList.length;i++) {
                Range r = (Range) ranges.get(i);
                //System.out.println("rrr:"+r.start+ ":" + r.end);
                if (0==i) {
                    sortList[0]=0;
                    startPoints[0]=r.start;
                }
                else {

                    for (int j=i-1;j>-1;j--)  {
                        if (r.start <= startPoints[j]) {
                           sortList[j+1] = sortList[j];
                           startPoints[j+1]=startPoints[j];
                           sortList[j] = i;
                           startPoints[j] = r.start;
                        }
                        else {
                            sortList[j+1] = i;
                            startPoints[j+1]=r.start;
                            j=-1;
                        }


                    }

                }

            }

            //for (int i=0;i<ranges.size();i++) {
            //    System.out.println("\nxxx:"+sortList[i] + " : "+ startPoints[i]);
            //}
        }

        return sortList;
    }


    /**
     *   Figure out what the heap name should be
     *
     */
    private String establishHeapName(long obj) {

        String retString = null;


        // first we look at the stored ranges to figure out whether its in the
        // System or TLH's - we could do it with a binary chop check but
        // because there are generally so few of these ranges we do it with
        // a straight scan (may change the code later to do both depending
        // on the size of the array.

        //for (int i=0;i<rangeArray.length;i++) {
        //    System.out.println(rangeArray[i][0]+":"+rangeArray[i][1]);
        //}

        for (int i=0;i<rangeArray.length;i++) {

            if (obj >= rangeArray[i][0] && obj <= rangeArray[i][1] ) {
                    // we are in the zone
                    if (rangeArray[i][2] == -1) {
                        return "System";
                    }
                    else
                    {
                        DvProcess p = (DvProcess) DvUtils.getAPT(theDump,"P");
                        DvThread[] threads = p.getThreads();

                        return ("TLH for thread " + threads[(int) rangeArray[i][2]].id()) ;
                    }

            }
            else {
                if (obj < rangeArray[i][0]) {
                    i = rangeArray.length;
                }
            }

         }

         // So it wasn't system or TLH
         // Lets try MiddlewareHeap and Transient
         //System.out.println("o:"+obj+ " s:"+ startOfMHeap + " e:" + endOfMHeap);
        if (obj>= startOfTransHeap && obj < endOfTransHeap) {
            return "Transient";

         }

         if (obj>= startOfMHeap && obj < endOfMHeap) {
            return "MiddleWare";

         }
         return "TBD";



    }

}
