/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvUtils.java, ras, dumpviewer 1.51
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502  PHILR:  New for feature 36524
 * 36524  200502  philr   Changed so that no knowledge of gui
 * 36524  060602  philr   Support for find command plus standard adherance
 * 36524  070602  hdrc    Hints for allocbits added.
 * 36524  120602  philr   filtrate method added
 * 36524  260602  philr   Support for resetting jvmhints when new dump opened
 * 36524  260602  philr   Support for trace
 * 36524  030702  philr   Javadoc changes to method headers
 * 36524  080702  philr   Removed type files header check
 * 36524  110702  philr   Trace added to setupJVMHints
 * 36524  110702  philr   Support for ebcdic formatting
 * 36524  250702  hdrc    Hint for dg dataP
 * 36524  050802  jkbar3  Change to loading plugins messages
 * 36524  060802  philr   Added getTime function
 * 36524  060802  philr   Upgrade setValue to support remove
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import javax.swing.*;
import java.awt.*;
import java.lang.reflect.Method;
import java.util.*;
import java.io.*;
import java.text.*;


/**
 *  Description: The methods in this class are generic methods to be used
 *               across the whole of the DvConsole and Dumpviewer application.
 *               As such all methods should be static, allowing invocation
 *               without the need to instantiate DvUtil.
 *
 */
public class DvUtils {

    // The three currentXXX variables are used as a fastpath for
    // working with things - in general they will correspond to
    // the ASID/PROC/THREAD values that are SET when a dump is
    // opened.
    private static DvThread  currentThread=null;
    private static DvProcess  currentProcess=null;
    private static DvAddressSpace currentAsid=null;
    private static boolean consoleOutput=true;
    private static Properties setValues=null;
    private static boolean   consoleTrace=false;
    private static Dvifm     theIFM;
    private static String    traceString,traceString1,traceString2;
    private static boolean order12 = true;
    private static Properties dvProperties;
    private static JTextArea tc_text = null;
    public  static String    byteToHex =
                  "000102030405060708090A0B0C0D0E0F" +
                  "101112131415161718191A1B1C1D1E1F" +
                  "202122232425262728292A2B2C2D2E2F" +
                  "303132333435363738393A3B3C3D3E3F" +
                  "404142434445464748494A4B4C4D4E4F" +
                  "505152535455565758595A5B5C5D5E5F" +
                  "606162636465666768696A6B6C6D6E6F" +
                  "707172737475767778797A7B7C7D7E7F" +
                  "808182838485868788898A8B8C8D8E8F" +
                  "909192939495969798999A9B9C9D9E9F" +
                  "A0A1A2A3A4A5A6A7A8A9AAABACADAEAF" +
                  "B0B1B2B3B4B5B6B7B8B9BABBBCBDBEBF" +
                  "C0C1C2C3C4C5C6C7C8C9CACBCCCDCECF" +
                  "D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF" +
                  "E0E1E2E3E4E5E6E7E8E9EAEBECEDEEEF" +
                  "F0F1F2F3F4F5F6F7F8F9FAFBFCFDFEFF"
                 ;

    private static  boolean verbose=false; // controls trace

    public  static String    byteToAscii =
                  "................" +
                  "................" +
                  " !\"#$%&'()*+'-./" +
                  "0123456789:;<=>?" +
                  "@ABCDEFGHIJKLMNO" +
                  "PQRSTUVWXYZ[\\]^_"+
                  "`abcdefghijklmno" +
                  "pqrstuvwxyz{|}~." +
                  "................" +
                  "................" +
                  "................" +
                  "................"+
                  "................" +
                  "................"+
                  "................" +
                  "................"
                 ;
    public  static String    byteToEbcdic =
                  "................" +
                  "................" +
                  "................" +
                  "................" +
                  "...........<(+|" +
                  "&.........!$*);"+
                  "-/.........,%_>?" +
                  "..........:#@'=\"" +
                  ".abcdefghi......" +
                  ".jklmnopqr......" +
                  "..stuvwxyz......" +
                  ".........`......"+
                  ".ABCDEFGHI......" +
                  ".JKLMNOPQR......"+
                  "..STUVWXYZ......" +
                  "0123456789......"
                 ;
    private static boolean typeFileLoaded=false;
    private static Vector jvmHints=null;

    private static DateFormat df = DateFormat.getTimeInstance();

    private static String[] hintListJvm =
        {
            "Jvm @ 0x,!,JVM,",
            "lk :,LockInterface @ 0x,LOCKINTERFACE",
            "lk :,dataP = 0x,LKGLOBALPTR",
            "dg :,dataP = 0x,DGGLOBALPTR",
            "st :,dataP = 0x,STGLOBALPTR"

        };

    private static String[] hintListSTGlobal =
        {
            "allocbits = 0x,!,ALLOCBITS",
            "MH_heapbase = 0x,!,MH_HEAPBASE",
            "MH_heaplimit = 0x,!,MH_HEAPLIMIT",
            "curHeapMin = 0x,!,CURHEAPMIN",
            "curHeapMax = 0x,!,CURHEAPMAX",
            "TH_heapbase = 0x,!,TH_HEAPBASE",
            "TH_heaplimit = 0x,!,TH_HEAPLIMIT",
            "loadedACSClasses = 0x,!,LOADEDACSCLASSES",
            "loadedSystemClasses = 0x,!,LOADEDSYSTEMCLASSES",
            "loadedClasses = 0x,!,LOADEDCLASSES"
        };


    /**
     * Generic error message handler output when
     * something unexpected goes wrong.
     *
     *
     */
    public static void errorMsg(String error, int severity) {

        System.err.println("\n***** ERROR **********");
        System.err.println(error);
        System.err.println("***** ERROR **********\n");

    }

    /**
     * Generic error message handler output when
     * something unexpected goes wrong.
     *
     *
     */
    public static void infoMsg(String info) {

        System.out.println("\n***** Information **********");
        System.out.println(info);
        System.out.println("***** Information **********\n");

    }


    /**
     * Inform trace of the text area that has been established
     * and added to the trace frame.
     *
     */
    public static void setTc_text(JTextArea ta) {
        tc_text = ta;
    }


    /**
     * Get a name property value from Dv.properties
     *
     */
    public static String getDvProperty(String propname) {

        String retstr = null;
        String sep = System.getProperty("file.separator");
        boolean foundit = false;

        if (dvProperties == null) {
            try {

                FileInputStream x=new FileInputStream("."+sep+"Dv.properties");
                dvProperties = new Properties();
                dvProperties.load(x);
                retstr = dvProperties.getProperty(propname);
                foundit = true;
            }
            catch (IOException e) {

                // Doesn't look like we could find it on current
                // directory = so lets permeate through classpath

                String path=  "com"+sep+"ibm"+sep+"jvm"+sep+
                            "dump"+sep+"Dv.properties";
                dvProperties = searchClassPathForPropFile(path);
                if (null == dvProperties) {
                    path=  "Dv.properties";
                    dvProperties = searchClassPathForPropFile(path);
                }
                if (null != dvProperties) {
                    retstr = dvProperties.getProperty(propname);
                    foundit = true;
                }
            } // end of catch

            if (foundit == false) {
                errorMsg("Error trying to access Dv.properties",0);
            }
            else {
                retstr = dvProperties.getProperty(propname);
            }
        }
        else {
            retstr = dvProperties.getProperty(propname);
        }

        return retstr;
    }

    /**
     * Search for a properties file through the classpath                   establish the list of supported plugins. These are then
     *
     *    properties object or null if
     *
     */

    public static Properties searchClassPathForPropFile(String propfile) {

        String path;
        Properties p=null;
        String cp = System.getProperty("java.class.path");
        StringTokenizer st = new StringTokenizer(cp,";");
        boolean foundit= false;
        while (st.hasMoreTokens() && foundit==false) {
            String sep = System.getProperty("file.separator");
            path = st.nextToken()+sep + propfile  ;

            //System.out.println("\ntrying " + path);
            try {
                FileInputStream x=new FileInputStream(path);
                p = new Properties();
                p.load(x);
                foundit = true;

            }
            catch (IOException e1) { // just try next one)
            }  // end of catch
        } // end of while

        return p;
    }

    /**
     * The CommandPlugins stanza in Dv.properties is used to
     * establish the list of supported plugins. These are then
     * found, validity checked and returned.
     *
     * Returned: Vector containing command plugin class instances.
     *
     */
    public static Vector  findCommandPlugins() {
        DvUtils.trace("Looking for command plugins",0,false);
        DvDump theDump;
        Vector theCommandClasses = new Vector();

        // its here that we figure out a plugin that supports this
        // file format (or url or whatever) .........


        String dp = getDvProperty("CommandPlugins");
        if (null == dp) {
            errorMsg("No CommandPlugins stanza found in Dv.properties",0);
            return null; // we are probably useless
        }

        theCommandClasses = DvUtils.findCommandClasses(dp);
        DvUtils.trace(" DvUtils:findCommandPlugins - "+theCommandClasses.size()+
                      " command plugins found",0,false);


        return theCommandClasses;
    }

    /**
     *
     *
     *
     *
     *
     *
     */
    public static DvDump findDumpPlugin(String filename) {
        DvDump theDump;
        boolean bResult = false; // lets assume that they have chosen a duff one
        // its here that we figure out a plugin that supports this
        // file format (or url or whatever) .........
        DvUtils.trace("Looking for dump plugin that supports "+filename,0,false);

        String dp = getDvProperty("DumpPlugins");
        if (null == dp) {
            dp = "SDFFDump"; // the default one
        }

        theDump = DvUtils.findDumpClass(filename,dp);
        if (null != theDump) {
            bResult=true; // Whahey someone owned up to wanting us
        }
        else {
            DvUtils.trace("No dump plugin owned up to supporting "+filename,
                            0,false);
        }


        return theDump;
    }

    /**
     *
     *
     */
    public static void enableConsoleTrace(boolean b) {
        consoleTrace = b;
    }

    /**
     *
     *
     *
     *
     *
     *
     */
    public static void setNativeLookAndFeel() {
        try {

            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
        catch(Exception e) {
            writetoTrace("Error setting native LAF: " + e);
        }
    }

    /**
     *
     *
     *
     *
     *
     *
     */
    public static void setJavaLookAndFeel() {
        try {
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
        }
        catch(Exception e) {
            writetoTrace("Error setting Java LAF: " + e);
        }
    }

    /**
     *
     *
     *
     *
     *
     *
     */
    public static void setMotifLookAndFeel() {
        try {
            UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
        }
        catch(Exception e) {
            writetoTrace("Error setting Motif LAF: " + e);
        }
    }

    /**
     * The internal frame manager is established elsewhere
     * but trace uses it when under Dumpviewer, this communcates
     * its whereabouts.
     *
     */
    public static void setTheIFM (Dvifm ifm) {
        theIFM = ifm;

    }

    /**
     * With the Dumpviewer many people want access to the IFM
     * at various times.
     *
     */
    public static Dvifm getTheIFM () {
        return theIFM;
    }


    /**
     * Writes a line out to our trace
     *
     */
    public static void writetoTrace(String traceEntry) {

        if (theIFM != null) {

            if (theIFM.trace_iframe != -1) {
                //System.out.println("\n"+traceEntry);
                if (order12) {
                    traceString2 = traceString2+"\n"+traceEntry;
                    if (traceString2.length() > 20000) {
                        traceString1 = traceString2;
                        traceString2 = "";
                        traceString = traceString1;
                        order12 = false;
                    }
                    else {
                        traceString=traceString+"\n"+traceEntry;
                    }

                }
                else {
                    traceString1 = traceString1+"\n"+traceEntry;
                    if (traceString1.length() > 20000) {
                        traceString2 = traceString1;
                        traceString1 = "";
                        traceString = traceString2;
                        order12 = true;
                    }
                    else {
                        traceString=traceString+"\n"+traceEntry;
                    }

                }

                tc_text.setText(traceString);


            }
            else  {

            }
        }

        if (true == consoleTrace) {
            trace(traceEntry,0,false);
        }

    }

    /**
     * Converts a hex string to a java long
     * - handles up to 16 hex characters, returns -1
     *   for bad input.
     *
     */
    public static long hexToLong(String hex) throws NumberFormatException {

        long ret=0;
        int i;
        int j = hex.length();
        if (j <= 16) {
            for (i=0;i<j;i++) {
                int digit = hexValue(hex.charAt(i));
                if (digit == -1) {
                    throw new NumberFormatException("Bad hex input");// bad input
                }
                ret = 16*ret + digit;
            }
        }
        else {
            throw new NumberFormatException("Too many hex digits");
        }


        return ret;
    }

    /**
     *
     *
     */
    public static int hexValue(char ch) {
        switch(ch) {
            case '0' : return 0; case '1' : return 1; case '2' : return 2;
            case '3' : return 3; case '4' : return 4; case '5' : return 5;
            case '6' : return 6; case '7' : return 7; case '8' : return 8;
            case '9' : return 9;
            case 'a' : case 'A' : return 10;
            case 'b' : case 'B' : return 11;
            case 'c' : case 'C' : return 12;
            case 'd' : case 'D' : return 13;
            case 'e' : case 'E' : return 14;
            case 'f' : case 'F' : return 15;
            default  : return -1;
        }

    }

    /**
     *
     *
     */
    private static Vector findCommandClasses(String theList)
    {
        Vector commandClassInstances = new Vector();
        StringTokenizer st = new StringTokenizer(theList,",");
        int count = st.countTokens(); // Now we know how big the list is.

        int pluginsLoadedOK = 0;
        boolean done = false;
        String workString = st.nextToken();

        System.out.println("*** Loading plugins ***\n");

        while(!done) {
            try {
                DvUtils.writetoTrace(" Trying .." + workString);
                Class clz = ClassLoader.getSystemClassLoader().loadClass(workString);

                try {
                    commandClassInstances.add( (CommandPlugin) clz.newInstance());
                    System.out.println("loaded "+workString);
                    pluginsLoadedOK++;
                } catch (java.lang.InstantiationException e) { }

            } catch (java.lang.ClassNotFoundException e) {
                System.out.println(   "ERROR: "+workString+".class - class not found");
                DvUtils.writetoTrace(" ERROR: "+workString+".class - class not found");
            } catch (java.lang.IllegalAccessException e) {
                System.out.println(   "ERROR: "+workString+" - IllegalAccessException");
                DvUtils.writetoTrace(" ERROR: "+workString+" - IllegalAccessException");
            } catch (java.lang.ClassCastException cce) {
                System.out.println(   "ERROR: "+workString+" - is not a CommandPlugin");
                DvUtils.writetoTrace(" ERROR: "+workString+" - is not a CommandPlugin");
            }

            try {
                workString = st.nextToken();
            } catch (NoSuchElementException e) {
                done = true;
            }
        } // end of while loop

        // finished loading all plugins - but did they load okay?
        if (pluginsLoadedOK == count) {
            System.out.println("\n*** All plugins loaded successfully ***");
        } else {
            System.out.println("\n*** ERROR: "+(count-pluginsLoadedOK)+
                               " plugin(s) failed to load ***");
        }

        return  commandClassInstances;
    }

    /**
     *
     *
     *
     *
     *
     *
     */
    public static DvDump findDumpClass(String theSource,String theList) {


        DvUtils.trace("Dumpviewer: findDumpClass entry - "+ theSource + "-" +
                           theList,0,false);
        DvDump theOne = null;


        //The idea is that we try each of the values in the incoming List in order...
        // So first lets load stuff from the property file --- note that if there is no
        // property file then we just assume SDFFDump is IT!

        // first we have to seperate the list into an array of strings
        StringTokenizer st = new StringTokenizer(theList,",");
        int count = st.countTokens(); // Now we know how big the list is.


        boolean done = false;
        String workString = st.nextToken();
        for (int i=0;done == false;i++) {
            try {
                DvUtils.trace(" Trying .." + workString,0,false);
                //DvUtils.writetoTrace(" Trying .." + workString);
                Class clz = ClassLoader.getSystemClassLoader().loadClass(workString);

                Method method = clz.getMethod("isSupportedSource",
                                               new Class[]{String.class});


                // Note we do not need an instantiation as isSupportedSource is static
                Object o = method.invoke(null,new Object[] {theSource});
                Boolean b = (Boolean) o;

                if (b.booleanValue() == true) {
                    try {
                        theOne = (DvDump) clz.newInstance();
                        done = true;
                    }
                    catch (java.lang.InstantiationException e) {
                        System.err.println("\nInstantiationException");
                    }
                    catch (java.lang.IllegalAccessException e) {
                        System.err.println("\nIllegalAccessException");
                    }
                } // end of if...

            } // end of try
            catch (java.lang.ClassNotFoundException e) {
                DvUtils.writetoTrace(" ERROR!!! "+workString+
                    ".class was not found !!!!");

            }
            catch (java.lang.NoSuchMethodException e) {
                // Somebody has provided us with a class that doesn't obey the
                // rules......................................................
                DvUtils.writetoTrace(" ERROR!!! "+ workString+
                    ".class has no isSupportedSource method !!!!");

            }
            catch (java.lang.IllegalAccessException e) {
                System.err.println("\nIllegalAccessException");
            }
            catch (java.lang.reflect.InvocationTargetException e) {
                System.err.println("\nInvocationTargetException");
            }

            try {
                workString = st.nextToken();
            }
            catch (NoSuchElementException e) {

                //System.out.println("\nNoSuchElementException");
                done = true;
            }

        } // end of for loop

        DvUtils.trace("Dumpviewer: findDumpClass exit - "+
                      ((null==theOne)?"Unrecognized format":"Recognized format"),0,false);
        //DvUtils.writetoTrace("Dumpviewer: findDumpClass exit - "+
        //                    ((null==theOne)?"Unrecognized format":"Recognized format"));

        return theOne;
    }

    /**
     *
     *
     */
    public static void removeSpecFromVector(Vector v,String s) {
        int found_it=-1;

        for (int i=0; i<v.size() && found_it==-1 ; i ++) {
            if (true == (v.get(i)).equals(s)) {
                found_it = i;
            }
        }
        if (-1 != found_it) {
            v.remove( found_it);
        }
    }

    /**
     *
     *
     */
    public static void  initSetCommand() {

        // if setvalues is not null then we are already inited
        if (null == setValues) {
            try {
                FileInputStream x=new FileInputStream("DvSetDefaults.properties");
                Properties setValues = new Properties();
                // rest of world uses DvUtil to look for assorted properties
                DvUtils.setProperties(setValues);
                setValues.load(x);

            }
            catch (IOException e) {
                // didn't find it on current directory
                String sep = System.getProperty("file.separator");
                String path=  "com"+sep+"ibm"+sep+"jvm"+sep+
                            "dump"+sep+"DvSetDefaults.properties";
                setValues =  DvUtils.searchClassPathForPropFile(path);
                if (null == setValues) {
                    path=  "DvSetDefaults.properties";
                    setValues = DvUtils.searchClassPathForPropFile(path);
                }
            }
            if (null == setValues) {
               System.err.println("Unable to find DvSetDefaults");
            }
        }
    }

    /**
     * This
     *
     */
    public static Enumeration getPropertyNames()
    {
        return setValues.propertyNames();
    }


    /**
     * This
     *
     */
    public static void setProperties(Properties p)
    {
        setValues = p;
    }

    /**
     * Sets a key/value pair in the global values
     * property set. Note: all keys are converted to
     * uppercase to avoid confusion.....
     *
     */
    public static void setValue(String key,String value)
    {
        if (null == value || value.equals("")) {
            setValues.remove(key.toUpperCase());
        }
        else {
            setValues.setProperty(key.toUpperCase(),value);
        }
    }

    /**
     * returns value of a previously set key.
     * Note: key is considered uppercased!
     *
     */
    public static String getValue(String key)
    {
        return (setValues.getProperty(key.toUpperCase()));
    }

    /**
     *
     *
     */
    public static int getPropertyInt(String prop) {
        int ret=-1;
        if (null != setValues) {
            String s = setValues.getProperty(prop.toUpperCase());
            try {
                ret = ((new Integer(s)).intValue());
            }
            catch (Exception e) {}
        }
        return ret;
    }

    /**
     *
     *
     */
    public static String getProperty(String prop) {
        String retString=null;
        if (null != setValues) {
            retString = setValues.getProperty(prop);
        }
        return retString;

    }


    /**
     *
     *
     */
    public static boolean isConsoleOutput() {
        return consoleOutput;
    }

    /**
     *
     *
     */
    public static void setConsoleOutput(boolean b) {
        consoleOutput = b;
    }

    /**
     *
     *
     */
    final protected static int convertEndian(int in)
    {
      return ( (in >>> 24) | (in << 24 ) |
               (( in << 8 ) & 0x00FF0000) | ((in  >> 8 ) & 0x0000FF00));
    }

    /**
     * Given the architecture (combination of endianess
     * and 32/64 bitness returns true if 64 bit system.
     *
     */
    public static boolean is64BitSystem(int arch) {
        boolean bRet=false; // assume 32 bit at first
        // cheats a little by looking simply at string describing
        // architecture for 64!!!!!!
        try {
            if (-1 != DvDump.ARCHITECTURE_DESCRIPTION[arch].indexOf("64")) {
                bRet=true;
            }
        }
        catch (java.lang.ArrayIndexOutOfBoundsException e) {
        }

        return bRet;
    }

    /**
     * Given the architecture (combination of endianess
     * and 32/64 bitness returns true if its a big
     * endian system.
     *
     */
    public static boolean isBigEndian(int arch) {
        boolean bRet=false; // assume 32 bit at first
        // cheats a little by looking simply at string describing
        // architecture for Big!!!!!!
        try {
            if (-1 != DvDump.ARCHITECTURE_DESCRIPTION[arch].indexOf("Big")) {
                bRet=true;
            }
        }
        catch (java.lang.ArrayIndexOutOfBoundsException e) {
        }


        return bRet;
    }

    /**
     * Given a 8 or 16 byte array convert its endianess
     * from big to little. If its not 8 or 16 bytes in
     * length then do nothing
     *
     *
     */
    public static byte[] convertToBigEndian(byte[] b) {

        if (b.length != 4 && b.length != 8) {
            return b;
        }
        // so if we get here we have 8 or 16 bytes
        if (b.length == 8) {
            // its 64 bits (8 bytes)
            byte[] returnBytes={0,0,0,0,0,0,0,0};

            returnBytes[0] = b[7];
            returnBytes[1] = b[6];
            returnBytes[2] = b[5];
            returnBytes[3] = b[4];
            returnBytes[4] = b[3];
            returnBytes[5] = b[2];
            returnBytes[6] = b[1];
            returnBytes[7] = b[0];

            return returnBytes;


        }
        else {
            // its 32 bits
            byte[] returnBytes={0,0,0,0};
            returnBytes[0] = b[3];
            returnBytes[1] = b[2];
            returnBytes[2] = b[1];
            returnBytes[3] = b[0];

            return returnBytes;
        }

    }


    /**
     *  Given an array of bytes figures out what
     *  it looks like as an hex string
     *
     */
    public static String bytesToHexString(byte[] b) {
        String retString="";
        int ival;
        for (int i=0, j=1; i<b.length; i++) {
            ival = (int) b[i];
            if (ival < 0) {
                ival=ival+256;
            }
            retString = retString + (DvUtils.byteToHex).substring(2*ival,2*ival+2)  ;
        }
        return retString;
    }

    /**
     *
     *
     */
    public static void output(String o) {
        System.out.println(o);
    }

    /**
     * Find sequence of bytes in a random access file.
     *
     *
     *
     */
    public static long findNextInFile(RandomAccessFile raf,
                                        byte[] whatFor,long startPoint,
                                        long endPoint, int boundary ) {

        // not to much checking of parameters


        DvUtils.trace("DvUtil:findNextInFile entry(sp=x"+
                        Long.toHexString(startPoint)+
                        " ep=x"+Long.toHexString(endPoint),
                        0,false);
        final int bufsize=4096;
        long retval = -1; // assume failure
        long len = whatFor.length;
        long trueStart = startPoint;
        byte[] buffer = new byte[bufsize];
        boolean isFound,bufferDone;
        int bufferOffset;
        Byte b = null;

        //System.out.println("\nstart =" +  startPoint + " :"+endPoint);

        if (0 != boundary) {
            trueStart = startPoint + startPoint%boundary;
        }
        if (trueStart<endPoint) {
            //System.out.println("\nyyyyyy");

            try {


                //System.out.println("\nxxxxxx");
                // Now start comparing bytes
                isFound = false;
                bufferDone = true;
                bufferOffset = 0;
                while (false == isFound && startPoint < endPoint) {

                    //System.out.println("Loop...." + trueStart + bufferDone);
                    if (true == bufferDone) {
                        DvUtils.trace(" Reading from file at offfset x"+
                                        Long.toHexString(trueStart)
                                        ,0,false);
                        raf.seek(trueStart);
                        raf.read(buffer);
                        bufferDone = false;
                        bufferOffset = 0;
                    }
                    isFound = true;
                    for (int i=0;i<whatFor.length && true == isFound;i++) {
                        //b = new Byte(buffer[i+bufferOffset]);
                        if (whatFor[i] != buffer[i+bufferOffset]) {
                            isFound = false;
                        }
                    } // end of for loop

                    if (isFound == false) {
                        if (0 != boundary) {
                            bufferOffset = bufferOffset + boundary;
                        }
                        else {
                            bufferOffset++;
                        }


                        if (bufferOffset > (bufsize -  len)) {
                            bufferDone = true;
                            trueStart = trueStart + bufsize - (len+len%boundary);
                            if (trueStart >= endPoint) {
                                // gets us out when nothing there
                                startPoint = endPoint;
                            }
                            // should now read next piece of buffer
                        }
                    }

                } // end of while loop

                if (true == isFound) {

                    retval = trueStart + bufferOffset;
                    //System.out.println("Found it :" + retval);
                }

            } // END OF TRY
            catch (IOException ioe) {
                DvUtils.trace("DvUtil:findNextInFile caught IOException",
                                0,false);
                //System.err.println(" Caught IOException");
            }
            catch (java.lang.ArrayIndexOutOfBoundsException aobEx) {
                DvUtils.trace(
                    "DvUtil:findNextInFile caught ArrayIndexOutOfBoundsException",
                     0,false);
                //System.err.println("Caught ArrayIndexOutOfBoundsException");
            }
        }


        return retval;
    }

    /**
     * Find sequence of bytes in memory crossing memranges if
     * necessary.
     *
     *
     */
    public static Vector findInMemory(DvDump theDump,byte[] whatFor,
                                     DvAddress startPoint,
                                     DvAddress endPoint, int boundary,
                                     int stopAt, boolean caseSensitive  ) {


        long bytesread=0;
        Vector vResult = new Vector(); // return a vector of DvAddress objects
        DvAddressSpace as = DvUtils.findAsid(theDump,DvUtils.getValue("ASID"));


        DvAddress retAddress = null;
        long startPointAsLong = startPoint.getAddressAsLong();
        long endPointAsLong = endPoint.getAddressAsLong();

        // Lets look at the boundary setting and check that  start and end
        // on the boundary

        if (0 != startPointAsLong%boundary) {
            startPointAsLong = startPointAsLong +
                         (boundary - startPointAsLong%boundary);
            try {
                startPoint.offsetBy((long)(boundary - startPointAsLong%boundary));
            }
            catch (DvAddressException e) {
                try {
                    startPoint.offsetBy((long)(-startPointAsLong%boundary));
                }
                catch (DvAddressException e1) {
                    // should never happen
                }

            }
        }
        if (0 != endPointAsLong%boundary) {
            endPointAsLong = endPointAsLong +
                      (boundary - endPointAsLong%boundary);
            try {
                endPoint.offsetBy((long)(boundary - endPointAsLong%boundary));
            }
            catch (DvAddressException e) {
                try {
                    endPoint.offsetBy((long)(-endPointAsLong%boundary));
                }
                catch (DvAddressException e1) {
                    // should never happen
                }

            }


        }


        if (null != theDump &&
                1 == startPoint.compare(endPoint)) {


            DvMemRanges[] mr = as.getMemRanges();
            // Note that we test on stopAt to limit number of finds when user
            // puts in something common and a large range......
            for (int i=0;i<mr.length && vResult.size()<stopAt ;i++) {
                long rangeStartAsLong =  mr[i].addressStart().getAddressAsLong();
                long rangeEndAsLong = rangeStartAsLong + mr[i].length();

                DvAddress rangeEnd = as.createAddress(rangeEndAsLong);
                DvAddress rangeStart = as.createAddress(rangeStartAsLong);


                if (0 != rangeStartAsLong%boundary) {
                    try {
                         rangeStart.offsetBy((long)(boundary - rangeStartAsLong%boundary));
                    }
                    catch (DvAddressException e) {
                        try {
                             rangeStart.offsetBy((long)(-rangeStartAsLong%boundary));
                        }
                        catch (DvAddressException e1) {
                            // should never happen
                        }

                    }

                }
                if (0 != rangeEndAsLong%boundary) {
                    try {
                         rangeEnd.offsetBy((long)(boundary - rangeEndAsLong%boundary));
                    }
                    catch (DvAddressException e) {
                        try {
                             rangeEnd.offsetBy((long)(-rangeEndAsLong%boundary));
                        }
                        catch (DvAddressException e1) {
                            // should never happen
                        }

                    }


                }

                // first check if this is within the range
                // the range end must be greater than our startpoint
                // and the range start must be less than our endpoint
                if (1 != endPoint.compare(startPoint) &&
                    -1 != startPoint.compare(rangeEnd) &&
                    -1 != rangeStart.compare(endPoint) &&
                    mr[i].length() !=0 ) {

                    DvAddress ongoing = as.createAddress(rangeStartAsLong);
                    if (1 == ongoing.compare(startPoint)) {
                    //if (rangeStartAsLong < startPoint.getAddressAsLong()) {
                         ongoing = as.createAddress(startPoint.getAddressAsLong());
                    }

                    boolean memRangeExhausted = false;
                    while (false == memRangeExhausted) {
                        // we read into a buffer
                        long whereWeAre = ongoing.getAddressAsLong();
                        int searchSize = 16000;
                        byte[] memoryBuffer = null;
                        try {
                            memoryBuffer = as.readBytes(ongoing,16384);
                        }
                        catch (DvAddressException e) {
                            // no need to do anything
                        }
                        if (null == memoryBuffer) {
                         // should never happen!!!!!
                         System.err.println("Error!!! - Ongoing:"+ongoing.getAddressAsLong());
                        }
                        bytesread = bytesread + memoryBuffer.length;
                        int mbLen = memoryBuffer.length;
                        if (16384==memoryBuffer.length) {
                            try {
                                ongoing.offsetBy(16000);

                                // if ongoing bigger than end of this mem range then
                                // we have finished and move onto next memrange
                                // also if ongoing is bigger than our specified
                                // end point then we set memRangExhausted as well
                                if (1 == rangeEnd.compare(ongoing) ||
                                    1 == endPoint.compare(ongoing)) {
                                //if ( rangeEndAsLong < ongoing.getAddressAsLong()) {
                                  memRangeExhausted = true;
                                }
                            }
                            catch (DvAddressException e) {
                                  memRangeExhausted = true;
                            }

                        }
                        else {
                            // we've read less than 16384 bytes
                            // which means we are at the end of this range

                            memRangeExhausted = true;
                            searchSize = mbLen;
                        }

                        boolean isFound = false;
                        if (mbLen > 0) {
                            //System.out.println("Searching some memory");
                            // finally search sonme memory, looking for the bytes
                             //&& (long) (whereWeAre+(long)j)<endPointAsLong
                            for (int j=0;
                                 j<searchSize;
                                 j = j+boundary) {
                                //System.out.println("Searching some memory");
                                isFound = true;
                                for (int k=0;k<whatFor.length && true == isFound;k++) {

                                    if (whatFor[k] != memoryBuffer[k+j]) {
                                       isFound = false;
                                    }
                                } // end of for loop
                                if (true == isFound) {
                                    DvAddress wwaAddress = as.createAddress(whereWeAre+j);
                                    if (-1 != wwaAddress.compare(endPoint)) {
                                        vResult.add(as.createAddress(whereWeAre+j));
                                        if (stopAt == vResult.size()) {
                                            // we use stopAt to control number
                                            // of results we will pass back
                                            j = searchSize;  // stop the loop
                                            memRangeExhausted = true;
                                        }
                                    }
                                    else {
                                        memRangeExhausted = true;
                                    }
                                }


                            }

                        }

                    }  // end of while
                } // end of if
            } // end of for
        }

        System.out.println("\nbytesread:"+bytesread);
        return vResult;
    }

    /**
     * Given a DvProcess and DvAddressSpace andd threadId return the
     * DvThread that hopefully exists (null if it doesn't)
     *
     *
     */
    public static DvThread findThread(DvDump theDump,String threadId,DvAddressSpace a,
                                    DvProcess p) {

        DvThread retThread =  currentThread; // assume current thread is good
        if (null != currentThread && currentThread.id().equals(threadId)) {
            // Check that owning as and owning process have same id....
        }
        else {
            retThread = null; // assume we ain't going to find one
            // go through all procs for address spaces til we find one with matching id!
            DvThread[] tArray =  p.getThreads();

            if (null != tArray) {
                boolean finito = false;
                for (int i=0;i<tArray.length && false == finito;i++) {
                    if (tArray[i].id().equals(threadId)) {
                        // found one

                        retThread = tArray[i];
                        finito  = true;
                    }
                }
            }

        }


        return retThread;
    }

    /**
     * Given a DvAddressSpace and a process id (PID) return the
     * Dvprocess that hopefully exists (null if it doesn't)
     *
     *
     */
    public static DvProcess findProc(DvDump theDump,String pid,DvAddressSpace a) {



        DvProcess retproc =  currentProcess; // assume current proc is good
        if (null != currentProcess && currentProcess.id().equals(pid)) {
            // Check that owning as has same id....
        }
        else {
            retproc = null; // assume we ain't going to find one
            // go through all procs for address spaces til we find one with matching id!
            DvProcess[] pArray =  a.getProcesses();

            if (null != pArray) {
                boolean finito = false;
                for (int i=0;i<pArray.length && false == finito;i++) {
                    if (pArray[i].id().equals(pid)) {
                        // found one

                        retproc = pArray[i];
                        finito  = true;
                    }
                }
            }

        }

        return retproc;
    }

    /**
     * Given a ASID return the DvAddressSpace
     *     that hopefully exists (null if it doesn't)
     *
     *
     */
    public static DvAddressSpace findAsid(DvDump theDump,String asid) {
        DvAddressSpace retAs =  currentAsid; // assume current ASID is good
        if (null != retAs && retAs.id().equals(asid)) {
            // we have a good one!!!
        }
        else {
            retAs = null; // assume we ain't going to find one
            // go through all address spaces til we find one with matching id!
            DvAddressSpace[] asArray =  theDump.getAddressSpaces();
            if (null != asArray) {
                boolean finito = false;
                for (int i=0;i<asArray.length && false == finito;i++) {
                    if (asArray[i].id().equals(asid)) {
                    // found one

                        retAs = asArray[i];
                        finito  = true;
                    }
                }
            }

        }

        return retAs;
    }

    /**
     * Given a filename xxx.yyy - search for it
     *     on current working dir
     *
     *
     */
    public static File searchForFile(String filename) {

        File retFile = null;
        String sep = System.getProperty("file.separator");

        retFile = new File("."+sep+filename)  ;
        if (null != retFile && true == retFile.canRead()) {
        }
        else {
            //lets try adding in package name
            String path=  "com"+sep+"ibm"+sep+"jvm"+sep+
                        "dump"+sep+filename;
            retFile = new File(path)  ;
            if (null != retFile && true == retFile.canRead()) {
            }
            else {
                // still stuffed - lets see if we can find it on classpath
                String cp = System.getProperty("java.class.path");
                StringTokenizer st = new StringTokenizer(cp,";");
                boolean foundit= false;
                while (st.hasMoreTokens() && foundit==false) {

                    path = st.nextToken()+sep + filename  ;
                    retFile=new File(path);
                    if (null != retFile && true == retFile.canRead()) {
                        foundit = true;
                    }
                    else {
                        retFile=null;
                    }

                } // end of while
            }
        }

        return retFile;
    }

    /**
     * Given a DvAddress and DvAddressSpace check whether its available
     * as memory in the dump
     *
     *
     */
    public static boolean checkAddressInDump(DvAddress a, DvAddressSpace as ) {
        boolean bRet = false; // assume invalid

        try {
            as.readInt(a);
            bRet = true;
        }
        catch (DvAddressException e) {
        // guess its not available!!!!
        }

        return bRet;
    } // end of method

    /**
     * Given a String and a seperator split it into
     * a vector of substrings
     *
     *
     */
    public static Vector splitIntoPieces(String s,String sep)  {


        Vector individualParms = new Vector();
        StringTokenizer st = new StringTokenizer(s,sep);

        if (st.countTokens() != 0) {
            while (st.hasMoreTokens()) {
                individualParms.add(st.nextToken());
            }
        }


        return individualParms;
    } // end of method


    /**
     * Have we already loaded a type file
     * so we can successfully format control blocks
     *
     *
     */
    public static boolean isTypeFileLoaded()  {
        return typeFileLoaded;
    }

    /**
     * mark type file as loaded
     *
     *
     *
     */
    public static void setTypeFileLoaded(boolean b)  {
        typeFileLoaded = b;
    }

    /**
     * remove 0x from front of address
     *
     *
     *
     */
    public static String remove0x(String in)  {
        String out=in.trim().toUpperCase();

        if (out.startsWith("0X")) {
            out = out.substring(2);
        }

        return out;
    }

    /**
      * remove @ from front of address
     *
     *
     *
     */
    public static String removeAtSign(String in)  {
        String out=in.trim().toUpperCase();

        if (out.startsWith("@")) {
            out = out.substring(1);
        }

        return out;
    }

    /**
     * find the current set of ASID,Process,Thread objects
     * using the set keywords
     *
     *
     *
     */
    public static Object getAPT(DvDump theDump,String type)  {
        String asidVal = getValue("ASID");
        String threadVal = getValue("THREAD");;
        String pidVal = getValue("PID");
        DvAddressSpace as = null;
        DvProcess proc = null;
        DvThread thread = null;
        as = findAsid(theDump,asidVal);
        if (null != as) {
            proc = findProc(theDump,pidVal,as);
            if (null != proc) {
                thread = findThread(theDump,threadVal,as,proc);
            }
        }
        if (type.toUpperCase().equals("A")) {
            return as;
        }
        if (type.toUpperCase().equals("P")) {
            return proc;
        }
        if (type.toUpperCase().equals("T")) {
            return thread;
        }
        return null;
    }

    /**
     * Lets get all the hints for the as
     *
     *
     *
     */
    public static Vector getAllJVMHints(DvAddressSpace as) {
        Vector rV = new Vector();



        JVMHints workingJVMHints = null;
        // So look through jvmHints vector for a JVMHints object matching this
        // asid.....
        if (0 != jvmHints.size()) {

            // lets look for a jvmHints that matches this AddressSpace

            for (int i = 0; i< jvmHints.size();i++) {
                workingJVMHints = (JVMHints) jvmHints.get(i);
                if ( workingJVMHints.getId().equals(as.id())) {
                    // we are in business
                    rV = workingJVMHints.getAllValues();
                    return rV;
                }
            }
        }



        return rV;

    }

    /**
     * Lets see if we have a hint for an address
     * - most of the important ones should be there
     *
     *
     *
     */
    public static String getJVMHint(DvAddressSpace as,String key)
    {
        String value = null;

        JVMHints workingJVMHints = null;
        // So look through jvmHints vector for a JVMHints object matching this
        // asid.....
        if (0 != jvmHints.size()) {

            // lets look for a jvmHints that matches this AddressSpace

            for (int i = 0; i< jvmHints.size();i++) {
                workingJVMHints = (JVMHints) jvmHints.get(i);
                if ( workingJVMHints.getId().equals(as.id())) {
                    // we are in business
                    value = workingJVMHints.getValue(key);
                    return value;
                }
            }
        }

        return value;
    }

    /**
     * When we establish where the JVM is then we
     * cheat and set up a lot of useful pointers out
     * of the JVM, this means that other plugins etc.
     * do not need to replicate a whole lot of code
     * to access fields they are interested in.
     * We save these hints using SET. Note we use an
     * imbedded class (JVMHints) to contain all this
     *
     *
     *
     *
     */
    public static void setupJVMHints(CTypeObject jvm_cTO,DvAddressSpace as) {

        DvUtils.trace(" Entry to DvUtils:setupJVMHints",0,false);

        if (null == jvmHints) {
            jvmHints = new Vector();
        }
        else {
            // covers us when they open a second dump
            jvmHints.clear();
        }
        // So we scan down through the JVM looking for impotant addresses and
        // then store these away in the JVMHints instance for this ASID
        JVMHints workingJVMHints = null;
        // So look through jvmHints vector for a JVMHints object matching this
        // asid.....
        if (0 != jvmHints.size()) {
            // lets look for a jvmHints that matches this one
            // in which case we will refresh the current one
            // this allows thenm to have had the wrong types file
            // when putting up the original hints - correct it and
            // get a better set of hints.....
            for (int i = 0; i< jvmHints.size();i++) {
                workingJVMHints = (JVMHints) jvmHints.get(i);
                if ( workingJVMHints.getId().equals(as.id())) {
                    // we need to refresh the values, so
                    workingJVMHints.setAllValues(jvm_cTO,hintListJvm);
                    // Now get the STGlobal and do hints for that as well
                    setupSTGHints(workingJVMHints,as);
                    DvUtils.trace(" Exit from DvUtils:setupJVMHints",0,false);
                    return;
                }
            }
            // if we have fallen through the if then we need to create a
            // new one......

        }

        DvUtils.trace("\tCreating new JVMHints",0,false);
        workingJVMHints = new JVMHints(as.id(),jvm_cTO);
        workingJVMHints.setAllValues(jvm_cTO,hintListJvm);
        // Now get the STGlobal and do hints for that as well
        jvmHints.add(workingJVMHints);
        setupSTGHints(workingJVMHints,as);

        DvUtils.trace(" Exit from DvUtils:setupJVMHints",0,false);
        return;
    }

    public static void setupSTGHints(JVMHints workingJVMHints, DvAddressSpace as) {
        // So we get the STGLOBALPTR hint and use this as the base address
        // for creating the stg_cTO.... and then its easy peasy
        String hex = workingJVMHints.getValue("STGLOBALPTR");

        if (null != hex && null != as) {
            //System.out.println("hex is:"+hex+"!!!");
            DvAddress a = as.createAddress(DvUtils.hexToLong(hex));

            if (checkAddressInDump(a,as)) {
                CTypeObject stg_cTO= formatAddrAsType(as,a,"STGlobal");
                workingJVMHints.setAllValues(stg_cTO,hintListSTGlobal);
            }
        }
    }  // end of method

    /**
     * Basically creates a CtypeObject, but with a little more error
     * resolution built in
     *
     */
    public static CTypeObject formatAddrAsType(DvAddressSpace as,DvAddress a,String type) {
        CTypeObject retCType = null;
        CType theType =  CType.find(type);
        DvUtils.trace("name: "+theType.name+"\ndescr:"+theType.description +
        "\naddress:"+ Long.toHexString(a.getAddressAsLong()),0,false);
        if (null != theType) {
            retCType = new CTypeObject(type,a,as);
        }

        return retCType;
    }

    private static class JVMHints {

        // this hintlist is an arry of searchable strings followed
        // by the key that will be set corresponding to the
        // data following the string (if found). The data will
        // be an address in hex format. Note the searchable strings
        // is two layered so that we can find same names in different
        // subsections of the CTypeObject, note '!' means missing


        protected String as_id; // what address space do I belong with
        protected Properties p; // where we store the interesting stuff

        // Constructor
        protected JVMHints(String id,CTypeObject jvm_cTO) {

            as_id = id;

            p = new Properties();


        }

        protected String getId() { return as_id;}

        protected String getValue(String key) {
            return p.getProperty(key.toUpperCase());
        }

        protected void setValue(String key,String value) {
            p.setProperty(key.toUpperCase(),value);
        }

        protected void setAllValues(CTypeObject jvm_cTO,String[] hintList) {
            // So we convert the jvm CTypebject to a string
            // and scan it for those things that we want to
            // store away.....
            String  stringJvm = jvm_cTO.toString();

            // now we go through the hintList and set up the
            // properties as needed
            for (int i=0;i<hintList.length;i++) {
                StringTokenizer st = new StringTokenizer(hintList[i],",");
                //System.out.println("#Tokens ="+st.countTokens());
                //System.out.println("i is:"+i+ "  "+hintList[i]);
                // There are hopefully 3 tokens!!!!
                String search1 = st.nextToken();
                String search2 = st.nextToken();
                if (search2.equals("!")) {
                    search2=null;
                }
                String key = st.nextToken();
                // the first being what to look for first,
                // the second is what to look for second
                // and the third is the property to set

                // so see if first string is a substring of stringJvm
                int p2 = 0;
                int p1 = stringJvm.indexOf(search1);
                if (-1 != p1) {
                    if (null != search2) {
                        p2 =  stringJvm.indexOf(search2,p1);
                        p2 =  p2+search2.length();
                    }
                    else {
                        p2 = p1+search1.length() ;
                    }
                }

                if (-1 != p1 && -1 != p2) {
                // we are in business
                    int p3 = stringJvm.indexOf(" ",p2);
                    if (-1 != p3) {
                    // whahey we are rolling
                        String val = stringJvm.substring(p2,p3);
                        p.setProperty(key.toUpperCase(),val.trim());

                    }
                }


            }

        } // end of method

    protected Vector getAllValues( ) {

        Vector rV = new Vector();
        String key = null;

        for (Enumeration names = p.propertyNames(); names.hasMoreElements() ;) {
            key = (String)names.nextElement();
            rV.add(key + " = "+ getValue(key));

        }

        return rV;

    }

 } // end of class JVMHints

    /**
     * We match up entries from CTypeFile stanzas in Dv.properties
     * to the version. Expandable to incorporate return of * when
     * the type information is available in the dump itself!!!!
     *
     */
    public static String getCtypeFile(DvDump theDump,String versionInfo) {

        String retString = null;

        // Placeholder for looking in the dump for the Ctype info!!!!!
        Vector v = getMultiplePropertyValues("CTypeFile","Dv.properties");

        if (v.size()>0) {
            for (int i=0;i<v.size();i++) {
                String s=(String) v.elementAt(i);
                StringTokenizer st = new StringTokenizer(s,",");
                String s1 = st.nextToken().toUpperCase();
                if (-1 != versionInfo.toUpperCase().indexOf(s1)) {
                    retString = st.nextToken();
                    retString = retString.trim();
                    //System.err.println("retString:" + retString);
                    i = v.size();
                }

            }
        }




        return retString;
    }

    /**
     * Goes through a property file and returns a vector of
     * string values matching keys xyz1 upwards where xyz is the base
     *
     */
    public static Vector getMultiplePropertyValues(String key,String propertyFile) {


        Vector vRet = new Vector();
        String key1 = null;
        //String value ="";
        String value1=null;
        Properties myProps = null;
        Enumeration myPropsEnum;

        try {
                FileInputStream myPropsFIS =
                    new FileInputStream(propertyFile);
                myProps  = new Properties();
                myProps.load(myPropsFIS);
        }
        catch (IOException e) {
                // didn't find it on current directory
                String sep = System.getProperty("file.separator");
                String path=  "com"+sep+"ibm"+sep+"jvm"+sep+
                        "dump"+sep+propertyFile;
                myProps =  DvUtils.searchClassPathForPropFile(path);
                if (null == myProps) {
                    path=  propertyFile;
                    myProps = DvUtils.searchClassPathForPropFile(path);
                }
        }


        if (null != myProps) {

            myPropsEnum = myProps.propertyNames();

            for (int i=1; i<999;i++) {
                key1 = key+i;
                value1 = myProps.getProperty(key1);

                if (null != value1) {
                    vRet.add(value1);
                }
                else {
                    i = 999;
                }

            }
        }

        return vRet;

    }


    /**
     *
     *
     *  @param
     *  @return
     */
    public static long getFieldAddress(long startAddr, String[] nameArray, DvAddressSpace as) throws DvAddressException
    {
        CTypeObject cTO;
        CTypeObject addr;

        if ((nameArray.length)==2) {
                cTO = new CTypeObject(nameArray[0], startAddr, as);
                addr = cTO.getField(nameArray[1]);
                startAddr = addr.getAddr();
            return startAddr;
        }
        else throw new DvAddressException("Odd number of arguments in String[]");
    }


    /**
     *
     *
     *  @param
     *  @return
     */
    public static long getFieldValue(long startAddr, String[] nameArray, DvAddressSpace as, DvDump theDump) throws DvAddressException
    {
        long p =0;
        CTypeObject cTO;
        CTypeObject addr;
        if (!DvUtils.checkAddressInDump(as.createAddress(startAddr),as))
            throw new DvAddressException("Address not in dump");

        if ((nameArray.length)==2) {
                cTO = new CTypeObject(nameArray[0], startAddr, as);
                addr = cTO.getField(nameArray[1]);
                p = addr.getAddr();
                if (!DvUtils.checkAddressInDump(as.createAddress(p),as))
                    throw new DvAddressException("Address not in dump");
                if (true == DvUtils.is64BitSystem(theDump.getArchitecture())) {
                    p = as.readLong(as.createAddress(p));
                }
                else {
                    p = (long) as.readInt(as.createAddress(p));
                }
        return p;

        }
        else throw new DvAddressException("More or Less than two elements in String[]");
    }


    /**
     * filter a string against a vector of filters
     *
     *
     */
     public static boolean filtrate(String name, Vector theFilters) {

        boolean bRet = false;

        String myName = name.toUpperCase();
        boolean bDone = false;
        // it only takes 1 excluding filter to stop it
        // however name can pass multiple accepting filters before
        // getting tossed
        for (int i=0;i<theFilters.size() && false == bDone;i++) {

            bRet = false;
            // so analyze the filter - first we look for it starting !
            // then we look for * at beginning and end
            // we are not getting too fancy and allowing * in the middle.
            String filter = (String) theFilters.get(i);
            boolean bNot = false;
            boolean bStartAsterisk = false;
            boolean bEndAsterisk = false;
            if (filter.equals("*") || filter.equals("**"))
            {
              // its passing (for now)
              bRet = true;
            }
            else {
                if (filter.equals("!")) {
                  // its failed!!!!
                  bDone = true;
                  bRet = false;
                }
                else {

                    // So lets look for ! at front
                    if (filter.startsWith("!")) {
                        bNot = true;
                        filter = filter.substring(1);
                    }
                    // Now lets look for * positions
                    if (filter.startsWith("*")) {
                        bStartAsterisk = true;
                        filter = filter.substring(1);
                    }
                    if (filter.endsWith("*")) {
                        bEndAsterisk = true;
                        filter = filter.substring(0,filter.length()-1);
                    }
                    //System.out.println("filter:" + filter);
                    filter = filter.toUpperCase();
                    if (bStartAsterisk && bEndAsterisk) {
                        if (-1 != myName.indexOf(filter)) {
                            if (bNot) {
                                bDone = true;
                                bRet = false;
                            }
                            else {
                                bRet = true;
                            }
                        }
                        else {
                            if (bNot) {
                                bRet = true;
                            }
                        }
                    }

                    if (bStartAsterisk && !bEndAsterisk) {
                        if (myName.endsWith(filter)) {
                            if (bNot) {
                                bDone = true;
                                bRet = false;
                            }
                            else {
                                bRet = true;
                            }
                        }
                        else {
                            if (bNot) {
                                bRet = true;
                            }
                        }
                    }

                    if (bEndAsterisk && !bStartAsterisk) {
                        if (myName.startsWith(filter)) {
                            if (bNot) {
                                bDone = true;
                                bRet = false;
                            }
                            else {
                                bRet = true;
                            }
                        }
                        else {
                            if (bNot) {
                                bRet = true;
                            }
                        }
                    }

                    if (!bEndAsterisk && !bStartAsterisk) {
                        if (myName.equals(filter)) {
                            if (bNot) {
                                bDone = true;
                                bRet = false;
                            }
                            else {
                                bRet = true;
                            }
                        }
                        else {
                            if (bNot) {
                                bRet = true;
                            }
                        }
                    }
                }
            }
        }


        return bRet;
    }

    /**
     * Extract null terminated string from byte array
     *
     *
     */
    public static String getStringFromBytes(byte[] b) {

     int i = 0;
     StringBuffer str=new StringBuffer();

     while (true && i<b.length) {
            if (b[i] == '\0' || b[0] == '\0')
                break;
            str.append((char)(b[i]));
            i++;
     }

     return str.toString();

    }

    /**
     * Extract pointer (long) from long address
     *
     *
     */
    public static long getPtrFromPtr(DvDump theDump,
                                     DvAddressSpace as,
                                     long p) throws DvAddressException  {

        long retLong;

        if (true == DvUtils.is64BitSystem(theDump.getArchitecture())) {
            retLong = as.readLong(as.createAddress(p));
                                       // Note readLong will have coped
                                       // with the endianess!!!!!

        }
        else {
            retLong = (long) as.readInt(as.createAddress(p));
                                 // Note readInt will have coped with the
                                 // endianess!!!!!
        }

        return retLong;


    }

    /**
     *
     *
     *
     */
   public static short bytesToShort(byte[] b,int offset,int arch) {

        short s=0;
        if (true == DvUtils.isBigEndian(arch)) {
            s = (short) (256*b[offset]+b[offset+1]);
        }
        else {
            s = (short) (256*b[offset+1]+b[offset]);
        }

        return s;

   }

    /**
     *
     *
     *
     */
    public static void trace(String traceText, int tracePoint, boolean forceOut) {

        // For the moment simply write the output to err
        if (true == forceOut || true == verbose) {
            System.err.println("("+tracePoint+")"+traceText);
        }
    }

    /**
     *
     *
     *
     */
    public static void setVerbose(boolean tf) {

        verbose = tf;
    }

    /**
     * Load the format file
     *
     *
     */
    public static void loadFormatFile(Vector output) {

        // so we go and look for FORMATFILE setting
        String formatFileName = DvUtils.getValue("FORMATFILE");
        if (null == formatFileName || formatFileName.equals("")) {
            if (null != output) {
                output.add(" No FORMATFILE setting.");
            }
        }
        else {
            // check it can be found.....
            // either full name or just last part allowed.....
            File formatFile = new File(formatFileName);
            if (false == formatFile.canRead())     {
                formatFile = DvUtils.searchForFile(formatFileName);
            }
            if (formatFile == null || false == formatFile.canRead()) {
                if (null != output) {
                    output.add(" FORMATFILE "+formatFileName + " not found.");
                }
            }
            else {
                if (null != output) {
                    output.add(" FORMATFILE is "+formatFile.getAbsolutePath());
                }
                try {
                    // ok so we have a format file - lets see what formats it supports
                    // first check that it has a valid eyecatcher JVMSYMS
                    // Check removed - 080702 - WPR
                    FileInputStream fis = new FileInputStream(formatFile);
                    //byte[] b=  new byte[8];
                    //fis.read(b); // read 8 bytes
                    //String s = new String(b);
                    //if (s.indexOf("JVMSYMS") == 0) {
                    //    if (null != output) {
                    //        output.add(" .... and it has a valid eyecatcher (JVMSYMS)\n");
                    //    }
                    //    DvUtils.setTypeFileLoaded(true);
                    //}
                    //else {
                    //    if (null != output) {
                    //        output.add(" .... although eyecatcher is not JVMSYMS !!!!\n");
                    //    }
                    //}
                    // even with invalid eyecatcher we will try an list the supported
                    // formats........................
                    CTypeFile c = new CTypeFile();
                    c.processTypes(formatFile.getAbsolutePath());

                }
                catch (java.io.FileNotFoundException e) {
                    /* should never happen */
                }
                catch (java.io.IOException e) {
                    if (null != output) {
                        output.add(" IOError for "+formatFileName);
                    }
                }
            }

        }
    } // end of method

    /**
     * Validate the JVM via eyecatchers - this will
     * hopefully allow the catching of mismatch between
     * types file and dump level.
     *
     *
     */
    public boolean validateJVM() {
        return true;
    }

    /**
     *  gets a string from an address in memory
     *
     *
     */
     public static String stringFromAddress(long nameAddr, DvAddressSpace as) {
        boolean finished = false;

        byte b= 0;
        int offset = 0;
        String name = "!!!! value could not be established !!!!";
        while(false == finished) {
            try {

                b  = as.readByte(as.createAddress(nameAddr+offset));
                if (0 == b ) {
                    finished = true;
                }
                else {

                    offset++;
                }

            }
            catch (DvAddressException e) {
                // copes with running out of memory area or
                // an invalid address
                finished = true;
            }
        }

        if (0 == offset) {
            // We can't find a signature

        }
        else
        {
            try {
                byte[] bArray =  as.readBytes(as.createAddress(nameAddr),offset);
                name = new String(bArray);
            }
            catch (DvAddressException e) {
                // should never happen.....
            }

        }
        return name;
     }

    /**
     *  Strips path info from filename string
     *
     *
     */
     public static String stripToFileName(String fullName) {
        StringTokenizer st = new StringTokenizer(fullName,"\\/");
        int count = st.countTokens();
        String s=null;
        for (int i=0;i<count;i++) {
            s = st.nextToken();
        }
        return s;
     }

    /**
     *  get the time
     *
     *
     */
    static final String getTime() {
        Calendar c = Calendar.getInstance();
        return df.format(c.getTime());
    }


}




