/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/ElfNoteS390.java, ras, dumpviewer 1.3
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 41452  210502 nmasson  New

 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.io.*;
import java.util.*;

 /**
  * ElfNoteS390 reads Notes sections in a Linux core file
  * The layout is derived from the kernel files
  * /usr/src/linux/include/linux/elfcore.h
  * /usr/src/linux/fs/binfmt_elf.c
  */
import java.util.*;
import java.io.*;

class ElfNoteS390 extends ElfNote {
    static final int NumGeneralPurposeRegisters = 16;
    static final int NumACRegisters = 16;


    ElfNoteS390(DvFileReader fileReader) throws IOException {
	super(fileReader);
	//Data from /usr/src/linux/include/asm-390/ptrace.h

	// Align to 8-byte boundary
	long pos = fileReader.getPos();
	pos = (pos + 7) & (~7);
	DvUtils.trace("Align from " + fileReader.getPos() + " to " + pos, 2, false);
	fileReader.setPos(pos);
	long mask = fileReader.readAddress();
	long addr = fileReader.readAddress();

	long gpr;
	String pad = "    ";
	for(int i=1; i<= NumGeneralPurposeRegisters; i++) {
	    gpr = fileReader.readAddress();
	    String regName = "gpr"+ Long.toString(i) + pad;
	    DvUtils.trace(regName + " = " + Long.toHexString(gpr), 2, false);
	    register.put((Object)regName, (Object)(new Long(gpr)));
	    if(i == 9) {
	    	pad = "   ";
	    }
	}
	pad = "    ";
	for(int i=1; i<= NumACRegisters; i++) {
	    gpr = fileReader.readWord();
	    String regName = "acr"+ Long.toString(i) + pad;
	    DvUtils.trace(regName + " = " + Long.toHexString(gpr), 2, false);
	    register.put((Object)regName, (Object)(new Long(gpr)));
	    if(i == 9) {
	    	pad = "   ";
	    }
	}
	long origgpr2 = fileReader.readAddress();
	long trap = fileReader.readWord();
	long old_ilc = fileReader.readWord();
	
	register.put((Object)"origgpr2", (Object)(new Long(origgpr2)));
	register.put((Object)"trap    ", (Object)(new Long(trap)));
	register.put((Object)"old_ilc ", (Object)(new Long(old_ilc)));

	DvUtils.trace("origgpr2 = " + Long.toHexString(origgpr2), 2, false);
	DvUtils.trace("trap = " + Long.toHexString(trap), 2, false);
	DvUtils.trace("old_ilc = " + Long.toHexString(old_ilc), 2, false);

    }

}    
