/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/ElfNotes.java, ras, dumpviewer 1.7    
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 41452  210502 nmasson  New

 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.io.*;
import java.util.*;

 /**
  * ElfNotes reads Notes sections in a Linux core file
  * The layout is derived from the kernel files
  * /usr/src/linux/include/linux/elfcore.h
  * /usr/src/linux/fs/binfmt_elf.c
  */
import java.util.*;
import java.io.*;

class ElfNotes {
    DvFileReader reader;
    ElfHeader header;
    ElfNote note;			// Note for current process
    HashMap register = new HashMap();
    HashMap notesList = new HashMap();

    // These definitions are taken from sys/elf.h on Solaris
    static final int NT_PRSTATUS =   1;  /* prstatus_t   <sys/old_procfs.h> */
    static final int NT_PRFPREG  =   2;  /* prfpregset_t <sys/old_procfs.h> */
    static final int NT_PRPSINFO =   3;  /* prpsinfo_t   <sys/old_procfs.h> */
    static final int NT_PRXREG   =   4;  /* prxregset_t  <sys/procfs.h>     */
    static final int NT_PLATFORM =   5;  /* string from sysinfo(SI_PLATFORM)*/
    static final int NT_AUXV     =   6;  /* auxv_t array <sys/auxv.h>       */
    static final int NT_GWINDOWS =   7;  /* gwindows_t   SPARC only         */
    static final int NT_ASRS     =   8;  /* asrset_t     SPARC V9 only      */
    static final int NT_PSTATUS  =   10; /* pstatus_t    <sys/procfs.h>     */
    static final int NT_PSINFO   =   13; /* psinfo_t     <sys/procfs.h>     */
    static final int NT_PRCRED   =   14; /* prcred_t     <sys/procfs.h>     */
    static final int NT_UTSNAME  =   15; /* struct utsname <sys/utsname.h>  */
    static final int NT_LWPSTATUS=   16; /* lwpstatus_t  <sys/procfs.h>     */
    static final int NT_LWPSINFO =   17; /* lwpsinfo_t   <sys/procfs.h>     */

    //IBM.FIXME This is a Linux definition
    static final int NT_TASKSTRUCT = 4;

    ElfNotes(DvFileReader reader, ElfHeader header) throws IOException {
	this.reader = reader;
	this.header = header;
	boolean noProblem = true;
	long nextNote = reader.getPos();
	long fudgeFactor = 0;

	if(header.machine == 2) {
	    // Linux and Solaris disagree on the length of Notes in
	    // the core file.  On Solaris the next note appears after
	    // an extra four bytes
	    fudgeFactor = 4;
	}
	 do {
	    if(reader.getPos() != nextNote) {
		System.err.println("Expected to be at offset " +
		    nextNote + " after reading Note, but find myself at " +
		    "offset " + reader.getPos() + ".  Adjusting offset");
		reader.setPos(nextNote);
	    }
	    ElfNote note = null;
            long nameLen = reader.readWord();
            long dataSize = reader.readWord();
            long type = reader.readWord();
            String name = reader.readString(4);
            DvUtils.trace("nameLen = " + nameLen, 2, false);
            DvUtils.trace("dataSize = " + dataSize, 2, false);
            DvUtils.trace("type = " + type, 2, false);
            DvUtils.trace("name = " + name, 2, false);
	    DvUtils.trace("Current file offset = " + 
		Long.toHexString((reader.getPos() - 16)), 2, false);

	    if(!name.equals("CORE")) {
	    	break;
	    }
	    nextNote = reader.getPos() + dataSize + fudgeFactor; 
	    DvUtils.trace("Next file offset = " + 
		Long.toHexString(nextNote), 2, false);

	    switch((int)type) {
	    case ElfNotes.NT_PRSTATUS:
		DvUtils.trace("Process status note", 2, false);
	        if(header.machine == 3) { //IBM.FIXME Intel Ugly
	    	    note = new ElfNoteIA32(reader);
		} else if(header.machine == 50) {
	    	    note = new ElfNoteIA64(reader);
	    	} else if(header.machine == 20) { //IBM.FIXME PPC Ugly
	    	    note = new ElfNotePPC32(reader);
	    	} else if(header.machine == 22) { //IBM.FIXME s390 Ugly
					//IBM.FIXME also check the value of 22
	    	    note = new ElfNoteS390(reader);
	   	}
		if(note != null) {
		    DvUtils.trace("Store registers for process " + 
			note.pid, 2, false);
		    register.put((Object)note.pid, (Object)note);
		    //IBM.FIXME Read the pid from a different Elf note
		    // For now assume that the first note is for the
		    // current pid
		    this.note = note;
		}
	    	break;

	    case ElfNotes.NT_PRFPREG:
		DvUtils.trace("Floating point note", 2, false);
		reader.setPos(nextNote);
	  	break;

	    case ElfNotes.NT_PRPSINFO:
		DvUtils.trace("Process info note", 2, false);
		reader.setPos(nextNote);
	  	break;

	    case ElfNotes.NT_TASKSTRUCT:
		DvUtils.trace("Task structure note", 2, false);
		reader.setPos(nextNote);
	  	break;

	    case ElfNotes.NT_PSTATUS:
		DvUtils.trace("Process Status note", 2, false);
		note = new ElfNotePStatus(reader);
	  	break;

	    case ElfNotes.NT_PSINFO:
		DvUtils.trace("Process Info note", 2, false);
//		note = new ElfNotePsInfo(reader);
	  	break;

	    default:
		System.err.println("Unrecognised note in core file: type=" +
		    type);
		reader.setPos(nextNote);
		break;

	    } //switch(type)

	} while(noProblem);
    }

}    
