/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/HPIModule.java, ras, dumpviewer 1.5    
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 41452  210502 nmasson  New

 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

 /**
  *
  */

import java.util.*;
import java.io.*;

class HPIModule extends Segment {
    long hpiInfo;
    long threadQueue;
    long threadNum;
    long threadNext;
    long execEnvOffset;
    long pidOffset;
    long tidOffset;
    long pfmData;

    HPIModule() {}

    HPIModule(Segment seg) {
	super(seg.moduleName, seg.moduleAddress.getAddressAsLong(), 
		seg.moduleSize, seg.moduleType, seg.reader, seg.offset);
    }

    boolean populate() throws IOException {
	DvUtils.trace("populate HPIModule", 0, false);
	int length = (int) moduleSize;
	((DvFileReader)reader).setPos(offset);
	String buffer = ((DvFileReader)reader).readString(length);

        //long buffOffset = buffer.indexOf("HPIRAS\0\0\uaa55\uaa55");
        //long buffOffset = buffer.indexOf("HPIRAS\0\0\u55aa\u55aa");
        int buffOffset = buffer.indexOf("HPIRAS");  
	byte[] bitPattern = new byte[4];
	while(buffOffset != -1) {
	    // Check the bit pattern for correct identification
	    DvUtils.trace("Match at " + Long.toHexString(buffOffset), 1, false);
	    buffOffset += 8;
	    ((DvFileReader)reader).setPos(offset+buffOffset);
	    ((DvFileReader)reader).read(bitPattern);
	    String s = new String(bitPattern);
	    DvUtils.trace("Bit pattern = <" + s +">", 1, false);
	    DvUtils.trace("Try <\uaa55\uaa55> and <\u55aa\u55aa> (unicode)", 1, false);
	    DvUtils.trace("Try <\252\125\252\125> and <\125\252\125\252> (octal)", 
		1, false);
	    if(s.equals("\252\125\252\125") || s.equals("\125\252\125\252")) {
		break;
	    }
	    // Find the next prospective match
	    buffOffset = buffer.indexOf("HPIRAS", buffOffset);
	}
	if(buffOffset == -1) {
	    // Tagged data not found
	    System.err.println("Bitpattern not found");
	    return false;
	}
	long fileOffset = this.offset + buffOffset;
	fileOffset += 2*8;			// IBM.FIXME read 4 ints
	((DvFileReader)reader).setPos(fileOffset);
        threadQueue = ((DvFileReader)reader).readAddress();
        threadNext = ((DvFileReader)reader).readAddress();
        execEnvOffset = ((DvFileReader)reader).readAddressOffset();
        pidOffset = ((DvFileReader)reader).readAddressOffset();
        tidOffset = ((DvFileReader)reader).readAddressOffset();
        pfmData = ((DvFileReader)reader).readAddressOffset();
	DvUtils.trace("threadQueue=" + Long.toHexString(threadQueue), 2, false);
	DvUtils.trace("threadNext=" + Long.toHexString(threadNext), 2, false);
	DvUtils.trace("execEnvOffset=" + Long.toHexString(execEnvOffset), 2, false);
	DvUtils.trace("pidOffset=" + Long.toHexString(pidOffset), 2, false);
	DvUtils.trace("tidOffset=" + Long.toHexString(tidOffset), 2, false);
	DvUtils.trace("pfmData=" + Long.toHexString(pfmData), 2, false);

	return true;
    }
    boolean populateThread(ExProcess process, HashMap noteList) 
	throws IOException {
	// Now read the Java threads
	long thread;
	DvUtils.trace("populateThread: process="+process + 
		" reader="+ reader, 0, false);
	((DvFileReader)reader).setAddress(threadQueue);
	thread = ((DvFileReader)reader).readAddress();
	DvUtils.trace("Thread: " + Long.toHexString(thread), 2, false);
	while(thread != 0) {
	    long execEnv = thread + execEnvOffset;
	    DvUtils.trace("thread="+thread+" execEnvOffset=" + execEnvOffset
	    	+ " execEnv=" + execEnv, 1, false);
	    DvUtils.trace("execEnv : " + Long.toHexString(execEnv), 2, false);
	    ((DvFileReader)reader).setAddress(thread);
	    long pid = ((DvFileReader)reader).readAddress();
	    DvUtils.trace("pid address: " + 
		Long.toHexString(thread) + 
		" value = " + Long.toHexString(pid), 2, false);
	    DvUtils.trace("Thread: " + Long.toHexString(thread), 2, false);
	    ((DvFileReader)reader).setAddress(thread + tidOffset);
	    long tid = ((DvFileReader)reader).readAddress();
	    ExThread jvmThread = new ExThread(pid, tid, execEnv, noteList);
	    process.javaThread.add(jvmThread);

	    ((DvFileReader)reader).setAddress(thread + threadNext);
	    thread = ((DvFileReader)reader).readAddress();
	    DvUtils.trace("Thread: " + Long.toHexString(thread), 2, false);
	}
	return true;
    }
}    
