/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/SDFFAddressSpace.java, ras, dumpviewer 1.25
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502 PHILR:  New for feature 36524
 * 41452  230502 nmasson Add printSDFF
 * 36524  050802 philr   fix to readPointer for 32 bits
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.io.*;
import java.util.*;

/**
 *  Description: This classe is concerned with allowing analysis of the
 *               address spaces contained with in the sddf. In addition it
 *               acts as the anchor position for obtaining memory from the
 *               dump.
 */
public class SDFFAddressSpace extends DvAddressSpace  {

    private long mmapOffset;
    private long mmapLength;
    private SDFFMemRanges[] theMemRanges = null;
    private RandomAccessFile theRAF;
    private Vector processes = new Vector(); // SDFFProcess objects for this
                                           // AddressSpace....

    private Vector process_offsets = new Vector();
    private long id;
    private long processCount;
    private SDFFDump owningDump; // what SDFFDump object owns us


    //private long currentThread;
    //private long signal;
    //private String command_line;



    // private MetaData metaData;   later later

    /**
     * Description: Constructor
     *
     */
    public SDFFAddressSpace(long asID,long prCount,
                          long mmo, long mml,
                          RandomAccessFile raf ) {
        id = asID;
        processCount = prCount;
        mmapOffset = mmo;
        mmapLength = mml;
        theRAF = raf;

    }


    /**
     * Description: Reads memory from the memory area of the dump
     *
     */
    public  byte[] readBytes(DvAddress addr, long len) {

        //System.out.println("addr/len is:"+addr+","+len);
        byte[] b = null;
        long longAddress = addr.getAddressAsLong();


        // So we need to use theMemRanges arrayto determine offset
        if (null == theMemRanges) {
            // we are stuffed
        }
        else {

            int foundit = findAddressRangeIndex( addr);

            // so if we've found a memrange containing our address lets try and read it
            // note we may only be able to do a partial read!!!!!!
            // Need to improve this so we can do cacheing of some sort, especially when they
            // want gigantic number of bytes (say >32K),
            if (-1 != foundit) {
                long start=theMemRanges[foundit].addressStart().getAddressAsLong();;
                // figure out how many bytes we should be able to read
                // we may want to be clever and be able to cross contiguous
                // memory ranges.......
                // At this point start of the range in question is here
                // as in the start of the memory we are looking for (longAddress)
                long offset_into_range = longAddress - start;
                //System.out.println("\nFound offset_into_range:  "+offset_into_range);
                try {
                    long where_to_read_from =
                    theMemRanges[foundit].getOffset()+ offset_into_range;
                    //System.out.println("\nWhere to read from:  "+where_to_read_from);
                    // Now figure out how many bytes we can read bearing in mind that the
                    // asker has asked for len bytes
                    // The maximum we can provide is 2G as this is the maximum size
                    // of an array

                    //if (len>(2^32-1)) len = 2^32-1; // do we really want to ask for 2GB
                                                // we need to implement a cache!!!!

                    int howmany = (int) len;

                    //System.out.println("\n !!! "+ theMemRanges[foundit].length() + " -- "
                    // + offset_into_range + " -- " + len);
                    long available = theMemRanges[foundit].length() - offset_into_range;
                    if (len > available) {
                        howmany = (int) available;
                    }

                    //System.out.println("\n !!! "+ where_to_read_from + " -- " + howmany);
                    b = new byte[howmany];
                    theRAF.seek(where_to_read_from);
                    theRAF.read(b,0,howmany);

                }
                catch (IOException e) {
                    System.out.println("\nSDFFAddressSpace:readBytes got IOException!!!"  );
                }


            }


         }


         return b;  // if b is null then we didn't find the memory

    }

    /**
     * Description: Loads the mem ranges array
     *
     */
    public  DvMemRanges[] getMemRanges()  {
        // Mote we only load the memranges array once!!!

        if (null == theMemRanges && mmapOffset != -1) {

            // lets figure out how many we want
            // using static method in SDFF memRanges
            int count = SDFFMemRanges.countThem(mmapOffset,mmapLength,theRAF);

            // static method in SDFF memRanges temporarily
            theMemRanges = new SDFFMemRanges[count];
            // So we know how many - lets load them up!
            for (int i = 0;i<count;i++)
            {
                theMemRanges[i] = new SDFFMemRanges();
                theMemRanges[i].loadFromFile(mmapOffset,mmapLength,theRAF,i);
            }
            //System.out.println("\nMemranges[0] has... " +
            //                     (theMemRanges[0].addressStart()).getAddressAsLong() + " " +
            //                     theMemRanges[0].length() + " " +
            //                     theMemRanges[0].getOffset());

        }


        return theMemRanges;
    }

    /**
     * Description: Given a long converts it to an address
     *
     */
    public DvAddress createAddress(long x) {

        SDFFAddress retAddr = new SDFFAddress();
        retAddr.setAddress(x);

        return retAddr;
    }

    /**
     * Description: Loads the SDFFAddressSpace instance from the dump file
     *
     */
    public static SDFFAddressSpace[] loadFromFile(long o ,long length,
                              long mmo, long mml, RandomAccessFile raf) {

    // We do this by looking at the SDDFPRTH segment and chaining through
    // the ADDRSPC segments within it, checking on the eyecatchers to ensure
    // that we are on track...... top of each address space will contain
    // its length and its eyecatcher, hence we can find next one (if there
    // is one)
    // First incoming parameter should be the offset of the sdffprth
    // second the length of the prth, 3rd mmap offset, 4th mmap length
    // fifth the dumpfile link
    //
    // Note this method establishes the total content of the PRTH and hence
    // is very crucial to the successful working of the dump formatter.....

    // First skip the SdffStd header
    // We have a lot of trace facility in here as if an SDFF dump is not the
    // correct format probably nothing will work later.......
        long  asOffset_in_file = o  + SdffStd.getSize();
        boolean format_problem_found=false;
        long  prOffset_in_file;
        long  asLength,prLength;
        long  asID,prID;
        long  asCount;
        long  prCount;
        byte  asEyeCatcher[]= new byte[8];
        byte  prEyeCatcher[]= new byte[8];
        byte  signalName[]= new byte[8];
        byte  command_line[]= new byte[256];
        long  threadArrayOffset;
        SDFFAddressSpace[] asArray=null;
        SDFFProcess tempProcess=null;
        long numThreads ;
        long currThread ;
        long signalNumber ;
        int  currProcessIndex ;

        try {
            raf.seek(asOffset_in_file); // Now we are at right place
            asCount = raf.readLong();
            DvUtils.trace("Count of address spaces is: "+asCount,
                            0,false);
            // set up enough array iems
            asArray = new SDFFAddressSpace[(int) asCount];

            //System.out.println("\nNumber of AddressSpaces =  "+asCount);
            for (int i=0;i<asCount;i++) {
                asLength = raf.readLong();
                DvUtils.trace("Length of address space (index " + i + ") is: "+asLength,
                               0,false);

                DvUtils.trace("Checking ADDRSPC eyecatcher @ offset "+
                                raf.getFilePointer(),
                                0,false);
                raf.readFully(asEyeCatcher);

                // Now check eyecatcher for being valid

                String y = new String(asEyeCatcher);
                if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffAs)) {
                    DvUtils.trace("ADDRSPC eyecatcher appears valid",0,false);
                }
                else {
                    DvUtils.errorMsg("ADDRSPC eyecatcher invalid"+
                                    "\nFurther failures expected", 0);
                    format_problem_found = true;

                }

                if (false == format_problem_found) {
                    asID = raf.readLong();
                    DvUtils.trace("ADDRSPC id is: "+asID,
                                0,false);
                    //DvUtils.trace("Checking PROCESS eyecatcher @ offset "+raf.getFilePointer(),
                    //        0,false);
                    //System.out.println("\nAS ID is "+asID);
                    // Now we should be able to see number of processes
                    prCount = raf.readLong();
                    // !!! HACK WARNING
                    // hack to support "old" extracted dumps during development
                    // !!! HACK WARNING
                    long temp_pos =  raf.getFilePointer();
                    long temp = raf.readLong();  // size of process section
                    if (raf.readByte() != (byte) 'P') {
                        raf.seek(temp_pos);
                        currProcessIndex = raf.readInt();
                        DvUtils.trace("There appear to be "+prCount+
                                    " processes and current process index is "+currProcessIndex,
                                0,false);

                        // Addition of 60 spare bytes..... which we throw away for now...
                        byte[] ba = new byte[60];
                        raf.read(ba);
                    }
                    else {

                       DvUtils.trace("There appear to be "+prCount+
                                    " processes ",0,false);
                       raf.seek(temp_pos);
                    }


                    // set up
                    asArray[i] = new SDFFAddressSpace(asID,prCount,mmo,mml,raf);

                    prOffset_in_file= raf.getFilePointer();

                    format_problem_found = loadProcesses(prOffset_in_file,raf,prCount,asArray,i);
                    if (true == format_problem_found) {
                        // hack for late change to sdff format
                        // with 64 spare bytes added to AddressSpace
                        format_problem_found = loadProcesses(prOffset_in_file+64,
                                                            raf,prCount,asArray,i);
                        if (true == format_problem_found) {
                            DvUtils.errorMsg("PROCESSS eyecatcher invalid"+
                                        "\nFurther failures expected", 0);
                        }

                    }


                    asOffset_in_file += asLength;   // so we will point at next one!
                                                    // if we ever have more than 1
                }


            } // end of for loop indexed by i

        }
        catch (java.io.IOException e) {
            System.out.println("\nIOException caught in SDFFAddressSpace:loadFromFile");
        }


        return asArray;

    }

    /**
     * Description: Determines how many address spaces are contained in the dump
     *
     */
    public static int countThem(long o,RandomAccessFile raf) {

        byte  asEyeCatcher[]= new byte[8];
        int count=0;
        long  asLength;
        // count is simple as its stored as part of the prth segment
        // but we also check eyecatcher as part of this
        long  asOffset_in_file = o  + SdffStd.getSize();
        try {
            raf.seek(asOffset_in_file); // Now we are at right place
            count = (int) raf.readLong();
            asLength = raf.readLong();

            //System.out.println("\nLen of address space hopefully.. "+asLength);

            raf.readFully(asEyeCatcher);
            // Now check eyecatcher for being valid

            String y = new String(asEyeCatcher);
            if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffAs)) {
                //System.out.println("\nLooks like we have a valid ADDRSPC eyecatcher");
            }
            else {
                DvUtils.errorMsg("ADDRSPC eyecatcher invalid", 0);
                count = 0;
            }
        }
        catch (java.io.IOException e) {
            System.out.println("\nIOException caught in SDFFAddressSpace:countThem");
            count = 0;
        }


        return count;
    }

    /**
     * Description: gets all the processes under this array
     *
     */
    public DvProcess[] getProcesses() {
        DvProcess [] retArray=null;
        if (processes.size() != 0) {
            retArray = new DvProcess[processes.size()];
            for (int i=0;i<processes.size();i++) {
                retArray[i] = (DvProcess) processes.elementAt(i);
            }
        }


        return retArray;
    }

    /**
     * Description:returns the address space id
     *
     */
    public String id() {

        return Long.toString(id);
    }

    /**
     * Description: Read a byte from memory
     *
     */
    public  byte readByte(DvAddress address) throws DvAddressException {
        byte b;
        byte bArray[];

        bArray = readBytes(address,1);
        if (null == bArray) {
            throw  new DvAddressException("No range containing this address");
        }
        else {
            b=bArray[0];
        }

        return b;
    }

    /**
     * Description: Read a java int from memory (4 bytes)
     *
     */
    public   int readInt(DvAddress address) throws DvAddressException {

        int retval=0;
        try {
            long longAddress = address.getAddressAsLong();
            int foundit = findAddressRangeIndex(address);
            if (-1 != foundit) {
                long start=theMemRanges[foundit].addressStart().getAddressAsLong();
                // figure out how many bytes we should be able to read, lets
                // hope its at least 4
                // we may want to be clever and be able to cross contiguous
                // memory ranges....... later later
                long offset_into_range = longAddress - start;

                long where_to_read_from =
                      theMemRanges[foundit].getOffset()+ offset_into_range;
                theRAF.seek(where_to_read_from);
                DvDump theDump = DvConsole.getCurrentDump();
                if (false == DvUtils.isBigEndian(theDump.getArchitecture())) {
                    int accum = 0;
                    for (int shiftBy = 0; shiftBy <32; shiftBy+=8) {
                        accum |=  (theRAF.readByte() & 0xff) << shiftBy;
                    }

                retval = accum;

                }
                else retval = theRAF.readInt();
            }
            else {
                throw new DvAddressException("No range containing this address");
            }
        }
        catch (java.io.IOException ex) {
            // convert this into a DvAddressException
            throw new DvAddressException("IOException behind this");
        }


        return retval;
    }

    /**
     * Description: Read a java short from memory (2 bytes)
     *
     */
    public short readShort(DvAddress address) throws DvAddressException {

        short retval=0;
        try {
            long longAddress = address.getAddressAsLong();
            int foundit = findAddressRangeIndex(address);
            if (-1 != foundit) {
                long start=theMemRanges[foundit].addressStart().getAddressAsLong();
                // figure out how many bytes we should be able to read, lets
                // hope its at least 4
                // we may want to be clever and be able to cross contiguous
                // memory ranges....... later later
                long offset_into_range = longAddress - start;

                long where_to_read_from =
                    theMemRanges[foundit].getOffset()+ offset_into_range;
                theRAF.seek(where_to_read_from);
                DvDump theDump = DvConsole.getCurrentDump();
                if (false == DvUtils.isBigEndian(theDump.getArchitecture())) {
                    short accum = 0;
                    for (int shiftBy = 0; shiftBy <16; shiftBy+=8) {
                        accum |=  (theRAF.readByte() & 0xff) << shiftBy;
                    }

                    retval = accum;

                }
                else retval = theRAF.readShort();
            }
            else {
                throw new DvAddressException("No range containing this address");
            }
        }
        catch (java.io.IOException ex) {
            // convert this into a DvAddressException
            throw new DvAddressException("IOException behind this");
        }


        return retval;
    }

    /**
     * Description: readPointer gets a long or int from the address given and
     *                   converts it into a DvAddress taking into account the
     *                   endianess and bitness of the dump....
     *
     *
     */
    public   DvAddress readPointer(DvAddress address) {
        DvAddress retval = null;
        DvDump theDump = DvConsole.getCurrentDump();
        long addrAsLong;
        try {
            if (true == DvUtils.is64BitSystem(theDump.getArchitecture())) {
                addrAsLong = readLong(address); // Note readLong will have coped
                                        // with the endianess!!!!!

            }
            else {

                addrAsLong = (long) readInt(address);
                // if the int was -ve then need to get rid of the
                // leading f's
                addrAsLong &=  0x00000000ffffffffL;

            }
            //now create a DvAddress from the long
            retval = createAddress(addrAsLong);

        }
        catch (DvAddressException ex) {
            // will return null retval!!! indicating error
        }

        return retval;
    }

    /**
     * Description: reads a java long from memory given the address. Converts
     *                   the endianess and bitness where necessart
     *
     */
    public long readLong(DvAddress address) throws DvAddressException {
        long retval=0;
        try {
            long longAddress = address.getAddressAsLong();
            int foundit = findAddressRangeIndex(address);
            if (-1 != foundit) {
                long start=theMemRanges[foundit].addressStart().getAddressAsLong();;
                // figure out how many bytes we should be able to read, lets
                // hope its at least 8
                // we may want to be clever and be able to cross contiguous
                // memory ranges....... later later
                long offset_into_range = longAddress - start;

                long where_to_read_from =
                    theMemRanges[foundit].getOffset()+ offset_into_range;
                theRAF.seek(where_to_read_from);
                DvDump theDump = DvConsole.getCurrentDump();
                if (false == DvUtils.isBigEndian(theDump.getArchitecture())) {
                    long accum = 0;
                    for (int shiftBy = 0; shiftBy <64; shiftBy+=8) {
                        accum |= (long) (theRAF.readByte() & 0xff) << shiftBy;
                    }
                    retval = accum;
                }
                else {
                    retval = theRAF.readLong();
                }
            }
            else {
                throw new DvAddressException("Address not found in an address range");
            }
        }
        catch (java.io.IOException ex) {
            // convert this into a DvAddressException
            throw new DvAddressException("IOException behind this");
        }

        return retval;
    }

    /**
     * Description: Find the current process (i.e. the one in
     *                   control when the dump was taken
     *
     */
    public  DvProcess getCurrentProcess() {

        int i=DvConsole.getCurrentProcessIndex();

        return (DvProcess) processes.get(i);
    }

    /**
     * Description: Find index into address range table for an address
     *                   -1 means that it isn't in the dump
     *
     */
    private int findAddressRangeIndex(DvAddress address) {
        long start=0;
        int foundit=-1; // assume we are going to fail
        long longAddress = address.getAddressAsLong();
        if (null == theMemRanges) {
            // we are stuffed
        }
        else {
            for (int i=0;i<theMemRanges.length && foundit == -1 ;i++) {
                start = theMemRanges[i].addressStart().getAddressAsLong();
                if (longAddress >= start  &&
                    longAddress < start+theMemRanges[i].length()) {
                    //System.out.println("\nFound one !!! - "+i);
                    foundit = i;
                }
            }
        }
        return foundit;
    }

    /**
     * loadProcesses
     */
    private static boolean loadProcesses(long start,
                               RandomAccessFile raf,
                               long prCount,
                               SDFFAddressSpace[] asArray,
                               int i) {

        boolean format_problem_found = false;
        long prOffset_in_file = start;
        long prLength;
        long prID;
        long numThreads;
        long currThread;
        long signalNumber;
        byte  prEyeCatcher[]= new byte[8];
        byte  signalName[]= new byte[8];
        byte  command_line[]= new byte[256];
        long threadArrayOffset;
        SDFFProcess tempProcess=null;

      try {
        raf.seek(prOffset_in_file);
        for (int j=0;j<prCount;j++) {
            DvUtils.trace("Checking PROCESS with index "+j,0,false);
            asArray[i].processes.clear();
            prLength = raf.readLong();
            DvUtils.trace("Process length is "+prLength,0,false);
            DvUtils.trace("Checking PROCESS eyecatcher @ offset "+
                           raf.getFilePointer(),
                          0,false);
            raf.readFully(prEyeCatcher);
            String y = new String(prEyeCatcher);
            if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffProc)) {
                 DvUtils.trace("PROCESS eyecatcher is valid",0,false);
            }
            else {
                format_problem_found = true;
                j = (int) prCount; // so as we abandon loop at earliest
                                        // opportunity
            }

            if (false == format_problem_found) {
                prID = raf.readLong();
                //System.out.println("\nProcess ID is "+prID);
                // Now read rest of process info.... this gets very complex
                // as structure is quite complicated.......
                numThreads = raf.readLong();
                currThread = raf.readLong();
                signalNumber = raf.readLong();
                raf.readFully(signalName);
                String sigName = new String(signalName);
                //System.out.println("signalName " + yyy);
                raf.readFully(command_line);
                threadArrayOffset = raf.readLong();

                // Next we need to set up the envdata... but and loaded sections
                // but will skip that of the moment.....

                // we cheat on finding the envdata

                // TO BE DONE
                // TO BE DONE
                // TO BE DONE
                // TO BE DONE


                // Note the instantiation of this new process causes thread objects
                // to be instantiated
                tempProcess = new SDFFProcess(prID,numThreads,currThread,signalNumber,
                                            sigName,"command_line",
                                            prOffset_in_file,prLength,
                                            raf,asArray[i]);
                asArray[i].processes.add(tempProcess);
                prOffset_in_file += prLength;
            }

        } // end of for loop indexed by j
      }
      catch (java.io.IOException e) {
            System.err.println("\nIOException caught in SDFFAddressSpace:loadProcesses");
            format_problem_found = true;
      }

      return format_problem_found;
    }

    /**
     * Serialise to SDFF file
     */
    void printSdff(Sdff sdff, MemoryMap map) throws IOException {
        long start = sdff.getFilePointer();
        sdff.writeLong(0l);
        sdff.writeBytes("ADDRSPC ");
        sdff.writeLong(id);
        sdff.writeLong(processes.size());
        Enumeration list = processes.elements();
        while(list.hasMoreElements()) {
            SDFFProcess process = (SDFFProcess)list.nextElement();
            process.printSdff(sdff, map);
        }
        sdff.writeLong(-1l);
        sdff.printLength(start);
    }

    public int pointerSize() {
        DvDump theDump = DvConsole.getCurrentDump();
        if (theDump != null && DvUtils.is64BitSystem(theDump.getArchitecture()))
            return 8;
        else
            return 4;
    }
}

