/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/SDFFDump.java, ras, dumpviewer 1.16
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502 PHILR:  New for feature 36524
 * 41452  230502 nmasson Add printSdff
 * 41452  230502 nmasson Make members protected
 * 36724  090602 philr   Tidy up
 * 36724  260602 philr   Trace added  #
 * 36724  080702 philr   Reset added plus tidy up
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.io.*;
import java.util.*;


 /**
  *   This class encapsulates the concept of a dump formatted as an SDFF dump.
  *
  */
public class SDFFDump extends DvDump  {

    protected static RandomAccessFile theRAF; // This is how we get into the guts of the
                                        // SDFF
    protected static long prthOffset = -1;  // Offset of SDFFPRTH segment in the file
    protected static long prthLength = -1;  // Length of SDFFPRTH segment in the file
    protected static long mmapOffset = -1;  // Offset of SDFFMMEM segment in the file
    protected static long mmapLength = -1;  // Length of SDFFMMEM segment in the file
    protected static long memOffset  = -1;  // Offset of SDFFMEM  segment in the file
    protected static long memLength  = -1;  // Length of SDFFMEM segment in the file
    protected static long sizeOfFile = -1;  // Size of the SDFF (in bytes)
    protected static String nameOfFile;

    protected Vector as_offsets = new Vector();  // and their offsets
    protected SDFFAddressSpace[] theASArray;

    // SDFFHEAD data

    protected static int system=-1;
    protected static int architecture=-1; // endianess and 32/64 bitness
    protected static int subsystem=-1;
    protected static int processor=-1;
    protected static int processorSubtype=-1;
    protected static int numProcessors=-1;
    protected static int currentProcessor=-1;



    /**
     *   Find the AddressSpaces hidden within the dump.
     *
     */
    public DvAddressSpace[] getAddressSpaces() {


        if (null == theASArray) {
            if (prthOffset == -1) {
                return null; // what no prth section!
            }

            int asCount = SDFFAddressSpace.countThem(prthOffset,theRAF);
            if (0 != asCount) {
                theASArray = new SDFFAddressSpace[asCount];
            }
            else {
                return null; // we are in trouble.....
            }

            for (int i=0; i<asCount; i++) {
                theASArray = SDFFAddressSpace.loadFromFile(prthOffset,prthLength,
                                                  mmapOffset,mmapLength,
                                                  theRAF);
            }


        }

        return theASArray;
    }

    /**
     *   Return the endianess and bitness for the dump.
     *
     */
    public  int getArchitecture() {
        return architecture;
    }

    /**
     *   Set the endianess and bitness for the dump.
     *
     */
    public void setArchitecture(WordLength wordLength, WordType endType) {
        architecture = 0;
        if(wordLength.size == 4) {
            architecture += DvConstants.ARCH_IS_32BIT;
        }
        else {
            if(wordLength.size == 8) {
                architecture += DvConstants.ARCH_IS_64BIT;
            }
        }

        if(endType instanceof LittleEnd) {
            architecture += DvConstants.ARCH_IS_LITTLEENDIAN;
        }
        else {
            architecture += DvConstants.ARCH_IS_BIGENDIAN;
        }
    }

    /**
     *   Return the System Type. i.e. Windows/AIX etc.
     *
     */
    public  int getSystemType() {
        return system;
    }

    /**
     *   Enables determination of the sub system type.
     *   i.e. Windows NT 4, windows XP
     *
     */
    public  int getSubSystemType() {
        return subsystem;
    }

    /**
     *
     *
     */
    public  int getProcessor() {
        return -1;  // unknown
    }

    /**
     *   Static method determining whether the dump source
     *   is compatible with this plugin class.For an SDFF this
     *   means assuming that the incoming parameter is a file
     *   name that can be used to create a FileInputStream.
     *   The contents of this input stream are then checked to
     *   establish whether this is in SDFF format.
     *
     */
    public static Boolean isSupportedSource (String theSource) {

        DvUtils.trace("Entry to SDFFDump:isSupportedSource("+theSource+")",0,false);
        File file=null;
        SdffStd sdffstd;
        Boolean bRet = Boolean.FALSE;

        try {

            file = new File(theSource);
            theRAF=new RandomAccessFile(file,"r"); // no writing to it

            byte[] bytes = new byte[32];
            theRAF.read(bytes,0,32);

            sdffstd = new SdffStd(bytes);

            String y = new String(sdffstd.sectionEye);
            if (y.equals(SdffConstants.SEGMENT_Eye_SdffHead) != true) {
                // DVUtils.writetoTrace("Error in SdffHead for Eyecatcher");
                //DvUtils.output("SDFFHEAD Eyecatcher not nice");
                DvUtils.trace("Exit from SDFFDump:isSupportedSource(false)",0,false);
                return bRet;
            }
            else {
                // read rest of fields from SDFFHEAD
                theRAF.seek(32);
                system = theRAF.readInt();
                architecture =  theRAF.readInt();
                subsystem = theRAF.readInt();
                processor =  theRAF.readInt();
            }
            if (sdffstd.checkMatch() != true) {
                // This is a format error in the SDFF most probably
                // We need to tell someone
                DvUtils.output("Looks like SDFF format Error");
                DvUtils.trace("Exit from SDFFDump:isSupportedSource(false)",3,false);
                return bRet;
            }


        }
        catch (IOException e) {
            DvUtils.output("\nIOException captured in SDFFDump for "+file);
            DvUtils.output("This may be fine if dump is not a file "+file);
            DvUtils.trace("Exit from SDFFDump:isSupportedSource(false)",1,false);
            return new Boolean(false);
        }

        DvUtils.output("Dump recognised as SDFF dump");

        reset(); // so we can support reopening of a new dump


        // Now we do some useful legwork to save time later
        //  We also fill in the 'helper' offsets to SDFFMMEM and SDFFMEM segments
        //  if these are -1 then it means they were not found.

        sizeOfFile = file.length();
        nameOfFile = theSource;

        // Lets go and look at the next segment
        boolean done = false;
        long ongoing_offset = sdffstd.sectionLen;
        while (done == false) {
            try {


                theRAF.seek(ongoing_offset); // OK so skip over this section
                // there should be a SddfStd section on top of here
                byte[] bytes = new byte[32];
                // Its important to remember that you can only read
                // a java ints worth at a time (2G or 4G?).
                int n = theRAF.read(bytes,0,32);
                // System.out.println("\n" + n+ " bytes read from "+ongoing_offset);

                sdffstd = new SdffStd(bytes);
                // so check out this one to see what type it is
                if (sdffstd.checkMatch() != true) {
                    // This is a format error in the SDFF
                    // We need to tell someone
                    // can occur when length is greater than it should be....!!!!
                    DvUtils.output(" checkMatch came back false");
                    done = true;
                }
                else {
                    // ok we have a "good" segment
                    if (sdffstd.sectionType == SdffConstants.SEGMENT_SdffHead) {
                        // Two SdffHeads - something wrong here splat 'em
                        DvUtils.output("Two Sdffheads - not nice");
                        done = true;
                    }
                    else {
                        if (sdffstd.sectionType == SdffConstants.SEGMENT_Sdffprth) {
                            if (prthOffset == -1) {

                                prthOffset = ongoing_offset;
                                prthLength = sdffstd.sectionLen;

                            }
                            else {
                            // two prth sections - scuppered!
                                DvUtils.output("Two Sdffprths - not nice");
                                done = true;
                            }


                        }
                        else {
                            if (sdffstd.sectionType ==
                                SdffConstants.SEGMENT_Sdffmmap) {
                                if (mmapOffset == -1) {
                                    mmapOffset = ongoing_offset;
                                    mmapLength = sdffstd.sectionLen;
                                }
                                else {
                                    // two mmap sections - scuppered!
                                    DvUtils.output("Two Sdffmmaps - not nice");
                                    done = true;
                                }
                            }
                            else {
                                if (sdffstd.sectionType ==
                                        SdffConstants.SEGMENT_Sdffmem) {
                                    if (memOffset == -1) {
                                        memOffset = ongoing_offset;
                                        memLength = sdffstd.sectionLen;
                                    }
                                    else {
                                        // two mem sections - scuppered!
                                        DvUtils.output("Two Sdffmems - not nice");
                                        done = true;
                                    }
                                }
                            }
                        }
                    }

                    ongoing_offset = ongoing_offset+sdffstd.sectionLen;
                }

            }
            catch (IOException e) {
                done = true;
            }

        } // end of while loop

        // Now we should have offsets and lengths for the major sections
        // lets print em out if they are screwy (debug help!!!)
        if (-1 == prthOffset) {
            DvUtils.trace("\n*** Error in dump format\nPrth offset/length is "+
                          prthOffset+ ":" + prthLength,
                          2,true);
            //System.err.println("\n*** Error in dump format\nPrth offset/length is "+
            //              prthOffset+ ":" + prthLength );
        }
        if (-1 == mmapOffset) {
            DvUtils.trace("\n*** Error in dump format\nMmap offset/length is "+
                          mmapOffset+ ":" + mmapLength,
                          2,true);
            //System.err.println("\n*** Error in dump format\nMmap offset/length is "+
            //               mmapOffset+ ":" + mmapLength );
        }
        if (-1 == memOffset) {
            DvUtils.trace("\n*** Error in dump format\nMem  offset/length is "+
                          memOffset+ ":" + memLength ,
                          2,true);
            //System.err.println("\n*** Error in dump format\nMem  offset/length is "+
            //           memOffset+ ":" + memLength );
        }

        bRet = Boolean.TRUE;
        DvUtils.trace("Exit from SDFFDump:isSupportedSource(true)",2,false);
        return bRet;
    }

    /**
     *   The processor subtype - index into DvDump.
     *
     */
    public  int getProcessorSubType() {
        return processorSubtype;
    }

    /**
     *   The processor that produced the dump - normally the same as
     *   the one for the failing thread
     *
     */
    public int getCurrentProcessor() {
        return currentProcessor;
    }

    /**
     *   Gives the number of processors for the machine producing the dump
     *
     */
    public int getProcessorCount(){
        return numProcessors;
    }

    /**
     *
     *
     *
     */
    public DvAddressSpace getCurrentAddressSpace() {
        int i = DvConsole.getCurrentAsIndex();

        DvAddressSpace x = (DvAddressSpace) theASArray[i];

        return x;


    }

    /**
     *   Description: reset saved data when new dump opened
     *
     *
     */
    public static void reset() {
        prthOffset = -1;  // Offset of SDFFPRTH segment in the file
        prthLength = -1;  // Length of SDFFPRTH segment in the file
        mmapOffset = -1;  // Offset of SDFFMMEM segment in the file
        mmapLength = -1;  // Length of SDFFMMEM segment in the file
        memOffset  = -1;  // Offset of SDFFMEM  segment in the file
        memLength  = -1;  // Length of SDFFMEM segment in the file
        sizeOfFile = -1;
    }

    /**
     * Serialise to an SDFF file
     */
    void printSdff(Sdff sdff) throws IOException {
        long start = sdff.getFilePointer();
        SdffStd Std = new SdffStd(SdffConstants.SEGMENT_SdffHead,
        SdffConstants.VERSION1_0_0);
        Std.printSdff(sdff);
        sdff.writeInt(system);
        sdff.writeInt(architecture);
        sdff.writeInt(subsystem);
        sdff.writeInt(processor);
        sdff.writeInt(processorSubtype);
        sdff.writeInt(numProcessors);
        sdff.writeInt(currentProcessor);
        sdff.writeLong(-1l);
        sdff.printLength(start);
    }

}
