/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/SDFFEnvData.java, ras, dumpviewer 1.7
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524   220502 philr   New for feature 36524
 * 41452   230502 nmasson add printSDFF
 * 36524   100602 philr   Tidy up
 *
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.util.*;
import java.io.*;

 /**
  *  Description: Whats the purpose of this class
  *
  *
  *
  *
  *
  */
public class SDFFEnvData extends DvEnvData {


    private int    java_version_size;
    private byte[] java_version;
    private int    jhd_size;
    private byte[] java_home_dir;
    private int    jdd_size;
    private byte[] java_dll_dir;
    private int    env_var_size;
    private byte[] env_var;
    private int    classpath_size;
    private byte[] classpath;
    private byte[] metadata;
    private long   metadata_offset;



    /**
     *    Get value of java home directory
     *
     */
    public  String javaHomeDir() {
        String retString = null;
        if (null != java_home_dir) {
            retString = new String(java_home_dir);
        }
        return retString;
    }

    /**
     *    Get value of java dll directory
     *
     */
    public  String javaDllDir( ) {
        String retString = null;
        if (null != java_dll_dir) {
            retString = new String(java_dll_dir);
        }
        return retString;
    }

    /**
     *    Get value of java version
     *
     */
    public  String javaVersion() {
        String retString = null;
        if (null != java_version) {
            retString = new String(java_version);
        }
        return retString;
    }

    /**
     *    Get environment variables
     *
     */
    public  Vector envVars() {
        // Here we break up the bytes into seperate strings
        Vector v = new Vector();

        int count = 0;
        int ongoing=0;
        byte[] ba=null;
        for (int i=0;i<env_var.length;i++) {

            if (env_var[i] != 0x00) {
                count++;
            }
            else {
                if (count != 0) {
                   ba = new byte[count+1];
                   for (int j=0;j<count+1;j++) {
                        ba[j] = env_var[j+ongoing];
                   }
                   String s = new String(ba);
                   DvUtils.trace("\tAdding environment variable:"+s,0,false);
                   v.add(s);
                   ba = null;

                }
                count = 0;
                ongoing = i;
            }
        }

        return v;

    }

    /**
     *    Get java classpath
     *
     */
    public  String classpath() {
        String retString = null;
        if (null != classpath) {
            retString = new String(classpath);
        }
        return retString;
    }

    /**
     *    metadata retrieval
     *
     */
    public byte[] getMetadata() {

        return metadata;

    }

    /**
     *   <p>Description: Load an EnvData object from an SDFFFile
     *   </p>
     */
    public static  SDFFEnvData loadFromFile(long o,RandomAccessFile raf) {

        DvUtils.trace("Entry to SDFFEnvData:loadFromFile...",0,false);
        long offset_in_file = o; // This is offset of envdata subsection
        DvUtils.trace("  .... Loading SDFFEnvData object from file :"+o,
                        0,false);
        SDFFEnvData envData = new SDFFEnvData();

        try {
            raf.seek(offset_in_file);
            long length = raf.readLong();
            DvUtils.trace("\tEnvData length is:"+length,
                        0,false);
            // Hack to cover missing Envdata !!!!!
            if (length < 48) {
            }
            else {
                raf.skipBytes(8); // skip the eyecatcher (assumed caller has already verified)

                envData.java_version_size=32; // replace this with a readInt soon!!!
                envData.java_version = new byte[envData.java_version_size];
                raf.read(envData.java_version);
                envData.jhd_size = raf.readInt();
                envData.java_home_dir = new byte[envData.jhd_size];
                raf.read(envData.java_home_dir);
                envData.jdd_size = raf.readInt();
                envData.java_dll_dir = new byte[envData.jdd_size];
                raf.read(envData.java_dll_dir);
                envData.env_var_size = raf.readInt();
                DvUtils.trace("\t Environment variable length is:"+envData.env_var_size,
                            0,false);
                // Temporray Hack to cope with alternative method of loading envData
                if (0 == envData.env_var_size) {
                    DvUtils.trace("\tAlternative loading of EnvData",
                        0,false);
                    envData.env_var = new byte[(int)length - 52];
                    raf.read(envData.env_var);
                }
                else {
                    envData.env_var = new byte[envData.env_var_size];
                    raf.read(envData.env_var);
                }
                envData.classpath_size = raf.readInt();
                envData.classpath = new byte[envData.classpath_size];
                raf.read(envData.classpath);
                // System.out.println("emdo:"+offset_in_file+":"+length);
                raf.seek(offset_in_file+length-8);
                envData.metadata_offset = raf.readLong();
                if (-1 != envData.metadata_offset) {
                    // if theres some metadata
                    //System.out.println("emdo:"+envData.metadata_offset+":"+length);
                    envData.metadata = new byte[(int)(envData.metadata_offset)];
                    raf.seek(offset_in_file+length-8-envData.metadata_offset);
                    raf.read(envData.metadata);
                    //String s = new String(envData.metadata);
                    //System.out.println("emdo:"+s);
                }

            }

        }
        catch (IOException e) {
            DvUtils.errorMsg("IOException in SDFFEnvData: loadFromFile",0);
            String s = e.getMessage();
            if (null != s) {
                System.out.println(s);
            }
            e.printStackTrace(System.out);
        }
        catch (OutOfMemoryError e) {
            DvUtils.errorMsg("OutOfMemoryError in SDFFEnvData: loadFromFile",0);
        }

        DvUtils.trace("Exit from SDFFEnvData:loadFromFile",0,false);
        return envData;
    }

    void printSdff(Sdff sdff) throws IOException {
        System.out.println("print Environment TBD");
    }

}
