/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/SDFFProcess.java, ras, dumpviewer 1.17
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502 PHILR:  New for feature 36524
 * 41452  230502 nmasson Add printSDFF
 * 36524  100602 philr   tidy up plus activate envInfo
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;


import java.util.*;
import java.io.*;

 /**
  *  Description: Whats the purpose of this class
  *
  *
  *
  *
  *
  */
public class SDFFProcess extends DvProcess {


    long id;
    long numThreads;
    long currentThread;
    long signalNumber;
    String signalName;
    String commandLine;

    boolean hasJavaThreads; // Set to true if it has a java thread

    Vector threads= new Vector();

    DvAddressSpace aSpace; // address space that owns this process

    long processOffset; // used in search for THREADS
    long processLength; // used in search for THREADS
    SDFFEnvData envInfo;  // envdata for this process
    SDFFLoaded  loadList;
    // private Thread[] threads;   later later
    // private MetaData metaData;   later later

    /**
     *   Description:
     *
     */
    public SDFFProcess(long idIn, long tCount, long cThread,
                       long sig, String sigName,String cLine,
                       long offset,long length,RandomAccessFile raf,
                       DvAddressSpace a) {

        id = idIn;
        numThreads = tCount;
        currentThread = cThread;
        signalNumber = sig;
        signalName = sigName;
        commandLine = cLine;
        processOffset = offset;
        processLength = length;
        aSpace = a;
        hasJavaThreads = false; // asume it has no java threads

        // First we look for EnvData subsection - we cheat by searching for
        // ENVDATA eyecatcher on an 4 byte boundary. we assume there is
        // only one!!!!
        byte[] envEye = SdffConstants.SEGMENT_Eye_SdffEnvData.getBytes();
        long whereToEndSearch= processOffset + processLength;
        long envOffset = processOffset;
        DvUtils.trace("Looking for EnvData eyecatcher...",0,false);
        envOffset = DvUtils.findNextInFile( raf,
                                            envEye,
                                            envOffset,
                                            whereToEndSearch,
                                            1);
        DvUtils.trace("... returned offset is x"+Long.toHexString(envOffset)
                        ,0,false);
        if (-1 != envOffset) {
            envInfo = SDFFEnvData.loadFromFile(envOffset-8,raf);
        }
        else {
            envInfo = null; // shuld we report a format error?
        }

        // Next we look for the Loaded subsection - afgain we cheat by searching
        // for LOADED eyecatcher on an 4 byte boundary - we assume ther is only 1

        byte[] envLoaded = SdffConstants.SEGMENT_Eye_SdffLoaded.getBytes();
        whereToEndSearch= processOffset + processLength;
        long loadedOffset = processOffset;
        DvUtils.trace("Looking for Loaded eyecatcher...",0,false);
        loadedOffset = DvUtils.findNextInFile(  raf,
                                                envLoaded,
                                                loadedOffset,
                                                whereToEndSearch,
                                                1);
        DvUtils.trace("... returned offset is x"+ Long.toHexString(loadedOffset),
                      0,false);
        if (-1 != loadedOffset) {
            loadList = SDFFLoaded.loadFromFile(loadedOffset-8,raf);
        }
        else {
            loadList = null; // shuld we report a format error?
        }


        // Now we go and look for the threads that go with this process
        whereToEndSearch= processOffset + processLength;
        int threadCount=0;
        byte[] eyeCatcher = SdffConstants.SEGMENT_Eye_SdffThread.getBytes();

        // We cheat by searching for THREAD eyecatcher.... on 8 byte boundary
        // this gets us first one and we then find rest
        long x = processOffset;
        //DvUtils.findNextInFile(raf,
        //                       eyeCatcher,
        //                       processOffset,
        //                       whereToEndSearch,
        //                       8);

        //System.out.println("!"+x+" : "+threadCount);
        while (x != -1) {
            DvUtils.trace("  Looking for Thread eyecatcher...",0,false);
            x= DvUtils.findNextInFile(raf,
                               eyeCatcher,
                               x,
                               whereToEndSearch,
                               1);
            DvUtils.trace("  ... returned offset is x"+ Long.toHexString(x),
                            0,false);
            if (x != -1) {
                threadCount++;
                // Now create a new SDFFThread object
                // loaded from file.......
                SDFFThread t = SDFFThread.loadFromFile(x-8,raf);
                threads.add(t);
                if (false == hasJavaThreads && true == t.isJava()) {
                    hasJavaThreads = true;
                }
                try {
                    raf.seek(x-8);
                    // each thread object has its length at the start - we use
                    // this to reset the search area
                    x += raf.readLong();
                }
                catch (java.io.IOException ioE) {
                    System.out.println("IOException in SDFFProcess");
                }


            } // end of if


        } // end of while
     //System.out.println(" Thread count is " +threadCount);


    }

    /**
     *   Description:
     *
     */
    public DvAddressSpace getAddressSpace() {return aSpace;}

    /**
     *   Description:
     *
     */
    public String id() {
        return Long.toString(id);
    }

    /**
     *   Description:
     *
     */
    public DvThread[] getThreads() {

        DvThread[] threadArray = new DvThread[threads.size()];

        for (int i=0;i<threads.size();i++) {
            threadArray[i]=(DvThread) threads.get(i);
        }



        return threadArray;

    }

    /**
     *   Description:
     *
     */
    public boolean isJava() {
        return hasJavaThreads;
    }

    /**
     *   Description:
     *
     */
    public byte[] getMetaData() {
        return null;
    }

    /**
     *   Description: Returns the signal name associated with the failure
     *
     */
    public String getSignal() {
        return signalName;
    }

    /**
     *   Description: Returns the current thread for this process. This is usually
     *                   the most likely candidate for where the problem lies.
     *
     *   Returned   : Thread object, null if current thread is unknown
     *
     */
    public DvThread getCurrentThread() {


        if (-1 == currentThread) return null;

        DvThread retThread = null;

        for (int i=0;i<threads.size();i++) {
            retThread = (com.ibm.jvm.dump.DvThread) threads.get(i);

            if ((retThread.id()).equals(Long.toString(currentThread))) {
                return retThread;
            }
            else {
                retThread = null;
            }
        }  // end of for

        return retThread;
    }

    /**
     *   Description: Returns the current thread for this process. This is usually
     *                   the most likely candidate for where the problem lies.
     *
     *   Returned   : Thread object, null if current thread is unknown
     *
     */
    public DvEnvData getEnvData() { return envInfo;}

    /**
     *   Description: Returns the current thread for this process. This is usually
     *                   the most likely candidate for where the problem lies.
     *
     *   Returned   : Thread object, null if current thread is unknown
     *
     */
    public  DvLoaded getLoaded() { return loadList;}

    /**
     * Serialises to an SDFF file
     */
    void printSdff(Sdff sdff, MemoryMap map) throws IOException {
        System.out.println("Printing Process");
        long start = sdff.getFilePointer();
        sdff.writeLong(0l);
        sdff.writeBytes("PROCESS ");
        sdff.writeLong(threads.size());
        sdff.writeLong(-1l);
        sdff.writeLong(signalNumber);
        sdff.writeBytes(signalName);
        long threadOffset = sdff.getFilePointer();
        envInfo.printSdff(sdff);
        map.printLoaded(sdff);
        sdff.printLength(threadOffset);
        Enumeration list = threads.elements();
        while(list.hasMoreElements()) {
            SDFFThread jvmThread = (SDFFThread)list.nextElement();
            jvmThread.printSdff(sdff);
        }
        // Write Metadata
        sdff.writeLong(-1l);
        sdff.printLength(start);

    }


}


