/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/SDFFThread.java, ras, dumpviewer 1.10
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502 PHILR:  New for feature 36524
 * 41452  220502 nmasson  Add printSdff
 * 36524  240702 philr:  Add in loading for Register subsections
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */
package com.ibm.jvm.dump;

import java.io.*;
import java.util.*;



 /**
  *  Description: Encapsulation of the thread information contained in an
  *               SDFF format dump
  *
  *
  *
  *
  *
  */
class SDFFThread extends DvThread {

    private static long   length;   // length of thread segment in SDFF file

    //String id;  // string representation of thread id
    private long   threadID; // thread id as a java long
    private int    threadType;
    private int    threadState;
    private long   stackptr;
    private long   execEnv; // address in memory of ExecEnv
    private long   pc;      // program counter
    private byte[] metadata;
    private long   metadata_offset;
    private DvProcess owner;
    private SDFFJStackTr stackTrace;
    private Vector registerSections=new Vector(); // there can be multiple SDFFRegister sections
    private Vector jStackSections=new Vector(); // there can be multiple SDFFJStackTR sections
    //MetaData ?????


    /**
     *  Constructor
     *
     */
    SDFFThread() {
        stackTrace = new SDFFJStackTr();
        //registerInfo = new SDFFRegister();
    }

    /**
     *  Constructor
     *
     */
    SDFFThread(long thread, long tid, long execEnv) {
        super();
        //address = thread;
        this.execEnv = execEnv;
    }



    /**
     *   Find the Process object that owns us.
     *
     */
    public DvProcess getProcess() {return owner;}

    /**
     *   The id of this thread (as a string)
     *
     */
    public String id() {return Long.toString(threadID);}


    /**
     * Is this thread a java thread - has it got an execenv!
     */
    public  boolean isJava() {
        if (-1 != execEnv) return true;
        else return false;
    }

    /**
     *  Retrieve address of execenv
     */
    public   DvAddress eeAddress() {

        SDFFAddress retAddr = new SDFFAddress();
        retAddr.setAddress(execEnv);
        return retAddr;
    }

    /**
     *
     */
    public DvNativeFrame[] getNativeFrames() {return null;}

    /**
     */
    public DvJavaFrame[] getJavaFrames() {return null;}

    /**
     */
    public  DvAddress getPSW() {return null;}


    /**
     *      Returns an array of DvRegister objects (sets of registers)
     *      for the thread
     */
    public  DvRegister[] getRegisters() {

        DvRegister[] theArray = null;
        int count = registerSections.size();
        if (0 == count) {
            //return null;
        }
        else {
            theArray = new DvRegister[count];
            for (int i=0;i<count;i++) {
                theArray[i] = (DvRegister) registerSections.get(i);
            }
        }

        return theArray;
    }


    /**
     */
    public byte[] getMetadata() {

        return metadata;
    }


    /**
     *    Load an SDFFThread object from an SDFFFile
     *
     */
    public static SDFFThread loadFromFile(long o,RandomAccessFile raf) {

        long offset_in_file = o; // This is offset of thread subsection
        DvUtils.trace(" Loading SDFFThread object from file offset "+o,0,false);
        SDFFThread retThread = new SDFFThread();

        try {
            raf.seek(offset_in_file);
            retThread.length=raf.readLong();
           // first we deal with any metadata
            raf.seek(offset_in_file+length-8);
            retThread.metadata_offset = raf.readLong();
            if (-1 != retThread.metadata_offset) {
                    // if theres some metadata
                    //System.out.println("emdo:"+Thread.metadata_offset+":"+length);
                    retThread.metadata = new byte[(int)(retThread.metadata_offset)];
                    raf.seek(offset_in_file+length-8-retThread.metadata_offset);
                    raf.read(retThread.metadata);
                    //String s = new String(retThread.metadata);
                    //System.out.println("Thread metadata:"+s);
            }
            raf.seek(offset_in_file+8);  // back to start point
            raf.skipBytes(8); // skip the eyecatcher (assumed caller has already verified)
            retThread.threadID=raf.readLong();
            //System.out.println("retThread.threadID:" + retThread.threadID);
            retThread.threadType = raf.readInt();
            retThread.threadState = raf.readInt();
            retThread.stackptr = raf.readLong();
            retThread.pc = raf.readLong();
            retThread.execEnv = raf.readLong();


        }
        catch (IOException e) {
            DvUtils.errorMsg("IOException in SDFFThread: loadFromFile",0);
        }

        // Now we look for REGISTER sections to load
        DvUtils.trace(" Looking for REGISTER sections "+o,0,false);

        // we cheat here and scan memory for REGISTER eyecatcher between
        // offset_in_file and end ( offset_in_file+retThread.length)

        long x = offset_in_file;
        long whereToEndSearch = offset_in_file+retThread.length;
        int regSectionsCount=0;
        byte[] eyeCatcher = SdffConstants.SEGMENT_Eye_SdffRegister.getBytes();

        while (x != -1) {
            DvUtils.trace("  Looking for Register eyecatcher...",0,false);
            x= DvUtils.findNextInFile(raf,
                               eyeCatcher,
                               x,
                               whereToEndSearch,
                               1);
            DvUtils.trace("  ... returned offset is x"+ Long.toHexString(x),
                            0,false);
            if (x != -1 && x < whereToEndSearch) {

                // Now create a new SDFFThread object
                // loaded from file.......
                SDFFRegister r = SDFFRegister.loadFromFile(x-8,raf);
                retThread.registerSections.add(r);


                try {
                    raf.seek(x-8);
                    // each thread object has its length at the start - we use
                    // this to reset the search area
                    x += raf.readLong();
                }
                catch (java.io.IOException ioE) {
                    System.out.println("IOException in SDFFThread");
                }


            } // end of if


        } // end of while

        // Now we look forJSTACKTR recordss to load
        DvUtils.trace(" Looking for JSTACKTR subsections "+o,0,false);

        // we cheat here and scan memory for JSTACKTR eyecatcher between
        // offset_in_file and end ( offset_in_file+retThread.length)

        x = offset_in_file;
        whereToEndSearch = offset_in_file+retThread.length;
        int jStackSectionsCount=0;
        eyeCatcher = SdffConstants.SEGMENT_Eye_SdffJstackTr.getBytes();

        while (x != -1) {
            DvUtils.trace("  Looking for JStackTR eyecatcher...",0,false);
            x= DvUtils.findNextInFile(raf,
                               eyeCatcher,
                               x,
                               whereToEndSearch,
                               1);
            DvUtils.trace("  ... returned offset is x"+ Long.toHexString(x),
                            0,false);
            if (x != -1 && x < whereToEndSearch) {

                // Now create a new SDFFThread object
                // loaded from file.......
                SDFFJStackTr jStack = SDFFJStackTr.loadFromFile(x-8,raf);
                retThread.jStackSections.add(jStack);


                try {
                    raf.seek(x-8);
                    // each thread object has its length at the start - we use
                    // this to reset the search area
                    x += raf.readLong();
                }
                catch (java.io.IOException ioE) {
                    System.out.println("IOException in SDFFThread");
                }


            } // end of if


        } // end of while


        return retThread;
    }

    /**
     * Serialises to SDFF file
     */
    void printSdff(Sdff sdff) throws IOException {
    System.out.println("Printing Thread");
        long start = sdff.getFilePointer();
        sdff.writeLong(0l);
        sdff.writeBytes("THREAD  ");
        sdff.writeLong(threadID);
        sdff.writeInt(-1); //IBM.FIXME
        sdff.writeInt(-1); //IBM.FIXME
        sdff.writeLong(stackptr);
        sdff.writeLong(pc);
        sdff.writeLong(execEnv);
        stackTrace.printSdff(sdff);
        for (int i=0;i<registerSections.size();i++) {
            SDFFRegister registerInfo = (SDFFRegister) registerSections.get(i);
             registerInfo.printSdff(sdff);
        }
        //registerInfo.printSdff(sdff);
    // Print Metadata
        sdff.writeLong(-1l);
        sdff.printLength(start);
    }
}

