/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/SdffStd.java, ras, dumpviewer 1.8
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524  020502 PHILR:  New for feature 36524
 * 41452  230502 nmasson Add printSdff()
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

package com.ibm.jvm.dump;


import java.lang.String;
import java.io.*;


 /**
  *  Description: The SdffStd class supports the manipulation of the header
  *               at the start of each major section of an SDFF format file.
  *
  */
public class SdffStd  {
   long sectionLen;
   long sectionType;
   byte sectionEye[]= new byte[8];
   long sectionVer;

 /**
  *   <p>Description: This returns the length in bytes of the data
  *                   making up the header.
  *   </p>
  */
   public static long getSize() {

      return 32; // this needs to be kept in line with data structure
                 // in an SdffStd !!!!
   }

 /**
  *   <p>Description: Instantiate an SdffStd object given type and version
  *   </p>
  */
   public SdffStd(long type,long ver) {
        String[] SEGMENT_EYECATCHER =
        {
            SdffConstants.SEGMENT_Eye_SdffHead,
            SdffConstants.SEGMENT_Eye_SdffPrth,
            SdffConstants.SEGMENT_Eye_SdffMmem,
            SdffConstants.SEGMENT_Eye_SdffMem

        };
       sectionType = type;
       sectionEye = SEGMENT_EYECATCHER[(int)type].getBytes();

       //sectionEye = eye;
       sectionVer = ver;
   }

 /**
  *   <p>Description: Instantiate an SdffStd object given a byte array
  *                   normally from a file.
  *   </p>
  */
   public SdffStd (byte[] bytes) throws IOException  {
    // This mechanism of creating an SdffStd takes an array off bytes and
    // creates it.... you need enough bytes otherwise bang!

    try {
      ByteArrayInputStream ba = new ByteArrayInputStream(bytes);
      DataInputStream da = new DataInputStream(ba);
      sectionLen =  da.readLong();
      sectionType = da.readLong();
      da.readFully(sectionEye);
      sectionVer = da.readLong();
     }
     catch (IOException e)
     {
       System.out.println("\nIn IOException for SdffStd");
       throw e;
     }
   }

 /**
  *   <p>Description: Makes sure that the type and eyecatcher match
  *   </p>
  */
   public boolean checkMatch() {



       boolean bresult = true; // assume good

       String y = new String(sectionEye);

       if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffHead)) {
         if ((int)sectionType != SdffConstants.SEGMENT_SdffHead) {
              bresult = false;
         }
       }
       else if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffPrth)) {
         if ((int)sectionType != SdffConstants.SEGMENT_Sdffprth) {
              bresult = false;
         }
       }
       else if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffMmem)) {
         if ((int)sectionType != SdffConstants.SEGMENT_Sdffmmap) {
              bresult = false;
         }
       }
       else if (true == y.equals(SdffConstants.SEGMENT_Eye_SdffMem)) {
         if ((int)sectionType != SdffConstants.SEGMENT_Sdffmem) {
              bresult = false;
         }
       }
       else {
           bresult = false;
           //System.out.println("\ny is !"+y+"!");

       }

       return bresult;

   }

    /**
     * serialises itself to an SDFF file
     */
    void printSdff(Sdff sdff) throws IOException{
        DvUtils.trace("Writing SdffStd", 0, false);
        long start = sdff.getFilePointer();
        sdff.writeLong(0);
        sdff.writeLong(sectionType);
        sdff.write(sectionEye);
        sdff.writeLong(sectionVer);
        sdff.printLength(start);
    }
   

}

