/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/WinExtract.c, ras, dumpviewer 1.1
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 41451  100702  philv   New for feature
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

/*
Extracts data from a minidump.  The dump mnust be an
IBM Windows extractor minidump.  The data are pulled
out and stuffed into a format suitable for the cross
platform dump formatter
*/

#include <windows.h>
#include <stdio.h>
#include <dbghelp.h>       // minidump API
#include <winnt.h>         // for sys info etc

#include "Win_Extract.h"       // us

#include "WinExtract.h"    // JNI

typedef int intptr_t;   //needed for hpiras.h

#include <hpiras.h>

/*
// fix Rajini's stuff which is not backwards compat
typedef struct {
    union {
        double hpiRasAlign;
        struct {
            char eyecatcher[8];
            int bitpattern1;
            int bitpattern2;
//            int length;
//            int version;
            intptr_t threadQueuePtr;
            intptr_t nextOffset;
            intptr_t eeOffset;
        } data;
    } ras;
} hpi_ras_t;
*/

int trace = 0;

HANDLE map;
PVOID  pmap;
HANDLE hfile;            // minidump file handle

#define IA32 1

#ifdef IA32
   #define PFORMAT "08l"
   #define PVALUE PULONG32
#else
   #define PFORMAT "016I64"
   #define PVALUE PULONG64
#endif


static void closefile         (void);
static void analyseModuleData (PMINIDUMP_MODULE_LIST pml);

static PSZ extractComment     (void);
static PSZ readMiniDumpString (RVA rvaString);

static BOOL openfile                (void);
static BOOL readThreadList          (PMINIDUMP_MEMORY64_LIST pMem);
static BOOL readThreadListEx        (void);
static BOOL readExceptionStream     (void);
static BOOL readHandleDataStream    (void);
static BOOL readFunctionTableStream (void);

static RVA findRVA (PULONG64 start, ULONG32 length, PMINIDUMP_MEMORY64_LIST pMem);

static PMINIDUMP_MODULE_LIST extractModule (void);

static PMINIDUMP_SYSTEM_INFO readSystemInfo (void);

static PMINIDUMP_MEMORY64_LIST readDumpMemory (void);

// helpers
PVOID Rva2Address (RVA64 rva);
RVA64 Address2Rva (PVOID address);


/*-------------------------------Globals------------------------------*/
   PMINIDUMP_SYSTEM_INFO        pSysInfo          = NULL;
   PMINIDUMP_MEMORY64_LIST      pMem              = NULL;
   PMINIDUMP_THREAD_EX_LIST     pThreadExList     = NULL;
   PMINIDUMP_THREAD_LIST        pThreadList       = NULL;
   PMINIDUMP_MODULE_LIST        pml               = NULL;
   PMINIDUMP_EXCEPTION_STREAM   pExceptionStream  = NULL;
   PMINIDUMP_HANDLE_DATA_STREAM pHandleDataStream = NULL;

   hpi_ras_t * phpiras = NULL;
/*---------------------------------------------------------------------*/


BOOL APIENTRY DllMain ( HINSTANCE hInst
                      , DWORD     reason_for_call
                      , LPVOID    lpReserved
                      )
{
   hInst=hInst;lpReserved=lpReserved;

   switch (reason_for_call)
   {
   case DLL_PROCESS_ATTACH :
      //printf("DLL process attach\n");
      //extract (0, NULL, NULL);
      break ;
   case DLL_THREAD_ATTACH  :
      printf("DLL thread attach\n");
      break ;
   case DLL_THREAD_DETACH  :
      printf("DLL thread detach\n");
      break ;
   case DLL_PROCESS_DETACH :
      //printf("DLL process detach\n");
      closefile ();
      break ;
   }
   return TRUE;
}


/* As the DLL loads this function is called
** It extracts all the goodies from the minidump
** and leaves a set of PMINIDUMP pointers referencing
** the file memory map.  Nice and quick !
*/

JNIEXPORT jint JNICALL Java_WinExtract_doExtract (JNIEnv * e, jobject o, jint level)
{
   e=e;o=o;
   return extract (level);
}

/**
  Initialise all the streams...
*/
int extract (int level)
{

   PMINIDUMP_MODULE        pModule = NULL;
   HANDLE                  p;                // process handle
   DWORD                   pid;              // process id
   PSZ                     pComment;

   trace = level;

   /*
   typedef DWORD RVA;
   typedef struct _MINIDUMP_LOCATION_DESCRIPTOR
   {
       ULONG32 DataSize;
       RVA Rva;
   } MINIDUMP_LOCATION_DESCRIPTOR;
   typedef struct _MINIDUMP_DIRECTORY
   {
       ULONG32 StreamType;
       MINIDUMP_LOCATION_DESCRIPTOR Location;
   } MINIDUMP_DIRECTORY, *PMINIDUMP_DIRECTORY;
   */

   printf ("Windows dump extractor v0.5 ....\n");

   // open dump file
   if (!openfile ())
   {
      printf("Cannot open the extract file !\n");
      return -1;
   }

   // get process id
   pid = GetCurrentProcessId ();
   if (trace == 3)
   {
      printf ("WinEx: The current pid is %d(0x%lx)\n", pid, pid);
   }
   // get process handle
   p = GetCurrentProcess ();
   if (trace == 3)
   {
      printf ("WinEx: The current process handle is 0x%x\n", p);
   }

   /* Read the streams
   ** available types are :-
   ** typedef enum _MINIDUMP_STREAM_TYPE
   ** {
   **
   **     UnusedStream                = 0,
   **     ReservedStream0             = 1,
   **     ReservedStream1             = 2,
   **     ThreadListStream            = 3,  //pv
   **     ModuleListStream            = 4,  //pv
   **     MemoryListStream            = 5,  //NotRequired
   **     ExceptionStream             = 6,  //knp
   **     SystemInfoStream            = 7,  //pv
   **     ThreadExListStream          = 8,  //pv
   **     Memory64ListStream          = 9,  //pv
   **     CommentStreamA              = 10, //pv
   **     CommentStreamW              = 11, //NotRequired
   **     HandleDataStream            = 12, //knp
   **     FunctionTableStream         = 13, //knp TBT
   **     UnloadedModuleListStream    = 14,
   **     MiscInfoStream              = 15,
   **
   **     LastReservedStream          = 0xffff
   **
   ** } MINIDUMP_STREAM_TYPE;
   */

   // Get comment, verify this is an IBM dump
   pComment = extractComment ();
   if (pComment == NULL || strcmp (pComment, IBM_EXTRACT_ID) != 0)
   {
      printf ("This does not appear to be an IBM minidump extraction !\n");
      return -1;
   }


   // Get system info
   pSysInfo = readSystemInfo ();


   // get list of modules
   pml = extractModule ();
   if (pml == NULL)
   {
      // something is awry !
      printf ("Cannot read module list from dump !\n");
      return -2;
   }

   analyseModuleData (pml);

   // get the memory
   pMem = readDumpMemory ();

   // get thread info
   readThreadList (pMem);
   readThreadListEx ();

   // get Exception info
   readExceptionStream ();

   // get Handle info
   readHandleDataStream ();

   // get FunctionTable info
   readFunctionTableStream ();

   printf (".. done\n");
   return 0;
}

/*                          EXTRACTOR ROUTINES
**                          ------------------
**
** These routines extract data from the mindump and make it
** available via a set of PMINIDUMP_<> pointers
*/
/*
typedef struct _MINIDUMP_SYSTEM_INFO
{
      USHORT ProcessorArchitecture;
      USHORT ProcessorLevel;
      USHORT ProcessorRevision;
      USHORT Reserved0;
      ULONG32 MajorVersion;
      ULONG32 MinorVersion;
      ULONG32 BuildNumber;
      ULONG32 PlatformId;
      RVA CSDVersionRva;
      ULONG32 Reserved1;
      union _CPU_INFORMATION
      {
          struct
          {
              // CPUID Subfunction 0, EAX (VendorId[0]),
              // EBX (VendorId[1]) and ECX (VendorId[2]).

              ULONG32 VendorId[3];

              // CPUID Subfunction 1, register EAX

              ULONG32 VersionInformation;

              // CPUID Subfunction 1, register EDX

              ULONG32 FeatureInformation;

              // CPUID, Subfunction 80000001, register EBX.
              // Supported only if VendorId is "AuthenticAMD".

              ULONG32 AMDExtendedCpuFeatures;
          } X86CpuInfo;

          struct
          {
              ULONG64 ProcessorFeatures[2];
          } OtherCpuInfo;

      } Cpu;
  } MINIDUMP_SYSTEM_INFO, *PMINIDUMP_SYSTEM_INFO;
*/
static PMINIDUMP_SYSTEM_INFO readSystemInfo (void)
{
   BOOL                  brc        = FALSE;
   ULONG                 streamSize = 0;
   PMINIDUMP_DIRECTORY   pmdDir     = NULL;
   PMINIDUMP_SYSTEM_INFO pSysInfo   = NULL;

   brc = MiniDumpReadDumpStream ( pmap
                                , SystemInfoStream
                                , &pmdDir        // directory structure
                                , &pSysInfo
                                , &streamSize
                                );

   if (trace > 0)
       printf("Minidump read SystemInfoStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read SystemInfoStream Failed:: GetLastError: %u\n", GetLastError());
      return NULL;
   }

   if (pSysInfo == NULL) return NULL;

   if (trace > 2)
   {
      printf ("System architecture = ");

      switch (pSysInfo->ProcessorArchitecture)
      {
      case PROCESSOR_ARCHITECTURE_INTEL :
         printf("Intel\n");
         break;

      case PROCESSOR_ARCHITECTURE_IA64 :
      // documented on ms website but not in header file..
      //case PROCESSOR_ARCHITECTURE_IA32_ON_WIN64 :
      //case PROCESSOR_ARCHITECTURE_AMD64 :
         printf("IA64\n");
         break;

      default :
         printf("unknown (%d)\n", pSysInfo->ProcessorArchitecture);
         break;
      }
   }
   return pSysInfo;
}

/*
  typedef struct _MINIDUMP_THREAD_EX_LIST {
    ULONG32 NumberOfThreads;
    MINIDUMP_THREAD_EX Threads [0];
} MINIDUMP_THREAD_EX_LIST, *PMINIDUMP_THREAD_EX_LIST;

typedef struct _MINIDUMP_THREAD_EX {
    ULONG32 ThreadId;
    ULONG32 SuspendCount;
    ULONG32 PriorityClass;
    ULONG32 Priority;
    ULONG64 Teb;
    MINIDUMP_MEMORY_DESCRIPTOR Stack;
    MINIDUMP_LOCATION_DESCRIPTOR ThreadContext;
    MINIDUMP_MEMORY_DESCRIPTOR BackingStore;
} MINIDUMP_THREAD_EX, *PMINIDUMP_THREAD_EX;

  typedef struct _MINIDUMP_MEMORY_DESCRIPTOR {
      ULONG64 StartOfMemoryRange;
      MINIDUMP_LOCATION_DESCRIPTOR Memory;
  } MINIDUMP_MEMORY_DESCRIPTOR, *PMINIDUMP_MEMORY_DESCRIPTOR;

  typedef struct _MINIDUMP_LOCATION_DESCRIPTOR {
    ULONG32 DataSize;
    RVA Rva;
  } MINIDUMP_LOCATION_DESCRIPTOR;


*/
static BOOL readThreadListEx ()
{
   BOOL                     brc         = FALSE;
   ULONG                    streamSize  = 0;
   ULONG32                  count;
   PMINIDUMP_DIRECTORY      pmdDir      = NULL;

   brc = MiniDumpReadDumpStream ( pmap
                                , ThreadExListStream
                                , &pmdDir        // directory structure
                                , &pThreadExList
                                , &streamSize
                                );

   if (trace > 0)
       printf("Minidump read ThreadExListStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read ThreadExListStream Failed:: GetLastError: %u\n", GetLastError());
      return FALSE;
   }


   if (trace > 2)
   {
      printf("Type ptr returned = %p, stream size = %lx\n", pThreadExList, streamSize);
      printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
             , pmdDir->StreamType
             , pmdDir->Location.DataSize
             , pmdDir->Location.Rva
             );


      for (count = 0 ; count < pThreadExList->NumberOfThreads ; count++)
      {
           printf ( "%03d tid = % 4d, TEB = %016I64x\n"
                  , count + 1
                  , pThreadExList->Threads[count].ThreadId
                  , pThreadExList->Threads[count].Teb
                  );
           printf ( "  Thread stack, start %016I64x, size %08lx, RVA %08lx\n"
                  , pThreadExList->Threads[count].Stack.StartOfMemoryRange
                  , pThreadExList->Threads[count].Stack.Memory.DataSize
                  , pThreadExList->Threads[count].Stack.Memory.Rva
                  );
           printf ( "  Thread context, size %08lx, RVA %08lx\n"
                  , pThreadExList->Threads[count].ThreadContext.DataSize
                  , pThreadExList->Threads[count].ThreadContext.Rva
                  );
      }
   }
   return TRUE;
}

/*
 typedef struct _MINIDUMP_THREAD_LIST {
      ULONG32 NumberOfThreads;
      MINIDUMP_THREAD Threads [0];
  } MINIDUMP_THREAD_LIST, *PMINIDUMP_THREAD_LIST;

  typedef struct _MINIDUMP_THREAD {
      ULONG32 ThreadId;
      ULONG32 SuspendCount;
      ULONG32 PriorityClass;
      ULONG32 Priority;
      ULONG64 Teb;
      MINIDUMP_MEMORY_DESCRIPTOR Stack;
      MINIDUMP_LOCATION_DESCRIPTOR ThreadContext;
  } MINIDUMP_THREAD, *PMINIDUMP_THREAD;

  typedef struct _MINIDUMP_MEMORY_DESCRIPTOR {
      ULONG64 StartOfMemoryRange;
      MINIDUMP_LOCATION_DESCRIPTOR Memory;
  } MINIDUMP_MEMORY_DESCRIPTOR, *PMINIDUMP_MEMORY_DESCRIPTOR;

  typedef struct _MINIDUMP_LOCATION_DESCRIPTOR {
    ULONG32 DataSize;
    RVA Rva;
  } MINIDUMP_LOCATION_DESCRIPTOR;
*/
static BOOL readThreadList (PMINIDUMP_MEMORY64_LIST pMem)
{
   BOOL                  brc         = FALSE;
   ULONG                 streamSize  = 0;
   ULONG32               count;
   ULONG64               stackEnd;
   PMINIDUMP_DIRECTORY   pmdDir      = NULL;

   brc = MiniDumpReadDumpStream ( pmap
                                , ThreadListStream
                                , &pmdDir        // directory structure
                                , &pThreadList
                                , &streamSize
                                );


   if (trace > 0)
      printf("Minidump read ThreadListStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read ThreadListStream Failed:: GetLastError: %u\n", GetLastError());
      return FALSE;
   }


   if (trace > 2)
   {
      printf("Type ptr returned = %p, stream size = %lx\n", pThreadList, streamSize);
      printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
             , pmdDir->StreamType
             , pmdDir->Location.DataSize
             , pmdDir->Location.Rva
             );
   }

   for (count = 0 ; count < pThreadList->NumberOfThreads ; count++)
   {

      /* A full memory dump will NOT extract the thread
      ** stacks all nit and tiddy for us ....
      ** If the Stack.Memory.Rva field is zero then we must
      ** run down the memory dump MINIDUMP_MEMORY_DESCRIPTOR64
      ** structures to find address corresponding to the
      ** Stack.StartOfMemoryRange field.  Compute the offset
      ** of Stack.StartOfMemoryRange (stack end) from
      ** StartOfMemoryRange in the descriptor and compute
      ** the RVA for the stack.  Stack is then this address
      ** to (+ Stack.Memory.DataSize)
      */

        if (pThreadList->Threads[count].Stack.Memory.Rva == 0)
        {
            stackEnd = pThreadList->Threads[count].Stack.StartOfMemoryRange;
            pThreadList->Threads[count].Stack.Memory.Rva = findRVA ( &stackEnd
                                                                   , pThreadList->Threads[count].Stack.Memory.DataSize
                                                                   , pMem
                                                                   );
        }

        if (trace > 2)
        {
            printf ( "%03d tid = % 4d, TEB = %016I64x\n"
                   , count + 1
                   , pThreadList->Threads[count].ThreadId
                   , pThreadList->Threads[count].Teb
                   );
            printf ( "  Thread stack, start %016I64x, size %08lx, RVA %08lx\n"
                   , pThreadList->Threads[count].Stack.StartOfMemoryRange
                   , pThreadList->Threads[count].Stack.Memory.DataSize
                   , pThreadList->Threads[count].Stack.Memory.Rva
                   );
            printf ( "  Thread context, size %08lx, RVA %08lx\n"
                   , pThreadList->Threads[count].ThreadContext.DataSize
                   , pThreadList->Threads[count].ThreadContext.Rva
                   );
        }
   }

   return TRUE;
}

/*
typedef struct MINIDUMP_EXCEPTION_STREAM {
    ULONG32 ThreadId;
    ULONG32  __alignment;
    MINIDUMP_EXCEPTION ExceptionRecord;
    MINIDUMP_LOCATION_DESCRIPTOR ThreadContext;
} MINIDUMP_EXCEPTION_STREAM, *PMINIDUMP_EXCEPTION_STREAM;


typedef struct _MINIDUMP_EXCEPTION  {
    ULONG32 ExceptionCode;
    ULONG32 ExceptionFlags;
    ULONG64 ExceptionRecord;
    ULONG64 ExceptionAddress;
    ULONG32 NumberParameters;
    ULONG32 __unusedAlignment;
    ULONG64 ExceptionInformation [EXCEPTION_MAXIMUM_PARAMETERS];
} MINIDUMP_EXCEPTION, *PMINIDUMP_EXCEPTION;

typedef struct _MINIDUMP_LOCATION_DESCRIPTOR {
  ULONG32 DataSize;
  RVA Rva;
} MINIDUMP_LOCATION_DESCRIPTOR;

*/


static BOOL readExceptionStream ()
{
   BOOL                  brc         = FALSE;
   ULONG                 streamSize  = 0;
   PMINIDUMP_DIRECTORY   pmdDir      = NULL;
   ULONG32               count;


   brc = MiniDumpReadDumpStream ( pmap
                                , ExceptionStream
                                , &pmdDir        // directory structure
                                , &pExceptionStream
                                , &streamSize
                                );


   if (trace > 0)
      printf("Minidump read ExceptionStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read ExceptionStream Failed:: GetLastError: %u\n", GetLastError());
      return FALSE;
   }

   if (trace > 2)
   {
      printf("Type ptr returned = %p, stream size = %lx\n", pExceptionStream, streamSize);
      printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
             , pmdDir->StreamType
             , pmdDir->Location.DataSize
             , pmdDir->Location.Rva
             );
   }
   if (!(pExceptionStream->ThreadContext.Rva))   /* Thread context RVA is NULL, we need to find it ourselves!!*/
   {
       if (trace > 0)
           printf ( "Need to findout rva for the thread context ourselves!!");
       for (count = 0 ; count < pThreadList->NumberOfThreads ; count++)
       {
           if ( pThreadList->Threads[count].ThreadId == pExceptionStream->ThreadId )
           {
                pExceptionStream->ThreadContext.Rva =  pThreadList->Threads[count].ThreadContext.Rva;
                break;
           }
       }
   }

   if (trace > 2)
   {
      printf ( "tid = % 4d\n"
             , pExceptionStream->ThreadId
             );

      printf ( "ExceptionRecord: ExceptionCode %08lx, ExceptionFlags %08lx, ExceptionRecord %016I64x,\n  ExceptionAddress %016I64x, NumberParameters %08lx\n"
             , pExceptionStream->ExceptionRecord.ExceptionCode
             , pExceptionStream->ExceptionRecord.ExceptionFlags
             , pExceptionStream->ExceptionRecord.ExceptionRecord
             , pExceptionStream->ExceptionRecord.ExceptionAddress
             , pExceptionStream->ExceptionRecord.NumberParameters
             );

      printf ( "Thread context, size %08lx, RVA %08lx\n"
             , pExceptionStream->ThreadContext.DataSize
             , pExceptionStream->ThreadContext.Rva
             );

      printf ( "Esp:: %08lx   Eip:: %08lx  \n"
             ,((PCONTEXT) ((PBYTE)pmap + pExceptionStream->ThreadContext.Rva))->Esp
             ,((PCONTEXT) ((PBYTE)pmap + pExceptionStream->ThreadContext.Rva))->Eip
             );

      for (count = 0 ; count < pExceptionStream->ExceptionRecord.NumberParameters ; count++)
      {
            printf ( "%03d ExceptionInformation = %016I64x\n"
                , count + 1
                , pExceptionStream->ExceptionRecord.ExceptionInformation[count]
                );
      }

   }


   return TRUE;
}

/*
typedef struct _MINIDUMP_HANDLE_DATA_STREAM {
    ULONG32 SizeOfHeader;
    ULONG32 SizeOfDescriptor;
    ULONG32 NumberOfDescriptors;
    ULONG32 Reserved;
} MINIDUMP_HANDLE_DATA_STREAM, *PMINIDUMP_HANDLE_DATA_STREAM;

typedef struct _MINIDUMP_HANDLE_DESCRIPTOR {
    ULONG64 Handle;
    RVA TypeNameRva;
    RVA ObjectNameRva;
    ULONG32 Attributes;
    ULONG32 GrantedAccess;
    ULONG32 HandleCount;
    ULONG32 PointerCount;
} MINIDUMP_HANDLE_DESCRIPTOR, *PMINIDUMP_HANDLE_DESCRIPTOR;


*/

static BOOL readHandleDataStream ()
{
   BOOL                  brc         = FALSE;
   ULONG                 streamSize  = 0;
   PMINIDUMP_DIRECTORY   pmdDir      = NULL;
   PMINIDUMP_HANDLE_DESCRIPTOR  pHandleDataDescriptor = NULL;
   ULONG32               count;

   brc = MiniDumpReadDumpStream ( pmap
                                , HandleDataStream
                                , &pmdDir        // directory structure
                                , &pHandleDataStream
                                , &streamSize
                                );

   if (trace > 0)
      printf("Minidump read HandleDataStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read HandleDataStream Failed:: GetLastError: %u\n", GetLastError());
      return FALSE;
   }


   if (trace > 2)
   {
      printf("Type ptr returned = %p, stream size = %lx\n", pHandleDataStream, streamSize);
      printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
             , pmdDir->StreamType
             , pmdDir->Location.DataSize
             , pmdDir->Location.Rva
             );
   }

   if (trace > 2)
   {
      printf ( "HandleDataStream Header: SizeOfHeader %08lx, SizeOfDescriptor %08lx, NumberOfDescriptors %08lx\n"
             , pHandleDataStream->SizeOfHeader
             , pHandleDataStream->SizeOfDescriptor
             , pHandleDataStream->NumberOfDescriptors
             );

      pHandleDataDescriptor = (PMINIDUMP_HANDLE_DESCRIPTOR)((PBYTE) pHandleDataStream + pHandleDataStream->SizeOfHeader);
      for (count = 0 ; count < pHandleDataStream->NumberOfDescriptors ; count++)
      {
            printf ( "%03d HandleDescriptor:: Handle = %016I64x  "
                   , count + 1
                   , pHandleDataDescriptor->Handle
                   );
            wprintf(L"Type name = %s \n", readMiniDumpString (pHandleDataDescriptor->TypeNameRva));

            if(pHandleDataDescriptor->ObjectNameRva)
               wprintf(L"Object name = %s \n", readMiniDumpString (pHandleDataDescriptor->ObjectNameRva));
            else
               /* No names for Section, Event, Thread and File !!!*/
               printf("Object name = NULL \n");

            printf ( "Attributes %08lx, GrantedAccess %08lx, HandleCount %08lx, PointerCount %08lx\n\n"
                   , pHandleDataDescriptor->Attributes
                   , pHandleDataDescriptor->GrantedAccess
                   , pHandleDataDescriptor->HandleCount
                   , pHandleDataDescriptor->PointerCount
                   );
            pHandleDataDescriptor = (PMINIDUMP_HANDLE_DESCRIPTOR)((PBYTE) pHandleDataDescriptor + pHandleDataStream->SizeOfDescriptor);

      }

   }

   return TRUE;
}
/*
typedef struct _MINIDUMP_FUNCTION_TABLE_STREAM {
    ULONG32 SizeOfHeader;
    ULONG32 SizeOfDescriptor;
    ULONG32 SizeOfNativeDescriptor;
    ULONG32 SizeOfFunctionEntry;
    ULONG32 NumberOfDescriptors;
    ULONG32 SizeOfAlignPad;
} MINIDUMP_FUNCTION_TABLE_STREAM, *PMINIDUMP_FUNCTION_TABLE_STREAM;

typedef struct _MINIDUMP_FUNCTION_TABLE_DESCRIPTOR {
    ULONG64 MinimumAddress;
    ULONG64 MaximumAddress;
    ULONG64 BaseAddress;
    ULONG32 EntryCount;
    ULONG32 SizeOfAlignPad;
} MINIDUMP_FUNCTION_TABLE_DESCRIPTOR, *PMINIDUMP_FUNCTION_TABLE_DESCRIPTOR;

*/


static BOOL readFunctionTableStream ()
{
   BOOL                  brc         = FALSE;
   ULONG                 streamSize  = 0;
   PMINIDUMP_DIRECTORY   pmdDir      = NULL;
   PMINIDUMP_FUNCTION_TABLE_STREAM pFunctionTableStream = NULL;
   PMINIDUMP_FUNCTION_TABLE_DESCRIPTOR  pFunctionTableDescriptor = NULL;
   ULONG32               count;

   brc = MiniDumpReadDumpStream ( pmap
                                , FunctionTableStream
                                , &pmdDir        // directory structure
                                , &pFunctionTableStream
                                , &streamSize
                                );


   if (trace > 0)
      printf("Minidump read FunctionTableStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read FunctionTableStream Failed:: GetLastError: %u\n", GetLastError());
      return FALSE;
   }

   if (trace > 2)
   {
      printf("Type ptr returned = %p, stream size = %lx\n", pFunctionTableStream, streamSize);
      printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
             , pmdDir->StreamType
             , pmdDir->Location.DataSize
             , pmdDir->Location.Rva
             );
      printf ( "FunctionTableStream Header: SizeOfHeader %08lx, SizeOfDescriptor %08lx, SizeOfNativeDescriptor %08lx, SizeOfFunctionEntry %08lx, NumberOfDescriptors %08lx, SizeOfAlignPad %08lx\n"
             , pFunctionTableStream->SizeOfHeader
             , pFunctionTableStream->SizeOfDescriptor
             , pFunctionTableStream->SizeOfNativeDescriptor
             , pFunctionTableStream->SizeOfFunctionEntry
             , pFunctionTableStream->NumberOfDescriptors
             , pFunctionTableStream->SizeOfAlignPad
             );

      pFunctionTableDescriptor = (PMINIDUMP_FUNCTION_TABLE_DESCRIPTOR) ((PBYTE)pFunctionTableStream
                               + pFunctionTableStream->SizeOfHeader + pFunctionTableStream->SizeOfAlignPad);

      for (count = 0 ; count < pFunctionTableStream->NumberOfDescriptors ; count++)
      {
            printf ( "%03d FunctionTableDescriptor:: \n "
                   , count + 1
                   );
            printf ( "MinimumAddress %016I64x, MaximumAddress %016I64x, BaseAddress %016I64x, EntryCount %08lx, SizeOfAlignPad %08lx\n"
                   , pFunctionTableDescriptor->MinimumAddress
                   , pFunctionTableDescriptor->MaximumAddress
                   , pFunctionTableDescriptor->BaseAddress
                   , pFunctionTableDescriptor->EntryCount
                   , pFunctionTableDescriptor->SizeOfAlignPad
                   );
            pFunctionTableDescriptor += pFunctionTableStream->SizeOfDescriptor
                                      + pFunctionTableStream->SizeOfNativeDescriptor
                                      + (pFunctionTableDescriptor->EntryCount * pFunctionTableStream->SizeOfFunctionEntry)
                                      + pFunctionTableDescriptor->SizeOfAlignPad;

      }

   }


   return TRUE;
}



/*
typedef struct _MINIDUMP_MEMORY_LIST {
      ULONG32 NumberOfMemoryRanges;
      MINIDUMP_MEMORY_DESCRIPTOR MemoryRanges [0];
  } MINIDUMP_MEMORY_LIST, *PMINIDUMP_MEMORY_LIST;

typedef struct _MINIDUMP_MEMORY64_LIST {
      ULONG64 NumberOfMemoryRanges;
      RVA64 BaseRva;
      MINIDUMP_MEMORY_DESCRIPTOR64 MemoryRanges [0];
  } MINIDUMP_MEMORY64_LIST, *PMINIDUMP_MEMORY64_LIST;

  typedef struct _MINIDUMP_MEMORY_DESCRIPTOR64 {
      ULONG64 StartOfMemoryRange;
      ULONG64 DataSize;
  } MINIDUMP_MEMORY_DESCRIPTOR64, *PMINIDUMP_MEMORY_DESCRIPTOR64;
*/
static PMINIDUMP_MEMORY64_LIST readDumpMemory ()
{
   BOOL                    brc        = FALSE;
   ULONG                   streamSize = 0;
   ULONG64                 nRanges;
   ULONG64                 count;
   ULONG64                 currRVA;
   PMINIDUMP_DIRECTORY     pmdDir     = NULL;
   PMINIDUMP_MEMORY64_LIST pmd        = NULL;

   brc = MiniDumpReadDumpStream ( pmap
                                , Memory64ListStream
                                , &pmdDir        // directory structure
                                , &pmd
                                , &streamSize
                                );
   if (trace > 0)
      printf("Minidump read Memory64ListStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read Memory64ListStream Failed:: GetLastError: %u\n", GetLastError());
      return NULL;
   }

   if (brc)
   {
      if (trace > 2)
      {
         printf("Type ptr returned = %p, stream size = %lx\n", pmd, streamSize);
         printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
                , pmdDir->StreamType
                , pmdDir->Location.DataSize
                , pmdDir->Location.Rva
                );
         nRanges = pmd->NumberOfMemoryRanges;
         printf("no of ranges = %I64d, base RVA = %I64x\n", nRanges,pmd->BaseRva);
         currRVA = pmd->BaseRva;
         count = 0;
         while (count < nRanges)
         {
            printf ( "  %03I64d start = %016I64x, size = %016I64x, RVA = %08I64x\n"
                   , count
                   , pmd->MemoryRanges[count].StartOfMemoryRange
                   , pmd->MemoryRanges[count].DataSize
                   , currRVA
                   );
            currRVA += pmd->MemoryRanges[count].DataSize;
            count++;
         }

      }
   }
   return pmd;
}




/*
*/
static PSZ extractComment ()
{
   /* Note: Although we can write an array of comment streams
   **       into the dump (using MINI_DUMP_USERSTREAM []), the
   **       readstream only permits us to extract ONE comment
   **       stream.  More comment streams could be extracted
   **       by reading the file map directly.
   **
   ** The comment is currently used to positively identify
   ** IBM dump extractions
   **
   */
   BOOL                brc        = FALSE;
   PVOID               pComment   = NULL;
   ULONG               streamSize = 0;
   PMINIDUMP_DIRECTORY pmdDir     = NULL;

   brc = MiniDumpReadDumpStream ( pmap
                                , CommentStreamA
                                , &pmdDir        // directory structure
                                , &pComment
                                , &streamSize
                                );

   if (trace > 0)
      printf("Minidump read CommentStreamA rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read CommentStreamA Failed:: GetLastError: %u\n", GetLastError());
      return NULL;
   }

   if (brc)
   {
      if (trace > 2)
      {
         printf("Type ptr returned = %p, stream size = %lx\n", pComment, streamSize);
         printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
                , pmdDir->StreamType
                , pmdDir->Location.DataSize
                , pmdDir->Location.Rva
                );
         printf("The comment is \"%s\"\n", pComment);
      }
   }
   return (PSZ)pComment;
}



/*
typedef struct _MINIDUMP_DIRECTORY
{
    ULONG32 StreamType;
    MINIDUMP_LOCATION_DESCRIPTOR Location;
} MINIDUMP_DIRECTORY, *PMINIDUMP_DIRECTORY;

typedef struct _MINIDUMP_MODULE_LIST
{
    ULONG32 NumberOfModules;
    MINIDUMP_MODULE Modules [ 0 ];
} MINIDUMP_MODULE_LIST, *PMINIDUMP_MODULE_LIST;
*/

static PMINIDUMP_MODULE_LIST extractModule ()
{
   BOOL                  brc        = FALSE;
   ULONG                 streamSize = 0;
   PMINIDUMP_DIRECTORY   pmdDir     = NULL;
   PMINIDUMP_MODULE_LIST pml        = NULL;



   brc = MiniDumpReadDumpStream ( pmap
                                , ModuleListStream
                                , &pmdDir        // directory structure
                                , &pml
                                , &streamSize
                                );
   if (trace > 0)
      printf("Minidump read ModuleListStream rc = %d\n", brc);

   if(!brc)
   {
      printf("Minidump read ModuleListStream Failed:: GetLastError: %u\n", GetLastError());
      return NULL;
   }

   if (brc)
   {
      if (trace > 2)
      {
         printf("Type ptr returned = %p, stream size = %lx\n", pml, streamSize);
         printf ( "Directory stream type = %d, datasize = %lx, RVA = %lx\n"
                , pmdDir->StreamType
                , pmdDir->Location.DataSize
                , pmdDir->Location.Rva
                );
         printf("No of modules = %d\n", pml->NumberOfModules);
      }
   }
   return pml;
}



/*
** Given an Rva find what address it is
*/
PVOID Rva2Address (RVA64 rva)
{
   // scan mem ranges until we find the one containing this rva
   ULONG64 count   = pMem->NumberOfMemoryRanges;
   ULONG64 i       = 0;
   RVA64   currRva = pMem->BaseRva;
   RVA64   diff;
   PVOID   address;

   while (i < count)
   {
      if (rva < currRva) break;
      diff = currRva;
      currRva += pMem->MemoryRanges[i].DataSize;
      i++;
   }

   currRva = pMem->MemoryRanges [i - 1].StartOfMemoryRange + (rva - diff);
   address = (PVOID)currRva;
   return (PVOID)address;
}

/*
** Given an address, find the rva
*/
RVA64 Address2Rva (PVOID address)
{
   ULONG64 count = pMem->NumberOfMemoryRanges;
   ULONG64 i = 0;
   RVA64 currRva = pMem->BaseRva;
   ULONG64 addr = (ULONG64)address;

   while (i < count)
   {
      if (addr > pMem->MemoryRanges [i].StartOfMemoryRange
      &&  addr < (pMem->MemoryRanges [i].StartOfMemoryRange + pMem->MemoryRanges [i].DataSize)) break;
      currRva += pMem->MemoryRanges[i].DataSize;
      i++;
   }
   if (i == count) return NULL;  // invalid address
   return currRva + (addr - pMem->MemoryRanges [i].StartOfMemoryRange);
}


/*
Return the RVA value for the given MemoryRange
*/
static RVA findRVA (PULONG64 start, ULONG32 length, PMINIDUMP_MEMORY64_LIST pMem)
{
   ULONG64 nRanges = pMem->NumberOfMemoryRanges;
   ULONG64 count;
   ULONG64 currRVA;
   ULONG64 prevRVA;


      if (trace > 2 )
      {
         printf("  Finding RVA for the Memory base %I64x, size %lx\n", *start, length);
         printf("no of ranges = %I64d, base RVA = %I64x\n", nRanges,pMem->BaseRva);
      }
      currRVA = pMem->BaseRva;
      count = 0;

      /*Run through all the MemoryRanges to findout where the required MemoryRange falls*/
      while (count < nRanges)
      {
             if (trace > 2)
             {
                 printf ( "  %03I64d start = %016I64x, size = %016I64x, RVA = %08I64x\n"
                        , count
                        , pMem->MemoryRanges[count].StartOfMemoryRange
                        , pMem->MemoryRanges[count].DataSize
                        , currRVA
                        );
             }
             if (pMem->MemoryRanges[count].StartOfMemoryRange > *start)
             {
                 /*The required MemoryRange falls in this Range*/
                 break;
             }
             prevRVA = currRVA;
             currRVA += pMem->MemoryRanges[count].DataSize;
             count++;
     }
     if (trace > 2)
     {
        printf ( "RVA for Memory base %I64x IS:: %08I64x\n "
               , *start
               , (prevRVA + (*start - pMem->MemoryRanges[count-1].StartOfMemoryRange))
               );
     }

   return (prevRVA + (*start - pMem->MemoryRanges[count-1].StartOfMemoryRange));
}

/*
** typedef struct _MINIDUMP_MODULE
** {
**       ULONG64 BaseOfImage;
**       ULONG32 SizeOfImage;
**       ULONG32 CheckSum;
**       ULONG32 TimeDateStamp;
**       RVA ModuleNameRva;
**       VS_FIXEDFILEINFO VersionInfo;
**       MINIDUMP_LOCATION_DESCRIPTOR CvRecord;
**       MINIDUMP_LOCATION_DESCRIPTOR MiscRecord;
**       ULONG64 Reserved0;
**       ULONG64 Reserved1;
**   } MINIDUMP_MODULE, *PMINIDUMP_MODULE;
*/
static void analyseModuleData (PMINIDUMP_MODULE_LIST pml)
{
   FILETIME filetime;

   if (trace >= 2)
   {
      ULONG32 count = 0;
      /* Run through all the modules and print the info.....*/
      while (count < pml->NumberOfModules)
      {
         printf("Module name offset = %08lx\n", pml->Modules[count].ModuleNameRva);
         wprintf(L"Module name = %s\n", readMiniDumpString (pml->Modules[count].ModuleNameRva));
         printf(" base = %08lx\n", pml->Modules[count].BaseOfImage);
         printf(" size = %08lx\n", pml->Modules[count].SizeOfImage);
         if (TimeDateStampToFileTime ( (DWORD)pml->Modules[count].TimeDateStamp
                                     , &filetime
                                     ))
         {
            char buffer [128];

            GetFileDateAsString (&filetime, buffer, 127);
            if (trace == 3) printf("   date/time = %s : ", buffer);
            GetFileTimeAsString (&filetime, buffer, 127, FALSE);
            if (trace == 3) printf("%s\n", buffer);
         }
         count++;
      }
   }
}


/*
typedef struct _MINIDUMP_STRING
{
      ULONG32 Length;
      WCHAR   Buffer [0];
  } MINIDUMP_STRING, *PMINIDUMP_STRING;
*/
static PSZ readMiniDumpString (RVA rvaString)
{
  PSZ string = (PSZ)((PBYTE)pmap + rvaString + sizeof(ULONG32));
   return string;
}


/*
Open the file and map it to memory..
*/
static BOOL openfile (void)
{
   hfile = CreateFile ( USERDUMP
                      , GENERIC_READ | GENERIC_WRITE
                      , 0
                      , NULL
                      , OPEN_EXISTING
                      , FILE_ATTRIBUTE_NORMAL
                      , NULL
                      );
   if (trace > 0) printf ("WinEx: open file returned handle %lx\n", hfile);
   if (hfile == (void *)0xffffffff) return FALSE;

   // create mem map
   map = CreateFileMapping ( hfile
                           , NULL            // security attributes, no inherit
                           , PAGE_READWRITE
                           , 0               // hi dword file size
                           , 0               // lo dword file size
                           , NULL            // map object name
                           );

   if (trace > 0)
   {
      printf("Minidump filemap rc = %lx\n", map);
   }

   // now map the file
   pmap = MapViewOfFile ( map
                        , FILE_MAP_WRITE
                        , 0      // start offset (hi)
                        , 0      // start offset (lo)
                        , 0      // map entire file
                        );

   if (trace > 0)
   {
      printf("Minidump map view rc = %p\n", pmap);
   }
   return TRUE;
}

/*
*/
static void closefile ()
{
   UnmapViewOfFile (pmap);
   CloseHandle (map);
   CloseHandle (hfile);
}


/*
  Convert a TimeDateStamp (i.e., # of seconds since 1/1/1970) into a FILETIME
*/

BOOL TimeDateStampToFileTime (DWORD timeDateStamp, LPFILETIME pFileTime)
{
   __int64 t1970 = 0x019DB1DED53E8000; // Magic... GMT...  Don't ask....

   __int64 timeStampIn100nsIncr = (__int64)timeDateStamp * 10000000;

   __int64 finalValue = t1970 + timeStampIn100nsIncr;

   memcpy( pFileTime, &finalValue, sizeof( finalValue ) );

   return TRUE;
}

/*
Convert FileDate to String
*/

BOOL GetFileDateAsString( LPFILETIME pFt, char * pszDate, unsigned cbIn )
{
   FILETIME   ftLocal;
   SYSTEMTIME st;
   char       szTemp [12];

   if ( !FileTimeToLocalFileTime( pFt, &ftLocal ) )
       return FALSE;

   if ( !FileTimeToSystemTime( &ftLocal, &st ) )
       return FALSE;

   wsprintf(   szTemp, "%02u/%02u/%04u",
               st.wDay, st.wMonth, st.wYear );
   lstrcpyn( pszDate, szTemp, cbIn );

   return TRUE;
}

/*
Convert FileTime to String
*/


BOOL GetFileTimeAsString ( LPFILETIME pFt
                         , char     * pszTime
                         , unsigned   cbIn
                         , BOOL       fSeconds
                         )
{
   FILETIME   ftLocal;
   SYSTEMTIME st;
   char       szTemp[12];

   if ( !FileTimeToLocalFileTime( pFt, &ftLocal ) )
       return FALSE;

   if ( !FileTimeToSystemTime( &ftLocal, &st ) )
       return FALSE;


   if ( fSeconds ) // Want seconds???
   {
       wsprintf( szTemp, "%02u:%02u:%02u", st.wHour, st.wMinute,st.wSecond );
   }
   else    // No thanks..  Just hours and minutes
   {
       wsprintf( szTemp, "%02u:%02u", st.wHour, st.wMinute );
   }

   lstrcpyn( pszTime, szTemp, cbIn );

   return TRUE;
}


/*                        EXTRACT API
**                        -----------
**
** The API provides systemic access to the mindump data
** The API requires that the extractor code has been invoked.
*/

ULONG32 __stdcall GetOS (PMINIDUMP_SYSTEM_INFO pSysInfo);
ULONG32 __stdcall GetArchitecture (PMINIDUMP_SYSTEM_INFO pSysInfo);
ULONG32 __stdcall GetMemRangeCount (PMINIDUMP_MEMORY64_LIST pMemList);

ULONG32 __stdcall GetOS (PMINIDUMP_SYSTEM_INFO pSysInfo)
{
   ULONG32 architecture = 0xffffffff;

   // are we 32 or 64 ?
   switch (pSysInfo->ProcessorArchitecture)
   {
   case PROCESSOR_ARCHITECTURE_INTEL :
      architecture = 9;    // 32b little endian
      break;

   case PROCESSOR_ARCHITECTURE_IA64 :
   // documented on ms website but not in header file..
   //case PROCESSOR_ARCHITECTURE_IA32_ON_WIN64 :
   //case PROCESSOR_ARCHITECTURE_AMD64 :
      architecture = 10;    // 64b little endian
      break;

   }
   return architecture;
}


ULONG32 __stdcall GetArchitecture (PMINIDUMP_SYSTEM_INFO pSysInfo)
{
   ULONG32 os;
   ULONG32 subsystem;

   #define NT3 (ULONG32)0
   #define NT4 (ULONG32)1
   #define W95 (ULONG32)2
   #define W98 (ULONG32)3
   #define ME  (ULONG32)4
   #define W2K (ULONG32)5
   #define XP  (ULONG32)6
   #define UNK (ULONG32)0XFFFFFFFF;

   /* what's the OS ?
   ** Major version :-
   ** 3 = NT 3.51
   ** 4 = 95, 98, ME or NT4
   ** 5 = 2K, XP or .NET
   **
   ** To sort out v4
   ** Minor version
   ** 0 = 95, NT4 or 2K
   ** 1 = XP or .NET
   ** 10 = 98
   ** 90 = ME
   **
   ** 3.x  = NT3.51
   ** 4.0  = either 95 or NT4
   **        Need to read PlatformId to sort these two out
   ** 4.10 = 98
   ** 4.90 = ME
   ** 5.0  = 2K
   ** 5.1  = XP or .NET
   **
   **
   */

   os = UNK;
   switch (pSysInfo->MajorVersion)
   {
   // nasty constants, can't find definitions, look harder sometime
   case 3 :
      os = NT3;
      break;

   case 4 :
      switch (pSysInfo->MinorVersion)
      {
      case 0 :
         switch (pSysInfo->PlatformId)
         {
         case VER_PLATFORM_WIN32_NT :
            os = NT4;
            break;

         case VER_PLATFORM_WIN32_WINDOWS :
            os = W95;
         }
         break;

      case 10 :
         os = W98;
         break;

      case 90 :
         os = ME;
         break;
      }
      break;

   case 5 :
      switch (pSysInfo->MinorVersion)
      {
      case 1 :
         os = XP;
         break;

      case 0 :
         os = W2K;
         break;
      }
      break;
   }
   if (trace > 2)
   {
      printf ("The dump OS is type %d\n", os);
   }

   // convert OS into SDFF format ..
   switch (os)
   {
   case NT3 :
      subsystem = 1;
      break;

   case NT4 :
      subsystem = 1;
      break;

   case W95 :
      subsystem = 3;
      break;

   case W98 :
      subsystem = 4;
      break;

   case W2K :
      subsystem = 2;
      break;

   case XP :
      subsystem = 6;
      break;

   case ME :
      subsystem = 5;
      break;

   default :
      subsystem = 0xffffffff;
      break;
   }

   return subsystem;

   #undef NT3
   #undef NT4
   #undef W95
   #undef W98
   #undef ME
   #undef W2K
   #undef XP
}

ULONG32 __stdcall GetMemRangeCount (PMINIDUMP_MEMORY64_LIST pMemList)
{
   return (ULONG32)pMemList->NumberOfMemoryRanges;
}


/*                       JAVA INTERFACE
**                       --------------
**
** All the JNI routines are mere wrappers for API functions
** The API functions use the data supplied by the extractor
*/
hpi_ras_t * findhpiras ();

/* ExecEnv true address vs thread id
*/
typedef struct _EXECENVS
{
   PVALUE pee;
   unsigned long id;
}EXECENVS;
typedef EXECENVS *PEXECENVS;;

PEXECENVS pExecEnvs = NULL;

/*
** Utility routine to allow pointers of any time to increment
*/
PVOID IncPointer (PVOID p, int offset)
{
   PBYTE pb = (PBYTE)p;
   pb += offset;
   return (PVOID)pb;
}

/*
** Utility routine to read tid
*/
unsigned long readTID (PVALUE pmap)
{
   // temp code pending proper hpiras with idoffset
   PBYTE pb = (PBYTE)pmap;
   PULONG32 p;
   pb += 4;
   p = (PULONG32)pb;
   return *p;
}


/*
**
*/

JNIEXPORT jlong JNICALL Java_WinExtract_getExecEnv (JNIEnv *e, jobject o, jint index)
{
   PBYTE     pee = NULL;          // execenv ptr
   PBYTE     p;                   // general ptr
   RVA64     rvaSysThread;        // offset in minidump to sys_thread_t structure
   PVALUE    pNext;               // the 'next' ptr in the sys thread
   PVALUE    pSysThread;          // address within the minidump (32/64b)
   PVALUE    pMap_SysThread;      // address within the minidump (32/64b)
   ULONG32   tid;                 // tid
   ULONG32   id;                  // TID
   PEXECENVS pees;                // stored ee's

   e=e;o=o;   // compiler pacifiers

   // get tid of nominated thread
   tid = pThreadList->Threads[index].ThreadId;

   // if not done then scan for the hpiras eye catcher
   if (phpiras == NULL)
   {
      // ppoint to the HPI structure
      phpiras = findhpiras ();
      if (phpiras == NULL)
      {
         printf("Cannot find the HPI RAS structure !\n");
         return NULL;
      }

      if (trace >= 3)
      {
         printf("The HPIRAS structure is :-\n");
         //printf("length = %d, version = %d\n", phpiras->ras.data.length, phpiras->ras.data.version);
         printf("thread Q ptr  = %08lx\n", phpiras->ras.data.threadQueuePtr);
         printf("'next' offset = %d\n", phpiras->ras.data.nextOffset);
         printf("'ee' offset   = %d\n", phpiras->ras.data.eeOffset);
         //printf("'id' offset   = %d\n", phpiras->ras.data.idOffset);
      }

      // make list of execenvs
      pExecEnvs = malloc (sizeof (EXECENVS) * pThreadList->NumberOfThreads);
      memset (pExecEnvs, 0, sizeof (EXECENVS) * pThreadList->NumberOfThreads);
      p = phpiras->ras.data.threadQueuePtr;  // JVM pointer to sys thread list
      // convert this address to an RVA
      rvaSysThread = Address2Rva (p);

      if (rvaSysThread == NULL)
      {
         // bad address
         free (pExecEnvs);
         pExecEnvs = NULL;
         return (jlong)-1;
      }

      // First ptr value at this RVA is the address of a sys_thread_t structure
      if (trace >= 3) printf("HPIRAS ptr to sys thread Q =%"PFORMAT"x\n",p);

      pSysThread = IncPointer (pmap, (int)rvaSysThread);  // map address of sys thread queue
      pSysThread = (PVALUE)*pSysThread;                   // deref to get true address of first sys_thread_t

      // convert sys_thread true address to rva
      rvaSysThread = Address2Rva (pSysThread);

      // get the address in the map of the structure
      pMap_SysThread = IncPointer (pmap, (int)rvaSysThread);

      /* OK ! rvaSysThread now indexes to the first sys_thread_t
      ** whose real address is pSysThread.  'pMap_SysThread' points
      ** to the start of the sys_thread_t structure in the  file
      ** memory-map
      */

      id = readTID (pMap_SysThread);
      if (trace >= 3)
         printf ( "The first sys_thread_t is at address %"PFORMAT"x, rva = %016I64x, map address = %"PFORMAT"x, , tid = %08lx\n"
                , pSysThread
                , rvaSysThread
                , pMap_SysThread
                , id
                );

      // ave data in structure for future ref
      pees      = pExecEnvs;
      pees->pee = IncPointer (pSysThread, (int)phpiras->ras.data.eeOffset);
      pees->id  = id;
      pees++;

      // now follow the 'next' links in the sys_thread_t
      for (;;)
      {
         // point to map address of sys_thread 'next' ptr
         pNext = IncPointer (pMap_SysThread, (int)phpiras->ras.data.nextOffset);
         // deref to get true address of next sys_thread
         pSysThread = (PVALUE)*pNext;
         if (pSysThread == NULL) break;    // done

         // convert to rva
         rvaSysThread = Address2Rva (pSysThread);
         // .. and to address in memory map
         pMap_SysThread = IncPointer (pmap, (int)rvaSysThread);

         // save new set of data (ee & tid)
         pees->pee = IncPointer (pSysThread, (int)phpiras->ras.data.eeOffset);
         pees->id  = readTID (pMap_SysThread);
         pees++;

      }

      if (trace >= 3)
      {
         pees = pExecEnvs;
         while (pees->pee != NULL)
         {
            printf("ExecEnv %"PFORMAT"x found for tid %08lx\n", pees->pee, pees->id);
            pees++;
         }
      }

   }

   // find the tid corresponding to thread index
   if (pExecEnvs == NULL)
   {
      // something is wrong
      return (jlong)-1;
   }

   pees = pExecEnvs;
   while (pees->pee != NULL)
   {
      if (pees->id == tid)
      {
         break;
      }
      pees++;
   }

   return (jlong)pees->pee;
}

/*
**
*/
hpi_ras_t * findhpiras ()
{
   ULONG64  count   = pMem->NumberOfMemoryRanges;
   RVA64    start   = pMem->BaseRva;
   PBYTE    pbstart = (PBYTE)pmap + pMem->BaseRva;
   PULONG64 pstart  = (PULONG64)pbstart;

   /* OK mem is just dumped in a huge smash in the
   ** mem mapped file, so just go scan it
   ** Crash & burn if not found - it's a disaster anyway
   */
   for (;;)
   {
      if (*pstart == 0xaa55aa55aa55aa55)
      {
         if (trace >= 2)
         {
            printf("Found HPIRAS eyecatcher at %08lx\n", pstart);
         }
         break;
      }
      pstart++;
   }
   pbstart  = (PBYTE)pstart;
   pbstart -= 8;    // skip back to make structure pointer

   return (hpi_ras_t *)pbstart;
}

JNIEXPORT jint JNICALL Java_WinExtract_getSubsystem (JNIEnv *e, jobject o)
{
   e=e;o=o;   // compiler pacifiers

   return (jint)GetOS (pSysInfo);
}

JNIEXPORT void JNICALL Java_WinExtract_setTraceLevel (JNIEnv * e, jobject o, jint level)
{
   e=e;o=o;   // compiler pacifiers

   if (level < 0 || level > 4)
      trace = 0;
   else
      trace = (int)level;
}

JNIEXPORT jint JNICALL Java_WinExtract_getArchitecture (JNIEnv * e, jobject o)
{
   e=e;o=o;
   return (jint)GetArchitecture (pSysInfo);
}

JNIEXPORT jint JNICALL Java_WinExtract_getMemoryRangeCount (JNIEnv * e, jobject o)
{
   e=e;o=o;
   return (jint)GetMemRangeCount (pMem);
}


   /**
    * Gets the start address of the next memRange
    * The value of memRangeId is 0 - n
    * ie return start of memrange [n]
    *
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getMemRangeStart  (JNIEnv * e, jobject o, jint id)
{
   e=e;o=o;

   return (jlong)pMem->MemoryRanges[id].StartOfMemoryRange;
}


   /**
    * Gets the length of the next memRange
    * The value of memRangeId is 0 - n
    * ie return start of memrange [n]
    *
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getMemRangeLength  (JNIEnv * e, jobject o, jint id)
{
   e=e;o=o;

   return (jlong)pMem->MemoryRanges[id].DataSize;

}


   /**
    * Gets the next mem range
    * The value of memRangeId is 0 - n
    * ie return bytes in data and length in return
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getNextMem  (JNIEnv * e, jobject o, jint id, jbyteArray ba)

{

   jboolean isCopy;
   char   * elements = (char *) (*e)->GetByteArrayElements (e, ba, &isCopy);
   jint     size     = (*e)->GetArrayLength (e, ba);
   jint     datasize = 0;
   char   * pData    = (char *)pmap;
   jint loop;

   o=o;

   pData += (DWORD)pMem->BaseRva;

   for (loop = 0 ; loop <= id ; loop++)
   {
      pData           += datasize;
      datasize = (jint) pMem->MemoryRanges[loop].DataSize;
   }


   if (size >= datasize)
   {
       memcpy(elements, pData, datasize );
       return (jlong)datasize;
   }

   return (jlong)0;
}

   /**
    * stack trace for thread indexed by 'index'
    *   indexThread 0 based thread index
    *   'data' is used to return data as follows
    *   0 = stack base true address
    *   1 = stack size
    *
    * returns index of thread
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getStackTrace  (JNIEnv * e, jobject o, jint index, jlongArray data)
{

   jboolean isCopy;
   jlong * elements = (jlong *) (*e)->GetLongArrayElements (e, data, &isCopy);
   jint    size     = (*e)->GetArrayLength (e, data);

   e=e; o=o;

   if (size < 2)
       return (jlong)-1;


   elements[0] = (jlong) pThreadList->Threads[index].Stack.StartOfMemoryRange;
   elements[1] = (jlong) pThreadList->Threads[index].Stack.Memory.DataSize;
   return index;
}

   /**
    * thread data
    *   indexThread 0 based thread index
    *   'data' is used to return data as follows
    *   0 = tid
    *   1 = stack ptr (0 if unknown)
    *   2 = eip
    *   3 = address of ExecEnv
    *   NB these are all 'true' addresses
    * returns thread index or -1 when no more threads
    */


JNIEXPORT jlong JNICALL Java_WinExtract_getThreadData  (JNIEnv * e, jobject o, jint index, jlongArray data)
{

   jboolean isCopy;
   jlong *  elements = (jlong *) (*e)->GetLongArrayElements (e, data, &isCopy);
   jint     size     = (*e)->GetArrayLength (e, data);
   PCONTEXT pcontext = (PCONTEXT) ((PBYTE) pmap + pThreadList->Threads[index].ThreadContext.Rva);

   e=e; o=o;

   if (size < 4)
       return (jlong)-1;

/*         printf ( "  Thread context, size %08lx, RVA %08lx\n"
                , pThreadList->Threads[count].ThreadContext.DataSize
                , pThreadList->Threads[count].ThreadContext.Rva
                );
*/


   elements[0] = (jlong) pThreadList->Threads[index].ThreadId;
   elements[1] = (jlong) pcontext->Esp;
   elements[2] = (jlong) pcontext->Eip;
   elements[3] = (jlong) 0;    //TBD:: After getting info about ee

   if (index++ < (jlong) pThreadList->NumberOfThreads)
       return index;
   else
       return (jlong)-1;

}

   /**
    * returns tid of the current thread
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getCurrThread  (JNIEnv * e, jobject o)
{
   e=e;o=o;

   if (pThreadList->NumberOfThreads == 1)
      return (jlong) (pThreadList->Threads[0].ThreadId);
   else
      /*return the tid of the thread on which exception occurred*/
      return (jlong) -1;

}

   /**
    * Returns no of thread
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getThreadCount  (JNIEnv * e, jobject o)
{
   e=e;o=o;

   return (jlong) (pThreadList->NumberOfThreads);

}

   /**
    * Returns no of loaded modules
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getModuleCount  (JNIEnv * e, jobject o)
{
   e=e;o=o;

   return (jlong) (pml->NumberOfModules);
}

   /**
    * Returns length of module name (-1 if last module)
    * Returns start address in data[0]
    * Returns module size in data[1]
    */
JNIEXPORT jlong JNICALL Java_WinExtract_getModuleData  (JNIEnv * e, jobject o, jint index, jlongArray data)
{

   PMINIDUMP_STRING pmds;
   jboolean         isCopy;
   jlong *          elements = (jlong *) (*e)->GetLongArrayElements (e, data, &isCopy);
   jint             size     = (*e)->GetArrayLength (e, data);

   if(index >= (jint)pml->NumberOfModules | size < 2)
      return (jlong) -1;

   elements[0] = (jlong) pml->Modules[index].BaseOfImage;
   elements[1] = (jlong) pml->Modules[index].SizeOfImage;

   o=o;

   pmds =(PMINIDUMP_STRING) ((PBYTE) pmap + pml->Modules[index].ModuleNameRva);

   if (trace > 2)
   {
      printf ("WinEx: Java_WinExtract_getModuleData:Length = %lx\n", pmds->Length);
      wprintf(L"module name = %s\n", pmds->Buffer);
   }

   return (jlong) (pmds->Length);
}

   /**
    * Returns module name
    */

JNIEXPORT jlong JNICALL Java_WinExtract_getModuleName  (JNIEnv * e, jobject o, jint index, jcharArray ca)
{
   jboolean         isCopy;
   char           * elements = (char *) (*e)->GetCharArrayElements (e, ca, &isCopy);
   jint             size     = (*e)->GetArrayLength (e, ca);
   jint             datasize = 0;
   PMINIDUMP_STRING pmds;

   o=o;

   if(index >= (jint)pml->NumberOfModules )
      return (jlong) -1;

   pmds  = (PMINIDUMP_STRING) ((PBYTE) pmap + pml->Modules[index].ModuleNameRva);

   if (size >= (jint) pmds->Length)
   {
       memcpy(elements, pmds->Buffer, pmds->Length );
       return (jlong)pmds->Length;
   }

   return (jlong)0;

}


   /**
    * Returns
    * Gs, Fs, Es, Ds, edi, esi, ebx, edx, ecx, eax, ebp, eip, Cs, eflags, esp, Ss
    */
JNIEXPORT void JNICALL Java_WinExtract_getRegisters  (JNIEnv * e, jobject o, jint index, jlongArray data)
{

   jboolean isCopy;
   jlong  * elements = (jlong *) (*e)->GetLongArrayElements (e, data, &isCopy);
   jint     size     = (*e)->GetArrayLength (e, data);
   PCONTEXT pcontext ;

   o=o;

/*   if (size < 16)
       return (jlong)-1;
*/
/*         printf ( "  Thread context, size %08lx, RVA %08lx\n"
                , pThreadList->Threads[count].ThreadContext.DataSize
                , pThreadList->Threads[count].ThreadContext.Rva
                );
*/
   pcontext = (PCONTEXT) ((PBYTE) pmap + pThreadList->Threads[index].ThreadContext.Rva);

   memcpy(elements, &(pcontext->Esp), 64);

   //elements[0] = (jlong) pcontext->Esp;;
   /*elements[1] = (jlong) pcontext->Esp;
   elements[2] = (jlong) pcontext->Eip;
   elements[3] = (jlong) 0;    //TBD:: After getting info about ee

   if (index++ < (jlong) pThreadList->NumberOfThreads)
       return index;
   else
       return (jlong) -1;
   */
}

/*
typedef struct _CONTEXT {

    //
    // The flags values within this flag control the contents of
    // a CONTEXT record.
    //
    // If the context record is used as an input parameter, then
    // for each portion of the context record controlled by a flag
    // whose value is set, it is assumed that that portion of the
    // context record contains valid context. If the context record
    // is being used to modify a threads context, then only that
    // portion of the threads context will be modified.
    //
    // If the context record is used as an IN OUT parameter to capture
    // the context of a thread, then only those portions of the thread's
    // context corresponding to set flags will be returned.
    //
    // The context record is never used as an OUT only parameter.
    //

    DWORD ContextFlags;

    //
    // This section is specified/returned if CONTEXT_DEBUG_REGISTERS is
    // set in ContextFlags.  Note that CONTEXT_DEBUG_REGISTERS is NOT
    // included in CONTEXT_FULL.
    //

    DWORD   Dr0;
    DWORD   Dr1;
    DWORD   Dr2;
    DWORD   Dr3;
    DWORD   Dr6;
    DWORD   Dr7;

    //
    // This section is specified/returned if the
    // ContextFlags word contians the flag CONTEXT_FLOATING_POINT.
    //

    FLOATING_SAVE_AREA FloatSave;

    //
    // This section is specified/returned if the
    // ContextFlags word contians the flag CONTEXT_SEGMENTS.
    //

    DWORD   SegGs;
    DWORD   SegFs;
    DWORD   SegEs;
    DWORD   SegDs;

    //
    // This section is specified/returned if the
    // ContextFlags word contians the flag CONTEXT_INTEGER.
    //

    DWORD   Edi;
    DWORD   Esi;
    DWORD   Ebx;
    DWORD   Edx;
    DWORD   Ecx;
    DWORD   Eax;

    //
    // This section is specified/returned if the
    // ContextFlags word contians the flag CONTEXT_CONTROL.
    //

    DWORD   Ebp;
    DWORD   Eip;
    DWORD   SegCs;              // MUST BE SANITIZED
    DWORD   EFlags;             // MUST BE SANITIZED
    DWORD   Esp;
    DWORD   SegSs;

} CONTEXT;


*/

#undef IA32
