/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/WinExtract.java, ras, dumpviewer 1.1
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 41451  100702  philv   New for feature
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:
 * ===========================================================================
 */

/*
** tbd
**
** Need information on CPU type, no. of processors,
** current processor number, current address space
** Probably best to shove these into a minidump user stream
*/
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.File;
import java.io.BufferedWriter;
import java.io.RandomAccessFile;

// import the cross platform debugger classes
import com.ibm.jvm.dump.DvDump;
import com.ibm.jvm.dump.SdffStd;          // header (sys info)
//import SdffMemRanges;    // mem range

public class WinExtract
{
   private static int trace = 0;

   //private java.io.BufferedReader    dIn;
   //private java.io.FileInputStream   fhIn;
   //PrintWriter    out;                      // output file stream
   RandomAccessFile file = null;

   private long currFileLengthPtr = 0;

   WinExtract (int trace)
   {
      // load the native extractor code
      try
      {
         System.loadLibrary ("WinExtract");
      }
      catch (Exception e)
      {
         System.out.println ("Can't load the native extractor code");
         return;
      }
      this.trace = trace;

      // create file for dump ...
      if (!createSDFFfile ())
      {
         System.out.println ("Can't create dump file !");
         return;
      }

      // get the data ready to hand ...
      if (doExtract (trace) < 0)
      {
         System.out.println ("The minidump file is missing or cirrupt !");
         return;
      }
      //doExtract (3);

      // dump the SDFF sections
      dumpSysInfo ();
      dumpProcessThread ();
      dumpMemory ();

      // close file, tidy up etc
      //file.close ();
   }

   private long getFilePointer ()
   {
      long l = -1;
      try
      {
         l = file.getFilePointer ();
      }
      catch (Exception eseek)
      {
      }
      return l;
   }

   /**
    * This function is used to write an SDFF length
    * field once we know the value to write into it.
    * assumes that we have saved 'currFileLengthPtr'
    * when the record header was writ
    */
   private void writeLength (long length)
   {
      long l = -1;
      try
      {
         l = file.getFilePointer ();
         file.seek (currFileLengthPtr);
         file.writeLong (length);
         file.seek (l);
      }
      catch (Exception eseek)
      {
      }
   }

   private boolean createSDFFfile ()
   {
      File sdff = new File ("SDFF");

      // force creation of new file
      if (sdff.exists ())
      {
         sdff.delete ();
      }
      try
      {
         sdff.createNewFile ();
      }
      catch (Exception ee)
      {
         System.out.println ("Can't make new SDFF");
         return false;
      }

      // use random access class - has spiffy binary writes
      try
      {
         file = new RandomAccessFile (sdff, "rw");
      }
      catch (Exception e)
      {
         System.out.println ("Can't make new SDFF");
         return false;
      }
      //try
      //{
      //   out = new PrintWriter (new BufferedWriter (new FileWriter ("SDFF")));
      //}
      //catch (java.io.IOException e)
      //{
      //   System.out.println("Can't open t file");
      //   return false;
      //}
      //System.out.println("out="+out);

      //try
      //{
      //   fhIn = new java.io.FileInputStream  ("SDFF");
      //   dIn  = new java.io.BufferedReader (new java.io.InputStreamReader (fhIn));
      //
      //}
      //catch (java.io.FileNotFoundException e)
      //{
      //   System.out.println ("Cannot open output stream");
      //   return false;        Cannot open output stream
      //}
      return true;
   }

   /**
    * Get the native code to extract all the
    * goodies.
    */
   private native int doExtract (int trace);

   /**
    * The system section
    *
    * From the C code we need :-
    * subsystem (32 bit / 64 bit)
    * architecture (flavour of windows)
    */
   private native int getSubsystem ();
   private native int getArchitecture ();

   /**
    * Sets the native code trace level
    * 0 no trace (default)
    * 1 minimal trace
    * 2 trace the dump
    * 3 trace everything, function ret codes etc etc
    */
   private native void setTraceLevel (int trace);

   /**
    * From the native code I need :-
    * - count of memory ranges
    */
   private native int getMemoryRangeCount ();

   private void writeNullMetaData ()
   {
      try
      {
         file.writeLong (-1);
      }
      catch (Exception e)
      {
         System.out.println ("Error writing meta data");
      }
   }

   /**
    * Gets the start address of the next memRange
    * The value of memRangeId is 0 - n
    * ie return start of memrange [n]
    *
    */
   private native long getMemRangeStart (int memRangeId);
   /**
    * Gets the length of the next memRange
    * The value of memRangeId is 0 - n
    * ie return start of memrange [n]
    *
    * May combine getMemRange Start/Length later on
    *
    */
   private native long getMemRangeLength (int memRangeId);
   /**
    * Gets the next mem range
    * The value of memRangeId is 0 - n
    * ie return byes
    */
   private native long getNextMem (int memRangeId, byte [] data);
   /**
    * The SDFF sys info section
    */
   private void dumpSysInfo ()
   {
      SDFFSTD header = new SDFFSTD (com.ibm.jvm.dump.SdffConstants.SEGMENT_SdffHead);
      //public static int  SEGMENT_SdffHead = 0;
      byte [] spare = new byte [128];

      header.setLength ( header.length   // default length
                       + 32              // sys info fields
                       + 128             // spare
                       + 8               // metadata
                       );
      // dump the header part
      header.dump (file);

      // now the sys info bits
      for (int i = 0 ; i < spare.length ; i++) spare [i] = 0x7f;
      try
      {
         file.writeInt (com.ibm.jvm.dump.DvDump.SYSTEM_IS_WINDOWS);      // system
         file.writeInt (getSubsystem ());               // architecture (size/endianess)
         file.writeInt (getArchitecture ());            // subsystem (type of Windoze)
         file.writeInt (com.ibm.jvm.dump.DvDump.PROCESSOR_IS_INTEL);     // processor generic arcitecture
         file.writeInt (com.ibm.jvm.dump.DvDump.PROC_SUBTYPE_UNKNOWN);// processor subtypre
         file.writeInt (1);                             // processors
         file.writeInt (0);                             // current processor
         file.writeInt (0);                             // current address space
         file.write (spare);                            // spare
      }
      catch (Exception esi)
      {
         System.out.println ("Error writing sys info");
      }
      // Dump meta data - version nos etc etc              ------------------------
      writeNullMetaData ();
   }

   /**
    * Given a true 32 bit address, find the memory
    * range it is in, and return the offset in the
    * file where this memory is
    */
   private void findMemRange32 (int address)
   {
      int difference = 0;
      int diff;
      int range = -1;

      for (int i = 0 ; i < nRanges ; i++)
      {
         if (ranges [i].start_address < address)
         {
            diff = address - (int)ranges[i].start_address;
            if (difference == 0)
            {
               difference = diff;
               range = i;
            }
            else
            {
               if (diff < difference)
               {
                  difference = diff;
                  range = i;
               }
            }
         }
      }
      //if (trace == 3)
      {
         System.out.println ("True address " + Integer.toHexString (address) + " found in mem range " + range + ", offset " + Integer.toHexString(difference));
         System.out.println (" Offset in file = " + Long.toHexString (ranges[range].offset + (long)difference));
      }
   }

   /**
    * The SDFF mem and mem ranges section
    */
   private SDFFMemRanges [] ranges  = null;
   private int              nRanges = 0;

   private void dumpMemory ()
   {
      // get number of memory ranges we are dummping
      long memLength = 0;
      long length;
      long start;
      long offset = 0;
      byte [] data;

      nRanges = getMemoryRangeCount ();
      // allocate mem range space
      ranges = new SDFFMemRanges [nRanges];

      if (trace >= 2)
         System.out.println ("No of mem ranges = " + getMemoryRangeCount ());

      // make SDFFMEM header
      SDFFSTD memHeader = new SDFFSTD (com.ibm.jvm.dump.SdffConstants.SEGMENT_Sdffmem);

      // remember where the section length field is
      currFileLengthPtr = getFilePointer ();
      // .. and dump the header
      memHeader.dump (file);

      // OK, now loop & get each mem section from the native code
      for (int i = 0 ; i < nRanges ; i++)
      {
         // get the next mem range
         ranges [i] = new SDFFMemRanges ();
         length     = getMemRangeLength (i);
         start      = getMemRangeStart (i);

         if (trace == 3)
         {
            System.out.println ("Mem range " + i + ", start = " + Long.toHexString (start) + ", length = " + Long.toHexString (length));
         }
         // fill in the mem range data
         ranges[i].start_address = start;
         ranges[i].memory_length = length;
         ranges[i].offset        = getFilePointer ();

         // update the mem header length
         memHeader.incLength (length);

         // make space for data
         data = new byte [(int)length];

         // write mem out
         long len;
         if ((len = getNextMem (i, data)) != length)
         {

            System.out.println (" ** err * data returned length = " + len);
            break;
         }
         try
         {
            offset = getFilePointer ();
            file.write (data);
         }
         catch (Exception edata)
         {
         }
         if (trace == 3)
         {
            System.out.println (" mem range " + i + " written at offset " + Long.toHexString (offset));
         }

      }
      writeLength (memHeader.getLength () + 8);   // write SDFFMEM section length + null metadata
      writeNullMetaData ();                       // terminate SDFFMEM section..

      // create SDFFMAP header
      SDFFSTD rangeHeader = new SDFFSTD (com.ibm.jvm.dump.SdffConstants.SEGMENT_Sdffmmap);
      rangeHeader.setLength ( rangeHeader.getLength ()  // default length
                            + (64 * nRanges)            // address range space
                            + 32                        // spare
                            + 8                         // null metadata
                            );

      rangeHeader.dump (file);

      byte [] spare = new byte [32];
      for (int i = 0 ; i < spare.length ; i++) spare [i] = 0x7f;
      try
      {
         file.write (spare);
      }
      catch (Exception err)
      {
      }

      for (int i = 0 ; i < nRanges ; i++)
      {
         try
         {
            file.writeLong (ranges [i].memory_id);
            file.writeLong (ranges [i].spare);
            file.writeLong (ranges [i].start_address);
            file.writeLong (ranges [i].memory_length);
            file.writeLong (ranges [i].offset);
            file.writeLong (ranges [i].flags1);
            file.writeLong (ranges [i].flags2);
            file.writeLong (ranges [i].flags3);
         }
         catch (Exception er)
         {
         }
      }
      writeNullMetaData ();

   }

   /**
    * stack trace for thread indexed by 'index'
    *   indexThread 0 based thread index
    *   'data' is used to return data as follows
    *   0 = stack base true address
    *   1 = stack size
    *
    * returns index of thread
    */
   private native long getStackTrace (int indexThread, long [] data);

   /**
    * returns count of threads
    */
   private native long getThreadCount ();

   /**
    * thread data
    *   indexThread 0 based thread index
    *   'data' is used to return data as follows
    *   0 = tid
    *   1 = esp (0 if unknown)
    *   2 = eip
    *   3 = address of ExecEnv
    *   NB these are all 'true' addresses
    * returns thread index or -1 when no more threads
    */
   private native long getThreadData (int threadIndex, long [] data);

   /**
    * returns tid of the current thread (-1 if unknown)
    */
   private native long getCurrThread ();


   private void dumpProcessThread ()
   {
      SDFFPRTH prth = new SDFFPRTH ();

      prth.dump (file);
   }

   /**
    * -- TEMP ------------
    * Phil hasn't done one of these in a usable form yet
    */
   class SDFFMemRanges
   {
      private long memory_id = 0;  // currently not used
      private long spare     = 0;  // currently not used
      private long start_address;
      private long memory_length;
      private long offset;
      private long flags1 = 0; // currently not used
      private long flags2 = 0; // currently not used
      private long flags3 = 0; // currently not used

      SDFFMemRanges ()
      {
      }

   }

   /**
    * All the SDFF sections have a class pertaining to herein.
    * Each class defines the section and has the ability to dump
    * itself and return how long its section is
    */


   /**
    * class
    */

   class   SDFFSTD
   //extends SdffStd
   {
      private long    length;
      private long    type;
      private byte [] eyecatcher = new byte [8];
      private long    version = com.ibm.jvm.dump.SdffConstants.VERSION1_0_0;

      private String [] eyecatchers
      =
      {
         com.ibm.jvm.dump.SdffConstants.SEGMENT_Eye_SdffHead,
         com.ibm.jvm.dump.SdffConstants.SEGMENT_Eye_SdffPrth,
         com.ibm.jvm.dump.SdffConstants.SEGMENT_Eye_SdffMmem,
         com.ibm.jvm.dump.SdffConstants.SEGMENT_Eye_SdffMem
      };

      SDFFSTD (long type)
      {
         this.type = type;
         eyecatcher = eyecatchers [(int)type].getBytes ();

         length = 32;     // initial length
      }

      SDFFSTD (long type, long version)
      {
         this (type);
         this.version = version;
      }

      protected void incLength (long length)
      {
         this.length += length;
      }

      protected void setLength (long length)
      {
         this.length = length;
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("SDFFSTD: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return this.length;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.writeLong (type);
            file.write     (eyecatcher);
            file.writeLong (version);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping SDFFSTD section");
            return;
         }

      }

   }

   /**
    * class
    */

   class SDFFJSTACK
   {
      private long    length       = 104;
      private byte [] eyecatcher   = {'J', 'S', 'T', 'A', 'C', 'K', 'T', 'R'};
      private long    frameId      = -1;
      private long    methodBlock  = -1;
      private long    lastPC       = -1;
      private long    argsPtr      = -1;
      private long    varsPtr      = -1;
      private long    isItJIT      = 0;
      private byte [] spare        = new byte [32];
      private long    nullMetaData = -1;

      SDFFJSTACK ()
      {
         for (int i = 0 ; i < spare.length ; i++) spare [i] = 0x7f;
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("JSTACK: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return length;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.write (eyecatcher);
            file.writeLong (frameId);
            file.writeLong (methodBlock);
            file.writeLong (lastPC);
            file.writeLong (argsPtr);
            file.writeLong (varsPtr);
            file.writeLong (isItJIT);
            file.write (spare);
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping JSTACKTR section");
            return;
         }

      }
   }


   /**
    * Returns no of loaded modules
    */
   private native long getModuleCount ();

   /**
    * Returns length od module name (-1 if last module)
    * Returns start address in data[0]
    * Returns module size in data[1]
    */
   private native long getModuleData (int index, long [] data);

   /**
    * Returns module name
    */
   private native long getModuleName (int index, char [] name);

   /**
    * class
    */

   class SDFFLOADED
   {
      /**
       * This class is needed 'cos the LOADED section
       * holds an array of module data that is not
       * otherwise defined
       */

      /**
       * class
       */

      class MODDATA
      {
         private long    startAddress;
         private long    moduleSize;
         private long    nameLength;
         private char [] name;

         private long length = 24;
         private transient long fillerLen;
         private transient long Clen = 0;

         MODDATA (int index)
         {
            long data [] = new long [2];

            nameLength   = getModuleData (index, data);
            if (trace >= 2)
            {
               System.out.println ("SDFFLOADED: module " + index + ", returned length " + nameLength);
               System.out.println ("            start Address =  " + Long.toHexString (data[0]) + ", size = " + Long.toHexString(data[1]));
            }
            startAddress = data [0];
            moduleSize   = data [1];
            name         = new char [(int)(nameLength)];
            Clen         = nameLength / 2; // length of C string (8 bit bytes)
            length      += Clen;
            fillerLen    = 8 - (Clen - ((Clen / 8) * 8));
            if (fillerLen == 8) fillerLen = 0;
            if (trace >= 2)
            {
               System.out.println ("  module " + index + ", str len added " + Clen + ", filler = " + fillerLen);
            }
            //length      += nameLength;
            length += fillerLen;
            getModuleName (index, name);
            if (trace == 3)
            {
               System.out.println ("SDFFLOADED: module " + index + ", name = " + new String (name));
            }
         }

         private long getLength ()
         {
            if (trace>= 2)
            {
               System.out.println ("MODDATA: length = " + length + "(0x" + Long.toHexString (length) + ")");
            }
            return this.length;
         }


         protected void dump (RandomAccessFile file)
         {
            byte [] filler = {0, 0, 0, 0, 0, 0, 0};
            long fp;
            int  count;
            int  len;
            String sname;
            byte b;

            len = name.length / 2;
            try
            {
               file.writeLong (startAddress);
               file.writeLong (moduleSize);
               file.writeLong ((length/2));
//fp = getFilePointer ();
//System.out.println ("writing mod name len "+name.length+" at "+fp);
               /* Gack; this is messy.  Write out the char array, char
               ** at a time, as a byte to make it easy for the dump
               ** formatter (they look like C strings).  Although
               ** we got wide bytes (= java chars) the length of the
               ** name is in ordinary bytes.  Yuketty Yuck
               */
               for (count = 0 ; count < len ; count++ )
               {
                  b = (byte)name[count];
                  file.write ((int)b);
                  //System.out.print (count +"byte "+b +"  ");
               }
               //System.out.println (" ");
               //sname = new String (name);
//System.out.println ("writing mod name str len "+sname.length()+" at "+fp);
               //file.writeChars (sname);
               // move to next 8 byte boundary
               fp = getFilePointer ();
//System.out.println ("written mod name at "+fp);
               if (fillerLen != 0) file.write (filler, 0, (int)fillerLen);
//System.out.println ("written filler "+fp);
            }
            catch (Exception e)
            {
               System.out.println ("Error dumping LOADED:MODDATA section");
               return;
            }
         }
      }

      private long       length       = 24;
      private byte    [] eyecatcher   = {'L', 'O', 'A', 'D', 'E', 'D', ' ', ' '};
      private MODDATA [] modData;
      private long       nullMetaData = -1;

      private long count = 0;

      SDFFLOADED ()
      {
         count = getModuleCount ();
         if (trace == 3)
         {
            System.out.println ("SDFFLOADED: module count = " + count);
         }
         modData = new MODDATA [(int)count];
         for (int i = 0 ; i < (int)count ; i++)
         {
            modData [i] = new MODDATA (i);
            length += modData[i].getLength ();
         }
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("LOADED: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return length;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.write (eyecatcher);
            for (int i = 0 ; i < (int)count ; i++)
            {
               modData[i].dump (file);
            }
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping LOADED section");
            return;
         }
      }
   }

   /**
    * Returns length and name (eg cn2001.....) full version
    * NB version string is max 32 bytes
    */
   private native void getVersion (String version);

   /**
    * Returns the length of the java home directory name
    */
   private native long getHomeLength ();

   /**
    * Returns the java home directory name
    */
   private native long getHomeName (byte [] homeName);

   /**
    * Returns length of the java lib (dll) directory
    */
   private native long getLibLength ();

   /**
    * Returns the java lib (dll) name
    */
   private native long getLibName (byte [] libName);

   /**
    * Returns size of buffer needed to hold environment variable set
    */
   private native long getEnvVarLength ();

   /**
    * Returns env vars as set of null terminated strings
    */
   private native long getEnvVars (byte [] env);

   /**
    * Returns size of classpath
    */
   private native long getClasspathLength ();

   /**
    * Returns the classpath
    */
   private native long getClasspath (byte [] classpath);


   /**
    * class
    */

   class SDFFENVDATA
   {
      private long    length       = 24;
      private byte [] eyecatcher   = {'E', 'N', 'V', 'D', 'A', 'T', 'A', ' '};
      private byte [] version      = new byte [32];
      private int     homeDirLen   = 0;
      private byte [] homeDir;
      private int     libDirSize   = 0;
      private byte [] libDir;
      private int     envSize      = 0;
      private byte [] env;
      private int     classSize    = 0;
      private byte [] classath;
      private long    nullMetaData = -1;

      SDFFENVDATA ()
      {
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("ENV: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return 24;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.write (eyecatcher);
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping ENVDATA section");
            return;
         }
      }
   }

   /**
    * Returns
    * Gs, Fs, Es, Ds, edi, esi, ebx, edx, ecx, eax, ebp, eip, Cs, eflags, esp, Ss
    */
   private native void getRegisters (int threadIndex, long [] regs);

   /**
    * class
    */

   class SDFFREGS
   {
      private long       length       = 24;
      private byte    [] eyecatcher   = {'R', 'E', 'G', 'I', 'S', 'T', 'E', 'R'};
      private long       nullMetaData = -1;

      SDFFREGS (int index)
      {
         long [] regs = new long [16];

         getRegisters (index, regs);

      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("REGS: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return 24;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.write (eyecatcher);
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping REGS section");
            return;
         }
      }

   }

   private native long getExecEnv (int index);

   /**
    * class
    */

   class SDFFTHREAD
   {
      private  long       length       = 104;
      private  byte []    eyecatcher   = {'T', 'H', 'R', 'E', 'A', 'D', ' ', ' '};
      private  long       tid;
      private  int        type         = -1;
      private  int        state        = -1;
      private  long       pStack;
      private  long       ip;
      private  long       ee;
      private  int        processor    = -1;
      private  int        spare1;
      private  SDFFJSTACK stack;
      private  SDFFREGS   registers;
      private  byte []    spare2       = new byte [32];
      private  long       nullMetaData = -1;

      SDFFTHREAD (int index)
      {
         stack     = new SDFFJSTACK ();
         registers = new SDFFREGS (index);

         length += stack.getLength ();
         length += registers.getLength ();

         spare1 = 0x7f7f;
         for (int i = 0 ; i < spare2.length ; i++) spare2 [i] = 0x7f;

         // find where this thread's stack is
         long [] data = new long [2];
         long    rc   = getStackTrace (index, data);

         if (trace == 3)
         {
            System.out.println ( "SDFFTHREAD: rc = " + rc
                               + ", stack base = " + Long.toHexString(data[0])
                               + ". stack size = " + Long.toHexString(data [1])
                               );
         }
         /* OK, now I need to scan the mem ranges to find
         ** the one containing the stack base and then
         ** convert it to a file offset and store in pStack
         ** The stack length is not saved in the SDFF,
         ** apparently the stack walkers will dynamically
         ** deduce the stack size !
         */

         // get the ee for this thread
         ee = getExecEnv (index);
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("THREAD: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return length;
      }


      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.write (eyecatcher);
            file.writeLong (tid);
            file.writeInt (type);
            file.writeInt (state);
            file.writeLong (pStack);
            file.writeLong (ip);
            file.writeLong (ee);
            file.writeInt (processor);
            file.writeInt (spare1);
            stack.dump (file);
            registers.dump (file);
            file.write (spare2);
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping THREAD section");
            return;
         }
      }
   }


   /**
    * class
    */

   class SDFFPROCESS
   {
      private long          length       = 328;
      private byte       [] eyecatcher   = {'P', 'R', 'O', 'C', 'E', 'S', 'S', ' '};
      private long          pid;
      private long          nThreads;
      private long          currThread;
      private long          signal       = 0;
      private byte       [] signalName   = {'0', '0', '0', '0', '0', '0', '0', '0'};
      private byte       [] commandLine  = new byte [256];
      private long          threadOffset; //
      private SDFFENVDATA   env;
      private SDFFLOADED    loaded;
      private SDFFTHREAD [] thread;
      private long          nullMetaData = -1;

      SDFFPROCESS ()
      {
         long rc;
         long [] data = new long [4];

         // get thread count
         nThreads = getThreadCount ();
         if (trace >= 2)
         {
            System.out.println ("SDFFPROCESS - threads = " + nThreads);
         }
         //nThreads = 2;
         // .. and create thread storage
         thread = new SDFFTHREAD [(int)nThreads];

         // create thread SDFF sections
         for (int i = 0 ; i < (int)nThreads ; i++)
         {
            thread [i] = new SDFFTHREAD (i);
            rc = getThreadData (i, data);
            thread[i].tid = data [0];
            thread[i].ip  = data [2];
            //thread[i].ee  = data [3];
            if (trace == 3)
            {
               System.out.println ("SDFFPROCESS:thread "+ i);
               System.out.println ("  tid = " + data[0] + "(0x" + Long.toHexString (data[0]) + ")");
               System.out.println ("  esp = " + Long.toHexString (data[1]));
               System.out.println ("  eip = " + Long.toHexString (data[2]));
               System.out.println ("  ee  = " + Long.toHexString (data[3]));
               System.out.println ("  len = " + thread[i].getLength ());
            }
            length += thread[i].getLength ();

         }
         currThread = getCurrThread ();
         if (trace == 3)
         {
            System.out.println ("SDFFPROCESS: curr thread = " + currThread + "(0x" + Long.toHexString (currThread) + ")");
         }

         env     = new SDFFENVDATA ();
         length += env.getLength ();
         loaded  = new SDFFLOADED ();
         length += loaded.getLength ();
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("PROCESS: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return length;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            //thread.dump (file);
            file.writeLong (length);
            file.write (eyecatcher);
            file.writeLong (pid);
            file.writeLong (nThreads);
            file.writeLong (currThread);
            file.writeLong (signal);
            file.write (signalName);
            file.write (commandLine);
            file.writeLong (threadOffset);
            env.dump (file);
            loaded.dump (file);
            for (int i = 0 ; i < (int)nThreads ; i++)
            {
               thread [i].dump (file);
            }
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping PROCESS section");
            return;
         }
      }
   }


   /**
    * class
    */

   class SDFFADDRESSSPACE
   {

      private  long          length       = 104;
      private  byte       [] eyecatcher   = {'A', 'D', 'D', 'R', 'S', 'P', 'C', ' '};
      private  long          id           = 0;
      private  long          nProcesses   = 1;
      private  int           index        = 0;
      private  byte       [] spare        = new byte [60];
      private  SDFFPROCESS   process;
      private  long          nullMetaData = -1;

      SDFFADDRESSSPACE ()
      {
         process = new SDFFPROCESS ();
         length += process.getLength ();

         for (int i = 0 ; i < spare.length ; i++) spare [i] = 0x7f;
      }

      protected long getLength ()
      {
         if (trace >= 2)
         {
            System.out.println ("ADDSPC: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return length;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            file.writeLong (length);
            file.write (eyecatcher);
            file.writeLong (id);
            file.writeLong (nProcesses);
            file.writeInt  (index);
            file.write     (spare);
            process.dump (file);
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping ADDRESS SPACE section");
            return;
         }
      }
   }


   /**
    * class
    */

   class SDFFPRTH
   {
      private  SDFFSTD          header         = null;
      private  long             nAddressSpaces = 1;
      private  SDFFADDRESSSPACE addressSpace   = null;
      private  byte          [] spare          = new byte [32];
      private  long             nullMetaData   = -1;

      private transient long length;

      SDFFPRTH ()
      {
         header       = new SDFFSTD (com.ibm.jvm.dump.SdffConstants.SEGMENT_Sdffprth, 1);
         addressSpace = new SDFFADDRESSSPACE ();

         header.setLength (getLength ());

         for (int i = 0 ; i < spare.length ; i++) spare [i] = 0x7f;

      }

      protected long getLength ()
      {
         length = header.getLength ()        // 'empty' length
                + 16                         // address space count + metadata
                + addressSpace.getLength ()  // address space
                + 32                         // spare space
                ;
         if (trace >= 2)
         {
            System.out.println ("PRTH: length = " + length + "(0x" + Long.toHexString (length) + ")");
         }
         return length;
      }

      protected void dump (RandomAccessFile file)
      {
         try
         {
            header.dump (file);
            file.writeLong (nAddressSpaces);
            addressSpace.dump (file);
            file.write (spare);
            file.writeLong (nullMetaData);
         }
         catch (Exception e)
         {
            System.out.println ("Error dumping PRTH section");
            return;
         }

      }
   }



   private static void help ()
   {
      System.out.println ("blah blah");
   }

   /**
    * Input parameters are
    * -f<name> Name of the minidump file
    * If absent a file called 'extract.dmp' will be looked for
    * (a) in the current directory
    * (b) in the path
    * -t<trace level>, default = 0 (no trace)
    * -h -? help
    */
   public static void main (String args [])
   {
      int trace       = 0;
      String dumpName = null;

      // Parameter analysis
      if (args.length > 0)
      {
         args[0].toLowerCase ();
         if (args.length == 1 && (args[0].startsWith ("-h") || args[0].startsWith ("-?")))
         {
            help ();
            System.exit (1);
         }

         int nParams = args.length - 1;
         while (nParams >= 0)
         {
            if (args [nParams].startsWith ("-f"))
            {
               // input filename supplied --- TBD
            }
            if (args [nParams].startsWith ("-t"))
            {
               // read and set trace
               String t = args[nParams].substring (args[nParams].length() - 1);
               try
               {
                  trace = Integer.parseInt (t.trim());
               }
               catch (NumberFormatException e)
               {
                  System.out.println ("Can't understand " + t);
                  System.exit (-1);
               }
               if (trace < 0 || trace > 3)
               {
                  System.out.println ("Invalid trace level");
                  System.exit (-1);
               }

            }
            nParams--;
         }
      }

      new WinExtract (trace);
   }
}
