package com.ibm.jvm.findroots;

import java.util.*;
import java.io.*;
import java.text.*;
import com.ibm.jvm.svcdump.Dump;

abstract class Base {

    static DateFormat df = DateFormat.getTimeInstance();
    static boolean xml = false;
    static boolean verbose = false;

    /**
     * Output a log message with time stamp. If xml flag is set, use xml formatting.
     */
    public final void log(String message) {
        if (!verbose())
            return;
        Calendar c = Calendar.getInstance();
        String time = df.format(c.getTime());

        if (xml)
            System.out.println("<log time='" + time + "'>" + message + "</log>");
        else
            System.out.println(time + ": " + message);
    }

    static boolean verbose() {
        return verbose;
    }

    static void setVerbose(boolean flag) {
        verbose = flag;
    }

    static void assert(boolean condition) {
        if (!condition)
            throw new Error("assert failed!");
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }

    String[] options() {
        return null;
    }

    String[] optionDescriptions() {
        return null;
    }

    boolean parseOption(String arg, String opt) {
        return false;
    }

    abstract String className();

    void usage() {
        System.err.print("Usage: java com.ibm.jvm.findroots." + className());
        for (int i = 0; i < options().length; i++)
            System.err.print(" " + options()[i]);
        System.err.println(" <filename>");
        for (int i = 0; i < options().length; i++)
            System.err.println("\t" + options()[i] + "\t" + optionDescriptions()[i]);
        System.exit(1);
    }

    void parseOptions(String args[]) {
        if (args.length == 0) {
            usage();
        }
        for (int i = 0; i < args.length; i++) {
            if (args[i].charAt(0) == '-') {
                if (args[i].equals("-verbose")) {
                    verbose = true;
                } else if (!parseOption(args[i], args[i + 1])) {
                    usage();
                }
            }
        }
    }
}

