package com.ibm.jvm.findroots;


import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;


/**
 *  This class parses simple test files which define directed graphs.
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class DigraphParser {
    HeapdumpContentHandler handler;
    LineNumberReader lnr;
    String currentLine;

    DigraphParser(HeapdumpContentHandler handler, String fileName) throws Exception {
        this.handler = handler;
        parse(fileName);
    }

    void parse(String fileName) throws Exception {
        LineNumberReader lnr = new LineNumberReader(new FileReader(fileName));

        for (String line = lnr.readLine(); line != null; line = lnr.readLine()) {
            if (line.length() == 0 || line.charAt(0) == '#')
                continue;
            int from = 0;

            for (;;) {
                while (from < line.length() && line.charAt(from) == ' ') from++;
                if (from == line.length()) break;
                int dash = line.indexOf('-', from);

                if (dash == -1)
                    handler.error("no dash found in: " + line);
                String fromString = line.substring(from, dash);
                int fromId = Integer.parseInt(fromString);
                int to = dash + 1;
                int space = line.indexOf(' ', to);
                String toString;

                if (space == -1)
                    toString = line.substring(to);
                else
                    toString = line.substring(to, space);
                int toId = Integer.parseInt(toString);

                //System.out.println("adding edge from " + fromId + " to " + toId);
                handler.addEdge(fromId, toId);
                if (space == -1) break;
                from = space + 1;
            }
        }
    }
}
