package com.ibm.jvm.findroots;


import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;


/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 *
 *  actually I nicked this code from somewhere...
 */

public class HeapdumpParser {
    HeapdumpContentHandler handler;
    LineNumberReader rin;
    String line;

    HeapdumpParser(HeapdumpContentHandler handler, String filename) throws Exception {
        this.handler = handler;
        parse(filename);
    }

    void parse(String filename) throws Exception {
        FileReader input = new FileReader(filename);
        rin = new LineNumberReader(input);

        /* Process the records */
        while ( (line = rin.readLine()) != null ) {
            if ( line.startsWith("//") ) {
              // A comment! Used to pass version but we ignore here
              continue;
            }
            if ( !line.startsWith("0x") ) {
                System.err.println( "Expected 0x... on line " +    /*ibm@0006*/
                                    rin.getLineNumber() +          /*ibm@0006*/
                                    " but got: " + line );         /*ibm@0006*/
                continue;
            }

            String saddr = line.substring( 2, 10 );
            long addr = Long.parseLong( saddr, 16 );

            if( line.charAt(11) != '[' ) {
                System.err.println( "Missing size marker '[' " +   /*ibm@0006*/
                                    " on line " + rin.getLineNumber() + /*ibm@0006*/
                                    " in: " + line );              /*ibm@0006*/
                continue;
            }
            int sizeEnd = 12;
            while( line.charAt(sizeEnd) != ']' ) { sizeEnd++; }
            int size = Integer.parseInt( line.substring( 12, sizeEnd ) );

            String name = line.substring(sizeEnd+2);

            if (name.startsWith("primitive array") ) {
                handler.primitiveArrayDump((int)addr, 1, size);
            } else {
                IntegerArray references = new IntegerArray();
                line = rin.readLine();

                if ( line == null || !line.startsWith("\t") ) {
                    System.err.println( "Expected \\t... on line "+/*ibm@0006*/
                                        rin.getLineNumber() +      /*ibm@0006*/ 
                                        " but got: " + line );     /*ibm@0006*/
                } else {
                    StringTokenizer st = new StringTokenizer(line);
                    while ( st.hasMoreTokens() ) {
                        String caddr = st.nextToken();
                        if ( caddr.equals("ignoring" ) ) {
                            break;
                        }
                        caddr = caddr.substring(2);
                        try {
                            references.add((int)Long.parseLong( caddr, 16 ) );
                        } catch (Exception e) {
                            e.printStackTrace();
                            System.err.println("something dodgy here on line " +
                                               rin.getLineNumber() + " in: " + line);
                        }
                    }
                }
                if (name.startsWith("class ")) {
                    handler.classDump((int)addr, name, references.toArray(), size);
                } else if (name.startsWith("array of ")) {
                    handler.objectArrayDump((int)addr, name, references.toArray(), size);
                } else {
                    handler.instanceDump((int)addr, name, references.toArray(), size);
                }
            }
        }
    }
}
