package com.ibm.jvm.findroots;


import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;


/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class JavatraceParser {
    HeapdumpContentHandler handler;
    LineNumberReader lnr;
    String currentLine;
    static int skip = 0;

    JavatraceParser(HeapdumpContentHandler handler, Reader input) throws Exception {
        this.handler = handler;
        parse(input);
    }

    void parse(Reader input) throws Exception {
        lnr = new LineNumberReader(input);

        skipHeader();

        /* Process the records */
        for (; currentLine != null; currentLine = lnr.readLine()) {

            int lineNumber = lnr.getLineNumber();

            if (currentLine.length() == 0) {
                System.out.println("blank line? " + lineNumber);
                continue;
            }

            char c = currentLine.charAt(0);

            if (c != 'n' && c != 'a') {
                System.out.println("strange line? " + lineNumber + ": " + currentLine);
                break;
            }

            int i1 = 3;
            int i2 = currentLine.indexOf(' ', i1);
            int i3 = currentLine.indexOf(' ', i2 + 1) + 1;
            int i4 = currentLine.indexOf(' ', i3);

            if (i4 == -1)
                i4 = currentLine.length();
            int id = 0;

            try {
                id = Integer.parseInt(currentLine.substring(i1, i2), 16);
            } catch (Exception e) {
                handler.error("bad address? " + lineNumber + ": " + currentLine);
            }

            if (c == 'a') {
                int size = 0;

                try {
                    size = Integer.parseInt(currentLine.substring(i3, i4), 10);
                } catch (Exception e) {
                    handler.error("bad array size? " + lineNumber + ": " + currentLine);
                }
                // We don't have the type info so assume it's char
                handler.primitiveArrayDump(id, 1, size);
            } else {
                String className = currentLine.substring(i3, i4);
                IntegerArray references = new IntegerArray();
                int i5 = currentLine.indexOf(' ', i4);

                if (i5 != -1) {
                    int start = i5 + 1;
                    int end = 0;

                    while (end != -1) {
                        end = currentLine.indexOf(' ', start);
                        int e = end;

                        if (end == -1) {
                            e = currentLine.length();
                            if (start == e)
                                break;
                        }
                        int rid = 0;

                        try {
                            rid = Integer.parseInt(currentLine.substring(start, e), 16);
                        } catch (NumberFormatException n) {
                            //handler.error("bad number? " + lineNumber + ": " + currentLine);
                            System.out.println("bad number? " + lineNumber + ": " + currentLine);
                            return;
                        }
                        references.add(rid);
                        start = end + 1;
                    }
                }
                if (currentLine.charAt(1) == 'c') {
                    handler.classDump(id, className, references.toArray(), 300);
                } else {
                    handler.instanceDump(id, className, references.toArray(), references.size() * 4 + 8);
                }
            }
        }
    }

    void skipHeader() throws Exception {

        String line;

        for (;;) {
            line = lnr.readLine();
            if (line.length() == 0)
                continue;
            if (line.charAt(0) == 'n') {
                currentLine = line;
                break;
            }
        }
        for (int i = 0; i < skip; i++) {
            for (;;) {
                line = lnr.readLine();
                if ((line.length() != 0) && (line.charAt(0) != 'n') && (line.charAt(0) != 'a'))
                    break;
            }
            currentLine = lnr.readLine();
        }
    }

}
