package com.ibm.jvm.findroots;

import java.util.*;
import java.io.*;
import java.text.*;
import java.lang.ref.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.util.html.*;

final public class NewGraph extends Base {

    static DateFormat df = DateFormat.getTimeInstance();
    static boolean xml;
    IntegerArray vertexIds = new IntegerArray();
    IntegerArray vertexIndexes = new IntegerArray();
    IntegerArray vertexSizes = new IntegerArray();
    IntegerArray orderedVertexIds;
    Edges edges = new Edges();
    HashMap indexCache;
    int dagIndex[];
    int reach[];
    NewGraph dag;
    PrintClient client;
    int size;
    BitSet isNotRoot = new BitSet();
    BitSet deleted = new BitSet();
    int lowestId = 0x7fffffff, highestId = 0;
    boolean recordParents;
    Edges parents;
    boolean complete = false;
    static boolean doFreeze = false;
    boolean sizeMatters = true;
    BitSet visited;
    int maxdepth = 50;
    int maxroots = 20;
    int prune = 100;
    Document doc = new Document();
    boolean exact = false;

    public NewGraph() {
        doc.setPlainText(!SvcdumpProperties.getBooleanProperty("svcdump.output.html", false));
        maxdepth = SvcdumpProperties.getIntProperty("findroots.depth", 1000);
        sizeMatters = SvcdumpProperties.getBooleanProperty("findroots.sizematters", false);
        prune = SvcdumpProperties.getIntProperty("findroots.prune", sizeMatters ? 10000 : 1000);
        maxroots = SvcdumpProperties.getIntProperty("findroots.maxroots", 20);
        exact = SvcdumpProperties.getBooleanProperty("findroots.exact", false);
        verbose = true;
    }

    /*
     * Following fields are junk.
     */
    public boolean nofinal;

    HashMap frozenObjects = new HashMap();

    String className() {
        return "NewGraph";
    }

    /**
     * Calculate the reachability from the given node. Note that this excludes nodes already
     * visited on previous calls.
     */
    public int reachFrom(int id) {
        complete();
        if (visited == null)
            visited = new BitSet();
        int index = idToIndex(id);
        Integer reach = (Integer)dfs(new Visitor() {
            int n = 0;
            void enterNode(int v, int depth) {
                n += vertexSizes.get(v);
            }
            Object result() {
                return new Integer(n);
            }
        }, index, visited);
        return reach.intValue();
    }

    public IntEnumeration getChildren(int id) {
        int index = idToIndex(id);
        assert(index >= 0);
        int root = edges.get(index);
        IntEnumeration e = TreeBitSet.elements(root);

        class WrapIntEnumeration implements IntEnumeration {
            IntEnumeration e;

            WrapIntEnumeration(IntEnumeration e) {
                this.e = e;
            }

            public boolean hasMoreElements() {
                return e.hasMoreElements();
            }

            public Object nextElement() {
                return new Integer(nextInt());
            }

            public int nextInt() {
                int i = e.nextInt();
//System.out.println("child index = " + i);
                return vertexIds.get(i);
            }

            public void reset() {
                e.reset();
            }
        }

        return new WrapIntEnumeration(e);
    }

    /**
     * Return the total number of vertices that are reachable from this one.
     */
    public int reachableVertices(int index) {
        assert(false);
        return 0;
    }

    /**
     * Return the total size occupied by all the vertices that are reachable from this one.
     */
    public int reachableSize(int index) {
        assert(false);
        return 0;
    }

    void freeze(Object object, String name) {
        log("freeze " + name);
        try {
            File file = File.createTempFile("tmp", ".freeze");
            FileOutputStream fos = new FileOutputStream(file);
            ObjectOutputStream oos = new ObjectOutputStream(fos);
            oos.writeObject(object);
            oos.close();
            frozenObjects.put(name, file);
            log("freeze " + name + " complete");
        } catch (Exception e) {
            throw new Error("Error freezing object: " + e);
        }
    }

    Object unfreeze(String name) {
        log("unfreeze " + name);
        try {
            File file = (File)frozenObjects.get(name);
            FileInputStream fis = new FileInputStream(file);
            ObjectInputStream ois = new ObjectInputStream(fis);
            Object object = ois.readObject();
            ois.close();
            frozenObjects.remove(name);
            file.delete();
            log("unfreeze " + name + " complete");
            return object;
        } catch (Exception e) {
            throw new Error("Error unfreezing object: " + e);
        }
    }

    void addEdgeByIndex(int v, int u) {
        edges.put(v, TreeBitSet.set(edges.get(v), u));
    }

    public void addEdge(int fromId, int toId) {
        int v = idToIndex(fromId);
        int u = idToIndex(toId);
        edges.put(v, TreeBitSet.set(edges.get(v), u));
    }

    void flushCache() {
        if (indexCache == null)
            return;
        Integer keys[] = (Integer[])indexCache.keySet().toArray(new Integer[0]);
        for (int i = 0; i < keys.length; i++) {
            Integer vertexIndex = (Integer)indexCache.get(keys[i]);
            vertexIndexes.add(vertexIndex.intValue());
            vertexIds.add(keys[i].intValue());
        }
        vertexIds.sort(vertexIndexes);
        indexCache = null;
    }

    void complete() {
        if (!complete) {
            flushCache();
            vertexIndexes.sort(vertexIds);
            vertexIndexes = null;
            complete = true;
        }
    }

    public int idToIndex(int id) {
        if (complete) {
            if (orderedVertexIds == null) {
                orderedVertexIds = new IntegerArray();
                vertexIndexes = new IntegerArray();
                for (int i = 0; i < vertexIds.size(); i++) {
                    orderedVertexIds.add(vertexIds.get(i));
                    vertexIndexes.add(i);
                }
                orderedVertexIds.sort(vertexIndexes);
            }
            int index = orderedVertexIds.indexOf(id);
            if (index == -1)
                System.out.println("warning! could not find id " + hex(id));
            return vertexIndexes.get(index);
        }
        int hint = -1;
        if (size > 0) {
            float r = ((float)id - (float)lowestId) / ((float)highestId - (float)lowestId);
            r *= size;
            hint = (int)r;
        }
        int index = vertexIds.indexOf(id, hint);
        if (index == -1) {
            Integer idInt = new Integer(id);
            if (indexCache == null)
                indexCache = new HashMap();
            Integer indexInt = (Integer)indexCache.get(idInt);
            if (indexInt == null) {
                edges.add(0);
                if (recordParents)
                    parents.add(0);
                vertexSizes.add(1);
                index = vertexIds.size() + indexCache.size();
                indexInt = new Integer(index);
                indexCache.put(idInt, indexInt);
                if (indexCache.size() == 500000) {
                    flushCache();
                }
                if (id < lowestId)
                    lowestId = id;
                else if (id > highestId)
                    highestId = id;
                size++;
            } else {
                index = indexInt.intValue();
            }
        } else {
            index = vertexIndexes.get(index);
        }
        return index;
    }

    public void setRecordParents(boolean recordParents) {
        this.recordParents = recordParents;
        parents = new Edges();
    }

    public int[] getParents(int id) {
        assert(recordParents);
        complete();
        int index = idToIndex(id);
        int p = parents.get(index);
        int pp[] = new int[TreeBitSet.numberOfElements(p)];
        int i = 0;
        for (IntEnumeration e = TreeBitSet.elements(p); e.hasMoreElements(); ) {
            int v = e.nextInt();
            pp[i++] = vertexIds.get(v);
        }
        return pp;
    }

    public int pruneLeaves() {
        assert(recordParents);
        complete();
        int leaves = 0;
        for (int i = 0; i < size; i++) {
            int p = parents.get(i);
            if (TreeBitSet.numberOfElements(p) == 0) {
                for (IntEnumeration e = TreeBitSet.elements(p); e.hasMoreElements(); ) {
                    int v = e.nextInt();
                    // temp edges.put(v, TreeBitSet.clear(edges.get(v), i));
                }
                parents.put(i, 0);
                vertexIds.put(i, -1);
                leaves++;
            }
        }
        return leaves;
    }

    public void addToVertex(int id, int size) {
        int v = idToIndex(id);
        vertexSizes.put(v, vertexSizes.get(v) + size);
    }

    public void addVertex(int id, int size) {
        int v = idToIndex(id);
        if (sizeMatters)
            vertexSizes.put(v, size);
    }

    public void addVertex(int id, int edgeIds[], int size) {
        int v = idToIndex(id);
        if (sizeMatters)
            vertexSizes.put(v, size);
        int root = edges.get(v);
        for (int i = 0; i < edgeIds.length; i++) {
            int u = idToIndex(edgeIds[i]);
            root = TreeBitSet.set(root, u);
            if (recordParents)
                parents.put(u, TreeBitSet.set(parents.get(u), v));
        }
        edges.put(v, root);
    }

    void printNode(int v, int depth, boolean firstTime) {
        int id = vertexIds.get(v);
        int r = reach[v];
        for (int i = 0; i < depth; i++) System.out.print("   ");
        System.out.println(r + " 0x" + hex(id) + " " + client.getName(id) + (firstTime ? "" : " (already visited)"));
    }

    void printTree(int node) {
        complete();
        log("begin printTree");
        printNode(node, 0, true);
        dfs(new Visitor() {
            int count = 0;

            void visitChild(int p, int v, boolean firstTime, int depth) {
                if (reach[v] < prune)
                    return;
                printNode(v, depth, firstTime);
            }

            boolean continueSearch(int v, int depth) {
                return depth < maxdepth && reach[v] > prune;
            }

            boolean sort() {
                return true;
            }

            int getCount(int v) {
                return reach[v];
            }
        }, node);
        log("end printTree");
    }

    int ancestorWithLowestSemi(int v, int ancestor[], BitSet anc, int dfnum[], int semi[]) {
        int u = v;
        while (anc.get(v)) {
            if (dfnum[semi[v]] < dfnum[semi[u]]) {
                u = v;
            }
            v = ancestor[v];
        }
        return u;
    }

    void link(int p, int n, BitSet anc) {
        anc.set(n);
    }

    /**
     * This method finds the dominator graph. It is copied from page 452 of "modern compiler
     * implementation in java" by andrew appel together with some mods by me to make it more
     * space efficient. I didn't bother with the path compression because it made no difference
     * to speed but required more space.
     */
    NewGraph findDominators() {
        log("begin findDominators");
        class AssignDfnum extends Visitor {
            int N = 1;
            int dfnum[];
            int vertex[];
            int parent[];
            int pre[];
            AssignDfnum(int dfnum[], int vertex[], int parent[], int pre[]) {
                this.dfnum = dfnum;
                this.vertex = vertex;
                this.parent = parent;
                this.pre = pre;
            }
            void visitChild(int p, int n, boolean firstTime, int depth) {
                if (firstTime) {
                    dfnum[n] = N;
                    vertex[N] = n;
                    parent[n] = p;
                    N++;
                }
               pre[n] = TreeBitSet.set(pre[n], p);
            }
            Object result() {
                return null;
            }
        }
        int dfnum[] = new int[size];
        int vertex[] = new int[size];
        int parent[] = new int[size];
        int pre[] = new int[size];
        int semi[];
        int idom[] = new int[size];
        int bucket[] = new int[size];
        BitSet anc = new BitSet();
        for (int i = 0; i < size; i++) {
            idom[i] = -1;
        }
        dfs(new AssignDfnum(dfnum, vertex, parent, pre), 0);
        // if in doubt make semi separate and init to -1 !!!
        semi = pre;     // can reuse since only one use at a time
        for (int i = size - 1; i >= 1; i--) {
            int n = vertex[i];
            assert(n != 0);
            int p = parent[n];
            int s = p;
            for (IntEnumeration e = TreeBitSet.elements(pre[n]); e.hasMoreElements(); ) {
                int v = e.nextInt();
                int s1;
                if (dfnum[v] <= dfnum[n]) {
                    s1 = v;
                } else {
                    s1 = semi[ancestorWithLowestSemi(v, parent, anc, dfnum, semi)];
                }
                if (dfnum[s1] < dfnum[s]) {
                    s = s1;
                }
            }
            semi[n] = s;
            bucket[s] = TreeBitSet.set(bucket[s], n);
            link(p, n, anc);
            for (IntEnumeration e = TreeBitSet.elements(bucket[p]); e.hasMoreElements(); ) {
                int v = e.nextInt();
                int y = ancestorWithLowestSemi(v, parent, anc, dfnum, semi);
                if (semi[y] == semi[v]) {
                    idom[v] = p;
                } else {
                    idom[v] = y | 0x80000000;   // equiv to "samedom[v] = y"
                }
            }
            TreeBitSet.free(bucket[p]);
            bucket[p] = 0;
        }
        for (int i = 1; i < size; i++) {
            int n = vertex[i];
            if (idom[n] < 0) {
                assert(idom[n] != -1);
                idom[n] = idom[idom[n] & 0x7fffffff];
            }
        }
        for (int i = 0; i < size; i++) {
            TreeBitSet.free(pre[i]);
        }
        dfnum = vertex = parent = pre = semi = bucket = null;
        log("end findDominators");
        NewGraph domtree = new NewGraph();
        for (int i = 0; i < size; i++) {
            domtree.idToIndex(vertexIds.get(i));      // must match
        }
        for (int i = 1; i < size; i++) {
            int fromId = vertexIds.get(idom[i]);
            int toId = vertexIds.get(i);
            domtree.addEdge(fromId, toId);
            domtree.vertexSizes.put(i, vertexSizes.get(i));
        }
        domtree.complete();
        return domtree;
    }

    public void deleteTree(int root) {
        dfs(new Visitor() {
            void exitNode(int v) {
                deleted.set(v);
            }
        }, root);
    }

    static void assert(boolean condition) {
        if (!condition)
            throw new Error("assert failed!");
    }

    NewGraph findStrongComponents() {
        log("begin findStrongComponents");
        int sc[] = (int[])dfs(new Visitor() {
            IntegerStack s = new IntegerStack();
            IntegerStack path = new IntegerStack();
            int pre[] = new int[size];                  
            int sc[] = new int[size];                   
            int cnt0 = 0, cnt1 = 1;                                 

            void init() {
                for (int i = 0; i < size; i++) {            
                    sc[i] = -1;                                         
                }                                                       
            }
            void enterNode(int v, int depth) {
                pre[v] = cnt0++;                        
                s.push(v);                              
                path.push(v);                           
            }
            void visitChild(int v, int u, boolean firstTime, int depth) {
                if (!firstTime && sc[u] == -1) {               
                    while (pre[path.peek()] > pre[u]) { 
                        path.pop();                     
                    }                                   
                }
            }
            void exitNode(int v) {
                if (path.peek() == v) {                 
                    path.pop();                         
                    int w;
                    do {                                
                        w = s.pop();
                        sc[w] = cnt1;            
                    } while (w != v);             
                    cnt1++;                             
                }                                       
            }
            Object result() {
                return sc;
            }
        });
        log("create dag");
        dag = new NewGraph();
        dagIndex = new int[size];
        dag.idToIndex(0);       // reserve 0 for the dummy root
        dag.isNotRoot.set(0);   // not a real root (protect transitiveClosure)
        for (int v = 0; v < size; v++) {
            if (deleted.get(v))
                continue;
            int root = edges.get(v);
            int dagv = dag.idToIndex(sc[v]);
            dagIndex[v] = dagv;
            dag.vertexSizes.put(dagv, dag.vertexSizes.get(dagv) + vertexSizes.get(v));
            if (root != 0) {
                for (IntEnumeration e = TreeBitSet.elements(root); e.hasMoreElements(); ) {
                    int u = e.nextInt();
                    dagIndex[u] = dag.idToIndex(sc[u]);
                    if (sc[v] != sc[u]) {
                        dag.addEdge(sc[v], sc[u]);
                        dag.isNotRoot.set(dagIndex[u]);
                    }
                }
            }
        }
        // create edges from dummy root to the other roots
        for (int i = 1; i < dag.size; i++) {
            // correct for initial vertex size of 1
            dag.vertexSizes.put(i, dag.vertexSizes.get(i) - 1);
            if (!dag.isNotRoot.get(i)) {
                dag.addEdgeByIndex(0, i);
            }
        }
        dag.complete();
        //dag.checkNoBackEdges();
        log("end findStrongComponents");
        return dag;
    }

    // get an id that has the given dag id
    int dagIdToId(int dagid) {
        return dagIdToIds(dagid)[0];
    }

    // get all the ids that have the given dag id
    int[] dagIdToIds(int dagid) {
        IntegerArray ids = new IntegerArray();
        int dagv = dag.idToIndex(dagid);
        for (int v = 0; v < size; v++) {
            if (deleted.get(v))
                continue;
            if (dagIndex[v] == dagv) {
                ids.add(vertexIds.get(v));
            }
        }
        return ids.toArray();
    }

    boolean isRoot(int index) {
        return !dag.isNotRoot.get(dagIndex[index]);
    }

    Object dfs(Visitor visitor) {
        BitSet grey = new BitSet();
        Stack enums = new Stack();
        IntegerStack stack = new IntegerStack();
        visitor.init();
        for (int i = 0; i < size; i++) {
            dfs(visitor, enums, stack, grey, i);
        }
        return visitor.result();
    }

    Object dfs(Visitor visitor, int node) {
        BitSet grey = new BitSet();
        return dfs(visitor, node, grey);
    }

    Object dfs(Visitor visitor, int node, BitSet grey) {
        Stack enums = new Stack();
        IntegerStack stack = new IntegerStack();
        visitor.init();
        dfs(visitor, enums, stack, grey, node);
        return visitor.result();
    }

    Object dfs(Visitor visitor, Stack enums, IntegerStack stack, BitSet grey, int node) {
        if (grey.get(node) || deleted.get(node)) {
            return visitor.result();
        }
        IntEnumeration e = visitor.elements(node);
        visitor.enterNode(node, stack.depth());
        stack.push(node);
        enums.push(e);
        grey.set(node);
        while (!stack.empty()) {
            int v = stack.peek();
            e = (IntEnumeration)enums.peek();
            boolean finished = true;
            if (e != null) {
                if (visitor.continueSearch(v, stack.depth()) && e.hasMoreElements()) {
                    int u = e.nextInt();
                    if (deleted.get(u)) {
                        continue;
                    }
                    boolean firstTime = !grey.get(u);
                    visitor.visitChild(v, u, firstTime, stack.depth());
                    if (!firstTime) {
                        continue;
                    }
                    grey.set(u);
                    visitor.enterNode(u, stack.depth());
                    e = visitor.elements(u);
                    stack.push(u);
                    enums.push(e);
                    finished = false;
                } else {
                    e.reset();                          
                    while (e.hasMoreElements()) {       
                        int u = e.nextInt();            
                        if (deleted.get(u)) {
                            continue;
                        }
                        visitor.visitChildAtExit(v, u);
                    }                                   
                }
            }
            if (finished) {
                visitor.exitNode(v);
                stack.pop();
                enums.pop();
            }
        }
        return visitor.result();
    }

    void transitiveClosure() {
        log("begin transitiveClosure");
        vertexIds = null;
        Visitor visitor = new Visitor() {
            int tc[] = new int[size];

            boolean ignoreDownEdges() {
                return true;
            }
            void visitChild(int v, int u, boolean firstTime, int depth) {
                tc[v] = TreeBitSet.set(tc[v], u);
            }
            void visitChildAtExit(int v, int u) {
                if (tc[u] != 0) {
                    tc[v] = TreeBitSet.or(tc[v], tc[u]);
                }           
            }
            void exitNode(int v) {
                tc[v] = TreeBitSet.close(tc[v]);
            }
            Object result() {
                return tc;
            }
        };
        BitSet grey = new BitSet();
        Stack enums = new Stack();
        IntegerStack stack = new IntegerStack();
        visitor.init();
        for (int i = 0; i < size; i++) {
            if (!isNotRoot.get(i)) {
                dfs(visitor, enums, stack, grey, i);
            }
        }
        int tc[] = (int[])visitor.result();
        log("calculate reachability");
        reach = new int[size];
        for (int i = 0; i < size; i++) {
            if (tc[i] != 0) {
                if (exact || sizeMatters) {
                    for (IntEnumeration e = TreeBitSet.elements(tc[i]); e.hasMoreElements(); ) {
                        int v = e.nextInt();
                        reach[i] += vertexSizes.get(v);
                    }
                } else {
                    reach[i] = TreeBitSet.numberOfElements(tc[i]);
                }
            }
        }
        for (int i = 0; i < size; i++) {
            TreeBitSet.free(tc[i]);
        }
        log("end transitiveClosure");
    }

    void gc() {
        log("enter gc");
        System.gc();
        for (int i = 0; i < 1000; i++) {
            byte b[] = new byte[100000];
            b[i] = (byte) (i << 1);
            if (b[99] == 33) System.out.println("yes");
        }
        System.gc();
        Runtime.getRuntime().gc();
        Runtime.getRuntime().runFinalization();
        log("exit gc");
    }

    void message(String msg) {
        gc();
        log(msg + ", bits heap usage " + TreeBitSet.usage() + " free " + TreeBitSet.freeMemory());
    }

    public void print(PrintClient client) {
        complete();
        message("begin analysis");
        this.client = client;
        log("number of vertexes = " + vertexIds.size());
        for (int i = 0; i < maxroots; i++) {
            int maxV = 0;
            int maxReach = 0;
            System.out.println("*** doing root " + i + " ***");
            System.out.println("");
            findStrongComponents();
            dag.transitiveClosure();
            reach = new int[size];
            for (int j = 0; j < size; j++) {
                reach[j] = dag.reach[dagIndex[j]];
            }
            dag = null;
            message("after trans closure");
            for (int j = 0; j < vertexIds.size(); j++) {
                int r = reach[j];
                if (r > maxReach) {
                    maxReach = r;
                    maxV = j;
                }
            }
            int maxId = vertexIds.get(maxV);
            log("max reach = " + maxReach + " for vertex " + hex(maxId) + " " + client.getName(maxId));
            printTree(maxV);
            deleteTree(maxV);
        }
    }

    public void print(PrintClient client, int id) {
        complete();
        this.client = client;
        log("number of vertexes = " + vertexIds.size());
        findStrongComponents();
        dag.transitiveClosure();
        reach = new int[size];
        for (int j = 0; j < size; j++) {
            reach[j] = dag.reach[dagIndex[j]];
        }
        dag = null;
        printTree(idToIndex(id));
    }

    public Vertex[] getRoots() {
        complete();
        findStrongComponents();
        dag.transitiveClosure();
        reach = dag.reach;
        dag = null;
        Vector vlist = new Vector();
        for (int i = 1; i < size; i++) {
            if (isRoot(i)) {
                Vertex v = new Vertex(this, i);
                vlist.add(v);
            }
        }
        return (Vertex[])vlist.toArray(new Vertex[0]);
    }

    public void printDomTree(PrintClient client) {
        message("begin analysis");
        complete();
        this.client = client;
        findStrongComponents();
        message("after find strong components");
        for (int i = 1; i < size; i++) {
            if (isRoot(i)) {
                addEdgeByIndex(0, i);
            }
        }
        NewGraph domTree = findDominators();
        domTree.deleted.set(0);
        domTree.print(client);
    }

    static NewGraph random(int seed, int vertices, int edges) {
        Random random = new Random(seed);
        NewGraph graph = new NewGraph();
        brute = new BruteGraph();
        brute.prune = 0;
        System.out.println("digraph G {");

        for (int i = 1; i <= vertices; i++) {
            graph.addVertex(i, 0);
            brute.addVertex(i);
        }
        brute.sort();
        for (int i = 0; i < edges;) {
            int fromId = random.nextInt(vertices) + 1;
            int toId = random.nextInt(vertices) + 1;

            if (fromId == toId)
                continue;
            graph.addEdge(fromId, toId);
            brute.addEdgeByIndex(fromId - 1, toId, true);
            System.out.println("    " + fromId + " -> " + toId + ";");
            i++;
        }
        System.out.println("}");
        return graph;
    }

    static BruteGraph brute;

    public static void main(String args[]) {
        /*
        int size = Integer.parseInt(args[0]);
        int seed = Integer.parseInt(args[1]);
        int ratio = Integer.parseInt(args[2]);
        log("create random graph, size " + size + " seed " + seed + " ratio " + ratio + "/" + (ratio - 1));
        Graph graph = random(seed, size, (size*ratio)/(ratio - 1));
        log("finished creating random graph");
        graph.print(new PrintClient() {
            public String getName(int id) {
                return "" + id;
            }
        });
        log("finished printing random graph");
        brute.print(new PrintClient() {
            public String getName(int id) {
                return "" + id;
            }
        });
        log("finished printing brute graph");
        //System.out.println("" + brute);
        */
        NewGraph graph = new NewGraph();
        graph.addEdge(0, 1);
        graph.addEdge(0, 2);
        graph.addEdge(1, 3);
        graph.addEdge(1, 6);
        graph.addEdge(2, 4);
        graph.addEdge(2, 7);
        graph.addEdge(3, 5);
        graph.addEdge(3, 6);
        graph.addEdge(4, 2);
        graph.addEdge(4, 7);
        graph.addEdge(5, 8);
        graph.addEdge(5, 10);
        graph.addEdge(6, 9);
        graph.addEdge(7, 12);
        graph.addEdge(8, 11);
        graph.addEdge(9, 8);
        graph.addEdge(10, 11);
        graph.addEdge(11, 12);
        graph.addEdge(11, 1);
        graph.print(new PrintClient() {
            char ids[] = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'};
            public String getName(int id) {
                return "" + ids[id];
            }
        });
        System.exit(1);
        graph.client = new PrintClient() {
            char ids[] = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'};
            public String getName(int id) {
                return "" + ids[id];
            }
        };
        System.out.println(graph);
    }

    public String toString() {
        String result = (String)dfs(new Visitor() {
            StringBuffer buff = new StringBuffer();
            void visitChild(int v, int u, boolean firstTime, int depth) {
                String idv = client.getName(vertexIds.get(v));
                String idu = client.getName(vertexIds.get(u));
                buff.append("    " + idv + " -> " + idu + ";\n");
            }
            Object result() {
                return buff.toString();
            }
        }, 0);
        return "digraph G {\n" + result + "}";
    }

    /*
     * Following methods are junk.
     */

    public void check() {
    }

    public void activate() {
    }

    public boolean active() {
        return true;
    }

    abstract class Visitor {

        void init() {
        }

        void enterNode(int v, int depth) {
        }

        boolean continueSearch(int v, int depth) {
            return true;
        }

        void visitChild(int v, int u, boolean firstTime, int depth) {
        }

        void visitChildAtExit(int v, int u) {
        }

        void exitNode(int v) {
        }

        boolean ignoreDownEdges() {
            return false;
        }

        Object result() {
            return null;
        }

        boolean sort() {
            return false;
        }

        int getCount(int v) {
            return 0;
        }

        IntEnumeration elements(int v) {
            IntEnumeration e = null;
            int root = edges.get(v);
            if (root != 0) {
                e = TreeBitSet.elements(root);
                if (sort()) {
                    SortedIntEnumeration sortedfore = new SortedIntEnumeration();
                    while (e.hasMoreElements()) {       
                        int u = e.nextInt();            
                        int count = getCount(u);
                        sortedfore.add(u, count);
                    }
                    sortedfore.sort();
                    e = sortedfore;
                }
            }
            return e;
        }
    }
}

class SortedIntEnumeration implements IntEnumeration {
    IntegerArray indexes = new IntegerArray();
    IntegerArray counts = new IntegerArray();
    int index = 0;

    void add(int v, int count) {
        indexes.add(v);
        counts.add(count);
    }

    void sort() {
        counts.reverseSort(indexes);
    }

    public boolean hasMoreElements() {
        return index < counts.size();
    }

    public Object nextElement() {
        return new Integer(nextInt());
    }

    public int nextInt() {
        return indexes.get(index++);
    }

    public void reset() {
        index = 0;
    }
}
        
class Edges implements Serializable {
    IntegerArray edges = new IntegerArray();

    public int get(int index) {
        return edges.get(index);
    }

    public void put(int index, int value) {
        edges.put(index, value);
    }

    public void add(int value) {
        edges.add(value);
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        out.writeInt(edges.size());
        for (int i = 0; i < edges.size(); i++) {
            TreeBitSet.writeRoot(out, edges.get(i));
        }
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        int size = in.readInt();
        edges = new IntegerArray();
        for (int i = 0; i < size; i++) {
            edges.add(TreeBitSet.readRoot(in));
        }
    }

    protected void finalize() throws Throwable {
        for (int i = 0; i < edges.size(); i++) {
            TreeBitSet.free(edges.get(i));
        }
    }
}
