package com.ibm.jvm.findroots;

import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;

/**
 *  This class forms the base of the Print family of classes.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class PrintBase extends Base implements HeapdumpListener, PrintClient {

    String className;
    Heapdump dump = new Heapdump();
    Vector strings = new Vector();
    NewGraph graph = new NewGraph();
    IntegerArray ids = new IntegerArray();
    IntegerArray indexes = new IntegerArray();
    IntegerArray types = new IntegerArray();
    IntegerArray excludeObjects = new IntegerArray();
    Vector excludeClasses = new Vector();
    int totalSize = 0;
    int totalCount = 0;

    String className() {
        return className;
    }

    void start(String args[], String className) {
        this.className = className;
        if (args.length == 0)
            usage();
        parseOptions(args);
        BufferedInputStream bis = null;
        String filename = args[args.length - 1];
        try {
            bis = new BufferedInputStream(new FileInputStream(filename));
        } catch (Exception e) {
            System.err.println("Could not open: " + filename);
            usage();
        }
        System.out.println("For more information on FindRoots visit http://w3.hursley.ibm.com/~dgriff/findroots.html");
        System.out.println("");
        System.out.println("Properties:");
        System.out.println("findroots.maxroots = " + graph.maxroots);
        System.out.println("findroots.depth = " + graph.maxdepth);
        System.out.println("findroots.prune = " + graph.prune);
        System.out.println("findroots.sizematters = " + graph.sizeMatters);
        System.out.println("findroots.exact = " + graph.exact);
        System.out.println("");
        parseStart();
        dump.parse(bis, this);
        ids.sort(new SortListener() {
            public void swap(int a, int b) {
                indexes.swap(a, b);
                types.swap(a, b);
            }
        });
        graph.complete();
        parseEnd();
    }

    public boolean parseOption(String arg, String opt) {
        if ("-excludeobject".equals(arg)) {
            try {
                int addr = Integer.parseInt(opt, 16);
                excludeObjects.add(addr);
            } catch (Exception e) {
                System.err.println(opt + " is not a valid number");
                usage();
            }
            return true;
        } else if ("-excludeclass".equals(arg)) {
            excludeClasses.add(opt);
            return true;
        }
        return false;
    }

    void parseStart() {
    }

    void parseEnd() {
    }

    String getString(int index) {
        return (String)strings.elementAt(index);
    }

    void addObject(int id, int nameIndex, int size, int type) {
        ids.add(id);
        indexes.add(nameIndex);
        types.add(type);
        totalSize += size;
        totalCount++;
    }

    boolean excludeObject(int id) {
        for (int i = 0; i < excludeObjects.size(); i++) {
            if (excludeObjects.get(i) == id) {
                System.out.println("exclude object match found for " + hex(id));
                return true;
            }
        }
        return false;
    }

    boolean excludeClass(int id, String classname) {
        for (int i = 0; i < excludeClasses.size(); i++) {
            String name = (String)excludeClasses.elementAt(i);
            if (name.equals(classname)) {
                System.out.println("exclude class match found for " + hex(id));
                return true;
            }
        }
        return false;
    }

    /*
     * HeapdumpListener methods
     */

    public void stringDump(String string) {
        strings.add(string);
    }

    public void objectDump(int id, int classStringIndex, int flags, int size, int refs[]) {
        if (excludeObject(id) || excludeClass(id, getString(classStringIndex)))
            return;
        graph.addVertex(id, refs, size);
        addObject(id, classStringIndex, size, 1);
    }

    public void classDump(int id, int nameIndex, int flags, int size, int refs[]) {
        if (excludeObject(id))
            return;
        graph.addVertex(id, refs, size);
        addObject(id, nameIndex, size, 2);
    }

    public void primitiveArrayDump(int id, int type, int flags, int size) {
        if (excludeObject(id))
            return;
        graph.addVertex(id, size);
        addObject(id, type, size, 3);
    }

    public void objectArrayDump(int id, int classStringIndex, int flags, int size, int refs[]) {
        if (excludeObject(id))
            return;
        graph.addVertex(id, refs, size);
        addObject(id, classStringIndex, size, 4);
    }

    public int classId(int index) {
        return (indexes.get(index) << 3) | types.get(index);
    }

    public String getClassName(int id) {
        int type = id & 7;
        int index = id >> 3;
        return getName(index, type);
    }

    public String getName(int id) {
        int index = ids.indexOf(id);
        if (index == -1)
            return "(unknown id " + hex(id) + ")";
        int nameIndex = indexes.get(index);
        int type = types.get(index);
        return getName(nameIndex, type);
    }

    public String getName(int index, int type) {
        switch (type) {
        case 1:
            return getString(index);
        case 2:
            return "class " + getString(index);
        case 3:
            return arrayName[index];
        case 4:
            return "array of " + getString(index);
        default:
            assert(false);
        }
        return null;
    }

    String arrayName[] = {"unknown array type", "primitive array", "array of references",
            "unknown array type", "array of boolean", "array of char", "array of float",
            "array of double", "array of byte", "array of short", "array of int",
            "array of long", "unknown array type", "unknown array type",
            "unknown array type", "unknown array type"};
}
