package com.ibm.jvm.findroots;

import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;

/**
 *  This class prints the strong components.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class PrintComponents extends PrintBase {

    NewGraph sc;

    public static void main(String args[]) {
        PrintComponents roots = new PrintComponents();
        roots.start(args, "PrintComponents");
    }

    String[] options() {
        return new String[] {"-excludeobject <addr>", "-excludeclass <name>"};
    }

    String[] optionDescriptions() {
        return new String[] {
            "\tExclude object with this address (hex)",
            "\tExclude all objects of this class (eg java/lang/Object)"
        };
    }

    void parseEnd() {
        sc = graph.findStrongComponents();
        sc.deleted.set(0);
        sc.print(this);
        System.out.println("");
        System.out.println("Now print the contents of the strong components referred to by the #n above");
        for (int v = 0; v < sc.size; v++) {
            int id = sc.vertexIds.get(v);
            if (sc.vertexSizes.get(v) == 1) {
                continue;
            }
            System.out.println("");
            System.out.println("*** #" + id + " ***");
            System.out.println("");
            int ids[] = graph.dagIdToIds(id);
            for (int i = 0; i < ids.length; i++) {
                System.out.println("0x" + hex(ids[i]) + " " + super.getName(ids[i]));
            }
        }
    }

    public String getName(int id) {
        int v = sc.idToIndex(id);
        int size = sc.vertexSizes.get(v);
        if (size == 1) {
            id = graph.dagIdToId(id);
            return "0x" + hex(id) + " " + super.getName(id);
        } else {
            return "#" + id + " " + size;
        }
    }
}

