package com.ibm.jvm.svcdump;


import java.io.*;
import java.util.*;
import java.lang.ref.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.dump.*;

/**
 *  This class represents a z/OS address space.
 */

public class AddressSpace extends DvAddressSpace implements Serializable, RandomAccessStream {
    static int ascbasxb;
    static AddressSpace root;
    IntegerArray addresses = new IntegerArray();
    IntegerArray blocks = new IntegerArray();
    SoftReference blockPointers[];
    int id;
    Dump dump;
    Hashtable traceTable = new Hashtable();
    Hashtable strings = new Hashtable();
    Function traceFunctions[];
    Tcb tcbs[];
    Tcb lastTcb;
    //WeakHashMap blockTable = new WeakHashMap();
    HashMap tcbMap = new HashMap();
    Jvm jvms[];
    int currentOffset;
    int firsttcb = 0;
    int lastAddress;
    BitSet pages = new BitSet();
    MemRange memRanges[];
    Process processes[];
    int cacheMisses;

    /* Start of DvAddressSpace methods */

    public DvMemRanges[] getMemRanges() {
        if (memRanges == null) {
            Vector v = new Vector();
            int startAddress = addresses.get(0);
            int lastAddress = startAddress;

            for (int i = 1; i < addresses.size(); i++) {
                int address = addresses.get(i);

                if (address != (lastAddress + 0x1000)) {
                    int length = lastAddress + 0x1000 - startAddress;
                    MemRange range = new MemRange(startAddress, length);

                    v.addElement(range);
                    startAddress = address;
                }
                lastAddress = address;
            }
            memRanges = (MemRange[]) v.toArray(new MemRange[0]);
        }
        return memRanges;
    }

    public DvProcess[] getProcesses() {
        if (processes == null) {
            processes = new Process[1];
            processes[0] = new Process(this);
        }
        return processes;
    }

    public DvProcess getCurrentProcess() {
        throw new Error("unimplemented method");
    }

    public DvAddress  createAddress(long theAddr) {
        Address address = new Address((int) theAddr);

        return (DvAddress) address;
    }

    public byte[] readBytes(DvAddress address, long len) throws DvAddressException {
        //Address addr = (Address)address;

        int a = (int) address.getAddressAsLong();
        byte b[] = new byte[(int) len];

        for (int i = 0; i < len; i++) {
            try {
                b[i] = readByte(a + i);
            } catch (IOException e) {
                // we allow partial readBytes - as long as they gor something
                // then return it.....  used primarily by find commands
                if (i == 0) {
                    throw new DvAddressException("Could not read: " + hex(a + i));
                } else {
                    i = (int) len;
                }
            }
        }
        return b;
    }

    public byte readByte(DvAddress address) throws DvAddressException {
        //Address addr = (Address)address;
        //int a = (int)addr.address;
        int a = (int) address.getAddressAsLong();
        byte b;

        try {
            b = readByte(a);
        } catch (IOException e) {
            throw new DvAddressException("Could not read: " + hex(a));
        }

        return b;
    }

    public short readShort(DvAddress address) throws DvAddressException {
        //throw new Error("unimplemented method");
        int a = (int) address.getAddressAsLong();

        try {
            return (short) readShort(a);
        } catch (IOException e) {
            throw new DvAddressException("Could not read: " + hex(a));
        }
        //return 0;
    }

    public int readInt(DvAddress address) throws DvAddressException {
        //Address addr = (Address)address;
        //int a = (int)addr.address;
        int a = (int) address.getAddressAsLong();

        try {
            return readInt(a);
        } catch (IOException e) {
            throw new DvAddressException("Could not read: " + hex(a));
        }
    }

    public long readLong(DvAddress address) throws DvAddressException {
        int a = (int) address.getAddressAsLong();

        try {
            long h = readInt(a);
            long l = readInt(a + 4);

            return (h << 32) | l;
        } catch (IOException e) {
            throw new DvAddressException("Could not read: " + hex(a));
        }
    }

    public String id() {
        return "" + id;
    }

    public DvAddress readPointer(DvAddress address) throws DvAddressException {
        // Address addr = (Address)address;
        //int i, a = (int)addr.address;
        int i, a = (int) address.getAddressAsLong();

        try {
            i = readInt(a);
        } catch (IOException e) {
            throw new DvAddressException("Could not read: " + hex(a));
        }
        Address contents = new Address(i);

        return (DvAddress) contents;
    }

    /* End of DvAddressSpace methods */

    /**
     * Add a Jvm to the list if jvmp has not been seen already.
     */
    public Jvm addJvm(int jvmp, boolean isPostRas) throws Exception {
        if (jvms == null) {
            jvms = new Jvm[0];
        }
        for (int i = 0; i < jvms.length; i++) {
            if (jvms[i].jvmp == jvmp)
                return jvms[i];
        }
        Jvm newjvms[] = new Jvm[jvms.length + 1];
        System.arraycopy(jvms, 0, newjvms, 0, jvms.length);
        Jvm jvm = new Jvm(this, jvmp, isPostRas);
        jvms = newjvms;
        jvms[jvms.length - 1] = jvm;
        return jvm;
    }

    public String getString(int address, int length) throws Exception {
        Integer a = new Integer(address);
        String s = (String)strings.get(a);
        if (s == null) {
            byte b[] = new byte[length];
            for (int i = 0; i < length; i++)
                b[i] = (byte)readUnsignedByte(address + i);
            s = Dump.getEbcdicString(b);
            strings.put(a, s);
        }
        return s;
    }

    public boolean isJava() {
        tcbs();
        return jvms != null;
    }

    public void mapTid(int tid, Tcb tcb) {
        tcbMap.put(new Integer(tid), tcb);
    }

    public int zeroBlocks() {
        int n = 0;

        for (int i = 0; i < addresses.size(); i++) {
            if (addresses.get(i) == 0)
                n++;
            else
                break;
        }
        return n;
    }

    public void addTraceEntry(int type, String typeString, int psw, int tcb, long tod) {
        Integer pswInt = new Integer(psw);
        Integer count = (Integer) traceTable.get(pswInt);

        traceTable.put(pswInt, new Integer(count == null ? 1 : count.intValue() + 1));
        if (lastTcb != null && lastTcb.tcbid == tcb) {
            lastTcb.addTraceEntry(type, psw);
        } else if (tcbs() != null && tcb != 0) {
            for (int i = 0; i < tcbs.length; i++) {
                lastTcb = tcbs[i];
                if (lastTcb.tcbid == tcb) {
                    lastTcb.addTraceEntry(type, psw);
                    break;
                }
            }
        }
    }

    public Function[] traceFunctions(Hashtable aTraceTable) {
        Hashtable table = new Hashtable();
        boolean verbose = Dump.debug;

        for (Enumeration num = aTraceTable.keys(); num.hasMoreElements();) {
            Integer iaddress = (Integer) num.nextElement();
            String name = null;

            // start at address plus 4 in case we are at the very first branch instruction
            int address = iaddress.intValue();
            address &= 0x7ffffffc;
            int i = 0;
            for (int a = address + 4; i < 0x1000; a -= 4, i++) {
                try {
                    int eyecatcher = readInt(a);

                    if (eyecatcher == 0xc3c5c5 || eyecatcher == 0x1c3c5c5) {
                        try {
                            name = Dsa.function(this, a - 4);
                        } catch (Exception e) {
                            if (verbose)
                                System.out.println("problem creating function at " + hex(a));
                        }
                        break;
                    }
                } catch (Exception e) {
                    if (verbose) {
                        System.out.println("problem reading " + hex(a));
                    }
                    break;
                }
            }
            if (i == 0x1000) {
                if (verbose) {
                    System.out.println("could not locate eyecatcher for " + 
                        hex(iaddress.intValue()));
                }
            }
            if (name == null)
                name = "unknown function (0x" + hex(address) + ")";
            if (verbose)
                System.out.println("address " + hex(iaddress.intValue()) + " maps to " + name);
            int count = ((Integer) aTraceTable.get(iaddress)).intValue();
            Function function = (Function) table.get(name);

            if (function == null)
                function = new Function(name, count);
            else
                function.addCount(count);
            table.put(name, function);
        }
        Function functions[] = (Function[]) table.values().toArray(new Function[0]);

        Arrays.sort(functions, new Comparator() {
                public int compare(Object o1, Object o2) {
                    Function f1 = (Function) o1;
                    Function f2 = (Function) o2;

                    return f2.entryCount - f1.entryCount;
                }
            }
        );
        return functions;
    }

    public Function[] traceFunctions() {
        if (traceFunctions == null)
            traceFunctions = traceFunctions(traceTable);
        return traceFunctions;
    }

    public boolean isTrace() {
        // read first block eyecatcher
        try {
            int block1 = addresses.get(0);
            int eyecatcher = readInt(block1);

            return eyecatcher == 0xe3e3c3c8;    // TTCH
        } catch (Exception e) {}
        return false;
    }

    AddressSpace(Dump dump, int id) {
        this.dump = dump;
        this.id = id;
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }

    String hexId() {
        return Integer.toHexString(id);
    }

    public String toString() {
        return hexId();
    }

    public boolean validTCB(int tcb, boolean lastDitch) {
        try {
            int eyecatcher = readInt(tcb + 0x100);

            if (eyecatcher == 0xe3c3c240)
                return true;
            if (!lastDitch && Dump.debug) System.out.println("bad eyecatcher: " + hex(eyecatcher));
            return false;
        } catch (Exception e) {
            if (!lastDitch && Dump.debug) System.out.println("could not read " + hex(tcb + 0x100));
            return false;
        }
    }

    public int nextTcb(int tcb) throws Exception {
        return readInt(tcb + 0x74);
    }

    public int prevTcb(int tcb) throws Exception {
        return readInt(tcb + 0xdc);
    }

    public int validTCBs() {
        int count = 0;
        int lasttcb = 0;

        try {
            firsttcb = readInt(ascbasxb + 0x4);
        } catch (Exception e) {
            if (Dump.debug) System.out.println("asid " + hexId() + ": could not read first tcb at " + (ascbasxb + 4));
            return 0;
        }
        try {
            lasttcb = readInt(ascbasxb + 0x8);
        } catch (Exception e) {
            if (Dump.debug) System.out.println("asid " + hexId() + ": could not read last tcb at " + (ascbasxb + 8));
            return 0;
        }
        if (firsttcb == lasttcb) {
            if (Dump.debug) System.out.println("first and last tcb the same! " + hex(firsttcb));
            try {
                firsttcb = root.readInt(ascbasxb + 0x4);
                lasttcb = root.readInt(ascbasxb + 0x8);
            } catch (Exception e) {
                if (Dump.debug) System.out.println("nice try");
                return 0;
            }
            if (firsttcb == lasttcb) {
                if (Dump.debug) System.out.println("first and last tcb STILL the same! " + hex(firsttcb));
                return 0;
            }
        }
        try {
            for (int tcb = firsttcb; tcb != 0; tcb = nextTcb(tcb)) {
                if (Dump.debug) System.out.println("asid " + hexId() + ": process tcb at " + hex(tcb));
                if (!validTCB(tcb, false)) {
                    if (Dump.debug) System.out.println("asid " + hexId() + ": invalid tcb at " + hex(tcb));
                    return 0;
                }
                count++;
                if (tcb == lasttcb)
                    break;
            }
            if (Dump.debug) System.out.println("asid " + hexId() + ": processed all tcbs from " + hex(firsttcb)
                    + " to " + hex(lasttcb));
            return count;
        } catch (Exception e) {
            if (Dump.debug) System.out.println("asid " + hexId() + ": error processing a tcb");
            return 0;
        }
    }

    void lastDitch() {
        // for corrupt tcb chain!
        try {
            int tcb = readInt(ascbasxb + 0x4);

            tcb &= 0xfff00000;
            int end = tcb + 0x100000;
            Vector v = new Vector();

            for (; tcb < end; tcb += 4) {
                if (validTCB(tcb, true)) {
                    v.add(new Tcb(this, tcb));
                }
            }
            tcbs = (Tcb[]) v.toArray(new Tcb[0]);
        } catch (Exception e) {
            if (Dump.debug) System.out.println("could not recover tcbs");
            tcbs = new Tcb[0];
        }
    }

    void addTcb(Tcb tcb) {
        Tcb t[] = new Tcb[tcbs.length + 1];

        System.arraycopy(tcbs, 0, t, 0, tcbs.length);
        t[tcbs.length] = tcb;
        tcbs = t;
    }

    public Tcb[] tcbs() {
        if (tcbs != null)
            return tcbs;
        int count = validTCBs();

        if (count != 0) {
            try {
                tcbs = new Tcb[count];
                int tcb = firsttcb;

                for (int i = 0; i < count; i++) {
                    tcbs[i] = new Tcb(this, tcb);
                    tcb = nextTcb(tcb);
                }
            } catch (Exception e) {
                if (Dump.debug) {
                    System.out.println("in asid " + hexId() + " problem creating tcbs:");
                    e.printStackTrace(System.out);
                }
                return null;
            }
        } else {
            if (Dump.debug) System.out.println("in asid " + hexId() + " corrupt tcb chain...");
            lastDitch();
        }
        if (tcbs == null) return null;
        // now add any zombie java tcbs
        for (int i = 0; i < tcbs.length; i++) {
            Tcb tcb = tcbs[i];

            if (tcb.scannedTid)
                continue;
            int tid = tcb.tid();

            if (tid == 0)
                continue;
            try {
                // follow the forward chain in sys_thread_t
                for (int otid = readInt(tid); otid != 0; otid = readInt(otid)) {
                    Tcb otcb = (Tcb) tcbMap.get(new Integer(otid));

                    if (otcb == null) {
                        otcb = new Tcb(this, 0, otid);
                        mapTid(otid, otcb);
                        if (otcb.isKnownType()) {
                            addTcb(otcb);
                        }
                    } else {
                        if (otcb.scannedTid)
                            break;
                    }
                    otcb.scannedTid = true;
                }
                // follow the backward chain in sys_thread_t
                for (int otid = readInt(tid + 4); otid != 0; otid = readInt(otid + 4)) {
                    Tcb otcb = (Tcb) tcbMap.get(new Integer(otid));

                    if (otcb == null) {
                        otcb = new Tcb(this, 0, otid);
                        mapTid(otid, otcb);
                        if (otcb.isKnownType()) {
                            addTcb(otcb);
                        }
                    } else {
                        if (otcb.scannedTid)
                            break;
                    }
                    otcb.scannedTid = true;
                }
            } catch (Exception e) {}
        }
        return tcbs;
    }

    public void add(int address, int block) {
        if ((address & 0xfff) != 0)
            throw new Error("address " + hex(address) + " is not page aligned!");
        int page = address >>> 12;
        if (pages.get(page)) {
            if (Dump.debug)
                System.out.println("duplicate address " + hex(address) + " for asid " + hexId());
            return;
        }
        pages.set(page);
        addresses.add(address);
        blocks.add(block);
    }

    public void sort() {
        addresses.sort(blocks);
    }

    public final void seek(long pos) throws IOException {
        currentOffset = (int) pos;
    }

    public final int readInt(long pos) throws IOException {
        int offset = (int) pos;
        if (offset == 0 || (offset & 0x80000000) != 0)
            throw new IOException("cannot read " + Integer.toHexString(offset));
        int blockAddress = offset & 0x7ffff000;
        int block[];
        if (blockPointers == null || blockPointers[blockAddress >> 12] == null ||
                (block = (int[])blockPointers[blockAddress >> 12].get()) == null) {
            int index = addresses.indexOf(blockAddress);
            if (index == -1) throw new IOException("cannot read " + Integer.toHexString(offset));
            long blockOffset = blocks.get(index);
            blockOffset &= 0xffffffffL;
            dump.seek(blockOffset + dump.HEADERSIZE);
            block = dump.readBlock();
            if (blockPointers != null) {
                blockPointers[blockAddress >> 12] = new SoftReference(block);
            }
            if (cacheMisses++ > 1000 && blockPointers == null) {
                blockPointers = new SoftReference[0x80000];
            }
        }
        int blockOffset = (offset & 0xfff) >> 2;
        int result = 0;

        switch (offset & 3) {
        case 0:
            result = block[blockOffset];
            break;

        case 1:
            result = (block[blockOffset] << 8) | (readInt((offset & 0xfffffffc) + 4) >>> 24);
            break;

        case 2:
            result = (block[blockOffset] << 16) | (readInt((offset & 0xfffffffc) + 4) >>> 16);
            break;

        case 3:
            result = (block[blockOffset] << 24) | (readInt((offset & 0xfffffffc) + 4) >>> 8);
            break;
        }
        return result;
    }

    public final int readInt() throws IOException {
        int result = readInt(currentOffset);

        currentOffset += 4;
        return result;
    }

    public final int readUnsignedByte() throws IOException {
        int result = readInt(currentOffset) >>> 24;

        currentOffset += 1;
        return result;
    }

    public final int readUnsignedByte(long offset) throws IOException {
        return readInt(offset) >>> 24;
    }

    public final byte readByte(long offset) throws IOException {
        return (byte) readUnsignedByte(offset);
    }

    public final byte readByte() throws IOException {
        return (byte) readUnsignedByte();
    }

    public int readShort(int offset) throws IOException {
        return readInt(offset) >>> 16;
    }

    public String readString(int offset) throws IOException {
        StringBuffer buff = new StringBuffer();
        int c;

        while ((c = readUnsignedByte(offset++)) != 0) {
            buff.append((char) c);
        }
        return buff.toString();
    }

    public String readStringObject(int str) throws IOException {
        StringBuffer buff = new StringBuffer();
        int value = readInt(str + 8);
        int offset = readInt(str + 12);
        int count = readInt(str + 16);
        for (int i = 0; i < count; i++) {
            char c = (char)readShort(value + 8 + offset + (i << 1));
            if (c != '\n')
                buff.append(c);
        }
        return buff.toString();
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        out.writeInt(id);
        out.writeObject(addresses);
        out.writeObject(blocks);
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        id = in.readInt();
        addresses = (IntegerArray) in.readObject();
        blocks = (IntegerArray) in.readObject();
        traceTable = new Hashtable();
        strings = new Hashtable();
        //blockTable = new WeakHashMap();
        //blockPointers = new SoftReference[addresses.size()];
        tcbMap = new HashMap();
    }

    /**
     *  Return the pointer size for this system (Philr 10/07/02)
     */
    public int pointerSize() {
        return 4;
    }
}

