package com.ibm.jvm.svcdump;

import java.util.*;
import com.ibm.jvm.util.*;

/**
 * This class represents an internal Java class structure.
 */
public class ClassClass extends Base {
    Jvm jvm;
    int cb;
    String type;
    String name;
    int staticReferences[];
    AddressSpace space;

    ClassClass(Jvm jvm, int cb, String name, String type) {
        this.jvm = jvm;
        this.cb = cb;
        this.type = type;
        this.name = name;
        space = jvm.space;
    }

    ClassClass(Jvm jvm, int cb) throws Exception {
        name = jvm.cbName(cb);
        this.jvm = jvm;
        this.cb = cb;
        space = jvm.space;
    }

    int[] staticReferences() throws Exception {
        if (staticReferences != null)
            return staticReferences;
        int fields = cbDeclaredFields();
        int fields_count = cbFieldsCount();
        IntegerArray refs = null;
        if (fields != 0 && fields_count > 0) {
            for (int i = 0; i < fields_count; i++) {
                int fb = fields + (i * 24);
                if (fbIsStatic(fb) && fbIsReference(fb)) {
                    int refp = fbStaticValueAddress(fb);
                    int ref = space.readInt(refp);
                    if (ref != 0 && jvm.stIsValidHandle(ref)) {
                        if (refs == null)
                            refs = new IntegerArray();
                        refs.add(ref);
                    }
                }
            }
        }
        staticReferences = (refs == null) ? new int[0] : refs.toArray();
        return staticReferences;
    }

    int getStaticField(String fieldName) throws Exception {
        int fields = cbDeclaredFields();
        int fields_count = cbFieldsCount();
        if (fields != 0 && fields_count > 0) {
            for (int i = 0; i < fields_count; i++) {
                int fb = fields + (i * 24);
                if (fbIsStatic(fb) && fbIsReference(fb) && fbName(fb).equals(fieldName)) {
                    int refp = fbStaticValueAddress(fb);
                    int ref = space.readInt(refp);
                    if (ref != 0 && jvm.stIsValidHandle(ref)) {
                        return ref;
                    }
                }
            }
        }
        return -1;
    }

    int cbObjectOffsets() throws Exception {
        return space.readInt(cb + 184);
    }

    int cbDeclaredMethods() throws Exception {
        return space.readInt(cb + 100);
    }

    int cbDeclaredFields() throws Exception {
        return space.readInt(cb + 104);
    }

    int cbMethodsCount() throws Exception {
        return space.readShort(cb + 146);
    }

    int cbFieldsCount() throws Exception {
        return space.readShort(cb + 148);
    }

    int cbSuperclass() throws Exception {
        return space.readInt(cb + 76);
    }

    boolean fbIsStatic(int fb) throws Exception {
        return (space.readShort(fb + 12) & 8) != 0;
    }

    boolean fbIsReference(int fb) throws Exception {
        return (space.readShort(fb + 14) & 2) != 0;
    }

    int fbStaticValueAddress(int fb) throws Exception {
        return space.readInt(fb + 20);
    }

    String mbName(int mb) throws Exception {
        int ptr = space.readInt(mb + 8);
        return space.readString(ptr);
    }

    String fbName(int fb) throws Exception {
        return mbName(fb);
    }

    public String toString() {
        return name;
    }

    public ClassMethod[] getMethods() throws Exception {
        int count = cbMethodsCount();
        ClassMethod m[] = new ClassMethod[count];
        int methods = cbDeclaredMethods();
        for (int i = 0; i < count; i++) {
            int mb = methods + (i * 100);
            m[i] = new ClassMethod(this, mb);
        }
        return m;
    }
}
