package com.ibm.jvm.svcdump;

import java.util.*;
import com.ibm.jvm.util.*;

/**
 * This class represents an internal Java method structure.
 */
public class ClassMethod extends Base {
    Jvm jvm;
    int mb;
    AddressSpace space;
    ClassClass classclass;

    ClassMethod(ClassClass classclass, int mb) {
        this.jvm = classclass.jvm;
        this.mb = mb;
        this.classclass = classclass;
        space = jvm.space;
    }

    ClassMethod(Jvm jvm, int mb) throws Exception {
        space = jvm.space;
        int cb = space.readInt(mb);
        // note should hashtable these
        classclass = new ClassClass(jvm, cb);
        this.jvm = jvm;
        this.mb = mb;
    }

    String fullname() throws Exception {
        return classclass + "." + this;
    }

    String mbName() throws Exception {
        int ptr = space.readInt(mb + 8);
        return space.readString(ptr);
    }

    String mbSignature() throws Exception {
        int ptr = space.readInt(mb + 4);
        return space.readString(ptr);
    }

    public String toString() {
        try {
            return mbName() + " " + mbSignature();
        } catch (Exception e) {
            throw new Error("error getting mbName: " + e);
        }
    }

    public int mbLocalVariableCount() throws Exception {
        return space.readShort(mb + 68);
    }
}
