package com.ibm.jvm.svcdump;

import java.util.*;

/**
 *  This class represents a single native stack frame (or Data Save Area, DSA)
 */

public class Dsa extends Base {
    int address;
    Dsa previous;
    AddressSpace space;
    Tcb tcb;
    int registers[] = new int[16];
    String function;
    String program;
    String service;
    static int micb = 0;
    boolean foundPrevious;
    int prevCount = -1;
    boolean foundPrevCount;
    boolean is118;
    static String knownFunctions[] = {
            // comment out the really common ones
            "CEEOPEF", "pthread exit forced",
            "CEEHSGL", "raise a condition",
            "CEEOSIGG", "deliver signal",
            "CEEOPCT", "pthread_cond_timedwait",
            "CEEOPML2", "pthread_mutex_lock",
            "CEEOPMU2", "pthread_mutex_unlock",
            "CEEPMDEL", "delete a load module",
            "CEEVGSTR", "get main storage",
            //"CEEOPCW",        "pthread_cond_wait",
            "CEEOPGS", "pthread_getspecific",
            "CEECZLOD", "process level load under CICS",
            "CEEPVDEL", "delete a module",
            "CEEVGQP", "get heap pool storage",
            "CEEOPCS", "pthread_cond_signal",
            "CEEOPMI", "pthread_mutex_init",
            "CEEPVLOD", "load module",
            "CEEVFQP", "free heap pool storage",
            "CEEVSTLO", "get heap latch",
            "CEEVCZST", "realloc",
            "CEEHDSP", "exception dispatcher",
            "CEEOPMD", "pthread_mutex_destroy",
            "CEEHHDLR", "register user condition handler",
            "CEEVGTST", "get heap storage",
            "CEEVGTSI", "extend stack",
            //"CEEPGTFN",        "get function address and swap WSA",
            "CEEVFSTR", "free main storage",
            "CEEPTLC", "load dll",
            //"CEEOPCMM",        "pthread_create mini main",
            "CEEBTERM", "batch termination",
            "CEEM@MOU", "output MVS message",
            "CEEV#GTS", "get heap storage",
            "CEEOGLAT", "get contended latch",
            "CEEORLAT", "release contended latch",
            "CEEBBEXT", "bootstrap routine",
            "CEEOTERM", "posix termination",
            "CEEZRET", "thread termination",
            "EDCZMINV", "main invocation event handler",
            "CEEHSFXS", "stack frame exit schedule",
            "mmipSelectInvokeJavaMethod", "java",
            "mmipSelectInvokeSynchronizedJavaMethod", "java",
            "mmipInvokeJavaMethod", "java",
            "mmipInvokeSynchronizedJavaMethod", "java"
        };
    static Hashtable knownFunctionsTable = new Hashtable();

    static {
        for (int i = 0; i < knownFunctions.length; i += 2)
            knownFunctionsTable.put(knownFunctions[i], knownFunctions[i + 1]);
    }

    public int stackPointer() {
        return address;
    }

    public String functionName() {
        return function();
    }

    Dsa(Tcb tcb, int address) throws Exception {
        this.tcb = tcb;
        this.space = tcb.space;
        this.address = address;
        registers[14] = space.readInt(address + 12);
        registers[15] = space.readInt(address + 16);
        for (int i = 0; i < 14; i++)
            registers[i] = space.readInt(address + 20 + (i * 4));
    }

    public Dsa previous() {
        if (!foundPrevious) {
            foundPrevious = true;
            try {
                int prev = space.readInt(address + 4);

                if (prev != 0 && prev != address) {
                    previous = new Dsa(tcb, prev);
                }
            } catch (Exception e) {}
        }
        return previous;
    }

    public static String function(AddressSpace space, int ep) throws Exception {
        int eyecatcher;
        String function = null;

        try {
            eyecatcher = space.readInt(ep + 4);
            if (eyecatcher == 0xc3c5c5 || eyecatcher == 0x1c3c5c5) {
                int ppa1 = ep + space.readInt(ep + 12);
                int offset = space.readUnsignedByte(ppa1);

                if (eyecatcher == 0x1c3c5c5) offset <<= 1;
                function = readFunctionName(space, ppa1, offset);
            } else if ((eyecatcher & 0xff0000) == 0xce0000) {
                int ppa1 = ep;
                int offset = eyecatcher >>> 24;

                function = readFunctionName(space, ppa1, offset);
            } else if ((space.readInt(ep) & 0xfffff000) == 0x47f0f000) {
                int length = space.readUnsignedByte(ep + 4);
                byte b[] = new byte[length];

                for (int i = 0; i < length; i++)
                    b[i] = (byte) space.readUnsignedByte(ep + 5 + i);
                function = Dump.getEbcdicString(b);
                for (int i = 0; i < function.length(); i++) {
                    if (!Character.isLetterOrDigit(function.charAt(i))) {
                        function = "(unknown)";
                        break;
                    }
                }
            } else
                function = "(unknown)";
        } catch (Exception e) {
            throw e;
        }
        String description = (String) knownFunctionsTable.get(function);

        if (description != null) {
            if (!description.equals("java"))
                function += "    (" + description + ")";
        }
        return function;
    }

    /**
     * Returns the function name for this Dsa entry or "(unknown)" if it can't be calculated.
     */
    public String function() {
        if (function != null)
            return function;

        function = "(unknown)";

        int ep = entryPoint();
        if (ep == 0) {
            return function;
        }
        try {
            function = function(space, ep);
            if (Dump.printArgs) {
                int r1 = previous().registers[1];

                function += "(";
                for (int i = 0; i < 4; i++) {
                    try {
                        int arg = space.readInt(r1 + (i << 2));

                        function += hex(arg);
                    } catch (Exception e) {
                        function += "?";
                    }
                    if (i == 3)
                        function += ") ";
                    else
                        function += ", ";
                }
            }
            if (Dump.debug) function += " (ep = " + hex(ep) + ")";
        } catch (Exception e) {}
        return function;
    }

    public String program() {
        if (program != null)
            return program;
        program = "";
        try {
            int ep = entryPoint();
            int eyecatcher = space.readInt(ep + 4);
            if (eyecatcher == 0xc3c5c5 || eyecatcher == 0x1c3c5c5) {
                int ppa1 = ep + space.readInt(ep + 12);
                int ppa2;
                if (eyecatcher == 0x1c3c5c5) {
                    ppa2 = ep + space.readInt(ppa1 + 4);
                    int offset = space.readInt(ppa2 + 0x8);
                    if (offset != 0) {
                        int ppa4 = ppa2 + offset;
                        int length = space.readShort(ppa4 + 12);
                        if (length > 1000)
                            throw new Exception("bad length: " + length);
                        program = space.getString(ppa4 + 14, length);
                    }
                } else {
                    ppa2 = space.readInt(ppa1 + 4);
                    int ppa4 = space.readInt(ppa2 + 0x8);
                    if (ppa4 != 0) {
                        int length = space.readShort(ppa4 + 12);
                        assert(length < 1000);
                        program = space.getString(ppa4 + 14, length);
                    }
                }
            }
        } catch (Exception e) {
        }
        return program;
    }

    public String service() {
        if (service != null)
            return service;
        try {
            int ep = entryPoint();
            int eyecatcher = space.readInt(ep + 4);
            if (eyecatcher == 0xc3c5c5 || eyecatcher == 0x1c3c5c5) {
                int ppa1 = ep + space.readInt(ep + 12);
                int ppa2;
                if (eyecatcher == 0x1c3c5c5) {
                    ppa2 = ep + space.readInt(ppa1 + 4);
                    int tsta = ppa2 + space.readInt(ppa2 + 0xc);
                    int length = space.readShort(tsta + 0x14);
                    byte b[] = new byte[length];

                    for (int i = 0; i < length; i++)
                        b[i] = (byte) space.readUnsignedByte(tsta + 0x16 + i);
                    service = Dump.getEbcdicString(b);
                } else {
                    ppa2 = space.readInt(ppa1 + 4);
                    int tsta = space.readInt(ppa2 + 0xc);
                }
                //program = function(space, ep);
            } else if ((eyecatcher & 0xff0000) == 0xce0000) {
                int ppa1 = ep;
                int ppa2 = space.readInt(ppa1 + 4);
                ep = space.readInt(ppa2 + 0x10);
                service = function(space, ep);
            } else
                service = "(unknown)";
        } catch (Exception e) {
            service = "(unknown)";
        }
        return service;
    }

    /** Search for the top dsa */
    public static Dsa findTop(Tcb tcb, int address) throws Exception {
        int topDsa = address;

        for (;;) {
            try {
                int forward = tcb.space.readInt(topDsa + 0x4c);

                if (tcb.space.readInt(forward + 4) != topDsa)
                    break;
                topDsa = forward;
            } catch (Exception e) {
                break;
            }
        }
        return new Dsa(tcb, topDsa);
    }

    /** Search for a stack around given address */
    public static Dsa findStack(Tcb tcb, int address) {
        Dsa bestDsa = null;

        for (int dsa = address; dsa > address - 0x1000; dsa -= 4) {
            try {
                int forward = tcb.space.readInt(dsa + 0x4c);

                if (tcb.space.readInt(forward + 4) != dsa)
                    continue;
                Dsa tryDsa = findTop(tcb, dsa);

                if (bestDsa == null || bestDsa.prevCount() < tryDsa.prevCount())
                    bestDsa = tryDsa;
            } catch (Exception e) {}
        }
        return bestDsa;
    }

    public int prevCount() {
        if (!foundPrevCount) {
            foundPrevCount = true;
            for (Dsa dsa = this; dsa != null; dsa = dsa.previous()) {
                prevCount++;
                //if (prevCount > 1000) {
                    //System.out.println("Warning! dodgy prev count in tcb " + tcb);
                    //break;
                //}
            }
        }
        return prevCount;
    }

    boolean badName(String name) {
        if (name.equals("."))
            return true;
        for (int i = 0; i < name.length(); i++) {
            if (Character.isIdentifierIgnorable(name.charAt(i))) {
                return true;
            }
        }
        return false;
    }

    /** 
     * Returns the Java method name if any.
     */
    public String method() {
        function();
        String method = "";
        try {
            // If it is an MMI function, add the Java method
            if (function.startsWith("mmipInvoke") || function.startsWith("mmisInvoke") ||
                function.startsWith("mmipSelectInvoke")) {
                method = cbName(false) + "." + mbName(false);
            } else if (function.equals("EXECJAVA") || function.equals("mmipExecuteJava")) {
                method = cbName(false) + "." + mbName(false);
                if (badName(method)) {
                    method = cbName(true) + "." + mbName(true);
                    if (badName(method)) {
                        method = "";
                    }
                }
            } else if (false && function.equals("sysMonitorWait")) {
                int r1 = previous().registers[1];
                int mid = space.readInt(r1 + 4);
                int tid = space.readInt(mid);

                function += "(owner = " + hex(tid) + ")";
            } else if (false && micb == 0 && function.equals("lkMonitorEnter")) {
                int r0 = previous().registers[0];

                System.err.println("found lk, wsa = " + hex(r0));
                String s = new String("monEnumerateCacheLocked");
                byte b[] = s.getBytes("Cp500");
                int ep = entryPoint();
                int monep = 0;
                int fp = ep;
                int bp = ep;

                for (;;) {
                    int i;

                    for (i = 0; i < b.length; i++) {
                        if ((byte) space.readUnsignedByte(fp + i) != b[i])
                            break;
                    }
                    if (i == b.length) {
                        monep = fp;
                        break;
                    }
                    for (i = 0; i < b.length; i++) {
                        if ((byte) space.readUnsignedByte(bp + i) != b[i])
                            break;
                    }
                    if (i == b.length) {
                        monep = bp;
                        break;
                    }
                    fp++;
                    bp--;
                }
                System.err.println("found mon! " + hex(monep));
                monep &= 0xfffffffe;
                int staticOff = 0;

                for (;; monep += 2) {
                    staticOff = space.readInt(monep);
                    if (staticOff == 0x10638) {
                        staticOff = space.readInt(monep + 8);
                        System.err.println("found " + hex(staticOff));
                        break;
                    }
                }
                micb = space.readInt(r0 + staticOff + 0x6c);
                System.err.println("found micb " + hex(micb));
            }
        } catch (Exception e) {
        }
        return method;
    }

    public static String readFunctionName(AddressSpace space, int ppa1, int offset) throws Exception {
        int length = space.readShort(ppa1 + offset);

        //System.out.println("ppa1 = " + hex(ppa1) + " offset = " + offset + " length = " + length);
        if (length > 500)
            throw new Exception("Impossibly long name: " + length);
        byte b[] = new byte[length];

        for (int i = 0; i < length; i++)
            b[i] = (byte) space.readUnsignedByte(ppa1 + offset + 2 + i);
        return Dump.getEbcdicString(b);
    }

    String mbName(boolean useJavaFrame) {
        String name = "";

        try {
            int mb = 0;

            if (useJavaFrame) {
                int r1 = previous().registers[1];
                int ee = space.readInt(r1);        // first arg
                int javaframe = space.readInt(ee + 8);
                int cp = space.readInt(javaframe + 0);

                //mb = space.readInt(javaframe + 28);
                mb = space.readInt(cp + 4);
            } else {
                mb = previous().registers[6];
            }
            int mbName = space.readInt(mb + 8);
            StringBuffer buff = new StringBuffer();
            int c;

            while ((c = space.readUnsignedByte(mbName++)) != 0) {
                buff.append((char) c);
            }
            name = buff.toString();
        } catch (Exception e) {}
        return name;
    }

    String cbName(boolean useJavaFrame) {
        String name = "";

        try {
            int mb = 0;

            if (useJavaFrame) {
                int r1 = previous().registers[1];
                int ee = space.readInt(r1);        // first arg
                int javaframe = space.readInt(ee + 8);
                // surely current frame isn't current any longer???
                int cp = space.readInt(javaframe + 0);

                //mb = space.readInt(javaframe + 28);
                mb = space.readInt(cp + 4);
            } else {
                mb = previous().registers[6];
            }
            int clazz = space.readInt(mb);

            name = tcb.jvm.cbName(clazz);
        } catch (Exception e) {}
        return name;
    }

    public int entryPoint() {
        if (previous() == null)
            return 0;
        if (previous().address == tcb.failingDsa()) {
            try {
                int forward = space.readInt(previous().address + 0x4c);
                int r15 = space.readInt(forward + 0x10);
                return r15 & 0x7fffffff;
            } catch (Exception e) {}
        }
        return previous().registers[15] & 0x7fffffff;
    }

    public int address() {
        return address;
    }

    public int returnAddress() {
        return registers[14] & 0x7fffffff;
    }

    public int offset() {
        if (address == tcb.failingDsa()) {
            int psw = tcb.failingPsw();
            if (psw != 0)
                return psw - entryPoint();
        }
        return returnAddress() - entryPoint() - 2;
    }

    public int reg(int reg) {
        return registers[reg];
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }
}
