package com.ibm.jvm.svcdump;

import java.io.*;
import java.util.*;
//import org.python.util.*;
//import org.python.core.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.util.html.*;
import com.ibm.jvm.dump.*;
import com.ibm.jvm.zseries.*;
import com.ibm.jvm.findroots.*;

/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Dump extends DvDump {

    RandomAccessFile raf;
    AddressSpace spaces[] = new AddressSpace[0];
    AddressSpace validSpaces[];
    static final int HEADERSIZE = 0x40;
    static final int BLOCKSIZE = 0x1000 + HEADERSIZE;
    static final int DR1 = 0xc4d9f140;
    static final int DR2 = 0xc4d9f240;
    boolean old;
    public static boolean verbose = false;
    public static boolean debug = false;
    static boolean printArgs = false;
    static boolean analyzeHeap = false;
    static boolean printAllocCache = false;
    static boolean printExceptionObject = false;
    static boolean disassemble = false;
    static int disassembleAddress;
    static int disassembleLength;
    static boolean dumpstorage = false;
    static int dumpstorageAddress;
    static int dumpstorageLength;
    static boolean dumpclasses = false;
    static int dumpclass = 0;
    static boolean dumpproperties = false;
    static boolean dumpmdata = false;
    static int mdata = 0;
    static boolean systrace = false;
    static int limit = 10;
    static Hashtable table = new Hashtable();
    static String dvDumpName;
    Graph graph;
    Document doc = new Document();

    /* Start of DvDump methods */

    public Dump() throws Exception {
        this(dvDumpName, null);
    }

    public static Boolean isSupportedSource(String theDumpSource) {
        dvDumpName = theDumpSource;
        return new Boolean(validFile(new File(theDumpSource)));
    }
    
    public DvAddressSpace[] getAddressSpaces() {
        return addressSpaces();
    }
    
    public DvAddressSpace getCurrentAddressSpace() {
        throw new Error("unimplemented method");
    }
    
    public int getArchitecture() {
        return DvDump.ARCHITECTURE_IS_32BIT | DvDump.ARCHITECTURE_IS_BIGENDIAN;
    }
    
    public int getSystemType() {
        return DvDump.SYSTEM_IS_OS390;
    }
    
    public int getSubSystemType() {
        return DvDump.SUBSYS_OS390;
    }
    
    public int getProcessor() {
        throw new Error("unimplemented method");
    }
    
    public int getProcessorSubType() {
        throw new Error("unimplemented method");
    }
    
    public int getCurrentProcessor() {
        throw new Error("unimplemented method");
    }
    
    public int getProcessorCount() {
        throw new Error("unimplemented method");
    }

    /* End of DvDump methods */

    boolean validBlock(int id) {
        if (id == DR2) {
            old = false;
        } else if (id == DR1) {
            old = true;
        } else {
            return false;
        }
        return true;
    }

    static boolean validFile(File file) {
        try {
            RandomAccessFile raf = new RandomAccessFile(file, "r");
            int magic = raf.readInt();

            raf.close();
            return magic == DR1 || magic == DR2;
        } catch (Exception e) {
            return false;
        }
    }
 
    static int maxBlocks(File file) {
        return (int) file.length() / BLOCKSIZE;
    }

    void scanFile() throws Exception {
        AddressSpace space = null;

        try {
            byte header[] = new byte[28];

            for (long offset = 0;; offset += BLOCKSIZE) {
                seek(offset);
                raf.readFully(header);        // No EOF if you just do read()
                int id = getInt(header, 0);

                if (!validBlock(id)) {
                    doc.println("No DR2 found at offset " + hex((int)offset) +
                        " instead found " + hex(id));
                    offset = offset - BLOCKSIZE + 0xb50;
                    seek(offset);
                    raf.readFully(header);
                    id = getInt(header, 0);
                    if (!validBlock(id)) {
                        offset -= 0xb50;
                        offset += 4;
                        seek(offset);
                        while (!validBlock(readInt())) offset += 4;
                        seek(offset);
                        raf.readFully(header);
                    }
                }
                int asid = getInt(header, 3);
                int address = getInt(header, old ? 5 : 6);

                space = getAddressSpace(asid);
                space.add(address, (int)offset);
            }
        } catch (EOFException e) {};
    }

    public Dump(String name, Graph graph) throws Exception {
        this.graph = graph;
        raf = new RandomAccessFile(name, "r");
        try {
            FileInputStream fis = new FileInputStream(name + ".cache");
            ObjectInputStream ois = new ObjectInputStream(fis);

            spaces = (AddressSpace[]) ois.readObject();
            ois.close();
        } catch (Exception e) {
            FileOutputStream fos = new FileOutputStream(name + ".cache");
            ObjectOutputStream oos = new ObjectOutputStream(fos);

            scanFile();
            oos.writeObject(spaces);
            oos.close();
        }
        AddressSpace space = null;

        if (debug) doc.println("finished scanning dump");

        /* Sort the addresses in the address spaces */
        for (int i = 0; i < spaces.length; i++) {
            spaces[i].dump = this;
            spaces[i].sort();
        }

        /* Figure out which one is the root address space and save the ascbasxb */
        for (int i = 0; i < spaces.length; i++) {
            space = spaces[i];
            try {
                int psaaold = space.readInt(0x224);

                try {
                    int ascbasxb = space.readInt(psaaold + 0x6c);

                    if (space.id == 1) {
                        if (debug) doc.println("ascb = " + hex(ascbasxb));
                        AddressSpace.ascbasxb = ascbasxb;
                        AddressSpace.root = space;
                    }
                } catch (Exception e) {
                    if (debug) doc.println("inner failed for " + space.hexId());
                }
            } catch (Exception e) {
                if (debug) doc.println("outer failed for " + space.hexId());
            }
        }
    }

    public Dump(String name) throws Exception {
        this(name, new Graph());
    }

    static int getInt(byte b[], int offset) {
        offset <<= 2;
        int b1 = b[offset] << 24;
        int b2 = (b[offset + 1] << 16) & 0xffffff;
        int b3 = (b[offset + 2] << 8) & 0xffff;
        int b4 = b[offset + 3] & 0xff;

        return b1 | b2 | b3 | b4;
    }

    void seek(long offset) throws IOException {
        raf.seek(offset);
    }

    AddressSpace getAddressSpace(int id) {
        AddressSpace space;

        /* No point in a hashtable, there are only a small number */
        for (int i = 0; i < spaces.length; i++) {
            space = spaces[i];
            if (space.id == id)
                return space;
        }
        AddressSpace newspaces[] = new AddressSpace[spaces.length + 1];

        System.arraycopy(spaces, 0, newspaces, 0, spaces.length);
        space = new AddressSpace(this, id);
        newspaces[spaces.length] = space;
        spaces = newspaces;
        return space;
    }

    public int readInt() throws IOException {
        return raf.readInt();
    }

    public int[] readBlock() throws IOException {
        int block[] = new int[0x400];
        byte b[] = new byte[0x1000];

        raf.readFully(b);
        for (int i = 0; i < block.length; i++)
            block[i] = getInt(b, i);
        return block;
    }

    static String hex(int word, int pad) {
        String s = Integer.toHexString(word);

        for (int i = s.length(); i < pad; i++)
            s = "0" + s;
        return s;
    }

    static String hex(int word) {
        return hex(word, 8);
    }

    static byte fix(int word) {
        int w = word & 0xff;

        if (w < 0x40 || (w > 0x40 && w < 0x4b) || (w > 0x50 && w < 0x5a) || (w > 0x61 && w < 0x6a) ||
            (w > 0x6f && w < 0x79) || (w > 0xa9 && w < 0xbc) || w > 0xf9)
            w = 0x40;
        byte b = (byte) w;

        return b;
    }

    public static String ebcdic(int word) {
        byte b[] = new byte[4];

        b[0] = fix(word >> 24);
        b[1] = fix(word >> 16);
        b[2] = fix(word >> 8);
        b[3] = fix(word >> 0);

        return getEbcdicString(b);;
    }

    public final boolean isEbcdic(short b) {
        b |= 0x40;
        return (b >= 0xc1 && b <= 0xc9) || (b >= 0xd1 && b <= 0xd9) || (b >= 0xa2 && b <= 0xa9) ||
            (b >= 0xf0 && b <= 0xf9);
    }

    public final boolean isEbcdic(int word) {
        short s1 = (short) (word & 0xff);
        short s2 = (short) ((word >> 8) & 0xff);
        short s3 = (short) ((word >> 16) & 0xff);
        short s4 = (short) ((word >> 24) & 0xff);

        return isEbcdic(s1) && isEbcdic(s2) && isEbcdic(s3) && isEbcdic(s4);
    }

    public AddressSpace[] addressSpaces() {
        if (validSpaces != null)
            return validSpaces;
        Vector v = new Vector();

        for (int i = 0; i < spaces.length; i++) {
            AddressSpace space = spaces[i];
            Tcb tcbs[] = space.tcbs();

            if (tcbs != null) {
                v.add(space);
            }
        }
        validSpaces = (AddressSpace[]) v.toArray(new AddressSpace[1]);
        Arrays.sort(validSpaces, new Comparator() {
                public int compare(Object o1, Object o2) {
                    AddressSpace s1 = (AddressSpace) o1;
                    AddressSpace s2 = (AddressSpace) o2;

                    return s2.tcbs().length - s1.tcbs().length;
                }
            }
        );
        return validSpaces;
    }

    public void processTrace(AddressSpace space) throws Exception {
        if (debug) doc.println("found trace in asid 0x" + space.hexId());
        int ttch = space.addresses.get(0);
        int numProcessors = space.readShort(ttch + 0x1c);
        int todtop = space.readInt(ttch + 0x20);
        long tod = todtop >> 16;
        tod <<= 48;
        AddressSpace lastSpace = null;

        for (int proc = 0; proc < numProcessors; proc++) {
            if (debug)
                doc.println("*** processor " + proc + " trace table ***");
            int procEntry = ttch + 0x38 + (proc * 8);
            int ttchnumb = space.readShort(procEntry);
            int ttchbptr = space.readInt(procEntry + 4);
            int tcb = 0;

            for (int buffer = 0; buffer < ttchnumb; buffer++) {
                int flags = space.readShort(ttchbptr + (buffer * 40) + 8);

                if (debug)
                    doc.println("*** doing buffer section " + buffer + " flags " + hex(flags));
                int ttchcptr = space.readInt(ttchbptr + (buffer * 40));
                int ttchclen = space.readInt(ttchbptr + (buffer * 40) + 4);
                int len = 0;

                for (int entry = ttchcptr; entry < (ttchcptr + ttchclen); entry += len) {
                    int id = space.readUnsignedByte(entry);

                    if (id < 0x70) {
                        switch (id) {
                        case 0x21:
                            //System.out.println("found program call control");
                            len = 8;
                            break;

                        case 0x32:
                            //System.out.println("found program return control");
                            len = 12;
                            break;

                        case 0x31:
                            //System.out.println("found program transfer control");
                            len = 8;
                            break;

                        case 0x10:
                            //System.out.println("found set secondary");
                            len = 4;
                            break;

                        default:
                            if (id != 0)
                                if (debug) doc.println("!!! unknown id " + hex(id));
                            len = 4;
                            break;
                        }
                        continue;
                    }
                    if (id >= 0x80) {
                        int branch = space.readInt(entry);

                        len = 4;
                        if (lastSpace != null) {
                            if (debug)
                                doc.println("found branch " + hex(branch) + " in asid " + lastSpace.hexId());
                            lastSpace.addTraceEntry(0, null, branch, tcb, 0);
                        }
                        continue;
                    }
                    // Note: 7 bytes following id contain the timestamp
                    int type = space.readShort(entry + 10);

                    tcb = space.readInt(entry + 12);
                    int asid = space.readShort(entry + 18);
                    int psw = space.readInt(entry + 28);

                    //int uniq1 = space.readInt(entry + 32);
                    len = 16 + ((id & 0xf) * 4);
                    String typeString = "unknown";

                    switch (type) {
                    case 0x1:
                        typeString = "SSCH";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 24);
                        break;

                    case 0x3:
                        typeString = "EXT";
                        break;

                    case 0x5:
                        typeString = "SVC";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 24);
                        break;

                    case 0x7:
                        typeString = "PGM";
                        break;

                    case 0xb:
                        typeString = "I/O";
                        break;

                    case 0xf:
                        typeString = "DSP";
                        break;

                    case 0x19:
                        typeString = "SUSP";
                        break;

                    case 0x7f:
                        typeString = "USR";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 24);
                        break;

                    case 0x103:
                        typeString = "EMS";
                        break;

                    case 0x105:
                        typeString = "SVCR";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 24);
                        break;

                    case 0x10f:
                        typeString = "SRB";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 24);
                        break;

                    case 0x205:
                        typeString = "SSRV";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 20);
                        break;

                    case 0x20f:
                        typeString = "SSRB";
                        break;

                    case 0x303:
                        typeString = "CALL";
                        break;

                    case 0x401:
                        typeString = "RSCH";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 24);
                        break;

                    case 0x403:
                        typeString = "CLKC";
                        break;

                    case 0xf0f:
                        typeString = "WAIT";
                        //uniq1 = psw;
                        psw = space.readInt(entry + 20);
                        continue;

                    default:
                        if (debug) doc.println("unknown type " + hex(type) + " id " + hex(id));
                        continue;
                    }
                    if (debug)
                        doc.println("found " + typeString + ", tcb = " + hex(tcb) + " asid = " +
                            hex(asid) + " psw = " + hex(psw));
                    if (psw != 0 && (psw & 1) == 1)
                        if (debug) doc.println("found dodgy " + typeString + ", tcb = " + hex(tcb) + " asid = " +
                                hex(asid) + " psw = " + hex(psw));
                    if (lastSpace == null || lastSpace.id != asid) {
                        int i;

                        for (i = 0; i < spaces.length; i++) {
                            AddressSpace thisSpace = spaces[i];

                            if (thisSpace.id == asid) {
                                lastSpace = thisSpace;
                                break;
                            }
                        }
                        if (i == spaces.length) {
                            lastSpace = null;
                            if (debug)
                                doc.println("Warning: could not locate asid " + hex(asid));
                            continue;
                        }
                    }
                    long t = 0;
                    for (int i = 0; i < 6; i++) {
                        t <<= 8;
                        t |= space.readUnsignedByte(entry + 2 + i);
                    }
                    t |= tod;
                    // Note TTE describes entries
                    lastSpace.addTraceEntry(type, typeString, psw, tcb, t);
                }
            }
        }
    }

    /**
     * Print the trace table.
     */
    private void printTrace(Function functions[]) {
        Table table = new Table(new String[] {"Count", "Function"});
        for (int j = 0; j < functions.length; j++) {
            Function function = functions[j];
            table.addRow(new String[] {"" + function.entryCount, function.name});
        }
        doc.addElement(table);
    }

    final static int arrayLength(int flags, int methods) {
        int type = flags >> 3;
        int length = 1 << (type & 3);

        length *= methods;
        length += 8;        // header
        return length;
    }

    static Hashtable counts = new Hashtable();

    static int arrayLength = 0;
    static int arrayCount = 0;
    static int objectCount = 0;
    static int reg = -1;
    static int caa;

    static void usage() {
        System.err.println("Usage: java com.ibm.jvm.svcdump.Dump [options] <filename>");
        System.err.println("Options:");
        System.err.println("\t-debug\tprint internal debugging info");
        System.err.println("\t-verbose\tprint extra info");
        System.err.println("\t-heap\tprint a table showing which classes have the most objects allocated");
        System.err.println("\t-cache\tprint alloc cache");
        System.err.println("\t-exception\tprint old exception objects");
        System.err.println("\t-dis <addr> <n>\tdisassemble <n> instructions starting at <addr> (hex)");
        System.err.println("\t-dump <addr> <n>\tdump <n> words of storage starting at <addr> (hex)");
        System.err.println("\t-dumpclasses\tprint info about all classes");
        System.err.println("\t-dumpclass <addr>\tprint info about class at given address");
        System.err.println("\t-dumpmdata <addr>\tprint info about mdata at given address");
        System.err.println("\t-dumpprops\tprint the system properties");
        System.err.println("\t-systrace\tprint the system trace");
        System.err.println("\t-caa <addr>\tspecify the caa to use when disassembling");
        System.err.println("\t-r<n>\tinclude saved register <n> in stack trace");
        System.err.println("\t-args\tprint first four function arguments");
        System.err.println("");
        System.err.println("Visit w3.hursley.ibm.com/~dgriff for more info");
        System.exit(1);
    }

    /*
    public static void xxxmain(String args[]) throws Exception {
        // Python expects command name to be the first arg
        String newargs[] = new String[args.length + 1];
        newargs[0] = "";
        System.arraycopy(args, 0, newargs, 1, args.length);
        PySystemState.initialize(System.getProperties(), null, newargs);
        PythonInterpreter interp = new PythonInterpreter();
        interp.execfile(Dump.class.getResourceAsStream("dumpmain.py"));
    }
    */

    public static void main(String args[]) throws Exception {
        if (args.length == 0)
            usage();
        String filename = null;
        Graph graph = new Graph();

        for (int i = 0; i < args.length; i++) {
            String arg = args[i];

            if (arg.startsWith("-")) {
                if (arg.equals("-verbose")) {
                    verbose = true;
                } else if (arg.equals("-debug")) {
                    debug = true;
                } else if (arg.equals("-xxxhelp")) {
                    InputStream is = Dump.class.getResourceAsStream("help.html");
                    int c;

                    while ((c = is.read()) != -1) {
                        System.out.print((char) c);
                    }
                    System.exit(0);
                } else if (arg.equals("-dumpprops")) {
                    dumpproperties = true;
                } else if (arg.equals("-dumpmdata")) {
                    dumpmdata = true;
                    mdata = Integer.parseInt(args[i + 1], 16);
                    i++;
                } else if (arg.equals("-systrace")) {
                    systrace = true;
                } else if (arg.equals("-dumpclasses")) {
                    dumpclasses = true;
                } else if (arg.equals("-dumpclass")) {
                    dumpclass = Integer.parseInt(args[i + 1], 16);
                    i++;
                } else if (arg.equals("-heap")) {
                    analyzeHeap = true;
                } else if (arg.equals("-limit")) {
                    limit = Integer.parseInt(args[i + 1]);
                    i++;
                } else if (arg.equals("-cache")) {
                    printAllocCache = true;
                } else if (arg.equals("-exception")) {
                    printExceptionObject = true;
                } else if (arg.equals("-args")) {
                    printArgs = true;
                } else if (arg.startsWith("-r")) {
                    reg = Integer.parseInt(arg.substring(2));
                } else if (arg.equals("-dis")) {
                    disassemble = true;
                    disassembleAddress = Integer.parseInt(args[i + 1], 16);
                    disassembleLength = Integer.parseInt(args[i + 2]);
                    i += 2;
                } else if (arg.equals("-dump")) {
                    dumpstorage = true;
                    dumpstorageAddress = Integer.parseInt(args[i + 1], 16);
                    dumpstorageLength = Integer.parseInt(args[i + 2]);
                    i += 2;
                } else if (arg.equals("-caa")) {
                    caa = Integer.parseInt(args[i + 1], 16);
                    i++;
                } else {
                    usage();
                }
            } else
                filename = arg;
        }
        Dump dump = new Dump(filename, graph);

        dump.doc.setPlainText(!SvcdumpProperties.getBooleanProperty("svcdump.output.html", false));
        if (disassemble) {
            dump.disassemble();
        } else if (dumpstorage) {
            dump.dumpstorage();
        } else if (dumpclasses) {
            dump.dumpclasses();
        } else if (dumpclass != 0) {
            dump.dumpclass(dumpclass);
        } else if (dumpproperties) {
            dump.dumpproperties();
        } else if (dumpmdata) {
            dump.dumpmdata(mdata);
        } else if (systrace) {
            dump.systrace();
        } else {
            dump.doStuff();
        }
        dump.flush();
    }

    void flush() {
        System.out.println(doc);
    }

    // Need to be able to specify the asid at some point.
    AddressSpace getDefaultSpace() {
        AddressSpace space = null;
        for (int i = 0; i < spaces.length; i++) {
            if (spaces[i].isJava()) {
                space = spaces[i];
                if (debug) System.out.println("found java asid " + space.hexId());
                break;
            }
        }
        if (space == null) {
            throw new Error("Could not find a Java address space!");
        }
        return space;
    }

    Jvm getDefaultJvm() {
        return getDefaultSpace().jvms[0];
    }

    void disassemble() throws Exception {
        System.out.println("Disassembly starting at 0x" + hex(disassembleAddress));
        AddressSpace space = getDefaultSpace();
        Engine engine = new Engine(space);
        Instruction instructions[];

        if (caa == 0)
            instructions = engine.disassemble(doc, disassembleAddress, disassembleLength);
        else
            instructions = engine.disassemble(doc, disassembleAddress, disassembleLength, caa);
    }

    void dumpstorage() throws Exception {
        doc.println("Storage dump starting at 0x" + hex(dumpstorageAddress));
        AddressSpace space = getDefaultSpace();

        if (dumpstorageLength == 0) {
            doc.println(space.readString(dumpstorageAddress));
        } else {
            for (int i = 0; i < dumpstorageLength; i++) {
                int a = dumpstorageAddress + (i << 2);
                doc.println("0x" + hex(a) + ": 0x" + hex(space.readInt(a)));
            }
        }
    }

    void dumpclasses() throws Exception {
        Jvm jvm = getDefaultJvm();

        System.out.println("Dump of all classes in asid " + jvm.space.hexId());
        try {
            jvm.dumpClasses(doc, new NullHandler());
        } catch (Exception e) {
            System.out.println("bad heap in asid " + jvm.space.hexId());
            e.printStackTrace();
        }
    }

    void dumpclass(int cb) throws Exception {
        Jvm jvm = getDefaultJvm();

        try {
            ClassClass classclass = jvm.addClass(cb, "dummy");
            jvm.dumpClass(doc, classclass);
        } catch (Exception e) {
            System.out.println("bad heap in asid " + jvm.space.hexId());
            e.printStackTrace();
        }
    }

    void dumpproperties() throws Exception {
        Jvm jvm = getDefaultJvm();

        System.out.println("Dump of system properties in asid " + jvm.space.hexId());
        try {
            jvm.dumpProperties(doc, new NullHandler());
        } catch (Exception e) {
            System.out.println("bad heap in asid " + jvm.space.hexId());
            e.printStackTrace();
        }
    }

    void dumpmdata(int mdata) throws Exception {
        Jvm jvm = getDefaultJvm();

        try {
            jvm.dumpMdata(doc, mdata);
        } catch (Exception e) {
            System.out.println("bad heap in asid " + jvm.space.hexId());
            e.printStackTrace();
        }
    }

    void systrace() throws Exception {
        Jvm jvm = getDefaultJvm();

        for (int i = 0; i < spaces.length; i++) {
            AddressSpace space = spaces[i];
            if (space.isTrace()) {
                try {
                    processTrace(space);
                } catch (Exception e) {
                    if (debug) doc.println("problem processing trace!");
                }
                if (debug) doc.println("finished processing trace");
            }
        }

        System.out.println("Dump of system trace");
        try {
            //jvm.dumpProperties(doc, new NullHandler());
        } catch (Exception e) {
            System.out.println("bad heap in asid " + jvm.space.hexId());
            e.printStackTrace();
        }
        throw new Error("not yet!");
    }

    void doStuff() {
        if (debug) doc.println("finished scan of dump");

        /* Process the trace table */
        for (int i = 0; i < spaces.length; i++) {
            AddressSpace space = spaces[i];
            if (space.isTrace()) {
                try {
                    processTrace(space);
                } catch (Exception e) {
                    if (debug) doc.println("problem processing trace!");
                }
                if (debug) doc.println("finished processing trace");
            }
        }

        if (debug) doc.println("finished creating Dump instance");
        for (int i = 0; i < spaces.length; i++) {
            AddressSpace space = spaces[i];

            if (debug) doc.println("processing asid " + space.hexId());
            Tcb tcbs[] = space.tcbs();

            if (tcbs == null || tcbs.length == 0) continue;
            doc.println("found " + tcbs.length + " TCBs in asid 0x" + space.hexId());
            doc.println("");
            int samplecaa = 0;
            int rootTid = 0;
            int longestTidChain = 0;
            boolean lastUnknown = false;
            for (int j = 0; j < tcbs.length; j++) {
                Tcb tcb = tcbs[j];

                if (tcb.isKnownType()) {
                    if (lastUnknown) {
                        doc.println("");
                        lastUnknown = false;
                    }
                    doc.print("TCB " + tcb.hexId() + " tid " + hex(tcb.tid()));
                    if (tcb.threadIndex() != 0)
                        doc.print(" tindex " + hex(tcb.threadIndex()));
                    if (tcb.tid_type() != -1)
                        doc.print(" tid type 0x" + hex(tcb.tid_type()));
                    if (tcb.tid_state() != -1)
                        doc.print(" tid state 0x" + hex(tcb.tid_state()));
                    doc.println(tcb.caa() == 0 ? "" : " caa " + hex(tcb.caa()));
                    doc.println("");
                    if (tcb.caa() != 0 && !tcb.dodgy && samplecaa == 0)
                        samplecaa = tcb.caa();
                    int regs[] = tcb.failingRegisters();
                    Table table;
                    if (regs != null) {
                        doc.println("Registers at point of failure:");
                        table = new Table();
                        table.setBorder(0);
                        table.addRow(new String[] {"r0:", hex(regs[0]), "r1:", hex(regs[1]), "r2:", hex(regs[2]), "r3:", hex(regs[3])});
                        table.addRow(new String[] {"r4:", hex(regs[4]), "r5:", hex(regs[5]), "r6:", hex(regs[6]), "r7:", hex(regs[7])});
                        table.addRow(new String[] {"r8:", hex(regs[8]), "r9:", hex(regs[9]), "r10:", hex(regs[10]), "r11:", hex(regs[11])});
                        table.addRow(new String[] {"r12:", hex(regs[12]), "r13:", hex(regs[13]), "r14:", hex(regs[14]), "r15:", hex(regs[15])});
                        table.addRow(new String[] {"psw:", hex(regs[16]), "", hex(regs[17])});
                        doc.addElement(table);
                    }
                    table = new Table();
                    table.suppressEmptyColumns(true);
                    table.addHeaderCell("Dsa");
                    table.addHeaderCell("Entry");
                    table.addHeaderCell("Offset");
                    if (reg != -1) {
                        String r = reg < 10 ? "r" + reg + " " : "r" + reg;
                        table.addHeaderCell(r);
                    }
                    table.addHeaderCell("Function");
                    if (tcb.isJava()) {
                        table.addHeaderCell("Method");
                    }
                    table.addHeaderCell("Module");
                    for (Dsa dsa = tcb.topDsa; dsa != null; dsa = dsa.previous()) {
                        Row row = new Row();
                        row.addCell(hex(dsa.address));
                        row.addCell(hex(dsa.entryPoint()));
                        row.addCell(hex(dsa.offset()));
                        if (reg != -1) {
                            row.addCell(hex(dsa.reg(reg)));
                        }
                        row.addCell(dsa.function());
                        if (tcb.isJava()) {
                            row.addCell(dsa.method());
                        }
                        row.addCell(dsa.program());
                        table.addRow(row);
                    }
                    doc.addElement(table);
                    if (tcb.hasPendingException() || printExceptionObject) {
                        tcb.printException(doc);
                    }
                    if (printAllocCache && tcb.tid() != 0)        // Java
                        tcb.printAllocCache(doc);
                    if (tcb.tid() != 0 && tcb.jvm != null) {
                        int tidChain = 0;
                        Jvm jvm = tcb.jvm;

                        try {
                            for (int tid = tcb.tid(); tid != 0; tid = space.readInt((long) tid)) {
                                tidChain++;
                            }
                        } catch (Exception e) {}
                        if (tidChain > jvm.longestTidChain) {
                            jvm.longestTidChain = tidChain;
                            jvm.rootTid = tcb.tid();
                        }
                    }
                    doc.println("");
                    Function functions[] = tcb.traceFunctions();

                    if (functions == null) continue;
                    doc.println("found trace table:");
                    printTrace(functions);
                    doc.println("");
                } else {
                    doc.println("TCB of unknown type " + tcb.hexId() + " caa " + hex(tcb.caa()));
                    lastUnknown = true;
                }
            }
            doc.println("");
            for (int j = 0; space.jvms != null && j < space.jvms.length; j++)
                space.jvms[j].checkTidChain(doc);
            for (int j = 0; j < tcbs.length; j++) {
                Tcb tcb = tcbs[j];
                if (tcb.tid() != 0) {
                    boolean found = false;
                    Jvm jvm = tcb.jvm;
                    try {
                        for (int tid = jvm.rootTid; tid != 0; tid = space.readInt(tid)) {
                            if (tcb.tid() == tid) {
                                if (debug)
                                    doc.println("found " + hex(tid) + " in tid chain");
                                found = true;
                                if (!tcb.isKnownType()) {
                                    doc.println("*** Warning! tcb " + tcb.hexId() + " with tid " + hex(tcb.tid()) + " is unknown TCB type but known by Java");
                                }
                                break;
                            }
                        }
                    } catch (Exception e) {}
                    if (!found)
                        doc.println("*** Warning! tid " + hex(tcb.tid()) + " not found in tid chain");
                }
            }
            Function functions[] = space.traceFunctions();

            if (functions != null && functions.length != 0) {
                doc.println("trace table for asid " + space.hexId() + ":");
                doc.println("");
                printTrace(functions);
                doc.println("");
            }
            if (samplecaa != 0) {
                try {
                    int ceeedb = space.readInt(samplecaa + 0x2f0);
                    int envar = space.readInt(ceeedb + 0x58);
                    doc.printHeader("environment variables:");
                    doc.println("");
                    Table table = new Table(new String[] {"key", "value"});
                    for (int env = 0; (env = space.readInt(envar)) != 0; envar += 4) {
                        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
                        int c;

                        while ((c = space.readUnsignedByte(env++)) != 0) {
                            bytes.write(c);
                        }
                        String name = getEbcdicString(bytes.toByteArray());
                        int eq = name.indexOf('=');
                        String key = name.substring(0, eq);
                        String value = name.substring(eq + 1);
                        table.addRow(new String[] {key, value});
                    }
                    doc.addElement(table);
                    doc.println("");
                    doc.printHeader("loaded dlls:");
                    doc.println("");
                    table = new Table(new String[] {"name", "wsa"});
                    for (int dlcb = space.readInt(ceeedb + 0x448); dlcb != 0; dlcb = space.readInt(dlcb)) {
                        int len = space.readShort(dlcb + 0x3c);
                        int ptr = space.readInt(dlcb + 0x40);
                        int wsa = space.readInt(dlcb + 0x14);
                        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
                        int c;
                        for (int k = 0; k < len; k++) {
                            c = space.readUnsignedByte(ptr++);
                            bytes.write(c);
                        }
                        String name = getEbcdicString(bytes.toByteArray());
                        table.addRow(new String[] {name, hex(wsa)});
                    }
                    doc.addElement(table);
                    doc.println("");
                } catch (Exception e) {
                }
            }

            Jvm[] jvms = space.jvms;

            if (jvms != null) {
                for (int j = 0; j < jvms.length; j++) {
                    Jvm jvm = jvms[j];
                    if (analyzeHeap) {
                        try {
                            jvm.analyzeHeap(doc, new NullHandler());
                        } catch (Exception e) {
                            doc.println("bad heap in asid " + space.hexId());
                            e.printStackTrace();
                        }
                    }
                }
            }
        }
    }

    public Jvm[] jvms() {
        Vector v = new Vector();
        for (int i = 0; i < spaces.length; i++) {
            AddressSpace space = spaces[i];
            Tcb tcbs[] = space.tcbs();

            for (int j = 0; j < tcbs.length; j++) {
                Tcb tcb = tcbs[j];

                if (tcb.isKnownType()) {
                    for (Dsa dsa = tcb.topDsa; dsa != null; dsa = dsa.previous()) {
                        dsa.function();
                    }
                }
            }
            if (space.jvms != null) {
                for (int j = 0; j < space.jvms.length; j++) {
                    v.addElement(space.jvms[j]);
                }
            }
        }
        return (Jvm[])v.toArray(new Jvm[0]);
    }

    /**
     * Convert ebcdic string to ascii. Sun don't support Cp1047 so we need to check for
     * some characters.
     */
    public static String getEbcdicString(byte b[]) {
        try {
            for (int i = 0; i < b.length; i++) {
                if (b[i] == (byte)0xad) 
                    b[i] = 0x4a;
                else if (b[i] == (byte)0xbd) 
                    b[i] = 0x5a;
            }
            return new String(b, "Cp500");
        } catch (Exception e) {
            throw new Error("No Cp500");
        }
    }
}
