package com.ibm.jvm.svcdump;

import java.util.*;
import java.text.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.util.html.*;
import com.ibm.jvm.findroots.*;

public final class Jvm extends Base {
    AddressSpace space;
    boolean checkedVersion = false;
    boolean is131 = false;
    boolean printstrings;
    Hashtable names = new Hashtable();
    Hashtable classes = new Hashtable();
    Hashtable livecounts = new Hashtable();
    Hashtable deadcounts = new Hashtable();
    int jvmp;
    int st;
    int heapbase;
    int heaplimit;
    int mh_heapbase;
    int mh_heaplimit;
    int th_heapbase;
    int th_heaplimit;
    int freeObjectCtr;
    int totalObjectCtr;
    int freeList;
    int allocbits;
    int lastaddress;
    HeapdumpContentHandler handler;
    int limit = 10;
    int totalLength = 0;
    int totalObjects = 0;
    static DateFormat df = DateFormat.getTimeInstance();
    boolean isPostRas;
    int clJavaLangClass;
    int stglobal;
    int st_jab;
    int longestTidChain;
    int rootTid;
    boolean includeDead;

    Jvm(AddressSpace space) {
        this.space = space;
        assert(false);
    }

    Jvm(AddressSpace space, int jvmp, boolean isPostRas) throws Exception {
        this.space = space;
        this.jvmp = jvmp;
        this.isPostRas = isPostRas;
        is131 = true;
        if (Dump.debug) System.out.println("jvmp = " + hex(jvmp));
        expect(jvmp, "MAIN", false);
        int eye;
        if (isPostRas) {
            st = jvmp + 340 + 8;
        } else {
            st = jvmp + 340;
        }
        if (Dump.debug) System.out.println("st = " + hex(st));
        if (isPostRas) {
            expect(st, "STOR", false);
            stglobal = space.readInt(st + 404 + 8);        // dataP, at the end of the structure
        } else {
            stglobal = space.readInt(st + 404);        // dataP, at the end of the structure
        }
        if (Dump.debug) System.out.println("stglobal = " + hex(stglobal));
        if (isPostRas) {
            expect(stglobal, "STDA", false);
            stglobal += 16;
        }
        st_jab = space.readInt(stglobal + 12);
        if (Dump.debug) System.out.println("st_jab = " + hex(st_jab));
        int real_heapbase = space.readInt(stglobal + 20);
        if (Dump.verbose) System.out.println("real_heapbase = " + hex(real_heapbase));
        int real_heaptop = space.readInt(stglobal + 24);
        if (Dump.verbose) System.out.println("real_heaptop = " + hex(real_heaptop));
        int real_TH_limit = space.readInt(stglobal + 32);
        if (Dump.debug) System.out.println("real_TH_limit = " + hex(real_TH_limit));
        int init_nursery_size = space.readInt(st_jab);
        if (Dump.debug) System.out.println("init_nursery_size = " + hex(init_nursery_size));
        int system_heap_next = space.readInt(st_jab + 0x14);
        if (Dump.debug) System.out.println("system_heap_next = " + hex(system_heap_next));
        int system_heap_limit = space.readInt(st_jab + 0x18);
        if (Dump.debug) System.out.println("system_heap_limit = " + hex(system_heap_limit));
        int ACS_heap_next = space.readInt(st_jab + 0x28);
        if (Dump.debug) System.out.println("ACS_heap_next = " + hex(ACS_heap_next));
        int ACS_heap_limit = space.readInt(st_jab + 0x2c);
        if (Dump.debug) System.out.println("ACS_heap_limit = " + hex(ACS_heap_limit));
        mh_heapbase = space.readInt(stglobal + 36);
        if (Dump.debug) System.out.println("mh_heapbase = " + hex(mh_heapbase));
        mh_heaplimit = space.readInt(stglobal + 40);
        if (Dump.debug) System.out.println("mh_heaplimit = " + hex(mh_heaplimit));
        th_heapbase = space.readInt(stglobal + 44);
        if (Dump.debug) System.out.println("th_heapbase = " + hex(th_heapbase));
        th_heaplimit = space.readInt(stglobal + 48);
        if (Dump.debug) System.out.println("th_heaplimit = " + hex(th_heaplimit));
        allocbits = space.readInt(stglobal + 100);
        if (Dump.debug) System.out.println("allocbits = " + hex(allocbits));
        int loadedSystemClasses = space.readInt(stglobal + 136);
        if (Dump.debug) System.out.println("loadedSystemClasses = " + hex(loadedSystemClasses));
        if (loadedSystemClasses != 0) {
            addClassChain(loadedSystemClasses, "System");
        }
        int loadedACSClasses = space.readInt(stglobal + 140);
        if (Dump.debug) System.out.println("loadedACSClasses = " + hex(loadedACSClasses));
        if (loadedACSClasses != 0) {
            addClassChain(loadedACSClasses, "ACS");
        }
        int gcctr = space.readInt(stglobal + 0xa8);
        if (Dump.debug) System.out.println("gcctr = " + hex(gcctr));
        int last_compactgc = space.readInt(stglobal + 0xac);
        if (Dump.debug) System.out.println("last_compactgc = " + hex(last_compactgc));
        int last_icompactgc = space.readInt(stglobal + 0xb0);
        if (Dump.debug) System.out.println("last_icompactgc = " + hex(last_icompactgc));
        int last_fullgc = space.readInt(stglobal + 0xb4);
        if (Dump.debug) System.out.println("last_fullgc = " + hex(last_fullgc));
        int onlineCPUs = space.readInt(stglobal + 0xb8);
        if (Dump.debug) System.out.println("onlineCPUs = " + hex(onlineCPUs));
        int maxedOutTLHs = space.readInt(stglobal + 0xbc);
        if (Dump.debug) System.out.println("maxedOutTLHs = " + hex(maxedOutTLHs));
        int maxTLHSize = space.readInt(stglobal + 0xc0);
        if (Dump.debug) System.out.println("maxTLHSize = " + hex(maxTLHSize));
        int num_hints = space.readInt(stglobal + 0x3c0);
        if (Dump.debug) System.out.println("num_hints = " + hex(num_hints));
        int hint_first = space.readInt(stglobal + 0x3c4);
        if (Dump.debug) System.out.println("hint_first = " + hex(hint_first));
        int hint_free = space.readInt(stglobal + 0x3c8);
        if (Dump.debug) System.out.println("hint_free = " + hex(hint_free));
        for (int h = hint_free; h != 0; h = space.readInt(h + 8)) {
            int len = space.readInt(h);
            int chunk = space.readInt(h + 4);
            int lru = space.readInt(h + 8);
            if (Dump.debug) System.out.println("hint chunk: " + hex(chunk) + " len: "
                + hex(len) + " lru: " + hex(lru));
        }
        if (isPostRas) {
            stglobal += 4;      // horrid horrid hack!
        }
        int afctr = space.readInt(stglobal + 0x3e4);
        freeObjectCtr = space.readInt(stglobal + 0x3f8);
        if (Dump.verbose) System.out.println("freeObjectCtr = " + freeObjectCtr);
        totalObjectCtr = space.readInt(stglobal + 0x3fc);
        if (Dump.verbose) System.out.println("totalObjectCtr = " + totalObjectCtr);
        freeList = space.readInt(stglobal + 0x40c);
        if (Dump.debug) System.out.println("freeList = " + hex(freeList));
        int freeTail = space.readInt(stglobal + 0x410);
        if (Dump.debug) System.out.println("freeTail = " + hex(freeTail));
        int deferredList = space.readInt(stglobal + 0x414);
        if (Dump.debug) System.out.println("deferredList = " + hex(deferredList));
        heapbase = mh_heapbase;
        heaplimit = mh_heaplimit;
    }

    void checkTidChain(Document doc) {
        boolean problem = false;
        if (rootTid != 0) {
            int tid = 0, lastTid = 0;
            try {
                for (tid = rootTid; tid != 0; tid = space.readInt(tid)) {
                    int prev = space.readInt(tid + 4);
                    if (prev != lastTid) {
                        doc.println("Warning! in tid " + hex(tid) + 
                            " prev pointer does not match! expected " + hex(lastTid) + 
                            " found " + hex(prev));
                        problem = true;
                    }
                    lastTid = tid;
                }
            } catch (Exception e) {
                doc.println("*** uh oh " + hex(tid));
            }
        }
        if (problem)
            doc.println("");
    }

    String cbName(int clazz) throws Exception {
        if (space.readInt(clazz + 28) == clazz) {        // handle_to_self
            if (checkedVersion) {
                if (is131) {

                    /*
                     * There is a small chance that last_subclass_of also points to us!
                     * So double check the self pointer.
                     */
                    if (space.readInt(clazz + 12) != clazz)         // self
                        throw new Error("Blooming heck!");
                }
            } else
                is131 = false;
        } else {
            if (checkedVersion) {
                if (!is131)
                    throw new Error("Blooming heck!");
            } else
                is131 = true;
        }
        checkedVersion = true;
        int cbName = is131 ? space.readInt(clazz + 64) : space.readInt(clazz + 12);
        String name = (String) names.get(new Integer(cbName));

        if (name == null) {
            StringBuffer buff = new StringBuffer();
            int c;

            try {
                while ((c = space.readUnsignedByte(cbName++)) != 0) {
                    buff.append((char) c);
                }
            } catch (Exception e) {
                return "(error reading " + hex(cbName) + ")";
            }
            name = buff.toString();
            names.put(new Integer(cbName), name);
        }
        return name;
    }

    void expect(int address, String eyecatcher, boolean ascii) throws Exception {
        byte b[];
        if (ascii)
            b = eyecatcher.getBytes("8859_1");
        else
            b = eyecatcher.getBytes("Cp500");
        int word = Dump.getInt(b, 0);
        int found = space.readInt(address);
        if (found != word)
            throw new Error("did not find expected eyecatcher! found " + hex(found) +
                " expected " + hex(word));
    }

    ClassClass getClass(int cb) throws Exception {
        Integer cbInt = new Integer(cb);
        ClassClass classclass = (ClassClass) classes.get(cbInt);
        return classclass;
    }

    ClassClass addClass(int cb, String string) throws Exception {
        Integer cbInt = new Integer(cb);
        ClassClass classclass = (ClassClass) classes.get(cbInt);

        if (classclass == null) {
            classclass = new ClassClass(this, cb, cbName(cb), string);
            classes.put(cbInt, classclass);
        }
        return classclass;
    }
    
    void addClassChain(int cb, String string) {
        try {
            for (; cb != 0; cb = space.readInt((long) (cb + 32)))
                addClass(cb, string);
        } catch (Exception e) {
            if (Dump.debug)
                System.out.println("Warning: dodgy cb in class chain: 0x" + hex(cb));
        }
    }
    
    boolean isLive(int address) {
        Tcb tcbs[] = space.tcbs();
        for (int i = 0; i < tcbs.length; i++) {
            if (tcbs[i].inAllocCache(address)) {
                System.out.println("yes! " + hex(address));
                System.exit(1);
                return true;
            }
        }
        return false;
    }

    void analyze131Heap() throws Exception {
        heapbase = mh_heapbase;
        heaplimit = mh_heaplimit;

        for (Enumeration e = classes.elements(); e.hasMoreElements();) {
            ClassClass classclass = (ClassClass)e.nextElement();

            if (handler != null)
                handler.classDump(classclass.cb, cbName(classclass.cb), classclass.staticReferences(), 256);
        }
        for (int chunk = heapbase, i = 0; chunk < heaplimit; i++) {
            int offset = chunk - heapbase;
            int allocindex = offset >> 8;
            int allocoffset = (offset >> 3) & 0x1f;
            int bits = space.readInt(allocbits + (allocindex << 2));
            int bit = 0x80000000 >>> allocoffset;
            boolean live = ((bits & bit) != 0);
            int length = space.readInt(chunk) & 0x3ffffff8;

            if (length > 0x10000000) {
                System.out.println("bad chunk at " + hex(chunk) + " chunk[0] = " + hex(space.readInt(chunk)));
                System.out.println("last address = " + hex(lastaddress));
            }
//System.out.println("process object " + hex(chunk + 4));
            if (live) {
                processObject(chunk + 4, length - 4);
            } else if (includeDead) {
                try {
                    processObject(chunk + 4, length - 4);
                } catch (Throwable e) {
                }
            }
            if (verbose() || ((i % 0x20000) == 0)) {
                if (handler != null)
                    handler.log("analyzing heap at " + hex(chunk) + " total objects so far = " + totalObjects);
                /*
                if (live && verbose()) {
                    int methods = space.readInt(chunk + 4);
                    int flags = space.readInt(chunk + 8);
                    if (flags == 0 && methods != 0) {
                        int clazz = space.readInt(methods);
                        System.out.print(" class = " + cbName(clazz));
                    } else {
                        System.out.print(" methods = " + hex(methods) + " flags = " + hex(flags));
                    }
                }
                System.out.println("");
                */
            }
            chunk += length;
            if (length == 0 || chunk < 0 || chunk > (heaplimit + 16))
                System.out.println("found bad length? " + hex(length));
            if (length == 0)
                break;
        }
        /*
         * Now do the thread local heaps
         */
        Tcb tcbs[] = space.tcbs();
        for (int i = 0; i < tcbs.length; i++) {
            int cacheBlock = tcbs[i].cacheBlock();
            int cacheSize = tcbs[i].cacheSize();
            int cacheOrigSize = tcbs[i].cacheOrigSize();
            if (cacheBlock == 0 || cacheSize == 0 || cacheOrigSize == 0)
                continue;
            try {
                for (int chunk = cacheBlock + cacheSize; chunk < cacheBlock + cacheOrigSize; ) {
                    int length = space.readInt(chunk) & 0x3ffffff8;
                    processObject(chunk + 4, length - 4);
                    chunk += length;
                    if (length == 0 || chunk < 0)
                        System.out.println("found bad length? " + hex(length));
                    if (length == 0)
                        break;
                }
            } catch (Exception e) {
            }
        }
    }

    int methodsToClass(int methods) throws Exception {
        if (isPostRas) {
            expect(methods, "CLMT", true);
            return space.readInt(methods + 8);
        } else {
            return space.readInt(methods);
        }
    }

    void searchHeap(int value) throws Exception {
        heapbase = mh_heapbase;
        heaplimit = mh_heaplimit;
        System.out.println("searchHeap: value = " + hex(value) + " base = " + hex(heapbase) + " limit = " + hex(heaplimit));
        for (int chunk = heapbase; chunk < heaplimit;) {
            int length = space.readInt(chunk) & 0x3ffffff8;

            for (int i = chunk + 12; i < chunk + length; i += 4) {
                if (space.readInt(i) == value) {
                    System.out.println("found it at " + hex(chunk + 4));
                    int methods = space.readInt(chunk + 4);
                    int flags = space.readInt(chunk + 8);

                    if ((flags & 2) == 2) {
                        System.out.println("found ref to " + hex(value) + " in array at " + hex(chunk + 4));
                    } else {
                        if (methods != 0) {
                            int clazz = methodsToClass(methods);
                            String name = cbName(clazz);

                            System.out.println("found ref to " + hex(value) + " in object " + name + " at " + hex(chunk + 4));
                        } else {
                            System.out.println("found ref to " + hex(value) + " in unknown at " + hex(chunk + 4));
                        }
                    }
                }
            }
            chunk += length;
        }
    }

    void analyzeHaifaHeap() throws Exception {
        IntegerArray addresses = space.addresses;

        for (int i = 0; i < addresses.size(); i++) {
            int address = addresses.get(i);

            if (space.readInt(address) == 0xc8c1d5c3) {
                int length = space.readInt(address + 24);

                if (verbose()) System.out.println("found HANC at " + hex(address) + " length = " + hex(length) + " in asid " + space.hexId());
                if (length > 0x6000000) {        // java heap
                    handler.log("found Haifa Java heap of size 0x" + hex(length));
                    heapbase = address + 0x1000;
                    heaplimit = address + length;
                    for (int a = address + 0x1000; a < address + length;) {
                        try {
                            int methods = space.readInt(a);
                            int flags = space.readInt(a + 4);

                            //System.out.println(hex(a) + ": " + hex(methods) + " " + hex(flags));
                            a += processBlock(a);
                        } catch (Exception e) {
                            if (verbose()) System.out.println("*** error reading " + hex(a) + " ***");
                            a += 0x1000;
                        }
                    }
                }
            }
        }
    }

    // Legacy!!!
Graph graph;
    public void findRoots(Graph g) throws Exception {
        graph = g;
        analyze(new HeapdumpContentHandler() {
            public void instanceDump(int id, String className, int references[], int size) {
                if (!graph.nofinal || !className.equals("java/lang/ref/Finalizer")) {
                    graph.addVertex(id, references, size);
                }
            }
            public void classDump(int id, String className, int references[], int size) {
                graph.addVertex(id, references, 300);
            }
            public void objectArrayDump(int id, String className, int references[], int size) {
                graph.addVertex(id, references, references.length << 2);
            }
            public void primitiveArrayDump(int id, int type, int size) {
                graph.addVertex(id, size);
            }
            public void error(String msg) {
                assert(false);
            }
            public void addEdge(int fromId, int toId) {
                assert(false);
            }
            public void log(String msg) {
                graph.log(msg);
            }
        });
        graph.log("");
        graph.log("end heap analysis");
        graph.print(new PrintClient() {
            public String getName(int id) {
                return xgetName(id);
            }
        });
    }

    void analyze(HeapdumpContentHandler handler) throws Exception {
        this.handler = handler;
        if (handler != null)
            handler.log("begin heap analysis");
        if (is131) {
            analyze131Heap();
        } else {
            analyzeHaifaHeap();
        }
    }

    public void scanHeap(HeapdumpContentHandler handler, boolean includeDead) throws Exception {
        this.includeDead = includeDead;
        analyze(handler);
        handler.log("");
        handler.log("end heap analysis, cache misses = " + space.cacheMisses);
    }

    public void analyzeHeap(Document doc, HeapdumpContentHandler handler) throws Exception {
        analyze(handler);
        doc.println("");
        doc.println("*** dump of live objects in the heap ***");
        doc.println("");
        printCounts(doc, livecounts);
        doc.println("");
        doc.println("end heap analysis, cache misses = " + space.cacheMisses);
        doc.println("total objects = " + totalObjects + " total length = " + totalLength);
        if (is131) {
            if (verbose()) {
                doc.println("*** dump of free list: ***");
                for (int chunk = freeList, free = 0; chunk != 0;) {
                    try {
                        free += space.readInt(chunk);
                        doc.println("chunk = " + hex(chunk) + " free total = " + free);
                        chunk = space.readInt(chunk + 4);
                    } catch (Exception e) {
                        doc.println("*** problem reading chunk " + hex(chunk));
                        break;
                    }
                }
            }
        }
    }

    public void dumpClass(Document doc, ClassClass classclass) throws Exception {
        ClassClass superclass = getClass(classclass.cbSuperclass());
        System.out.println("class " + classclass + " extends " + superclass);
        ClassMethod methods[] = classclass.getMethods();
        for (int i = 0; i < methods.length; i++) {
            ClassMethod method = methods[i];
            System.out.println("    method " + method);
            if (Dump.verbose) {
                System.out.println("        local variable count " + method.mbLocalVariableCount());
            }
        }
    }

    public void dumpMdata(Document doc, int mdata) throws Exception {
        int mb = space.readInt(mdata);
        ClassMethod m = new ClassMethod(this, mb);
        System.out.println("Dump of mdata at 0x" + hex(mdata));
        System.out.println("");
        System.out.println("Method: " + m.fullname());
        System.out.println("");
        int stptbl = space.readInt(mdata + 12);
        if (stptbl != 0) {
            int count = space.readInt(stptbl);
            System.out.println("Found " + count + " inlined methods");
            System.out.println("");
            for (int i = 0; i < count; i++) {
                int pc = space.readInt(stptbl + 4 + (i << 3));
                int sf = space.readInt(stptbl + 8 + (i << 3));
                System.out.print("0x" + hex(pc) + ": ");
                boolean first = true;
                while (sf != 0) {
                    int inlinemb = space.readInt(sf);
                    ClassMethod method = new ClassMethod(this, inlinemb);
                    System.out.print((first ? "" : " <- ") + method.fullname());
                    first = false;
                    sf = space.readInt(sf + 4);
                }
                System.out.println("");
            }
        }
    }

    public void dumpClasses(Document doc, HeapdumpContentHandler handler) throws Exception {
        System.out.println("Scanning heap, please wait...");
        analyze(null);
        System.out.println("");
        System.out.println("*** dump of classes in the heap ***");
        System.out.println("");
        for (Enumeration e = classes.elements(); e.hasMoreElements();) {
            ClassClass classclass = (ClassClass)e.nextElement();
            dumpClass(doc, classclass);
        }
    }

    void appendProperties(Table table, int props) throws Exception {
        assert(props > 0);
        assert(getName(props).equals("java/util/Properties"));
        int itable = space.readInt(props + 8);
        int count = space.readInt(props + 12);
        for (int i = 0; i < count; i++) {
            int entry = space.readInt(itable + 8 + (i << 2));
            while (entry != 0) {
                int ikey = space.readInt(entry + 12);
                int ivalue = space.readInt(entry + 16);
                String key = space.readStringObject(ikey);
                String value = space.readStringObject(ivalue);
                table.addRow(key, value);
                entry = space.readInt(entry + 20);
            }
        }
        int defaults = space.readInt(props + 40);
        if (defaults != 0) {
            appendProperties(table, defaults);
        }
    }

    public void dumpProperties(Document doc, HeapdumpContentHandler handler) throws Exception {
        System.out.println("Scanning heap, please wait...");
        analyze(null);
        doc.println("");
        doc.println("*** dump of system properties ***");
        doc.println("");
        Table table = new Table("key", "value");
        for (Enumeration e = classes.elements(); e.hasMoreElements();) {
            ClassClass classclass = (ClassClass)e.nextElement();
            if ("java/lang/System".equals(classclass.toString())) {
                int props = classclass.getStaticField("props");
                appendProperties(table, props);
                break;
            }
        }
        doc.addElement(table);
    }

    void printCounts(Document doc, Hashtable counts) {
        final String classes[] = new String[counts.size()];
        IntegerArray icounts = new IntegerArray();
        int x = 0;

        for (Enumeration keys = counts.keys(); keys.hasMoreElements();) {
            String name = (String) keys.nextElement();
            int n = ((Integer) counts.get(name)).intValue();

            classes[x++] = name;
            icounts.add(n);
        }
        icounts.reverseSort(new SortListener() {
                public void swap(int a, int b) {
                    String tmp = classes[a];

                    classes[a] = classes[b];
                    classes[b] = tmp;
                }
            }
        );
        Table table = new Table(new String[] {"count", "class"});
        for (x = 0; x < counts.size(); x++) {
            table.addRow(new String[] {"" + icounts.get(x), classes[x]});
        }
        doc.addElement(table);
    }

    final int arrayLength(int flags, int methods) {
        int type = flags >> 3;
        int length = 1 << (type & 3);

        length *= methods;
        length += 8;        // header
        return length;
    }

    int processBlock(int address) throws Exception {
        int methods = space.readInt(address);
        int flags = space.readInt(address + 4);
        int length;

        if ((flags & 2) == 2 && flags <= 0x5a) {
            // is an array
            length = arrayLength(flags, methods);
            if (length >= 0x800) {
                // not chunked
                processObject(address, length);
                return (length + 0xfff) & 0xfffff000;
            }
        } else if (flags != 0) {
            // could be a thin lock but lets not risk it
            if (verbose()) System.out.println("unknown block flags at " + hex(address));
            return 0x1000;        // skip block
        } else {
            length = 8;
        }
        int spacing = (length + 3) & 0xfffffffc;
        boolean valid = true;

        while (spacing < 0x800) {
            valid = true;
            for (int a = address + spacing; a < address + 0x1000; a += spacing) {
                methods = space.readInt(a);
                flags = space.readInt(a + 4);
                if ((flags & 2) == 2 && flags <= 0x5a) {
                    length = arrayLength(flags, methods);
                    if (length > spacing) {
                        valid = false;
                        break;
                    }
                } else if (flags != 0 || (methods != 0 && methods < 0x10000000) || (methods & 3) != 0) {
                    valid = false;
                    break;
                } else {
                    try {
                        if (methods != 0) {
                            int clazz = space.readInt(methods);
                            int cbName = space.readInt(clazz + 12);
                        }
                    } catch (Exception e) {
                        valid = false;
                        break;
                    }
                }
            }
            if (valid)
                break;
           spacing += 4;
        }
        if (valid) {
            for (int a = address; a < address + 0x1000; a += spacing) {
                processObject(a, spacing);
            }
        } else {
            if (verbose()) System.out.println("couldn't figure out " + hex(address));
        }
        return 0x1000;
    }

    boolean stIsValidHandle(int handle) {
        return (handle >= heapbase && handle < heaplimit || handle >= th_heapbase && handle < th_heaplimit);
    }

    void addCount(int address) {
        String name = getName(address);
        Integer count = (Integer) livecounts.get(name);

        if (count == null) {
            livecounts.put(name, new Integer(1));
        } else {
            livecounts.put(name, new Integer(count.intValue() + 1));
        }
    }

    boolean is_normal_object(int handle) {
        return (handle & 0x2) == 0;
    }
    
    boolean clJavaLangClass(int cb) throws Exception {
        if (clJavaLangClass != 0)
            return cb == clJavaLangClass;
        String name = cbName(cb);

        if (name.equals("java/lang/Class")) {
            clJavaLangClass = cb;
            return true;
        }
        return false;
    }

    void processObject(int address, int length) throws Exception {
        totalLength += length;
        totalObjects++;
        lastaddress = address;
        int methods = space.readInt(address);
        int flags = space.readInt(address + 4);
        addCount(address);

        if (is_normal_object(flags)) {
            try {
                int cb = methodsToClass(methods);

                if (clJavaLangClass(cb)) {
                    ClassClass classclass = addClass(address, "Middleware");

                    if (handler != null)
                        handler.classDump(address, cbName(address), classclass.staticReferences(), length);
                } else if (handler != null) {
                    IntegerArray integerarray = null;
                    int offsets = cbObjectOffsets(cb);

                    if (offsets != 0) {
                        for (;;) {
                            int offset = space.readShort(offsets);

                            if (offset == 0)
                                break;
                            assert(offset > 0);
                            assert(offset < length);
                            assert((offset & 0x1) != 0);
                            int ref = space.readInt(address + 7 + offset);

                            if (ref != 0 && stIsValidHandle(ref)) {
                                if (integerarray == null)
                                    integerarray = new IntegerArray();
                                integerarray.add(ref);
                            }
                            offsets += 2;
                        }
                    }
                    int[] is = (integerarray == null ? new int[0]
                            : integerarray.toArray());

                    handler.instanceDump(address, cbName(cb), is, length);
                }
            } catch (Exception e) {
                if (!includeDead)
                    e.printStackTrace();
                throw new Error("yuk");
            }
        } else if (flags > 0 && flags <= 0x5a) {
            if (handler != null) {
                if (flags == 0x12) {    // array of ref
                    IntegerArray refs = null;

                    for (int i = 0; i < methods; i++) {
                        int ref = space.readInt(address + 8 + (i << 2));

                        if (ref != 0 && stIsValidHandle(ref)) {
                            if (refs == null)
                                refs = new IntegerArray();
                            refs.add(ref);
                        }
                    }
                    int cb = space.readInt(address + 8 + (methods << 2));
                    int[] is = (refs == null ? new int[0]
                            : refs.toArray());

                    handler.objectArrayDump(address, cbName(cb), is, length);
                } else
                    handler.primitiveArrayDump(address, flags >> 3 & 0xf, length);
            }
        }
    }

    int cbObjectOffsets(int cb) throws Exception {
        return space.readInt(cb + 184);
    }

    String arrayName[] = {"unknown array type", "unknown array type", "array of references",
            "unknown array type", "array of boolean", "array of char", "array of float",
            "array of double", "array of byte", "array of short", "array of int",
            "array of long", "unknown array type", "unknown array type",
            "unknown array type", "unknown array type"};
    int finalClass;
    boolean foundFinalClass;

    public String xgetName(int id) {
        return getName(id);
    }

    public String getName(int id) {
        try {
            int flags = space.readInt(id + 4);

            if ((flags & 2) == 2) {
                int type = (flags >> 3) & 0xf;

                if (type == 5 && printstrings) {
                    int methods = space.readInt(id);
                    int alength = arrayLength(flags, methods);

                    if (alength > 40) alength = 40;
                    char cname[] = new char[alength];

                    for (int i = 0; i < alength; i++) {
                        cname[i] = (char) space.readShort(id + 8 + (i << 1));
                    }
                    String name = "array of char: " + new String(cname);

                    return name.intern();
                } else {
                    return arrayName[type];
                }
            } else {
                int methods = space.readInt(id);
                int clazz = methodsToClass(methods);

                return cbName(clazz);
            }
        } catch (Exception e) {
            return "error";
        }
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }

    public int asid() {
        return space.id;
    }
}

