package com.ibm.jvm.svcdump;

import java.util.*;
import com.ibm.jvm.dump.*;
import com.ibm.jvm.util.html.*;

/**
 *
 *
 */
public class Tcb extends DvThread {
    AddressSpace space;
    int topDsaAddress;
    Dsa topDsa;
    Dsa stack[];
    int tcbid;
    int tid;
    int ee;
    int state;
    int caa;
    String type;
    boolean isKnownType;
    boolean scannedTid;
    Hashtable traceTable;
    Function traceFunctions[];
    static final int STCB = 0x138;
    static final int OTCB = 0xd8;
    static final int THLI = 0xbc;
    static final int TID = 0x34;
    boolean is131 = false;
    boolean verbose = false;
    boolean isPostRas = false;
    int failingRegisters[];
    Jvm jvm;
    boolean dodgy = false;
    int cacheBlock = 0;
    int cacheSize = 0;
    int cacheOrigSize = 0;

    /* Start of DvThread methods */

    /**
     *
     *
     */
    public DvProcess getProcess() {
        return space.getProcesses()[0];
    }

    /**
     *
     *
     */
    public String id() {
        return hexId();
    }

    /**
     *
     *
     */
    public boolean isJava() {
        return tid != 0;
    }

    /**
     *
     *
     */
    public DvAddress eeAddress() {
        return new Address(ee);
    }

    /**
     *
     *
     */
    public DvNativeFrame[] getNativeFrames() {
        Vector v = new Vector();

        for (Dsa dsa = topDsa; dsa != null; dsa = dsa.previous()) {
            v.add(dsa);
        }

        return (DvNativeFrame[]) v.toArray(new DvNativeFrame[0]);
    }

    /**
     *
     *
     */
    public DvJavaFrame[] getJavaFrames() {
        throw new Error("unimplemented method");
    }

    /**
     *
     *
     */
    public DvAddress getPSW() {
        throw new Error("unimplemented method");
    }

    /**
     *
     *
     */
    public DvRegister[] getRegisters() {
        throw new Error("unimplemented method");
    }

    /* End of DvThread methods */

    /**
     *
     *
     */

    public int cacheBlock() {
        if (cacheBlock != 0) {
            return cacheBlock;
        }
        if (ee == 0)
            return 0;
        try {
            int c = space.readInt(ee + 0x28);
            space.readInt(c);
            cacheBlock = c;
        } catch (Exception e) {}
        return cacheBlock;
    }

    public int cacheSize() {
        if (cacheSize != 0 || cacheBlock() == 0) {
            return cacheSize;
        }
        if (ee == 0)
            return 0;
        try {
            cacheSize = space.readInt(ee + 0x24);
        } catch (Exception e) {}
        return cacheSize;
    }

    public int cacheOrigSize() {
        if (cacheOrigSize != 0 || cacheBlock() == 0) {
            return cacheOrigSize;
        }
        if (ee == 0)
            return 0;
        try {
            cacheOrigSize = space.readInt(ee + 0x2c);
        } catch (Exception e) {}
        return cacheOrigSize;
    }

    public boolean inAllocCache(int address) {
        try {
            if (address >= (cacheBlock() + cacheSize()) && address <= (cacheBlock() + cacheOrigSize()))
                return true;
        } catch (Exception e) {}
        return false;
    }

    public boolean hasPendingException() {
        if (ee != 0 && is131) {
            try {
                int exceptionKind = space.readInt(ee + 16);
                if (exceptionKind != 0) {
                    return true;
                }
            } catch (Exception e) {}
        }
        return false;
    }

    public void printException(Document doc) {
        if (ee != 0 && is131) {
            try {
                int exceptionObject = space.readInt(ee + 20);
                int exceptionKind = space.readInt(ee + 16);
                if (exceptionObject != 0) {
                    doc.println("");
                    String cbName = jvm.getName(exceptionObject);
                    String detail = space.readStringObject(space.readInt(exceptionObject + 12));
                    String type = exceptionKind == 0 ? "old" : "pending";
                    doc.println("found " + type + " exception: " + cbName + ": " + detail);
                }
            } catch (Exception e) {}
        }
    }

    public void printAllocCache(Document doc) {
        if (ee != 0 && is131) {
            try {
                doc.println("");
                doc.println("alloc cache info:");
                int cache_busy = space.readInt(ee + 0x20);
                int target_size = space.readInt(ee + 0x44);

                doc.println("cache_busy = 0x" + hex(cache_busy));
                doc.println("cache_block = 0x" + hex(cacheBlock()));
                doc.println("cache_size = 0x" + hex(cacheSize()) + " -> 0x" 
                    + hex(cacheBlock() + cacheSize()));
                doc.println("cache_orig_size = 0x" + hex(cacheOrigSize()) 
                    + " -> 0x" + hex(cacheBlock() + cacheOrigSize()));
                doc.println("target_size = 0x" + hex(target_size));
                int bad = 0;

                if (cacheBlock() != 0 && cacheSize() != 0) {
                    int used = cacheOrigSize() - cacheSize();
                    int p = cacheBlock() + cacheSize();
                    int tail = p + used;
                    int link_size = space.readInt(cacheBlock() - 12);
                    int obj_size = space.readInt(cacheBlock() - 8);
                    int a_flags = space.readInt(cacheBlock() - 4);
                    doc.println("cache link size = 0x" + hex(link_size));
                    doc.println("cache obj size = 0x" + hex(obj_size));
                    doc.println("cache flags = 0x" + hex(a_flags));
                    doc.println("");

                    Table table = new Table(new String[] {"address", "length", "methods", "flags", "class name"});

                    while (p < tail) {
                        Row row = new Row();
                        int len = space.readInt(p) & 0x3ffffff8;
                        row.addCell(hex(p));
                        row.addCell(hex(len));
                        int methods = space.readInt(p + 4);
                        int flags = space.readInt(p + 8);
                        row.addCell(hex(methods));
                        row.addCell(hex(flags));
                        if (methods != 0) {
                            String name = jvm.getName(p + 4);
                            row.addCell(name);
                        } else {
                            row.addCell("");
                        }
                        table.addRow(row);
                        if (len <= 0 || len > cacheOrigSize()) {
                            bad = p;
                            break;
                        }
                        p += len;
                    }
                    doc.addElement(table);
                }
                if (bad != 0) {
                    doc.println("*** dodgy alloc cache? dump follows ***");
                    for (int i = cacheBlock(); i < bad + 0x100; i += 4) {
                        doc.println("0x" + hex(i) + ": " + hex(space.readInt(i)));
                    }
                }
            } catch (Exception e) {
                //e.printStackTrace();
                doc.println("problem reading ee " + hex(ee) + " exception: " + e);
            }
        }
    }

    void checkEE() throws Exception {
        if (verbose) System.out.println("asid = " + space.hexId() + " ee = " + hex(ee));
        int jnienv = space.readInt(ee);
        if (verbose) System.out.println("jnienv = " + hex(jnienv));
        int reserved1 = space.readInt(jnienv + 4);
        if (verbose) System.out.println("reserved1 = " + hex(reserved1));
        if (reserved1 != 0)
            throw new Exception("reserved1 is not zero! " + reserved1);
        int thread = space.readInt(ee + 12);
        if (verbose) System.out.println("thread = " + hex(thread));
        if (thread == 0)
            throw new Exception("thread pointer in ee is zero");
        int methodtable = space.readInt(thread);
        if (verbose) System.out.println("methodtable = " + hex(methodtable));
        int eyecatcher = space.readInt(methodtable);
        if (verbose) System.out.println("eyecatcher = " + hex(eyecatcher));
        if (eyecatcher == 0x434c4d54) { // CLMT
            isPostRas = true;
        }
    }

    void findEE() throws Exception {
        ee = space.readInt(tid + 0x28);
        if (ee < 0x10000) {
            // 1.3
            ee = tid - 0x200;
            is131 = true;
            if (space.readInt(ee - 0x18) == 0x584d4545) {
                // found HpiRasHeader eyecatcher "XMEE", new in SR15
                ee -= 0x10;
                isPostRas = true;
            }
            int jvmp = space.readInt(ee + 452);
            try {
                jvm = space.addJvm(jvmp, isPostRas);
            } catch (Throwable e) {
                if (verbose) System.out.println("could not create jvm at " + hex(jvmp));
            }
        } else {
            throw new Exception("bad ee " + hex(ee));
        }
    }

    void findDsaFromLastFrame() {
        try {
            int last_frame = 0;

            if (is131) {
                last_frame = space.readInt(ee + 0x19c);
            } else {
                last_frame = space.readInt(ee + 0x40);
            }
            if (last_frame != 0) {
                topDsa = Dsa.findTop(this, last_frame);
                isKnownType = true;
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
    }

    /**
     *
     *
     */
    Tcb(AddressSpace space, int tcbid) {
        if (verbose) System.out.println("creating tcb " + hex(tcbid));
        this.space = space;
        this.tcbid = tcbid;
        try {
            int stcb = space.readInt(tcbid + STCB);
            int otcb = space.readInt(stcb + OTCB);
            int thli = space.readInt(otcb + THLI);

            tid = space.readInt(thli + TID);
            if (tid != 0) {
                space.mapTid(tid, this);
                findEE();
            }
        } catch (Exception e) {}

        if (isLE() || isRtwa() || isLinkage() || isLinkage2() || isIrb() || isCics() || isUsta()) {
            isKnownType = true;
            try {
                topDsa = new Dsa(this, topDsaAddress);
            } catch (Exception e) {}
            if (topDsa == null || topDsa.prevCount() < 3) {
                type = "dodgy";
                dodgy = true;
                try {
                    topDsaAddress = findBestDsa();
                    topDsa = new Dsa(this, topDsaAddress);
                } catch (Exception e) {}
            }
        } else {
            type = "unknown";
            findDsaFromLastFrame();
            findCaa();
if (false && tid != 0) {
try {
int stackbase = space.readInt(tid + 132);
int caa = space.readInt(tid + 212);
System.out.println("caa " + hex(caa));
if (isValidCaa(caa)) System.out.println("yes!");
int smcb = space.readInt(tid + 216);
int eye = space.readInt(smcb);
System.out.println("smcb " + hex(smcb));
System.out.println("smcb eye " + hex(eye));
System.out.println("stack base for tcb " + hex(tcbid) + " is " + hex(stackbase));
topDsa = Dsa.findStack(this, smcb + 1000);
isKnownType = true;
} catch (Exception e) {}
}
        }
        if (verbose) System.out.println("finished creating tcb " + hex(tcbid));
    }

    /**
     *
     *
     */
    // Constructor if tcbid is unkown
    Tcb(AddressSpace space, int tcbid, int tid) {
        this.space = space;
        this.tid = tid;
        try {
            findEE();
        } catch (Exception e) {}
        type = "Java thread with no TCB";
        findDsaFromLastFrame();
        findCaa();
    }

    /**
     *
     *
     */
    void findCaa() {
        try {} catch (Exception e) {}
    }

    /**
     *
     *
     */
    public void addTraceEntry(int type, int psw) {
        if (traceTable == null) {
            traceTable = new Hashtable();
        }
        Integer pswInt = new Integer(psw);
        Integer count = (Integer) traceTable.get(pswInt);

        traceTable.put(pswInt, new Integer(count == null ? 1 : count.intValue() + 1));
    }

    /**
     *
     *
     */
    public Function[] traceFunctions() {
        if (traceFunctions == null && traceTable != null) {
            traceFunctions = space.traceFunctions(traceTable);
        }
        return traceFunctions;
    }

    /**
     *
     *
     */
    public Dsa topDsa() {
        return topDsa;
    }

    /**
     *
     *
     */
    public Dsa[] stack() {
        if (stack == null) {
            Vector v = new Vector();

            for (Dsa dsa = topDsa(); dsa != null; dsa = dsa.previous()) {
                v.add(dsa);
            }
            stack = (Dsa[]) v.toArray(new Dsa[0]);
        }
        return stack;
    }

    /**
     *
     *
     */
    public String where() {
        try {
            return topDsa.function();
        } catch (Exception e) {
            return "";
        }
    }

    /**
     *
     *
     */
    public boolean isValidCaa(int caa) {
        try {
            int eye1 = space.readInt(caa - 0x18);
            int eye2 = space.readInt(caa - 0x14) & 0xffff0000;

            if (eye1 == 0xc3c5c5c3 && eye2 == 0xc1c10000) { // CEECAA
                this.caa = caa;
                if (verbose) System.out.println("found valid caa " + hex(caa));
                return true;
            }
            if (verbose) {
                System.out.println("invalid caa " + hex(caa) + ": " + hex(eye1) + ", " + hex(eye2));
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    public int caa() {
        return caa;
    }

    /**
     *
     *
     */
    public boolean isLE() {
        try {
            int caa = space.readInt(tcbid + 0x60);

            if (isValidCaa(caa)) {
                topDsaAddress = space.readInt(tcbid + 0x64);
                type = "LE";
                return true;
            }
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    int findBestDsa() throws Exception {
        int dummyDsa = space.readInt(caa + 0x2e0);
        int bottomDsa = space.readInt(dummyDsa + 0x4c);
        Dsa dsa = Dsa.findTop(this, bottomDsa);
        int bestDsa = dsa.stackPointer();
        int bestCount = dsa.prevCount();

        int topDsa = space.readInt(tcbid + 0x64);

        try {
            dsa = new Dsa(this, topDsa);
            if (dsa.prevCount() > bestCount) {
                bestDsa = dsa.stackPointer();
                bestCount = dsa.prevCount();
            }
        } catch (Exception e) {}

        try {
            int stcb = space.readInt(tcbid + STCB);
            int otcb = space.readInt(stcb + 0xd8);
            int cof = space.readInt(otcb + 0x38);
            int usta = space.readInt(cof + 0x4);

            topDsa = space.readInt(usta + 0x4c);
            dsa = new Dsa(this, topDsa);
            if (dsa.prevCount() > bestCount) {
                bestDsa = dsa.stackPointer();
                bestCount = dsa.prevCount();
            }
        } catch (Exception e) {}

        try {
            int rtwa = space.readInt(tcbid + 0xe0);

            topDsa = space.readInt(rtwa + 0x70);
            dsa = new Dsa(this, topDsa);
            if (dsa.prevCount() > bestCount) {
                bestDsa = dsa.stackPointer();
                bestCount = dsa.prevCount();
            }
        } catch (Exception e) {}

        try {
            int stcb = space.readInt(tcbid + STCB);
            int lsdp = space.readInt(stcb + 0x74);
            int estk = space.readInt(stcb + 0x80);

            if (lsdp != estk) {
                topDsa = space.readInt(estk + 0x8 + (13 * 4));
                dsa = new Dsa(this, topDsa);
                if (dsa.prevCount() > bestCount) {
                    bestDsa = dsa.stackPointer();
                    bestCount = dsa.prevCount();
                }
            }
        } catch (Exception e) {}

        try {
            int irb = space.readInt(tcbid);

            topDsa = space.readInt(irb + 0x54);
            dsa = new Dsa(this, topDsa);
            if (dsa.prevCount() > bestCount) {
                bestDsa = dsa.stackPointer();
                bestCount = dsa.prevCount();
            }
        } catch (Exception e) {}

        try {
            int r6 = space.readInt(tcbid + 0x48);

            if (space.readInt(r6 + 0x70) == tcbid) {
                topDsa = space.readInt(r6 + 0x84);
                dsa = new Dsa(this, topDsa);
                if (dsa.prevCount() > bestCount) {
                    bestDsa = dsa.stackPointer();
                    bestCount = dsa.prevCount();
                }
            }
        } catch (Exception e) {}

        try {
            int stcb = space.readInt(tcbid + STCB);
            int lsdp = space.readInt(stcb + 0x74);
            int estk = space.readInt(stcb + 0x80);

            if (lsdp != estk) {
                try {
                    int r5 = space.readInt(estk + 0x8 + (5 * 4));

                    topDsa = space.readInt(r5 + 0x54);
                    dsa = new Dsa(this, topDsa);
                    if (dsa.prevCount() > bestCount) {
                        bestDsa = dsa.stackPointer();
                        bestCount = dsa.prevCount();
                    }
                } catch (Exception e) {
                    int r5 = space.readInt(estk + 0x8 + (2 * 5 * 4 + 4)); // try 64-bit version

                    topDsa = space.readInt(r5 + 0x54);
                    dsa = new Dsa(this, topDsa);
                    if (dsa.prevCount() > bestCount) {
                        bestDsa = dsa.stackPointer();
                        bestCount = dsa.prevCount();
                    }
                }
            }
        } catch (Exception e) {}

        return bestDsa;
    }

    /**
     *
     *
     */
    public boolean isUsta() {
        try {
            int stcb = space.readInt(tcbid + STCB);
            int otcb = space.readInt(stcb + 0xd8);
            int cof = space.readInt(otcb + 0x38);
            int usta = space.readInt(cof + 0x4);
            int caa = space.readInt(usta + 0x48);

            if (isValidCaa(caa)) {
                topDsaAddress = findBestDsa();
                type = "Usta";
                return true;
            }
        } catch (Exception e) {
            if (verbose) {
                e.printStackTrace(System.out);
            }
        }
        return false;
    }

    /**
     *
     *
     */
    public boolean isLinkage() {
        try {
            int stcb = space.readInt(tcbid + STCB);
            int lsdp = space.readInt(stcb + 0x74);
            int estk = space.readInt(stcb + 0x80);

            if (lsdp != estk) {
                int caa = space.readInt(estk + 0x8 + (12 * 4));  // Register 12

                if (isValidCaa(caa)) {
                    topDsaAddress = space.readInt(estk + 0x8 + (13 * 4));
                    type = "Linkage";
                    return true;
                }
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    public boolean isLinkage2() {
        try {
            int stcb = space.readInt(tcbid + STCB);
            int lsdp = space.readInt(stcb + 0x74);
            int estk = space.readInt(stcb + 0x80);

            if (lsdp != estk) {
                // This stuff is a total guess!
                int usta = space.readInt(estk + 0x8 + (2 * 4));  // Register 2

                if (space.readInt(usta) == 0xe4e2e3c1) {
                    int caa = space.readInt(usta + 0x48);

                    if (isValidCaa(caa)) {
                        topDsaAddress = space.readInt(usta + 0x4c);
                        type = "Linkage 2";
                        return true;
                    }
                }
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    public boolean isRtwa() {
        try {
            int rtwa = space.readInt(tcbid + 0xe0);

            if (rtwa != 0) {
                int caa = space.readInt(rtwa + 0x6c);

                if (isValidCaa(caa)) {
                    topDsaAddress = space.readInt(rtwa + 0x70);
                    type = "Failing";
                    failingRegisters = new int[18];
                    for (int i = 0; i < failingRegisters.length; i++) {
                        failingRegisters[i] = space.readInt(rtwa + 0x3c + (i << 2));
                    }
                    return true;
                }
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    public boolean isIrb() {
        try {
            int irb = space.readInt(tcbid);
            int caa = space.readInt(irb + 0x50);

            if (isValidCaa(caa)) {
                topDsaAddress = space.readInt(irb + 0x54);
                type = "IRB";
                return true;
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    public boolean isCics() {
        // discovered by recursively searching dump for references to DSA
        try {
            int r6 = space.readInt(tcbid + 0x48);

            if (space.readInt(r6 + 0x70) == tcbid) {
                topDsaAddress = space.readInt(r6 + 0x84);
                // should create a Dsa method to check validity
                int prev = space.readInt(topDsaAddress + 4);

                prev = space.readInt(prev + 4);
                type = "Cics";
                return true;
            }
        } catch (Exception e) {
            if (verbose) e.printStackTrace(System.out);
        }
        return false;
    }

    /**
     *
     *
     */
    public boolean isKnownType() {
        return isKnownType;
    }

    /**
     *
     *
     */
    public String type() {
        return type;
    }

    /**
     *
     *
     */
    
    /** Return the sys_thread_t */
    public int tid() {
        return tid;
    }

    public int ee() {
        return ee;
    }

    public int threadIndex() {
        if (isJava() && is131) {
            try {
                return space.readInt(ee + 0x154);
            } catch (Exception e) {}
        }
        return 0;
    }

    public int failingDsa() {
        int regs[] = failingRegisters();
        if (regs != null)
            return regs[13];
        return 0;
    }

    public int failingPsw() {
        int regs[] = failingRegisters();
        if (regs != null)
            return regs[17] & 0x7fffffff;
        return 0;
    }

    public int[] failingRegisters() {
        if (failingRegisters != null)
            return failingRegisters;
        try {
            int cib = space.readInt(caa + 0x2d8);
            int zmch = space.readInt(cib + 0x24);
            int eye = space.readInt(zmch);
            if (eye != 0xe9d4c3c8)
                return null;
            failingRegisters = new int[18];
            for (int i = 0; i < failingRegisters.length; i++) {
                failingRegisters[i] = space.readInt(zmch + 8 + (i << 2));
            }
            return failingRegisters;
        } catch (Exception e) {}
        return null;
    }

    public int tid_type() {
        try {
            return space.readInt(tid + 36);
        } catch (Exception e) {
            return -1;
        }
    }

    public int tid_state() {
        try {
            return space.readInt(tid + 40);
        } catch (Exception e) {
            return -1;
        }
    }

    String hexId() {
        return Integer.toHexString(tcbid);
    }

    public String toString() {
        return hexId();
    }

    /**
     *
     *
     */
    static String hex(int i) {
        return Integer.toHexString(i);
    }
}

