#!/usr/local/bin/jython

# This is the main logic of Dump.main.

from com.ibm.jvm.svcdump import Dump
import sys

def usage():
    print "Usage: java com.ibm.jvm.svcdump.Dump [-options] <filename>"
    print "\nwhere options include:"
    print "        -cache        print alloc cache"
    sys.exit()

nargs = len(sys.argv)
if nargs < 2:
    usage()

verbose = 0
printcache = 0

dumpfile = sys.argv[nargs - 1]

for arg in sys.argv[1:nargs - 1]:
    if arg == "-verbose":
        verbose = 1
    elif arg == "-cache":
        printcache = 1
    else:
        usage()

dump = Dump(dumpfile)

spaces = dump.addressSpaces()

for space in spaces:
    print "space = ", space
    for tcb in space.tcbs():
        if tcb.isKnownType():
            tid = tcb.tid()
            print "\n*** TCB", tcb, "tid", hex(tid),
            if tcb.threadIndex():
                print "tindex", hex(tcb.threadIndex()),
            if tcb.tid_type() != -1:
                print "tid type", hex(tcb.tid_type()),
            if tcb.tid_state() != -1:
                print "tid state", hex(tcb.tid_state()),
            if tcb.caa():
                print "caa", hex(tcb.caa()),
            print "***\n"
            # if is is a failing tcb, print out the registers
            regs = tcb.failingRegisters()
            if regs:
                print "Registers at point of failure:"
                for y in range(0,4):
                    for x in range(0,4):
                        n = (y * 4) + x
                        if n > 9:
                            spaces = " "
                        else:
                            spaces = "  "
                        print "r%d:%s%08x" % (n, spaces, regs[n]), "   ",
                    print ""
                print "psw: %08x %08x" % (regs[16], regs[17])
            # now print the tracback
            print "Dsa         Entry       Offset      Function"
            print "---         -----       ------      --------"
            dsa = tcb.topDsa()
            while dsa:
                print "%08x    %08x    %08x    %s" % \
                    (dsa.address(), dsa.entryPoint(), dsa.offset(), dsa.function())
                dsa = dsa.previous()
            # optionally print the alloc cache
            if printcache and tid:
                # note: need to do this in python
                tcb.printAllocCache()
        elif verbose:
            print "unknown tcb: ", tcb
