/**
 * This class provides access to properties for the svcdump and findroots packages (which for
 * historical reasons are bundled together in svcdump.jar). The properties file is looked for
 * in the users home directory under the name ".svcdumprc". Properties may be overridden by
 * specifying them as system properties via -D on the command line.
 */

package com.ibm.jvm.util;

import java.util.*;
import java.io.*;

public class SvcdumpProperties {
    /** The properties object */
    static protected Properties props;

    /**
     * Initialize the class. This is where we open the .svcdumprc file and load the
     * properties.
     */
    static {
        try {
            props = new Properties();
            String home = System.getProperty("user.home");
            String sep = System.getProperty("file.separator");
            FileInputStream fis = new FileInputStream(home + sep + ".svcdumprc");
            props.load(fis);
        } catch (Exception e) {
        }
    }

    /**
     * Get the specified property. System properties (via -D) take precedence.
     * @param key the property key to get
     */
    public static String getProperty(String key) {
        String prop = System.getProperty(key);
        if (prop != null) {
            return prop;
        }
        return props.getProperty(key);
    }

    /**
     * Get the specified property as a boolean value or return the default value.
     */
    public static boolean getBooleanProperty(String key, boolean defaultValue) {
        String prop = getProperty(key);
        if (prop == null) {
            return defaultValue;
        } else {
            return prop.equals("") || prop.equalsIgnoreCase("true");
        }
    }

    /**
     * Get the specified property as an integer value or return the default value.
     */
    public static int getIntProperty(String key, int defaultValue) {
        String prop = getProperty(key);
        if (prop == null) {
            return defaultValue;
        } else {
            try {
                return Integer.parseInt(prop);
            } catch (NumberFormatException e) {
                throw new Error("bad property: " + key + " should be numeric but found " + prop);
            }
        }
    }
}
