package com.ibm.jvm.util.html;

import java.util.*;

/**
 *  This class represents a HTML table column.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Column {
    /** The containing Table. */
    protected Table table;
    /** The list of cells in this column. */
    protected ArrayList cells = new ArrayList();
    /** The width in characters of this column. */
    private int width = -1;

    /**
     * Sets the containing Table.
     */
    public void setTable(Table table) {
        this.table = table;
    }

    /**
     * Adds a new cell to the column.
     * @param cell the cell to be added.
     */
    public void addCell(Cell cell) {
        cells.add(cell);
        cell.setColumn(this);
    }

    /**
     * Adds a new cell to the column.
     * @param cell String version of the cell to be added.
     */
    public void addCell(String cell) {
        addCell(new Cell(cell));
    }

    /**
     * Returns the width of this column. This is calculated as the maximum number
     * of characters for any cell in the column.
     */
    public int width() {
        if (width == -1) {
            for (Iterator it = cells.iterator(); it.hasNext();) {
                Cell cell = (Cell)it.next();
                if (cell.width() > width) {
                    width = cell.width();
                }
            }
        }
        return width;
    }
}
