package com.ibm.jvm.util.html;

import java.util.*;

/**
 *  This class represents some text.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Text extends Element {

    /** The text string. */
    protected String text;
    /** Whether to output a newline (plain text mode) or break (html mode). */
    protected boolean newline;
    /** The header level (1-6). 0 indicates it is not a header. */
    protected int headerLevel;

    /**
     * Creates a new Text.
     */
    public Text() {
    }

    /**
     * Creates a new Text with the given text.
     * @param text the text
     */
    public Text(String text) {
        this.text = text;
    }

    /**
     * Creates a new Text with the given text and optional newline.
     * @param text the text
     * @param newline true if a newline is to be appended (or break in html mode)
     */
    public Text(String text, boolean newline) {
        this.text = text;
        this.newline = newline;
    }

    /**
     * Creates a new Text with the given text at the given header level.
     * @param text the text
     * @param level the header level
     */
    public Text(String text, int level) {
        this.text = text;
        headerLevel = level;
    }

    /**
     * Returns a html representation of this text string.
     */
    protected String toHtmlString() {
        String rep = text;
        if (headerLevel != 0) {
            rep = "<h" + headerLevel + ">" + text + "</h" + headerLevel + ">";
        }
        if (newline) {
            rep += "<br>";
        }
        return rep.length() == 0 ? rep : beautify(rep);
    }

    /**
     * Returns a text representation of this text string.
     */
    protected String toTextString() {
        String rep = text;
        if (headerLevel != 0) {
            if (headerLevel == 1) {
                rep += Document.lineSeparator;
                rep += Document.pad(text.length(), '-');
            } else {
                rep = "*** " + text + " ***";
            }
            rep += Document.lineSeparator;
        }
        if (newline) {
            rep += Document.lineSeparator;
        }
        return rep;
    }
}
