package com.ibm.jvm.zseries;


/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public abstract class RXInstruction extends Instruction {

    int word;

    RXInstruction(Engine engine, int word) {
        this.engine = engine;
        this.word = word;
    }

    public int length() {
        return 4;
    }

    int r1() {
        return (word >> 20) & 0xf;
    }

    int x2() {
        return (word >> 16) & 0xf;
    }

    int b2() {
        return (word >> 12) & 0xf;
    }

    int d2() {
        return word & 0xfff;
    }

    int address() throws Exception {
        int x2 = 0;

        if (x2() != 0) {
            if (engine.validRegister(x2()))
                x2 = engine.getRegister(x2());
            else
                throw new Exception("bad address");
        }
        if (engine.validRegister(b2()))
            return x2 + engine.getRegister(b2()) + d2();
        else
            throw new Exception("bad address");
    }

    public boolean isFloatingPoint() {
        return false;
    }

    public String toString() {
        String x2 = x2() == 0 ? "" : "$r" + x2() + ",";
        String r1 = isFloatingPoint() ? "$fpr" : "$r";

        return mnemonic() + "   " + r1 + r1() + ", x'" + hx(d2()) + "'(" + x2 + "$r" + b2() + ")";
    }
}
