/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/CommandPlugin.java, ras, dumpviewer 1.18
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 * 36524   020502 PHILR:  New for feature 36524
 * 36524   170502 philr:  extends Thread
 * 36524   270502 philr:  Plugin methods need to be non static.
 * 36524   300502 philr:  Tidied up for coding standards
 * 36524   040602 philr:  Added verbModifierForFind to support the Find
 *                        command syntax such as find "abc def",a,b,c
 * 36524  160602  philr   Added support for enhancers beyond the end bracket
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION: Abstract class
 * ===========================================================================
 */


// Note that there will be various methods in DvUtils that the plugins will
// comonly use to establish such things as whether this plugin is valid
// on this platform, they should not be used until the dump has been opened..
//
//  DvUtils.isBigEndian()
//  DvUtils.is32bit()
//  DvUtils.isDumpOpen()
//  DvUtils.getMemory()
//

package com.ibm.jvm.dump;

import java.util.*;
import java.io.*;
import java.lang.reflect.Method;


 /**
  *  Description: Whats the purpose of this class
  *
  *
  *
  *
  *
  */
abstract class CommandPlugin extends Thread {

protected Method method;
protected String paramString;
protected String verb=null;
protected String verbModifier=null;
protected String verbModifierForFind=null;
protected String seperator=null;  // was it = or (
protected Vector enhancers=null;
protected boolean forcedEnd = false;
protected CommandPluginResponse cpr; // the output
protected DvDump theDump=null;
protected boolean cpInProgress=true;


    /**
     *   <p>Description: Set the verb that will be used
     *                   when we run this instance.
     *   </p>
     */
    public CommandPlugin() {


        cpr = new  CommandPluginResponse();

    }



    /**
     *   <p>Description: Add an observer to the command plugin
     *                   response object.
     *   </p>
     */
    public void addObserverToResponse(Observer o) {

        cpr.addObserver(o);

    }

    /**
     *   <p>Description: Set the dump identifier so that the assorted
     *                   commands can get to memory etc....
     *   </p>
     */
    public void setTheDump(DvDump dump) {
        theDump = dump;
    }




    /**
     *   <p>Description: Set the verb that will be used
     *                   when we run this instance.
     *   </p>
     */
    public void setVerb(String v) {
        verb = v;
    }

    /**
     *   <p>Description: Set the verbModifier that will be used
     *                   when we run this instance.
     *   </p>
     */
    public  void setVerbModifier(String vm) {
    // strip off any leading trailing spaces
        if (null != vm) {
            if (null != paramString) {
              verbModifierForFind = vm + " "+ paramString;
            }
            verbModifier = vm.trim();
        }
        else {
            verbModifier=null;
        }
    }

    /**
     *   <p>Description: Set the seperator for the command
     *                   so that the launched methods can distinquish
     *                   between such things as SET xyz= and DISPLAY X(
     *                   type formats.....
     *   </p>
     */
    public void setSeperator(String sep) {
    // strip off any leading trailing spaces
        if (null != sep) {
            seperator = sep;
        }
        else {
            seperator = null;
        }
    }

    /**
     *   <p>Description: Set boolean that gives long running
     *                   requests chance to end prematurely
     *                   when asked from the console (! command)
     *   </p>
     */
    public void setForcedEnd(boolean torf) {
        forcedEnd = torf;
    }

    /**
     *   <p>Description: Get Boolean showing whether the command is still
     *                   going
     *   </p>
     */
    public boolean getInProgress() {
        return cpInProgress;
    }

    /**
     *   <p>Description: Set the methodName that will be used
     *                   when we run this instance.
     *   </p>
     */
    public  void setMethod(Method m) {
        method = m;
    }

    /**
     *   <p>Description: Set the parameters that will be used
     *                   when we run this command.
     *
     *   </p>
     */
    public void setParamString(String parms) {
        if (null != parms) {
            paramString = parms.trim();
        }
        else {
            paramString = null;
        }
     }

    /**
     *   <p>Description: Set the enhancers that will be used
     *                   when we run this command. Enhancers aare used
     *                   with commands of format
     *      Verb modifier(param1,Param2,param3...) enhancer1 enhancer2 ...
     *      and are stored in a vector.
     *
     *   </p>
     */
    public void setEnhancers(String enhancerString) {

        if (null != enhancerString) {
            enhancers = new Vector();
            StringTokenizer st = new StringTokenizer(enhancerString);

            int count = st.countTokens();
            for (int i =0;i<count; i++) {
                String s = st.nextToken();
                enhancers.add(s);
            }


        }
     }




    /**
     *   <p>Description:
     *   </p>
     */
    public abstract String  pluginName();


    /**
     *   <p>Description: Returns a set of strings showing gui
     *                    commands handled by this plugin. Format is:
     *
     *                    Verb,Verb Modifier,method
     *
     *       where method is the methodName within this plugin that
     *       handles the verb/modifier combination. If the verb Modifier is
     *       not applicable then need "verb, ,method" - note the space
     *   </p>
     */
    public abstract String[] getSyntax()  ;


    /**
     *   <p>Description:
     *   </p>
     */
    public abstract String[] guiPopupSyntax();


    /**
     *   <p>Description:
     *   </p>
     */
    public String   help() {
           return "No help: method help not overidden in " + pluginName();
    }

    //**********************************************************
    //*
    //*  console help, returns a string (with embedded \n's)
    //*  that will be presented when help is requested against
    //*  this plugin. The what string allows help to be displayed
    //*  for an individual command rather than a set of commands.
    //*
    //*
    //**********************************************************
    /**
     *   <p>Description:
     *   </p>
     */
    public String  help(String what) {
        return "No help: method help not overidden in " + pluginName();
    }

    /**
     *   <p>Description: Check that enhancers specified are valid
     *   </p>
     */

    public boolean enhancersCheck(String [] sa) {
        // so we go through each one in the enhancer vector and check that
        // its in the incoming string - if its in twice we let it pass
        // if theres an invalid one we return false - check is done after
        // triming and uppercasing....
        boolean bRet = true;
        if (null != sa && 0 != enhancers.size()) {
            for (int i=0;i<enhancers.size();i++) {
                String elem = ((String) enhancers.elementAt(i)).trim().toUpperCase();
                boolean bIsFound = false;
                for (int j=0;j<sa.length&&false == bIsFound;j++) {
                    String valid = sa[j].trim().toUpperCase();
                    if (valid.equals(elem)) {
                        bIsFound = true;
                    }
                }
                if (false == bIsFound) {
                    i = enhancers.size();
                    bRet = false;
                }
            }
        }
        return bRet;
    }

    /**
     *   <p>Description: Check whether an individual enhancers is specified
     *   </p>
     */

    public boolean enhancerIsPresent(String sa) {
        // so we go through each one in the enhancer vector and check whether
        // it  matches whats in incoming string

        boolean bRet = false; // assume it aint present
        if (null != sa && 0 != enhancers.size()) {

            for (int i=0;i<enhancers.size() && false == bRet;i++) {
                String elem = ((String) enhancers.elementAt(i)).trim().toUpperCase();
                String valid = sa.trim().toUpperCase();
                if (valid.equals(elem)) {
                        bRet = true;
                }


            }
        }
        return bRet;
    }


    /**
     *   <p>Description: Used to run a command under a sperate thread
     *   </p>
     */
    public void run() {

        // so lets construct a thread to do this under
        //System.out.println("\nYep I am running");
        forcedEnd = false;
        cpInProgress = true;

        try {
            Object o = method.invoke(this,null);
        }
        catch (java.lang.IllegalAccessException e) {
            String s = e.getMessage();
            if (null != s) {
                System.err.println(s);
                e.printStackTrace(System.out);
            }
            // don't rethrow this as we want the invoker to keep
            // on trucking ...................

            forcedEnd =true;

        }
        catch (java.lang.reflect.InvocationTargetException e) {
            System.err.println("InvocationTargetException!!! method = " + method);
            System.err.println("Target stack trace:");
            e.getTargetException().printStackTrace();
            String s = e.getMessage();
            if (null != s) System.out.println(s);
            System.err.println("Our stack trace:");
            e.printStackTrace(System.out);
            // don't rethrow this as we want the invoker to keep
            // on trucking ...................
            forcedEnd = true;
        }

        while (false == forcedEnd) {
            try {
                Thread.sleep(500);
            }
            catch  (java.lang.InterruptedException e) {
                System.out.println("\nInterruptedException");
            }
        } // end of while
        forcedEnd = false; // reset it to normal
        cpInProgress  = false;

     } // end of method  run()

    /**
    *   <p>Description: Used to run a command under a sperate thread
    *   </p>
    */
    public String toString() {
        String s;

        s = "\n Method:"+method +
            "\n paramString:"+paramString+
            "\n verb:" + verb +
            "\n verbModifier:" + verbModifier +
            "\n seperator:" + seperator;



        return s;

    }


}
